/* applyMove.h
 */
#ifndef _APPLY_MOVE_H
#define _APPLY_MOVE_H

#include "osl/position.h"
#include "osl/move.h"
#include "osl/piece.h"

namespace osl
{
  // Implementations are provided for evrey States.
  // It is important not to provide default implementations,
  // so as to detect missing implementation in compilation error
  template <Player P,class State>
  struct ApplyDoUndoSimpleMove;
  template <Player P,class State>
  struct ApplyDoUndoCaptureMove;
  template <Player P,class State>
  struct ApplyDoUndoDropMove;

  /**
   * doUndoMove Ƥ֤Υ󥿡ե
   * 
   * ApplyMove<BLACK>::doUndoMove(state, move, function);
   * Τ褦˻Ȥ
   * State  ApplyDoUndoXXXMove<State> ü첽Ǥ
   */  
  template<Player P>
  struct ApplyMove
  {
    template <class State, class Func>
    static void doUndoMove(State& state, Move move, Func &f)
    {
      assert(P == getPlayer(move));
      assert(move && isValid(move));
      Position from=getFrom(move);
      Position to=getTo(move);
      if (isOffBoard(from))
      {
	assert(state.getPieceAt(to) == PIECE_EMPTY);
	ApplyDoUndoDropMove<P,State>::template doUndoDropMove<Func>(state,to,getPtype(move),f);
      }
      else
      {
	assert(state.getPieceAt(from) != PIECE_EMPTY);
	Piece captured=state.getPieceAt(to);
	if (captured != PIECE_EMPTY)
	{
	  ApplyDoUndoCaptureMove<P,State>::template doUndoCaptureMove<Func>
	    (state,from,to,captured,getPromoteMask(move),f);
	}
	else
	{
	  ApplyDoUndoSimpleMove<P,State>::template doUndoSimpleMove<Func>
	    (state,from,to,getPromoteMask(move),f);
	}
      }
    }
  };

  struct ApplyMoveOfTurn
  {
    /**
     * doUndoMove Ƥ֤Υ󥿡ե
     * 
     * ApplyMoveOfTurn::doUndoMove(state, move, function);
     * Τ褦˻Ȥ
     * ApplyMove Ȱä Player template ˤɬפʤ
     */  
    template <class State, class Func>
    static void doUndoMove(State& state, Move move, Func &f)
    {
      assert(state.getTurn() == getPlayer(move));
      if (getPlayer(move) == BLACK)
	ApplyMove<BLACK>::doUndoMove(state, move, f);
      else
	ApplyMove<WHITE>::doUndoMove(state, move, f);
    }

    /**
     * undo ʤ move Ƥֻ interface
     */
    template<typename State>
    static void doMove(State& state, Move move)
    {
      assert(state.isAlmostValidMove(move));
      const Position from=getFrom(move);
      const Position to=getTo(move);
      if (isOffBoard(from))
      {
	state.doDropMove(to,getPtype(move));
      }
      else
      {
	const Piece captured = state.getPieceOnBoard(to);
	if (captured != PIECE_EMPTY)
	{
	  state.doCaptureMove(from,to,captured,getPromoteMask(move));
	}
	else
	{
	  state.doSimpleMove(from,to,getPromoteMask(move));
	}
      }
      state.changeTurn();
    }

    /**
     * ɾؿѹȿǤȤ interface
     *
     * ʤ EvalState::doMove ݤ򸫤뤬
     * EvalState Ȥ̤˥ƥȤȤѤ
     * @param Eval PieceEval Ȥ Progress ʤ
     */
    template<typename Eval, typename State>
    static void doMove(Eval& e, const State& state, Move move)
    {
      const Position from = getFrom(move);
      const Position to = getTo(move);
      if (isOffBoard(from))
      {
	e.doDropMove(state, to, getPtype(move));
      }
      else
      {
	const Piece captured = state.getPieceOnBoard(to);
	if (captured != PIECE_EMPTY)
	{
	  e.doCaptureMove(state, from, to, captured, getPromoteMask(move));
	}
	else
	{
	  e.doSimpleMove(state, from, to, getPromoteMask(move));
	}
      }
    }
  };
} // namespace osl


#endif /* _APPLY_MOVE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
