/* bitXmask.h
 */
#ifndef _BITXMASK_H
#define _BITXMASK_H

#include "osl/position.h"
#include "osl/carray.h"
#include <iosfwd>

namespace osl {

  enum BitXmask{
    XMASK_NULL=0,
    XMASK_MAX=0x3fe
  };
  
  inline void clearAll(BitXmask& mask){
    mask=static_cast<BitXmask>(0);
  }
  
  inline void set(BitXmask& mask,const int x){
    mask=static_cast<BitXmask>(static_cast<int>(mask)|(1<<x));
  }
  
  inline unsigned int *addr(BitXmask& mask,Position pos){
    return (unsigned int *)(&mask);
  }
  
  inline void clear(BitXmask& mask,const int x){
    mask=static_cast<BitXmask>(static_cast<int>(mask)& ~(1<<x));
  }
  
  inline bool isSet(const BitXmask mask,const int x){
    return (static_cast<int>(mask)&(1<<x))!=0;
  }

  class BoardTable{
  private:
    CArray<unsigned int,POSITION_SIZE> setMask;
  public:
    BoardTable();
    ~BoardTable();
    
    unsigned int getSetMask(Position pos) const{
      return setMask[pos];
    }
  };
  
  extern const BoardTable boardTable;
  
  template<>
  inline void setPos(BitXmask& mask,const Position pos){
    mask=static_cast<BitXmask>(static_cast<int>(mask)|boardTable.getSetMask(pos));
  }
  
  template<>
  inline void clearPos(BitXmask& mask,const Position pos){
    mask=static_cast<BitXmask>(static_cast<int>(mask)& ~boardTable.getSetMask(pos));
  }
  
  std::ostream& operator<<(std::ostream&,const BitXmask);
}

#endif /* _BITXMASK_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
