#ifndef _BOARD_H
#define _BOARD_H
#include "player.h"
#include "position.h"
#include <algorithm>
#include <cassert>
#include <iosfwd>

namespace osl {

  const int BOARD_HEIGHT=16;
  /*        ̤Υǥå                                */
  /* X, Y1-9ϰϤɽ */
  /* Xϱ롥YϾ夫 */
  /* ʤ0 */

  /*                                                     ()   */
  /* (A0)  ......................... (00) */
  /* (A1)  ......................... (01) */
  /* (A2) 92 82 72 62 52 42 32 22 12 (02) */
  /* (A3) 93 83 73 63 53 43 33 23 13 (03) */
  /* (A4) 94 84 74 64 54 44 34 24 14 (04) */
  /* (A5) 95 85 75 65 55 45 35 25 15 (05) */
  /* (A6) 96 86 76 66 56 46 36 26 16 (06) */
  /* (A7) 97 87 77 67 57 47 37 27 17 (07) */
  /* (A8) 98 88 78 68 58 48 38 28 18 (08) */
  /* (A9) 99 89 79 69 59 49 39 29 19 (09) */
  /* (AA) 9A 8A 7A 6A 5A 4A 3A 2A 1A (0A) */
  /* (AB) ...........................(0B) */
  /* (AC) ...........................(0C) */
  /* (AD) ...........................(0D) */
  /* (AE) ...........................(0E) */
  /* (AF) ...........................(0F) */

  inline Position newPosition(const int x,const int y){
    return static_cast<Position>((x*BOARD_HEIGHT)+y+1);
  }
  
  inline int getX(const Position pos){
    
    return static_cast<int>(pos)>>4;
  }
  inline int getY(const Position pos){
    return (static_cast<int>(pos)&0xf)-1;
  }
  
  
  inline bool isOnBoard(const Position pos){
    return (1<=getX(pos) && getX(pos) <=9 &&
	    1<=getY(pos) && getY(pos) <=9);
  }

  inline bool isValid(const Position pos){
    return isOffBoard(pos) ||
      isOnBoard(pos);
  }
  
  inline int reverseY(int y){
    return 10-y;
  }
  
  inline int reverseX(int x){
    return 10-x;
  }

  /** 
   * ξ̤ä֤
   * ®٤̵뤷 
   */
  inline Position positionForBlack(const Position pos,const Player pl){
    if(pl==BLACK) return pos;
    else return newPosition(reverseX(getX(pos)),reverseY(getY(pos)));
  }

  static inline Position onBoardMax(){ return newPosition(9,9);}
  static inline Position onBoardMin(){return newPosition(1,1);}
  
  /**
   * posONBOARD_MINONBOARD_MAXδ֤ˤ
   */
  inline bool isOnBoardRegion(const Position pos){
    return static_cast<unsigned int>(pos-onBoardMin()) <=
      static_cast<unsigned int>(onBoardMax()-onBoardMin());
  }

  static inline Position& operator++(Position& pos){
    pos=static_cast<Position>(static_cast<int>(pos)+1);
    return pos;
  }
  
  enum Offset{
    OFFSET_MIN=-0x100,
    OFFSET_ZERO=0,
    OFFSET_MAX=0x100
  };
  
#define NEW_OFFSET(dx,dy) static_cast<Offset>((static_cast<int>(dx)*BOARD_HEIGHT)+static_cast<int>(dy))

  inline Offset newOffset(int dx,int dy){
    return NEW_OFFSET(dx,dy);
  }
  
  inline Offset operator-(const Position pos0){
    return static_cast<Offset>(-static_cast<int>(pos0));
  }
  
  inline Offset operator-(const Position pos0,const Position pos1){
    return static_cast<Offset>(static_cast<int>(pos0)-static_cast<int>(pos1));
  }
  inline Position operator+(const Position pos,const Offset offset){
    return static_cast<Position>(static_cast<int>(pos)+static_cast<int>(offset));
  }
  inline Offset operator+(const Offset offset0,const Offset offset1){
    return static_cast<Offset>(static_cast<int>(offset0)+static_cast<int>(offset1));
  }
  inline Offset operator-(const Offset o1,const Offset o2){
    return static_cast<Offset>(static_cast<int>(o1)-static_cast<int>(o2));
  }
  inline Position operator-(const Position pos,const Offset offset){
    return static_cast<Position>(static_cast<int>(pos)-static_cast<int>(offset));
  }
  inline Offset operator*(const Offset offset0,const int mult){
    return static_cast<Offset>(static_cast<int>(offset0)*mult);
  }

  inline Position& operator+=(Position& pos,const Offset offset){
    pos=pos+offset;
    return pos;
  }
  
  inline Offset operator-(const Offset offset){
    return static_cast<Offset>(-static_cast<int>(offset));
  }
  inline Offset& operator+=(Offset& offset0,const Offset offset1){
    offset0=offset0+offset1;
    return offset0;
  }
  inline Position& operator-=(Position& pos,const Offset offset){
    pos=pos-offset;
    return pos;
  }
  inline Offset& operator-=(Offset& off1,const Offset off2){
    off1=static_cast<Offset>(static_cast<int>(off1)-static_cast<int>(off2));
    return off1;
  }
  inline Offset operator*(const Offset off1,const Offset off2){
    return static_cast<Offset>(static_cast<int>(off1)*static_cast<int>(off2));
  }
  

  /**
   *  uniqˤʤ褦
   * x*32+yƱΤκ
   * äȤƤɤ
   */
  enum Offset32{
    OFFSET32_MIN= -(8*32+8),
    OFFSET32_MAX= (8*32+8),
  };
  const size_t OFFSET32_SIZE=(OFFSET32_MAX-OFFSET32_MIN+1);

  inline Offset32 positionDiff(Position pos0, Position pos1){
    assert( isOnBoard(pos0) && isOnBoard(pos1) );
    int ret=pos0-pos1+(pos0&0xf0)-(pos1&0xf0);
    assert(OFFSET32_MIN<=ret && ret<=OFFSET32_MAX);
    return static_cast<Offset32>(ret);
  }
  inline Offset32 newOffset32(int dx,int dy){
    assert(-9<=dx && dx<=9 && -9<=dy && dy<=9);
    return static_cast<Offset32>(dx*32+dy);
  }

  /**
   * Player P ߤ offset ֤ΤΤѹ
   */
  template<Player P>
  inline Offset32 makeBlackOffset32(Offset32 offset32);
  template<>
  inline Offset32 makeBlackOffset32<BLACK>(Offset32 offset32){ 
    return offset32;
  }
  template<>
  inline Offset32 makeBlackOffset32<WHITE>(Offset32 offset32){ 
    return static_cast<Offset32>(-static_cast<int>(offset32));
  }


  template<Player P>
  inline bool canPromotePosition(const Position pos);
  
  template<>
  inline bool canPromotePosition<BLACK>(const Position pos){
    assert(isOnBoard(pos));
    return getY(pos)<=3;
    
  }
  template<>
  inline bool canPromotePosition<WHITE>(const Position pos){
    assert(isOnBoard(pos));
    return getY(pos)>=7;
    
  }

  inline bool canPromotePosition(Player player,Position pos){
    if(player==BLACK) return canPromotePosition<BLACK>(pos);
    else return canPromotePosition<WHITE>(pos);
  }

  /**
   * Player P ߤ offset ֤ΤΤѹ
   */
  template<Player P>
  inline Offset makeBlackOffset(Offset offset);
  template<>
  inline Offset makeBlackOffset<BLACK>(Offset offset){ 
    return offset;
  }
  template<>
  inline Offset makeBlackOffset<WHITE>(Offset offset){ 
    return static_cast<Offset>(-static_cast<int>(offset));
  }


  /**
   * ޤȤäƤʤ
   */
  template<Player P>
  inline int yForBlack(int y);

  template<>
  inline int yForBlack<BLACK>(int y){return y;}

  template<>
  inline int yForBlack<WHITE>(int y){return reverseY(y);}

  
  std::ostream& operator<<(std::ostream&,const Offset);
  std::ostream& operator<<(std::ostream&,const Position);
  std::ostream& csaShow(std::ostream&,const Position);
  
  //static const int BOARD_SIZE=0x100;
}
#endif /* _BOARD_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
