/* boardBitMask.h
 */
#ifndef _BOARD_BIT_MASK_H
#define _BOARD_BIT_MASK_H
#include "osl/position.h"
#include "osl/board.h"
#include "osl/carray.h"
#include "osl/carray2d.h"
#include <iosfwd>
#define USE_XMM
namespace osl{
  /**
   * SSE2ĤʤΤޤɤˡĤʤ
   */

#ifdef USE_XMM
    typedef int v4sf __attribute__ ((mode(V4SF))); // works
#endif
  struct BoardBitMask{
#ifdef USE_XMM
    union {
      v4sf xmm;
      unsigned long long mask[2];
      unsigned char bMask[16];
    };
#else
    union {
      unsigned long long mask[2];
      unsigned char bMask[16];
    };
#endif
    friend BoardBitMask& operator&=(BoardBitMask& lhs,BoardBitMask const& rhs);
    friend BoardBitMask& operator^=(BoardBitMask& lhs,BoardBitMask const& rhs);
    friend BoardBitMask operator^(BoardBitMask& src1,BoardBitMask const& src2);
  public:
    BoardBitMask(){}
    template<class State>
    explicit BoardBitMask(State const& st){
      clearAll();
      for(int y=1;y<=9;y++){
	for(int x=1;x<=9;x++){
	  Position position=newPosition(x,y);
	  if(isEmpty(st.getPieceAt(position)))
	    setBit(positionToOffset(position));
	}
      }
    }
    /**
     * ٤ƤΥӥåȤ0ΤΤ
     */
    void clearAll(){mask[0]=mask[1]=0ull;}
    void setAll(){mask[0]=mask[1]=static_cast<unsigned long long>(-1ll);}
    /**
     * 0-8 11-19 22-30
     * 33-41 44-52 55-63
     * 66-74 77-85 88-96
     * ǤäƤߤ
     * ֤ȳѤʤȤ륢ǥ⤢äΤ
     */
    static int positionToOffset(Position pos){
      assert(isOnBoard(pos));
      int x=getX(pos);
      int y=getY(pos);
      return (x-1)*11+(y-1);
    }
    /**
     * @param offset(0-96) - bit򥻥åȤ
     */
    void setBit(int offset){
      assert(0<=offset && offset<=96);
      int index=offset>>6;
      unsigned long long tmpMask=1ull<<(offset&63);
      mask[index]|= tmpMask;
    }
    void setBit(Position pos){
      setBit(positionToOffset(pos));
    }
    /**
     * @param offset(0-96) - bit򥯥ꥢ
     */
    void clearBit(int offset){
      assert(0<=offset && offset<=96);
      int index=offset>>6;
      unsigned long long tmpMask=1ull<<(offset&63);
      mask[index]&= ~tmpMask;
    }
    void clearBit(Position pos){
      clearBit(positionToOffset(pos));
    }
    bool isZero() const{
      return (mask[0]|mask[1])==0ull;
    }
    BoardBitMask& operator=(BoardBitMask const& rhs){
#ifdef USE_XMM
      xmm=rhs.xmm;
#else
      mask[0]=rhs.mask[0];
      mask[1]=rhs.mask[1];
#endif
      return *this;
    }

  };

		      
  inline BoardBitMask& operator^=(BoardBitMask& lhs,BoardBitMask const& rhs){
#ifdef USE_XMM
    lhs.xmm=__builtin_ia32_xorps(lhs.xmm,rhs.xmm);
#else
    lhs.mask[0]^=rhs.mask[0];
    lhs.mask[1]^=rhs.mask[1];
#endif
    return lhs;
  }

  inline BoardBitMask operator^(BoardBitMask const& lhs,BoardBitMask const& rhs){
#ifdef USE_XMM
    BoardBitMask ret=lhs;
    ret.xmm=__builtin_ia32_xorps(lhs.xmm,rhs.xmm);
    return ret;
#else
    BoardBitMask ret=lhs;
    ret.mask[0]^=rhs.mask[0];
    ret.mask[1]^=rhs.mask[1];
    return ret;
#endif
  }
  std::ostream& operator<<(std::ostream& os,BoardBitMask const& boardBitMask);

  class BoardBitMaskTable{
    CArray<BoardBitMask, POSITION_SIZE> maskOfPosition;
    /**
     * lance˴ؤƤϺʤƤɤ
     * lanceBetweenMask[from][to]  non all 0 ʤ
     */
    CArray2d<BoardBitMask,POSITION_SIZE,POSITION_SIZE> rookBetweenMask;
    /**
     *
     */
    CArray2d<BoardBitMask, POSITION_SIZE,POSITION_SIZE> lanceBetweenMask;
    CArray2d<BoardBitMask, POSITION_SIZE,POSITION_SIZE> bishopBetweenMask;
  private:
    void initMaskOfPosition();
    void initBetweenMask();
  public:
    BoardBitMaskTable();
    const BoardBitMask& getMask(Position pos) const{
      assert(isOnBoard(pos));
      return maskOfPosition[pos];
    }
    const BoardBitMask& getRookMask(Position from,Position to) const{
      assert(isOnBoard(from) && isOnBoard(to));
      return rookBetweenMask[from][to];
    }
    const BoardBitMask& getBishopMask(Position from,Position to) const{
      assert(isOnBoard(from) && isOnBoard(to));
      return bishopBetweenMask[from][to];
    }
    const BoardBitMask& getLanceMask(Position from,Position to) const{
      assert(isOnBoard(from) && isOnBoard(to));
      return lanceBetweenMask[from][to];
    }
  };
  extern BoardBitMaskTable boardBitMaskTable;
}
#endif // _BOARD_BIT_MASK_H
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
