/* categoryListUtil.h
 */
#ifndef _CATEGORYLISTUTIL_H
#define _CATEGORYLISTUTIL_H
#include "osl/category/categoryConcept.h"
#include "osl/category/categoryList.h"
#include "osl/category/categoryTraits.h"
#include "osl/category/categoryEnv.h"
#include "osl/category/categoryFlags.h"
#include "osl/loki.h"

namespace osl
{
  class MoveLogProbVector;
  class MoveLogProbSet;
  class MoveStack;

  namespace category
  {

    struct CategoryListUtil
    {
    public:
      /**
       * ʣΥƥ˻ dispatcher
       * @return cut ä(⤷ϲ)ɤ
       * @param State ɸŪʥƥǤ effect_state ɬ eval, hash 
       */
      template <typename Head, typename Tail, typename Function>
      static bool forEachCategory(CategoryList<Head,Tail>, 
				  const CategoryEnv& env, Function f);
      /**
       * CategoryList κǸξ
       */
      template <typename Function>
      static bool forEachCategory(NullCategory, const CategoryEnv&, Function)
      {
	return false;
      }
      /**
       * ʣΥƥ˻ dispatcher
       * ȤäɤФƤС
       */
      template <typename Head, typename Tail, typename Function>
      static bool forEachCategory(CategoryList<Head,Tail>, 
				  const CategoryEnv& env, Function f,
				  CategoryFlags& flags, size_t cur=0);      
      template <typename Function>
      static bool forEachCategory(NullCategory, 
				  const CategoryEnv& env, Function f,
				  CategoryFlags& flags, size_t cur)
      {
	return false;
      }

      /**
       * ƤΥƥμ out 
       */
      template <typename Categories>
      static void gatherAllMoves(const CategoryEnv& env, MoveLogProbSet& out);
    private:
      /**
       * ĤΥƥμ
       * @return Υƥɬפ뤫ɤ
       */
      template <typename Category, typename Function>
      static bool useOneCategory(const CategoryEnv& env, Function f)
      {
	if (category::MinProb<Category>::minProb() >= env.limit)
	  return true;
	MoveLogProbVector moves;
	generate<Category>(env, moves);
	if (moves.size() == 0)
	  return true;
      
	const bool tryNext 
	  = f.template operator()<Category::isKingEscape>(Category::getName(), moves);
	// ⤷꤬äƤ¾Υƥ꡼ϻƤ⤷礦ʤ
	return (! Category::isKingEscape) 
	  && tryNext;
      }
      /** 
       * Generator Ȥä out ˻ؼ롣
       */
      template <class Generator>
      static void generate(const CategoryEnv&, MoveLogProbVector& out);
    };

    template <typename Head, typename Tail, typename Function>
    bool CategoryListUtil::forEachCategory(CategoryList<Head,Tail>, 
					   const CategoryEnv& env, Function f)
    {
      const bool tryNext = 
	useOneCategory<Head,Function>(env, f);
      if (! tryNext)
	return true;
      
      return forEachCategory(Tail(), env, f);
    }

    template <typename Head, typename Tail, typename Function>
    bool CategoryListUtil::forEachCategory(CategoryList<Head,Tail>, 
					   const CategoryEnv& env, Function f,
					   CategoryFlags& flags, size_t cur)
    {
      if (! flags.isSet(cur))
      {
	flags.set(cur);
	const bool tryNext = 
	  useOneCategory<Head,Function>(env, f);
	if (! tryNext)
	  return true;
      }
      return forEachCategory(Tail(), env, f, flags, cur+1);
    }
    
  } // namespace category
  using category::CategoryListUtil;
} // namespace osl

#endif /* _CATEGORYLISTUTIL_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
