/* pawn.h
 */
#ifndef _PAWN_H
#define _PAWN_H

#include "osl/category/categoryTraits.h"
#include "osl/category/categoryEnv.h"
#include "osl/moveStack.h"
#include "osl/moveVector.h"
namespace osl
{
  namespace category
  {
    /**
     * Ѥ 
     *
     * Ѥ꤬ͤʵ⤹뤬Ϣ³á⤫ͤ
     */
    struct JoiningPawns : public FixedProbabilityCategory
    {
      static FixedValueTable probTable; // { return 263; } //0.161904
      static const char *getName() { return "JoiningPawns"; }
      static bool isInCategory(const CategoryEnv& env, Move move)
      {
	const Move lastMove=env.history->lastMove();
	return (getFrom(move) == OFFBOARD) 
	  && (getPtype(move) == PAWN)
	  && (getCapturePtype(lastMove) == PAWN)
	  && (directionTable.nextPosition(getPlayer(move), getTo(move) ,U)
	      == getTo(lastMove));
      }
      static void generate(const CategoryEnv& env, MoveVector& moves)
      {
	if (! env.history->hasLastMove())
	  return;
	const Move lastMove=env.history->lastMove();
	if (getCapturePtype(lastMove) != PAWN)
	  return;
	const Player P = env.state->getTurn();
	if (! env.state->hasOffBoard(P, PAWN))
	  return;
	const Position to
	  = directionTable.nextPosition(alt(P), getTo(lastMove), U);
	if (env.state->getPieceAt(to) != PIECE_EMPTY)
	  return;
	moves.push_back(newDrop(to, PAWN, P));
      }
    };

    /**
     * ư
     */
    struct DropAfterOpposingPawn : public FixedProbabilityCategory
    {
      static FixedValueTable probTable; // { return 385; } // 0.0692099
      static const char *getName() { return "DropAfterOpposingPawn"; }
      static bool isInCategory(const CategoryEnv& env, Move move)
      {
	if (! env.history->hasLastMove()) 
	  return false;
	const Move lastMove=env.history->lastMove();
	const Player P = getPlayer(move);
	return (getFrom(move) == OFFBOARD) 
	  && (getPtype(move) == PAWN)
	  && canPromotePosition(P, directionTable.nextPosition(P,getTo(move),U))
	  && (getCapturePtype(lastMove) == PAWN)
	  && (directionTable.nextPosition(alt(P),getFrom(lastMove),U)
	      == getTo(lastMove))
	  && (getTo(move) == getFrom(lastMove));
      }
      static void generate(const CategoryEnv& env, MoveVector& out)
      {
	if (! env.history->hasLastMove()) 
	  return;
	const Move lastMove=env.history->lastMove();
	const Player P = env.state->getTurn();
	const Position to = getFrom(lastMove);
	if ((getCapturePtype(lastMove) == PAWN)
	    && (directionTable.nextPosition(alt(P),getFrom(lastMove),U)
		== getTo(lastMove))
	    && isOnBoard(directionTable.nextPosition(P, to, U))
	    && canPromotePosition(P, directionTable.nextPosition(P, to, U))
	    && env.state->hasOffBoard(P, PAWN)
	    && (env.state->getPieceOnBoard(to) == PIECE_EMPTY)
	    && (canDropTo(P, PAWN, to)))
	{
	  out.push_back(newDrop(to, PAWN, P));
	}
      }
    };

    /**
     * 
     *
     * ˶𤬤äƤⵤˤʤǽɤ餷
     */
    struct ImmediateAntiPawnPromotion : public FixedProbabilityCategory
    {
      static FixedValueTable probTable;
      static const char *getName() { return "AntiPawnPromotion!"; }
      static bool isInCategory(const CategoryEnv& env, Move move)
      {
	if (! env.history->hasLastMove()) 
	  return false;
	const Move lastMove=env.history->lastMove();
	const Player P = getPlayer(move);
	if ((getFrom(move) == OFFBOARD) 
	    && (getPtype(move) == PAWN)
	    && (getPtype(lastMove) == PAWN))
	{
	  const Position focus = directionTable.nextPosition(P,getTo(move),U);
	  return canPromotePosition(alt(P),focus)
	    && (directionTable.nextPosition(P,focus,U)
		== getTo(lastMove));
	}
	return false;
      }
      static void generate(const CategoryEnv& env, MoveVector& out)
      {
	if (! env.history->hasLastMove()) 
	  return;
	const Move lastMove=env.history->lastMove();
	const Player P = env.state->getTurn();
	if (getPtype(lastMove) == PAWN)
	{
	  const Position focus = 
	    directionTable.nextPosition(alt(P),getTo(lastMove),U);
	  if (canPromotePosition(alt(P),focus))
	  {
	    const Position target =
	      directionTable.nextPosition(alt(P),focus,U);
	    if ((env.state->getPieceAt(target) == PIECE_EMPTY) // Ϥ߽Фǽ
		&& env.state->hasOffBoard(P, PAWN)
		&& (! env.state->isPawnMaskSet(P, getX(target))))
	    {
	      out.push_back(newDrop(target, PAWN, P));
	    }
	  }
	}
      }
    };
  } // namespace category
} // namespace osl

#endif /* _PAWN_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
