/* probOfEachCategory.h
 */
#ifndef _PROBOFEACHCATEGORY_H
#define _PROBOFEACHCATEGORY_H

#include "osl/category/categoryEnv.h"
#include "osl/category/categoryList.h"
#include "osl/moveLogProbVector.h"
#include "osl/pieceEval.h"
#include <iosfwd>
namespace osl
{
  namespace category
  {
    template <class T> class Classify;
    template <class T> class CategoryUtil;
    /**
     * ؼФƳƥƥ꤬ɤγΨդĴ٤
     */
    class ProbOfEachCategory
    {
    public:
      struct CallBack
      {
	virtual ~CallBack() {}
	virtual void found(const char *name, int prob){}
	virtual void notFound(const char *name) {}
      };
    private:
      const CategoryEnv& env;
      Move target;
      CallBack *function;
      int pieceValue;
    public:
      ProbOfEachCategory(const CategoryEnv& e, Move t, CallBack *f)
	: env(e), target(t), function(f)
      {
	pieceValue = PieceEval::
	  computeDiffAfterMoveForRP(*env.state, t);
      }
      template <typename Head, typename Tail>
      void analyze(CategoryList<Head, Tail>);
    private:
      template <typename Head>
      void analyze(Head)
      {
	MoveLogProbVector moves;
	CategoryUtil<Head>::generate(env,moves);
	for (size_t i=0; i<moves.size(); ++i)
	{
	  if (moves[i].getMove() == target)
	  {
	    function->found(Head::getName(), moves[i].getLogProb());
	    return;
	  }
	}
	function->notFound(Head::getName());
      }
      void analyze(NullCategory)
      {
      }
      template <typename SubCategories>
      void analyze(Classify<SubCategories>)
      {
	analyzeClassify(SubCategories());
      }
      template <typename Head, typename Tail>
      void analyzeClassify(CategoryList<Head,Tail>)
      {
	const std::pair<bool,int> r = 
	  CategoryUtil<Head>::getProbIf(env,target,pieceValue);
	if (r.first)
	  function->found(Head::getName(), r.second);
	else
	  function->notFound(Head::getName());
	analyzeClassify(Tail());
      }
      void analyzeClassify(NullCategory)
      {
      }
    public:
      struct Reporter : public ProbOfEachCategory::CallBack
      {
	std::ostream& os;
	explicit Reporter(std::ostream&);
	void found(const char *name, int prob);
	void notFound(const char *name);
      };
    };

    template <typename Head, typename Tail>
    void ProbOfEachCategory::analyze(CategoryList<Head, Tail>)
    {
      analyze(Head());
      analyze(Tail());
    }
  } // namespace category
} // namespace osl

#endif /* _PROBOFEACHCATEGORY_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
