/* silver.h
 */
#ifndef _SILVER_H
#define _SILVER_H

#include "osl/category/categoryTraits.h"
#include "osl/category/categoryEnv.h"
#include "osl/moveVector.h"
#include "osl/moveStack.h"
namespace osl
{
  namespace category
  {
    /**
     * Ǥ
     */
    struct StabInTheBack : public FixedProbabilityCategory
    {
      static FixedValueTable probTable; // { return 274; } // 0.149371
      static const char *getName() { return "StabInTheBack"; }
      static bool isInCategory(const CategoryEnv& env, Move move)
      {
	const CategoryEnv::effect_state_t& state = *env.state;
	const Player P = getPlayer(move);
	const Piece backL = 
	  state.getPieceAt(directionTable.nextPosition(P, getTo(move), DL));
	const Piece backR = 
	  state.getPieceAt(directionTable.nextPosition(P, getTo(move), DR));
	return (getPtype(move) == SILVER)
	  && (getFrom(move) == OFFBOARD)
	  && isPiece(backR)
	  && (getOwner(backR) == alt(P))
	  && isPiece(backL)
	  && (getOwner(backL) == alt(P))
	  && ((getPtype(backL) == GOLD)
	      || (getPtype(backL) == ROOK))
	  && ((getPtype(backR) == GOLD)
	      || (getPtype(backR) == ROOK))
	  && (! state.hasEffectBy(alt(P), getTo(move)));
      }
    };

    /**
     * ζ
     */
    struct SilverHeadOfKnight : public FixedProbabilityCategory
    {
      static FixedValueTable probTable; 
      static const char *getName() { return "SilverHeadOfKnight"; }

      static bool isInCategory(const CategoryEnv& env, Move move)
      {
	if (! ((getPtype(move) == SILVER)
	       && (getFrom(move) == OFFBOARD)))
	  return false;
	const Player P = getPlayer(move);
	const Position to = getTo(move);
	const Piece front = 
	  env.state->getPieceAt(directionTable.nextPosition(P, to, U));
	return isPiece(front)
	  && (getOwner(front) == alt(P))
	  && (getPtype(front) == KNIGHT);
      }
    };

    /**
     * ľηζ
     */
    struct SilverHeadOfLastKnight : public FixedProbabilityCategory
    {
      static FixedValueTable probTable;
      static const char *getName() { return "SilverHeadOfKnight!"; }

      static bool isEffective(const CategoryEnv& env)
      {
	return env.history->hasLastMove();	
      }
      static bool isInCategory(const CategoryEnv& env, Move move)
      {
	if (! isEffective(env))
	  return false;
	const Player P = getPlayer(move);
	const Move lastMove=env.history->lastMove();
	const Position to = getTo(move);
	return (getTo(lastMove) == directionTable.nextPosition(P, to, U))
	  && SilverHeadOfKnight::isInCategory(env, move);
      }
      static void generate(const CategoryEnv& env, MoveVector& moves)
      {
	if (! isEffective(env))
	  return;
	const Move lastMove=env.history->lastMove();
	if (getPtype(lastMove) != KNIGHT)
	  return;
	const Player P = env.state->getTurn();
	if (! env.state->hasOffBoard(P, SILVER))
	  return;
	const Position to
	  = directionTable.nextPosition(alt(P), getTo(lastMove), U);
	if (env.state->getPieceOnBoard(to) != PIECE_EMPTY)
	  return;
	moves.push_back(newDrop(to, SILVER, P));
      }
    };
  } // namespace category
} // namespace osl


#endif /* _SILVER_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
