/* checkMateRecorder.h
 */
#ifndef _CHECKMATERECORDER_H
#define _CHECKMATERECORDER_H

#include "osl/move.h"
#include <iosfwd>

namespace osl
{
  class PawnMaskState;
  namespace checkmate
  {
    class CheckHashRecord;
    class CheckMove;
    class CheckMateRecorder
    {
    public:
      static CheckMove *nextMove;
      static const char *leaveReason;

      static std::ostream& stream();
      static void dumpRecord(const PawnMaskState& state,
			     const CheckHashRecord *record, 
			     unsigned int proofLimit,
			     unsigned int disProofLimit,
			     unsigned int currentProofNumber,
			     unsigned int currentDisProofNumber);

      static void stat(const char *msg, Player P, unsigned int tableSize,
		       unsigned int totalNodeCount, unsigned int totalNodeLimit);
      /** root ǤξϿ */
      static void rootLog(const char *msg, unsigned int tableSize,
			  unsigned int continuousNoExpandLoop)
      {
#ifdef CHECKMATE_DEBUG
	writeRootLog(msg, tableSize, continuousNoExpandLoop);
#endif
      }
      static void writeRootLog(const char *msg, unsigned int tableSize,
			       unsigned int continuousNoExpandLoop);
      /** attack/defense Ͽ */
      static void enter(int depth, const char *name, 
			const CheckHashRecord *record, 
			unsigned int proofLimit, unsigned int disProofLimit);
      /** attack/defense 鵢Ͽ */
      static void leave(int depth, const char *name, 
			const CheckHashRecord *record);
      /** CheckHashRecord ݤ˵Ͽ */
      static void allocate(int depth, const CheckHashRecord *record);
      static void setNextMove(CheckMove *m)
      {
#ifdef CHECKMATE_DEBUG
	nextMove = m;
#endif
      }
      static void setLeaveReason(const char *reason)
      {
#ifdef CHECKMATE_DEBUG
	leaveReason = reason;
#endif
      }
      struct DepthTracer
      {
	static unsigned int depth;
	static unsigned int maxLogDepth;
	CheckHashRecord *& record;
	const char *name;
	DepthTracer(const char *n, 
		    CheckHashRecord *& r,
		    unsigned int proofLimit,
		    unsigned int disProofLimit)
	  : record(r), name(n)
	{
	  ++depth;
	  if (depth < maxLogDepth)
	    enter(depth, name, record, proofLimit, disProofLimit);
	}
	~DepthTracer()
	{
	  if (depth < maxLogDepth)
	  {
	    leave(depth, name, record);
	  }
	  depth--;
	}
	void allocate()
	{
	  if (depth < maxLogDepth)
	    CheckMateRecorder::allocate(depth, record);
	}
      };
      struct NullTracer
      {
	NullTracer(const char *n, 
		   const CheckHashRecord * r,
		   unsigned int proofLimit,
		   unsigned int disProofLimit)
	{
	}
	~NullTracer()
	{
	}
	void allocate()
	{
	}
      };
#ifdef CHECKMATE_DEBUG
      typedef DepthTracer Tracer;
#else
      typedef NullTracer Tracer;
#endif
    };  
  } // namespace checkmate
} // namespace osl

#endif /* _CHECKMATERECORDER_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
