/* checkMoveGenerator.tcc
 */
#ifndef _CHECKMOVEGENERATOR_TCC
#define _CHECKMOVEGENERATOR_TCC

#include "osl/checkmate/checkMoveGenerator.h"
#include "osl/checkmate/checkMoveList.h"
#include "osl/move_generator/escape.h"
#include "osl/move_action/store.h"
#include "osl/moveVector.h"

// #ifdef DELAY_INTERPOSE
// #  define DELAY_INTERPOSE_EXTENDED
// #endif

namespace osl
{
  namespace checkmate
  {
    template <Player P>
    template <class State>
    size_t CheckMoveGenerator<P>::generateEscape(const State& state, 
						 CheckMoveList& out, 
						 CheckMoveList& interposeOut,
						 Move lastMove)
    {
      MoveVector moves;
      {
	typedef typename State::effect_state_t effect_state_t;
	move_action::Store store(moves);
	move_generator::Escape<PlayerTraits<P>::opponent,effect_state_t,move_action::Store>::
	  generateKingEscape(state, lastMove,store);
      }
      assert(out.empty());
      out.reserve(moves.size());
#ifdef DELAY_INTERPOSE_EXTENDED
      // ̤ϼ֤ (ʤΤ礤Ʊ餤Ȳ)
      const Position kingPosition 
	= state.template getKingPosition<PlayerTraits<P>::opponent>();
      const Position lastTo = getTo(lastMove);
      bool isLastMoveCaptureByRookBishop = false;
      if (lastMove 
	  && (! state.hasEffectBy(P, lastTo))
	  && (getCapturePtype(lastMove) != PTYPE_EMPTY)
	  &&
	  (state.hasEffectNotBy(alt(P), state.getPieceOnBoard(kingPosition), lastTo)
#if 0 
	   // ͤǤϸŪͤޤʤǤϤޤ -> Ȥн褹٤
	   || (state.hasEffectBy(alt(P), lastTo)
	       && (unpromote(getCapturePtype(lastMove)) != PAWN)
	       && (unpromote(getCapturePtype(lastMove)) != LANCE)
	       && (unpromote(getCapturePtype(lastMove)) != KNIGHT)
	     )
#endif
	    ))
      {
	const Ptype ptype = unpromote(getPtype(lastMove));
	if ((ptype == ROOK) || (ptype == BISHOP))
	  isLastMoveCaptureByRookBishop = true;
      }
#endif
      for (size_t i=0; i<moves.size(); ++i)
      {
	const Move m = moves[i];
#ifdef DELAY_INTERPOSE
	const Position to = getTo(m);
#  ifdef DELAY_INTERPOSE_EXTENDED
	if (isLastMoveCaptureByRookBishop
	    && (to != lastTo))
	{
	  interposeOut.push_back(CheckMove(m));
	  continue;
	}
#  endif
	const Position from = getFrom(m);
	const Ptype ptype = getPtype(m);
	if (isOffBoard(from))
	{
	  // drop 
	  if ((ptype != PAWN)
	      && state.hasEffectBy(P, to)
	      && (! state.hasEffectBy(alt(P), to)))
	  {
	    interposeOut.push_back(CheckMove(m));
	    continue;
	  }
	}
	else if ((ptype != PAWN)
		 && (getCapturePtype(m) == PTYPE_EMPTY)
		 && state.hasEffectBy(P, to)
		 && (! state.hasEffectNotBy(alt(P), state.getPieceOnBoard(from), to)))
	{
	  // ư
	  interposeOut.push_back(CheckMove(m));
	  continue;
	}
#endif
	out.push_back(CheckMove(m));
      }
      if (out.empty())
	out.swap(interposeOut);
      return out.size();
    }
  } // namespace checkmate
} // namespace osl

#endif /* _CHECKMOVEGENERATOR_TCC */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
