/* dualCheckMateSearcher.h
 */
#ifndef _DUALCHECKMATESEARCHER_H
#define _DUALCHECKMATESEARCHER_H

#include "osl/checkmate/checkmateSearcher.h"
#include "osl/checkmate/checkHashTable.h"
#include "osl/checkmate/oraclePool.h"
#include "osl/checkmate/oracleProver.h"
#include "osl/carray.h"
#include <cassert>

/** ORACLE ǽ˻ */
#define USE_ORACLE

namespace osl
{
  namespace checkmate
  {
    /**
     * ξ Player εͤߤ򰷤;
     */
    template <class State>
    class DualCheckMateSearcher
    {
    public:
      typedef CheckmateSearcher<BLACK,State> black_checkmate_t;
      typedef CheckmateSearcher<WHITE,State> white_checkmate_t;
    private:
      CheckHashTable tableBlack, tableWhite;
      black_checkmate_t checkmateBlack;
      white_checkmate_t checkmateWhite;
      OraclePool blackOracles, blackOraclesAfterAttack, 
	whiteOracles, whiteOraclesAfterAttack;
      // statistics
      unsigned int proofByOracle, unknownByOracle;
      unsigned int proofBySearch, unknownBySearch, disProofBySearch;
    public:
      explicit DualCheckMateSearcher(size_t totalNodeCount=CHECKMATE_DEFAULT_TOTAL_NODE_LIMIT, bool verboseDestructor=true);
      ~DualCheckMateSearcher();
    
      static size_t limitToCheckCount(size_t limit){
	extern CArray<size_t,32> limitToCheckCountTable;
	return limitToCheckCountTable[limit>>7];
      }

    private:
      template <Player P, class Searcher>
      bool isWinningState(State& state, int nodeLimit, Move& bestMove,
			  Searcher& searcher, 
			  CheckHashTable& table, OraclePool& oracles)
      {
#ifdef USE_ORACLE
	ProofOracleAttack oracle(oracles.findOracle(state));
	if (oracle.isValid())
	{
	  OracleProver<P,State> prover(table);
	  if (prover.proofWin(state, oracle, bestMove))
	  {
	    ++proofByOracle;
	    assert(bestMove);
	    isValid(bestMove);
	    return true;
	  }
	  ++unknownByOracle;
	}
#endif
	const ProofDisProof proofDisProof
	  = searcher.hasCheckmateMove(state, nodeLimit, bestMove);
	if (isCheckmateSuccess(proofDisProof))
	{
	  assert(bestMove);
	  isValid(bestMove);
	  ++proofBySearch;
	  const CheckHashRecord *record = table.find(state.getHash());
	  assert(record);
	  oracles.addOracle(state, record);
	  return true;
	}
	if (isCheckmateFail(proofDisProof))
	  ++disProofBySearch;
	else
	  ++unknownBySearch;
	return false;
      }
    public:
      /**
       * @return ̤ͤ
       */
      template <Player P>
      bool isWinningState(int nodeLimit, State& state, Move& bestMove)
      {
	assert(state.getTurn() == P);
	if (P == BLACK)
	{
	  return isWinningState<BLACK,black_checkmate_t>
	    (state, nodeLimit, bestMove, 
	     checkmateBlack, tableBlack, blackOracles);
	}
	else
	{
	  return isWinningState<WHITE,white_checkmate_t>
	    (state, nodeLimit, bestMove, 
	     checkmateWhite, tableWhite, whiteOracles);
	}
      }
    private:
      /** @param P  */
      template <Player P, class Searcher>
      bool isLosingState(State& state, int nodeLimit,
			 Searcher& searcher, 
			 CheckHashTable& table, OraclePool& oracles)
      {
#ifdef USE_ORACLE
	ProofOracleDefense oracle(oracles.findOracle(state));
	if (oracle.isValid())
	{
	  OracleProver<PlayerTraits<P>::opponent,State> prover(table);
	  if (prover.proofLose(state, oracle))
	  {
	    ++proofByOracle;
	    return true;
	  }
	  ++unknownByOracle;
	}
#endif
	const ProofDisProof proofDisProof
	  = searcher.hasEscapeMove(state, nodeLimit, MOVE_INVALID);
	if (isCheckmateSuccess(proofDisProof))
	{
	  ++proofBySearch;
	  const CheckHashRecord *record = table.find(state.getHash());
	  assert(record);
	  oracles.addOracle(state, record);
	  return true;
	}
	++unknownBySearch;
	return false;
      }
    public:
      /**
       * @return ̤ͤ
       * : ̤˲꤬äƤ뤳
       */
      template <Player P>
      bool isLosingState(int nodeLimit, State& state)
      {
	assert(state.getTurn() == P);
	if (P == BLACK)
	{
	  return isLosingState<P,white_checkmate_t>
	    (state, nodeLimit, checkmateWhite, tableWhite, 
	     whiteOraclesAfterAttack);
	}
	else
	{
	  return isLosingState<P,black_checkmate_t>
	    (state, nodeLimit, checkmateBlack, tableBlack, 
	     blackOraclesAfterAttack);
	}
      }

      const black_checkmate_t& black() const { return checkmateBlack; }
      const white_checkmate_t& white() const { return checkmateWhite; }

      CheckHashTable& getTableBlack()  { return tableBlack; }
      CheckHashTable& getTableWhite()  { return tableWhite; }
      const CheckHashTable& getTableBlack() const { return tableBlack; }
      const CheckHashTable& getTableWhite() const { return tableWhite; }
      const CheckHashTable& getTable(Player P) const 
      {
	if (P == BLACK)
	  return tableBlack; 
	else
	  return tableWhite; 
      }
    };

  } // namespace checkmate
  using checkmate::DualCheckMateSearcher;
} // namespace osl

#endif /* _DUALCHECKMATESEARCHER_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
