/* compositEval.h
 */
#ifndef _COMPOSITEVAL_H
#define _COMPOSITEVAL_H

#include "osl/pieceEval.h"
#include "osl/piecePairEval.h"
#include "osl/progress.h"
namespace osl
{
  class CompositEval
  {
    PieceEval pieceEv;
    PiecePairEval pairEv;
    Progress progress;
  public:
    template<class State>
    CompositEval(State const& state);

    static const int pieceBias = 1;
    static const int progressMax = 300;
    static int composeValue(int val, int valPair, int progress)
    {
      // minusInfty η׻˻ȤΤΧˤ max Ȥ뤫
      // Ф˱ۤʤͤ progressMax ˻ꤹɬפ롥
      const int result = val*pieceBias
	+ valPair 
#if 0
	*std::max(progress, progressMax)
#endif
	;
      assert(result % 2 == 0);
      return result;
    }
    int getProgress() const { return progress.getVal(); }
    int getVal() const
    {
      return composeValue(pieceEv.getVal(), pairEv.getVal(), 
			  getProgress()); 
    }
    /** 1β MTDF Ȥ */
    int pawnValue() const
    {
      return composeValue(pieceEv.pawnValue(), 0, 0);
    }

    /**
     * SOMA  pieceEval Ǥ
     */
    template<class State,Player P>
    int computeValWithMove(const State& state,Move move) const
    {
      const int valEv = pieceEv.computeValWithMove<State,P>(state,move);
      return composeValue(valEv, 
			  pairEv.computeValWithMove<State,P>(state,move),
			  getProgress());
    }
    template<class State,Player P>
    int computeValWithTakeBack(const State& state, Position focus) const
    {
      const int valEv = pieceEv.computeValWithTakeBack<State,P>(state,focus);
      return composeValue(valEv, 
			  pairEv.computeValWithTakeBack<State,P>(state,focus),
			  getProgress());
    }

    template <class State>
    int diffWithMove(const State& state, Move move) const
    {
      assert(Progress::getProgress(state) == getProgress());
      return composeValue(PieceEval::diffWithMove(state, move),
			  PiecePairEval::diffWithMove(state, move),
			  getProgress());
    }

    template<class State> 
    void doSimpleMove(State const& state,Position from, Position to, 
		      int promoteMask)
    {
      pieceEv.doSimpleMove(state,from,to,promoteMask);
      pairEv.doSimpleMove(state,from,to,promoteMask);
      progress.doSimpleMove(state,from,to,promoteMask);
    }
    template<class State> 
    void doDropMove(State const& state,Position to,Ptype ptype)
    {
      pieceEv.doDropMove(state,to,ptype);
      pairEv.doDropMove(state,to,ptype);
      progress.doDropMove(state,to,ptype);
    }
    template<class State> 
    void doCaptureMove(State const& state,Position from, Position to, 
		       Piece victim,int promoteMask)
    {
      pieceEv.doCaptureMove(state,from,to,victim,promoteMask);
      pairEv.doCaptureMove(state,from,to,victim,promoteMask);
      progress.doCaptureMove(state,from,to,victim,promoteMask);
    }
  public:
    template <Player P>
    static int minusInfty()
    {
      return composeValue(PieceEval::minusInfty<P>(),
			  PiecePairEval::minusInfty<P>(),
			  progressMax);
    }
    static bool isEvalValueConsistent(int v)
    {
      return ((v % 2) == 0)
	&& EvalTraits<BLACK>::betterThan(v, minusInfty<BLACK>())
	&& EvalTraits<WHITE>::betterThan(v, minusInfty<WHITE>());
    }

    static int getCaptureVal(PtypeO ptypeO)
    {
      return composeValue(PieceEval::getCaptureVal(ptypeO), 0, progressMax);
    }
  };

  template<class State>
  CompositEval::CompositEval(State const& state) 
    : pieceEv(state), pairEv(state), progress(state)
  {
  }
};

#endif /* _COMPOSITEVAL_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
