#ifndef _DIRECTION_H
#define _DIRECTION_H
#include "player.h"
#include "board.h"
#include "carray.h"
#include <iosfwd>

namespace osl{
  enum Direction{
    UL=0,
    U=1,
    UR=2,
    L=3,
    R=4,
    DL=5,
    D=6,
    DR=7,
    UUL=8,
    UUR=9,
    LONG_UL=10,
    LONG_U=11,
    LONG_UR=12,
    LONG_L=13,
    LONG_R=14,
    LONG_DL=15,
    LONG_D=16,
    LONG_DR=17,
    DIRECTION_MIN=0,
    SHORT_DIRECTION_MAX=9,
    SHORT_DIRECTION_SIZE=10,
    DIRECTION_MAX=17,
    DIRECTION_SIZE=18
  };
  
  static bool inline isShort(Direction d){
    return d<LONG_UL;
  }

  static bool inline isLong(Direction d){
    return d>=LONG_UL;
  }
  
  static inline Direction longToShort(Direction d){
    assert(isLong(d));
    return static_cast<Direction>(static_cast<int>(d)-LONG_UL);
  }
  
  /**
   *  longDirͿƤϤʤ
   */
  static inline Direction shortToLong(Direction d){
    assert(isShort(d));
    return static_cast<Direction>(static_cast<int>(d)+LONG_UL);
  }

#define DIR_TO_MASK(dir) (1<<static_cast<int>(dir))

  static inline int dirToMask(Direction dir){
    return DIR_TO_MASK(dir);
  }

  template<Direction Dir>
    struct DirectionTraitsGen;
  
  template<>
    struct DirectionTraitsGen<UL>{
      //      static const Offset blackOffset=newOffset(1,-1);
      static const int blackDx=1;
      static const int blackDy=-1;
      static const bool canPromoteTo=true;
      static const Direction altDir=DR;
      static const Direction longDir=LONG_UL;
      static const Direction primDir=UL;
    };
  
  template<>
    struct DirectionTraitsGen<U>{
      static const int blackDx=0;
      static const int blackDy=-1;
      static const bool canPromoteTo=true;
      static const Direction altDir=D;
      static const Direction longDir=LONG_U;
      static const Direction primDir=U;
    };
  
  template<>
    struct DirectionTraitsGen<UR>{
      static const int blackDx=-1;
      static const int blackDy=-1;
      static const bool canPromoteTo=true;
      static const Direction altDir=DL;
      static const Direction longDir=LONG_UR;
      static const Direction primDir=UR;
    };
  template<>
    struct DirectionTraitsGen<L>{
      static const int blackDx=1;
      static const int blackDy=0;
      static const bool canPromoteTo=false;
      static const Direction altDir=R;
      static const Direction longDir=LONG_L;
      static const Direction primDir=L;
    };
  template<>
    struct DirectionTraitsGen<R>{
      static const int blackDx=-1;
      static const int blackDy=0;
      static const bool canPromoteTo=false;
      static const Direction altDir=L;
      static const Direction longDir=LONG_R;
      static const Direction primDir=L;
    };
  template<>
    struct DirectionTraitsGen<DL>{
      static const int blackDx=1;
      static const int blackDy=1;
      static const bool canPromoteTo=false;
      static const Direction altDir=UR;
      static const Direction longDir=LONG_DL;
      static const Direction primDir=UR;
    };
  template<>
    struct DirectionTraitsGen<D>{
      static const int blackDx=0;
      static const int blackDy=1;
      static const bool canPromoteTo=false;
      static const Direction altDir=U;
      static const Direction longDir=LONG_D;
      static const Direction primDir=U;
    };
  template<>
    struct DirectionTraitsGen<DR>{
      static const int blackDx=-1;
      static const int blackDy=1;
      static const bool canPromoteTo=false;
      static const Direction altDir=UL;
      static const Direction longDir=LONG_DR;
      static const Direction primDir=UL;
    };
  template<>
    struct DirectionTraitsGen<UUL>{
      static const int blackDx=1;
      static const int blackDy=-2;
      static const bool canPromoteTo=true;
      // no meaning
      static const Direction altDir=UUL;
      static const Direction longDir=UUL;
      static const Direction primDir=UUL;
    };
  template<>
    struct DirectionTraitsGen<UUR>{
      static const int blackDx=-1;
      static const int blackDy=-2;
      static const bool canPromoteTo=true;
      // no meaning
      static const Direction altDir=UUR;
      static const Direction longDir=UUR;
      static const Direction primDir=UUR;
    };
  
  template<>
    struct DirectionTraitsGen<LONG_UL>{
      //      static const Offset blackOffset=newOffset(1,-1);
      static const bool canPromoteTo=true;
      static const Direction altDir=LONG_DR;
      static const Direction longDir=LONG_UL;
      static const Direction primDir=UL;
      static const int blackDx=DirectionTraitsGen<UL>::blackDx;
      static const int blackDy=DirectionTraitsGen<UL>::blackDy;
    };
  
  template<>
    struct DirectionTraitsGen<LONG_U>{
      static const bool canPromoteTo=true;
      static const Direction altDir=LONG_D;
      static const Direction longDir=LONG_U;
      static const Direction primDir=U;
      static const int blackDx=DirectionTraitsGen<U>::blackDx;
      static const int blackDy=DirectionTraitsGen<U>::blackDy;
    };
  
  template<>
    struct DirectionTraitsGen<LONG_UR>{
      static const bool canPromoteTo=true;
      static const Direction altDir=LONG_DL;
      static const Direction longDir=LONG_UR;
      static const Direction primDir=UR;
      static const int blackDx=DirectionTraitsGen<UR>::blackDx;
      static const int blackDy=DirectionTraitsGen<UR>::blackDy;
    };
  template<>
    struct DirectionTraitsGen<LONG_L>{
      static const bool canPromoteTo=false;
      static const Direction altDir=LONG_R;
      static const Direction longDir=LONG_L;
      static const Direction primDir=L;
      static const int blackDx=DirectionTraitsGen<L>::blackDx;
      static const int blackDy=DirectionTraitsGen<L>::blackDy;
    };
  template<>
    struct DirectionTraitsGen<LONG_R>{
      static const bool canPromoteTo=false;
      static const Direction altDir=LONG_L;
      static const Direction longDir=LONG_R;
      static const Direction primDir=L;
      static const int blackDx=DirectionTraitsGen<R>::blackDx;
      static const int blackDy=DirectionTraitsGen<R>::blackDy;
    };
  template<>
    struct DirectionTraitsGen<LONG_DL>{
      static const bool canPromoteTo=false;
      static const Direction altDir=LONG_UR;
      static const Direction longDir=LONG_DL;
      static const Direction primDir=UR;
      static const int blackDx=DirectionTraitsGen<DL>::blackDx;
      static const int blackDy=DirectionTraitsGen<DL>::blackDy;
    };
  template<>
    struct DirectionTraitsGen<LONG_D>{
      static const bool canPromoteTo=false;
      static const Direction altDir=LONG_U;
      static const Direction longDir=LONG_D;
      static const Direction primDir=U;
      static const int blackDx=DirectionTraitsGen<D>::blackDx;
      static const int blackDy=DirectionTraitsGen<D>::blackDy;
    };
  template<>
    struct DirectionTraitsGen<LONG_DR>{
      static const bool canPromoteTo=false;
      static const Direction altDir=LONG_UL;
      static const Direction longDir=LONG_DR;
      static const Direction primDir=UL;
      static const int blackDx=DirectionTraitsGen<DR>::blackDx;
      static const int blackDy=DirectionTraitsGen<DR>::blackDy;
    };
  template<Direction Dir>
  struct DirectionTraits{
    // ؿˤȲ
    static const int mask=1<<static_cast<int>(Dir);
    static const bool isLong=(static_cast<int>(Dir) >= LONG_UL);
    static const int blackDx=DirectionTraitsGen<Dir>::blackDx;
    static const int blackDy=DirectionTraitsGen<Dir>::blackDy;
    static const Offset blackOffset=NEW_OFFSET(blackDx,blackDy);
      static const bool canPromoteTo=DirectionTraitsGen<Dir>::canPromoteTo;
      static const Direction longDir=DirectionTraitsGen<Dir>::longDir;
    };
  
  template<Direction Dir,Player P>
    struct DirectionPlayerTraits;
  
  template<Direction Dir>
    struct DirectionPlayerTraits<Dir,BLACK>{
      static const Offset offset=DirectionTraits<Dir>::blackOffset;
      static const Direction directionByBlack=Dir;
    };
  template<Direction Dir> 
  const Offset DirectionPlayerTraits<Dir,BLACK>::offset;
  template<Direction Dir> 
  const Direction DirectionPlayerTraits<Dir,BLACK>::directionByBlack;
  
  template<Direction Dir>
    struct DirectionPlayerTraits<Dir,WHITE>{
      static const Offset offset= static_cast<Offset>(-static_cast<int>(DirectionTraits<Dir>::blackOffset));
      static const Direction directionByBlack=DirectionTraitsGen<Dir>::altDir;
    };
  template<Direction Dir> 
  const Offset DirectionPlayerTraits<Dir,WHITE>::offset;
  template<Direction Dir> 
  const Direction DirectionPlayerTraits<Dir,WHITE>::directionByBlack;
  
  
  class DirectionTable
  {
  public:
    CArray<Direction,OFFSET32_SIZE> directions;
    CArray<Offset, OFFSET32_SIZE> shortOffsets;
    CArray<Offset, OFFSET32_SIZE> shortOffsetsNotKnight;
    static const CArray<Offset, DIRECTION_SIZE> offsets;
    static const CArray<int, DIRECTION_SIZE> dxs;
    static const CArray<int, DIRECTION_SIZE> dys;
    template<Direction Dir>
    void setDirections();
    template<Direction Dir>
    void setKnightDirections();
    void init();
    /**
     * ˤȤäƤoffset֤
     */
    Offset getOffsetForBlack(Direction dir) const{
      return offsets[static_cast<int>(dir)];
    }
    int getDxForBlack(Direction dir) const{
      return dxs[static_cast<int>(dir)];
    }
    int getDyForBlack(Direction dir) const{
      return dys[static_cast<int>(dir)];
    }
    template<Player P>
    Offset getOffset(Direction dir) const{
      return static_cast<Offset>
	(static_cast<int>(getOffsetForBlack(dir))*playerToMul(P));
    }
    Offset getOffset(Player pl,Direction dir) const{
      if(pl==BLACK)
	return DirectionTable::template getOffset<BLACK>(dir);
      else
	return DirectionTable::template getOffset<WHITE>(dir);
    }

    /**
     * next position from pos for player P
     *  isOnBoard Ȥϸ¤ʤ
     */
    Position nextPosition(Player P, Position pos, Direction dr) const
    {
      assert(! isOffBoard(pos));
      const Offset offset = getOffset(P, dr);
      return pos + offset;
    }

    DirectionTable();
    /** @param P ɤPlayerˤȤäƤ */
    template <Player P>
    Direction getLongDirection(Offset32 offset32) const
    {
      assert(OFFSET32_MIN <=offset32 && offset32 <= OFFSET32_MAX);
      const Offset32 blackOffset32 = makeBlackOffset32<P>(offset32);
      return directions[blackOffset32-OFFSET32_MIN];
    }
    /**
     * LongβǽΤoffsetξ, ȿ˻Ȥ offset
     * ShortoffsetξϤ켫Ȥ֤
     * βǽΤʤ0֤
     */
    Offset getShortOffset(Offset32 offset32) const{
      assert(OFFSET32_MIN <=offset32 && offset32 <= OFFSET32_MAX);
      return shortOffsets[offset32-OFFSET32_MIN];
    }
    /**
     * LongβǽΤoffsetξ, ȿ˻Ȥ offset
     * KnightʳShortoffsetξϤ켫Ȥ֤
     * Knight, βǽΤʤ0֤
     */
    Offset getShortOffsetNotKnight(Offset32 offset32) const{
      assert(OFFSET32_MIN <=offset32 && offset32 <= OFFSET32_MAX);
      return shortOffsetsNotKnight[offset32-OFFSET32_MIN];
    }
  };

  extern DirectionTable directionTable;
  
  std::ostream& operator<<(std::ostream& os,const Direction d);
}
#endif /* _DIRECTION_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
