/* doUndoMoveLock.h
 */
#ifndef _DOUNDOMOVELOCK_H
#define _DOUNDOMOVELOCK_H

#include "osl/move.h"
#include <boost/scoped_ptr.hpp>
namespace osl
{
  class PawnMaskState;

  // Implementations are provided for evrey States.
  // It is important not to provide default implementations,
  // so as to detect missing implementation in compilation error
  template <class State>  struct DoUndoMoveLockSimple;
  template <class State>  struct DoUndoMoveLockCapture;
  template <class State>  struct DoUndoMoveLockDrop;

  struct DoUndoMoveLockCleanUp
  {
    virtual ~DoUndoMoveLockCleanUp();
  };
  /**
   * slow but useful lock of doUndoMove.
   * {
   *    DoUndoMoveLock lock(state, move); // doMove(move) here
   *    ...
   *    // undoMove here at the end of the scope
   *  }
   * use ApplyMove::doUndoMove for game playing.
   */
  class DoUndoMoveLock
  {
    PawnMaskState& state;
    boost::scoped_ptr<DoUndoMoveLockCleanUp> cleaner;
  public:
    template<class State>
    DoUndoMoveLock(State& s, Move move);
    ~DoUndoMoveLock();
  };
  
} // namespace osl

#endif /* _DOUNDOMOVELOCK_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
