/* doUndoMoveLock.tcc
 */
#ifndef _DOUNDOMOVELOCK_TCC
#define _DOUNDOMOVELOCK_TCC

#include "osl/doUndoMoveLock.h"
#include "osl/piece.h"
#include "osl/position.h"
#include "osl/ptype.h"
#include "osl/player.h"

template <class State>
osl::DoUndoMoveLock::DoUndoMoveLock(State& s, osl::Move move) : state(s)
{
  assert(state.getTurn() == getPlayer(move));
  assert(move && isValid(move));
  const Position from=getFrom(move);
  const Position to=getTo(move);
  if (isOffBoard(from))
  {
    assert(state.getPieceAt(to) == PIECE_EMPTY);
    cleaner.reset(new DoUndoMoveLockDrop<State>(s,to,getPtype(move)));
  }
  else
  {
    assert(state.getPieceAt(from) != PIECE_EMPTY);
    const Piece captured=state.getPieceAt(to);
    if (captured != PIECE_EMPTY)
    {
      cleaner.reset(new DoUndoMoveLockCapture<State>
		      (s,from,to,captured,getPromoteMask(move)));
    }
    else
    {
      cleaner.reset(new DoUndoMoveLockSimple<State>
		      (s,from,to,getPromoteMask(move)));
    }
  }
  state.changeTurn();
}

#endif /* _DOUNDOMOVELOCK_TCC */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
