#ifndef _EFFECTUTIL_TCC
#define _EFFECTUTIL_TCC

#include "osl/effectUtil.h"
#include "osl/effect_action/selectCheapest.h"
#include "osl/effect_action/selectPromotable.h"
#include "osl/effect_action/showEffect.h"
#include <boost/static_assert.hpp>

template <class EffectState>
osl::Piece osl::EffectUtil::
cheapestPiece(Player P, const EffectState& state, Position pos)
{
  assert(isOnBoard(pos));
  effect_action::SelectCheapest selector;
  forEachEffect(P, state, pos, selector);
  return selector.result;
}

template <class EffectState>
osl::Piece osl::EffectUtil::
promotablePiece(Player P, const EffectState& state, 
		Position effectTarget)
{
  assert(isOnBoard(effectTarget));
  effect_action::SelectPromotable selector;
  forEachEffect(P, state, effectTarget, selector);
  return selector.result;
}

template <class EffectState>
void osl::EffectUtil::showEffect(const EffectState& state, Position target,
				 std::ostream& os)
{
  assert(isOnBoard(target));
  effect_action::ShowEffect printer(os);
  forEachEffect(BLACK, state, target, printer);
  forEachEffect(WHITE, state, target, printer);
}

namespace osl
{
  template <bool interestEmpty, Direction Dir>
  struct CheckEffectOfMove
  {
    template <class State, class Function>
    static void checkShort(Player P, const State& s, int mask, Position from, 
			   Function& f)
    {
      BOOST_STATIC_ASSERT(! DirectionTraits<Dir>::isLong);
      if (! (mask & DirectionTraits<Dir>::mask))
	return;

      const Offset offset = (P == BLACK) 
	? DirectionPlayerTraits<Dir,BLACK>::offset
	: DirectionPlayerTraits<Dir,WHITE>::offset;
      const Position target = from+offset;
      if (isOnBoard(target))
      {
	if (interestEmpty
	    || isPiece(s.getPieceOnBoard(target)))
	  f(target);
      }
    }
    template <class State, class Function>
    static void checkLong(Player P, const State& s, int mask, Position from, 
			  Function& f)
    {
      BOOST_STATIC_ASSERT(DirectionTraits<Dir>::isLong);
      if (! (mask & DirectionTraits<Dir>::mask))
	return;

      const Offset offset = (P == BLACK) 
	? DirectionPlayerTraits<Dir,BLACK>::offset
	: DirectionPlayerTraits<Dir,WHITE>::offset;

      Position toPos = from+offset;
      Piece toPiece = s.getPieceAt(toPos);
      while (isEmpty(toPiece))
      {
	if (interestEmpty)
	  f(toPos);		
	toPos = toPos+offset;
	toPiece = s.getPieceAt(toPos);
      }
      if (isPiece(toPiece))
      {
	f(toPos);
      }
    }
  };
} // namespace osl

template <class State, class Function, bool interestEmpty>
void osl::EffectUtil::
forEachEffectOfMove(const State& state, Move move, Function& f)
{
  const Ptype ptype = getPtype(move);
  const Player P = getPlayer(move);
  const Position from = getTo(move);
  const int mask = ptypeTable.getMoveMask(ptype);
  CheckEffectOfMove<interestEmpty,UL>::checkShort(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,U>::checkShort(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,UR>::checkShort(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,L>::checkShort(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,R>::checkShort(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,DL>::checkShort(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,D>::checkShort(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,DR>::checkShort(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,UUL>::checkShort(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,UUR>::checkShort(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,LONG_UL>::checkLong(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,LONG_U>::checkLong(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,LONG_UR>::checkLong(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,LONG_L>::checkLong(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,LONG_R>::checkLong(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,LONG_DL>::checkLong(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,LONG_D>::checkLong(P, state, mask, from, f);
  CheckEffectOfMove<interestEmpty,LONG_DR>::checkLong(P, state, mask, from, f);
}


#endif /* _EFFECTUTIL_TCC */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
