#ifndef _EVAL_H
#define _EVAL_H
#include "osl/piece.h"
#include "osl/move.h"
#include "osl/evalTraits.h"

namespace osl
{
  template <Player P,class State>  struct ApplyDoUndoSimpleMove;
  template <Player P,class State>  struct ApplyDoUndoCaptureMove;
  template <Player P,class State>  struct ApplyDoUndoDropMove;
  template <class State>  struct DoUndoMoveLockSimple;
  template <class State>  struct DoUndoMoveLockCapture;
  template <class State>  struct DoUndoMoveLockDrop;
  class PawnMaskState;
  /**
   * @param EvalT can be PieceEval
   */
  template<typename State,typename EvalT>
  class EvalState : public State
  {
  public:
    typedef EvalT eval_t;
  private:
    EvalT ev;

    friend class ApplyDoUndoSimpleMove<BLACK,EvalState>;
    friend class ApplyDoUndoCaptureMove<BLACK,EvalState>;
    friend class ApplyDoUndoDropMove<BLACK,EvalState>;
    friend class ApplyDoUndoSimpleMove<WHITE,EvalState>;
    friend class ApplyDoUndoCaptureMove<WHITE,EvalState>;
    friend class ApplyDoUndoDropMove<WHITE,EvalState>;
    friend class DoUndoMoveLockSimple<EvalState>;
    friend class DoUndoMoveLockCapture<EvalState>;
    friend class DoUndoMoveLockDrop<EvalState>;
  public:
    explicit EvalState(const PawnMaskState& state);
    virtual ~EvalState();
    /** դΥ᥽åɤƤѤ ApplyDoUndo{Simple,Drop,Capture}Move Ƥѹ뤳 */
    void doSimpleMove(Position from, Position to, int promoteMask)
    {
      ev.doSimpleMove(*this,from,to,promoteMask);
      State::doSimpleMove(from,to,promoteMask);
    }
    void doDropMove(Position to,Ptype ptype)
    {
      ev.doDropMove(*this,to,ptype);
      State::doDropMove(to,ptype);
    }
    void doCaptureMove(Position from, Position to, Piece target, int promoteMask)
    {
      ev.doCaptureMove(*this,from,to,target,promoteMask);
      State::doCaptureMove(from,to,target,promoteMask);
    }
    void doMove(Move move){ ApplyMoveOfTurn::doMove(*this,move); }
    bool isConsistent(bool showError=true) const;
    int getVal() const { return ev.getVal(); }
    int pawnValue() const { return ev.pawnValue(); }
    const EvalT& getEval() const { return ev; }
    /** SOMA 
     * @param P 
     */
    template <Player P>
    int computeValWithMove(Move move) {
      return ev.template computeValWithMove<EvalState,P>(*this,move);
    }
    /** SOMA
     * @param P ֤Ǥʤۤ
     */
    template <Player P>
    int computeValWithTakeBack(Position lastTo) {
      return ev.template computeValWithTakeBack<EvalState,P>(*this,lastTo);
    }
    int computeValWithTakeBack(Player P, Position lastTo) {
      if (P == BLACK)
	return computeValWithTakeBack<BLACK>(lastTo);
      else
	return computeValWithTakeBack<WHITE>(lastTo);
    }
    int computeValWithMove(Move move) 
    {
      if (getPlayer(move)==BLACK)
	return computeValWithMove<BLACK>(move);
      else
	return computeValWithMove<WHITE>(move);
    }
    /** SOMA ʤ */
    int computeValWithMoveWithoutSoma(Move move) const
    {
      return getVal() + ev.diffWithMove(*this, move);
    }
  };

  // doUndoMoves
  /**
   * default Ǥ٤ ɾؿκƷ׻򥵥ܤäƤ 
   * EvalT ˥ԡ˸ʤΤȤʤä memento ѥʤɤȤΤ
   */
  template <Player P,typename Base, typename EvalT>
  struct ApplyDoUndoSimpleMove<P,EvalState<Base,EvalT> >
  {
    typedef EvalState<Base,EvalT> state_t;
    template <typename F>
    static void doUndoSimpleMove(state_t& s,
				 Position from, Position to, int promoteMask,F& func)
    {
      EvalT oldEval = s.ev;
      s.ev.doSimpleMove(s,from, to, promoteMask);
      ApplyDoUndoSimpleMove<P,Base>::template doUndoSimpleMove<F>
	(s,from,to,promoteMask,func);
      s.ev = oldEval;
    }
  };
  template <Player P,class Base, class EvalT>
  struct ApplyDoUndoDropMove<P,EvalState<Base,EvalT> >
  {
    typedef EvalState<Base,EvalT> state_t;
    template <typename F>
    static void doUndoDropMove(state_t& s,
			      Position to, Ptype ptype,F func)
    {
      EvalT oldEval = s.ev;
      s.ev.doDropMove(s,to,ptype);
      ApplyDoUndoDropMove<P,Base>::template doUndoDropMove<F>
	(s,to,ptype,func);
      s.ev = oldEval;
    }
  };
  template <Player P,class Base, class EvalT>
  struct ApplyDoUndoCaptureMove<P,EvalState<Base,EvalT> >
  {
    typedef EvalState<Base,EvalT> state_t;
    template <typename F>
    static void doUndoCaptureMove(state_t& s, 
				  Position from, Position to, Piece target, int promoteMask, F func)
    {
      EvalT oldEval = s.ev;
      s.ev.doCaptureMove(s, from, to, target, promoteMask);
      ApplyDoUndoCaptureMove<P,Base>::template doUndoCaptureMove<F>
	(s, from, to, target, promoteMask, func);
      s.ev = oldEval;
    }
  };
}
#endif // _EVAL_H
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
