#ifndef _EVAL_TCC
#define _EVAL_TCC

#include "osl/eval.h"
#include "osl/applyMove.h"

namespace osl
{
  template <class BaseState, class EvalT>
  class DoUndoMoveLockSimple<EvalState<BaseState,EvalT> >
    : public DoUndoMoveLockCleanUp
  {
    typedef EvalState<BaseState,EvalT> state_t;
    state_t& state;
    EvalT oldEval;
    typedef DoUndoMoveLockSimple<BaseState> baselock_t;
    boost::scoped_ptr<baselock_t> baselock;
  public:
    DoUndoMoveLockSimple(state_t& s, 
			 Position from, Position to, int promoteMask)
      : state(s), oldEval(s.ev)
    {
      s.ev.doSimpleMove(s, from, to, promoteMask);
      baselock.reset(new baselock_t(s, from, to, promoteMask));
    }
    ~DoUndoMoveLockSimple()
    {
      baselock.reset();
      state.ev = oldEval;
    }
  };

  template <class BaseState, class EvalT>
  class DoUndoMoveLockDrop<EvalState<BaseState,EvalT> >
    : public DoUndoMoveLockCleanUp
  {
    typedef EvalState<BaseState,EvalT> state_t;
    state_t& state;
    EvalT oldEval;
    typedef DoUndoMoveLockDrop<BaseState> baselock_t;
    boost::scoped_ptr<baselock_t> baselock;
  public:
    DoUndoMoveLockDrop(state_t& s, Position to, Ptype ptype)
      : state(s), oldEval(s.ev)
    {
      s.ev.doDropMove(s, to, ptype);
      baselock.reset(new baselock_t(s, to, ptype));
    }
    ~DoUndoMoveLockDrop()
    {
      baselock.reset();
      state.ev = oldEval;
    }
  };
  template <class BaseState, class EvalT>
  class DoUndoMoveLockCapture<EvalState<BaseState,EvalT> >
    : public DoUndoMoveLockCleanUp
  {
    typedef EvalState<BaseState,EvalT> state_t;
    state_t& state;
    EvalT oldEval;
    typedef DoUndoMoveLockCapture<BaseState> baselock_t;
    boost::scoped_ptr<baselock_t> baselock;
  public:
    DoUndoMoveLockCapture(state_t& s, Position from, Position to, Piece target,
			  int promoteMask)
      : state(s), oldEval(s.ev)
    {
      s.ev.doCaptureMove(s, from, to, target, promoteMask);
      baselock.reset(new baselock_t(s, from, to, target, promoteMask));
    }
    ~DoUndoMoveLockCapture()
    {
      baselock.reset();
      state.ev = oldEval;
    }
  };

  template<typename State,typename EvalT>
  EvalState<State,EvalT>::EvalState(const PawnMaskState& state) 
    : State(state), ev(state)
  {
  }
  template<typename State,typename EvalT>
  EvalState<State,EvalT>::~EvalState() 
  {
  }

  template<typename State,typename EvalT>
  bool EvalState<State,EvalT>::isConsistent(bool showError) const
  {
    if (!State::isConsistent(showError)) 
      return false;
    if (getVal() != EvalT(*this).getVal())
    {
      if (showError) 
	std::cerr << "eval mismatch " << ev.getVal() <<
	  " != " << EvalT(*this).getVal() << std::endl <<
	  *this << std::endl;
      return false;
    }
    return true;
  }

} // namespace osl

#endif /* _EVAL_TCC */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
