/* fixedCapacityVector.h
 */
#ifndef _FIXED_CAPACITY_VECTOR_H
#define _FIXED_CAPACITY_VECTOR_H

#include "osl/carray.h"
#include <algorithm>
#include <cstddef>
#include <cassert>
namespace osl
{
  template <typename T, size_t Capacity>
  class FixedCapacityVector
  {
  protected:
    typedef CArray<T, Capacity> array_t;
    array_t elements;
    size_t index;
  public:
    typedef typename array_t::iterator iterator;
    typedef typename array_t::const_iterator const_iterator;
    
    FixedCapacityVector() : index(0) {}

    T& operator[] (size_t i)
    {
      assert(i < size());
      return elements[i];
    }

    iterator begin() { return &elements[0]; }
    iterator end() { return &elements[0]+index; }

    void push_back(const T& e) { elements[index++] = e; }
    void pop_back() { --index; }
    void clear() { index = 0; }

    /** ʣǤ */
    void unique()
    {
      std::sort(begin(),end());
      iterator last = std::unique(begin(), end());
      index = (last - begin());
    }

    size_t size() const { return index; }
    bool empty() const { return size() == 0; }
    size_t capacity() const { return Capacity; }
    
    T const& operator[] (size_t i) const
    {
      assert(i < size());
      return elements[i];
    }
    const_iterator begin() const { return &elements[0]; }
    const_iterator end()   const { return &elements[0]+index; }
    
    bool isMember(const T& e, const_iterator first, const_iterator last) const
    {
      return std::find(first, last, e) != last;
    }
    bool isMember(const T& e) const
    {
      return isMember(e, begin(), end());
    }
  };
}


#endif /* _FIXED_CAPACITY_VECTOR_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
