#ifndef _HASH_H
#define _HASH_H

#include "osl/hashKey.h"
#include "osl/board.h"
#include "osl/ptype.h"
#include "osl/piece.h"
#include "osl/move.h"
#include "osl/carray2d.h"
#include "osl/doUndoMoveLock.h"

namespace osl
{
  template <Player P,class State>  struct ApplyDoUndoSimpleMove;
  template <Player P,class State>  struct ApplyDoUndoCaptureMove;
  template <Player P,class State>  struct ApplyDoUndoDropMove;
  template <class State>  struct DoUndoMoveLockSimple;
  template <class State>  struct DoUndoMoveLockCapture;
  template <class State>  struct DoUndoMoveLockDrop;

  template<class State>
  class HashState : public State
  {
  public:
    /** Ѿ饹 hash ĤФ˻Ѥ */
    typedef HashState hash_state_t;
  private:
    HashKey hash;
    friend bool operator==(const HashState& s1,
			   const HashState& s2){
      if(s1.hash != s2.hash) return false;
      // assert줿
      return true;
    }
    friend class ApplyDoUndoSimpleMove<BLACK,HashState>;
    friend class ApplyDoUndoCaptureMove<BLACK,HashState>;
    friend class ApplyDoUndoDropMove<BLACK,HashState>;
    friend class ApplyDoUndoSimpleMove<WHITE,HashState>;
    friend class ApplyDoUndoCaptureMove<WHITE,HashState>;
    friend class ApplyDoUndoDropMove<WHITE,HashState>;
    friend class DoUndoMoveLockSimple<HashState>;
    friend class DoUndoMoveLockCapture<HashState>;
    friend class DoUndoMoveLockDrop<HashState>;
  public:
    typedef HashKey HashType;
    typedef State base_t;
    
    explicit HashState(const PawnMaskState& state) 
      : State(state), hash(HashKey::calcHash(state))
    {
    }
    virtual ~HashState(){}
  private:
  public:
    // move methods
    void doSimpleMove(Position from, Position to, int promoteMask);
    void doDropMove(Position to,Ptype ptype);
    void doCaptureMove(Position from, Position to, Piece target,int promoteMask);
    void doMove(Move move);
    HashKey const& getHash() const{ return hash;}
    bool isConsistent(bool showError=true) const;
    void print(std::ostream&) const;
  };

  // doUndoMoves
  /**
   * default Ǥ٤ hash κƷ׻򥵥ܤäƤ 
   */
  template <Player P,typename Base>
  struct ApplyDoUndoSimpleMove<P,HashState<Base> >
  {
    typedef HashState<Base> state_t;
    template <typename F>
    static void doUndoSimpleMove(state_t& s,
				 Position from, Position to, int promoteMask,F& func)
    {
      HashKey oldHash = s.hash;
      s.hash.doSimpleMove(s, from, to, promoteMask);
      ApplyDoUndoSimpleMove<P,Base>::template doUndoSimpleMove<F>
	(s,from,to,promoteMask,func);
      s.hash = oldHash;
    }
  };
  
  template <Player P,class Base>
  struct ApplyDoUndoDropMove<P,HashState<Base> >
  {
    typedef HashState<Base> state_t;
    template <typename F>
    static void doUndoDropMove(state_t& s,
			       Position to, Ptype ptype,F func)
    {
      HashKey oldHash = s.hash;
      s.hash.doDropMove(s, to,ptype);
      ApplyDoUndoDropMove<P,Base>::template doUndoDropMove<F>
	(s,to,ptype,func);
      s.hash = oldHash;
    }
  };

  template <Player P,class Base>
  struct ApplyDoUndoCaptureMove<P,HashState<Base> >
  {
    typedef HashState<Base> state_t;
    template <typename F>
    static void doUndoCaptureMove(state_t& s, Position from, Position to, 
				  Piece target, int promoteMask, F func)
    {
      HashKey oldHash = s.hash;
      s.hash.doCaptureMove(s, from, to, target, promoteMask);
      ApplyDoUndoCaptureMove<P,Base>::template doUndoCaptureMove<F>
	(s, from, to, target, promoteMask, func);
      s.hash = oldHash;
    }
  };

  template <class BaseState>
  class DoUndoMoveLockSimple<HashState<BaseState> >
    : public DoUndoMoveLockCleanUp
  {
    typedef HashState<BaseState> state_t;
    state_t& state;
    HashKey oldHash;
    typedef DoUndoMoveLockSimple<BaseState> baselock_t;
    boost::scoped_ptr<baselock_t> baselock;
  public:
    DoUndoMoveLockSimple(state_t& s, 
			 Position from, Position to, int promoteMask);
    ~DoUndoMoveLockSimple();
  };

  template <class BaseState>
  class DoUndoMoveLockDrop<HashState<BaseState> >
    : public DoUndoMoveLockCleanUp
  {
    typedef HashState<BaseState> state_t;
    state_t& state;
    HashKey oldHash;
    typedef DoUndoMoveLockDrop<BaseState> baselock_t;
    boost::scoped_ptr<baselock_t> baselock;
  public:
    DoUndoMoveLockDrop(state_t& s, Position to, Ptype ptype);
    ~DoUndoMoveLockDrop();
  };
  template <class BaseState>
  class DoUndoMoveLockCapture<HashState<BaseState> >
    : public DoUndoMoveLockCleanUp
  {
    typedef HashState<BaseState> state_t;
    state_t& state;
    HashKey oldHash;
    typedef DoUndoMoveLockCapture<BaseState> baselock_t;
    boost::scoped_ptr<baselock_t> baselock;
  public:
    DoUndoMoveLockCapture(state_t& s, Position from, Position to, Piece target,
			  int promoteMask);
    ~DoUndoMoveLockCapture();
  };
} // namespace osl

#endif /* _HASH_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
