#ifndef _HASH_TCC
#define _HASH_TCC

#include "osl/hash.h"
#include "osl/doUndoMoveLock.h"
#include "osl/applyMove.h"
#include <iostream>

template <class BaseState>
osl::DoUndoMoveLockSimple<osl::HashState<BaseState> >::
DoUndoMoveLockSimple(state_t& s, Position from, Position to, int promoteMask)
  : state(s), oldHash(s.hash)
{
  state.hash.doSimpleMove(s, from, to, promoteMask);
  baselock.reset(new baselock_t(s, from, to, promoteMask));
}
template <class BaseState>
osl::DoUndoMoveLockSimple<osl::HashState<BaseState> >::
~DoUndoMoveLockSimple()
{
  baselock.reset();
  state.hash = oldHash;
}

template <class BaseState>
osl::DoUndoMoveLockDrop<osl::HashState<BaseState> >::
DoUndoMoveLockDrop(state_t& s, Position to, Ptype ptype)
  : state(s), oldHash(s.hash)
{
  state.hash.doDropMove(s, to, ptype);
  baselock.reset(new baselock_t(s, to, ptype));
}
template <class BaseState>
osl::DoUndoMoveLockDrop<osl::HashState<BaseState> >::
~DoUndoMoveLockDrop()
{
  baselock.reset();
  state.hash = oldHash;
}
template <class BaseState>
osl::DoUndoMoveLockCapture<osl::HashState<BaseState> >::
DoUndoMoveLockCapture(state_t& s, Position from, Position to, Piece target,
		      int promoteMask)
  : state(s), oldHash(s.hash)
{
  state.hash.doCaptureMove(s, from, to, target, promoteMask);
  baselock.reset(new baselock_t(s, from, to, target, promoteMask));
}
template <class BaseState>
osl::DoUndoMoveLockCapture<osl::HashState<BaseState> >::
~DoUndoMoveLockCapture()
{
  baselock.reset();
  state.hash = oldHash;
}

template<class State>
bool osl::HashState<State>::isConsistent(bool showError) const
{
  if (! State::isConsistent(showError)) 
  {
    if (showError)
      std::cerr << "error before hash" << std::endl;
    return false;
  }
  if (hash != HashKey::calcHash(*this))
  {
    if(showError) 
      std::cerr << "hash mismatch" << std::endl;
    return false;
  }
  PieceStand blackStand;
  for (int num=0;num<40;num++)
  {
    if (getOffBoardMask(BLACK).isSet(num))
      blackStand.add(numTable.getPtype(num));
  }
  if (hash.blackStand() != blackStand)
  {
    if(showError) 
      std::cerr << "stand mismatch" << std::endl;
    return false;
  }
  return true;
}

template<class State>
void osl::HashState<State>::print(std::ostream& os) const {
  os << hash << std::endl;
  os<<static_cast<State>(*this);
}

template<class State>
void osl::HashState<State>::
doSimpleMove(Position from, Position to, int promoteMask)
{
#ifndef NDEBUG
  PtypeO newPtypeo = 
#endif
    hash.doSimpleMove(*this, from,to,promoteMask);
  State::doSimpleMove(from,to,promoteMask);
  assert(newPtypeo==getPtypeO(getPieceAt(to)));
}

template<class State>
void osl::HashState<State>::
doDropMove(Position to,Ptype ptype)
{
  hash.doDropMove(*this, to,ptype);
  State::doDropMove(to,ptype);
}

template<class State>
void osl::HashState<State>::
doCaptureMove(Position from, Position to, Piece p1,int promoteMask)
{
  hash.doCaptureMove(*this, from,to,p1,promoteMask);
  State::doCaptureMove(from,to,p1,promoteMask);
}

template<class State>
void osl::HashState<State>::
doMove(Move move)
{
  ApplyMoveOfTurn::doMove(*this,move); 
}

template<class State> inline
bool operator!=(const osl::HashState<State>& s1, 
		const osl::HashState<State>& s2)
{
  return !(s1==s2);
}

template<class State>
std::ostream& operator<<(std::ostream& os,const osl::HashState<State>& state)
{
  state.print(os);
  return os;
}


#endif /* _HASH_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
