/* hashKey.h
 */
#ifndef _HASH_KEY_H
#define _HASH_KEY_H

#include "osl/piece.h"
#include "osl/move.h"
#include "osl/pieceStand.h"
#include "osl/pawnMaskState.h"

#ifndef NDEBUG
#  define HASH_KEY_128_BIT
#endif

namespace osl
{
  /**
   * (֡ޤ)Բĵհ̤沽
   */
  class HashKey
  {
  private:
    /**
     * ׾ζ zoblist hash
     * v[0] β̰ӥåȤϼӤѤ 
     * v[2]  λȤƻѤ
     * (ξ󤬥ơ֥˺ʤ¤ꡤλδϾĹ) 
     */
#ifdef HASH_KEY_128_BIT
    unsigned int v[4];
#else
    unsigned int v[3];
#endif
  public:
    /** make zero initialized key */
    HashKey()
    {
      v[0]=0;
      v[1]=0;
      v[2]=0;
#ifdef HASH_KEY_128_BIT
      v[3]=0;
#endif
    }
    HashKey(int v0,int v1,int v2, int v3)
    {
      v[0]=v0; v[1]=v1; v[2]=v2; 
#ifdef HASH_KEY_128_BIT
      v[3]=v3;
#endif
    }

    void setRandom();
    void xorPlayer() { v[0]^=1; }
    void setPlayer(Player p) { v[0] |= playerToIndex(p); }

    HashKey& operator+=(const HashKey& r)
    {
      v[0] += r.v[0];      v[1] += r.v[1];      v[2] += r.v[2];
#ifdef HASH_KEY_128_BIT
      v[3] += r.v[3];
#endif
      return *this;
    }
    HashKey& operator-=(const HashKey& r)
    {
      v[0] -= r.v[0];      v[1] -= r.v[1];      v[2] -= r.v[2];
#ifdef HASH_KEY_128_BIT
      v[3] -= r.v[3];
#endif
      return *this;
    }

    bool isPlayerOfTurn(Player p) const
    {
      return static_cast<int>(v[0] & 1) == playerToIndex(p);
    }
    unsigned int value(int i) const { return v[i]; }
    PieceStand blackStand() const { return PieceStand(v[2]); }
      
    static HashKey calcHash(const PawnMaskState& state);

    HashKey newHashWithMove(Move move) const;

    PtypeO doSimpleMove(const PawnMaskState& state,
			Position from, Position to, int promoteMask);
    void doDropMove(const PawnMaskState&, Position to,Ptype ptype);
    void doCaptureMove(const PawnMaskState& state,
		       Position from, Position to, Piece p1, int promoteMask);
  };

  std::ostream& operator<<(std::ostream& os,const HashKey& h);

  inline bool operator==(const HashKey& l,const HashKey& r){
    return (l.value(0)==r.value(0))
      && (l.value(1)==r.value(1))
      && (l.value(2)==r.value(2))
#ifdef HASH_KEY_128_BIT
      && (l.value(3)==r.value(3))
#endif
      ;
  }
  inline bool operator<(const HashKey& l,const HashKey& r){
#ifdef HASH_KEY_128_BIT
    const int maxid = 3;
#else
    const int maxid = 2;
#endif    
    for (int i=0; i<maxid; ++i)
      if (l.value(i)!=r.value(i))
	return l.value(i)<r.value(i);
    return l.value(maxid)<r.value(maxid);
  }
  inline bool operator!=(const HashKey& h1,const HashKey& h2){
    return !(h1==h2);
  }

  class HashGenTable
  {
    CArray2d<HashKey,POSITION_SIZE,PTYPEO_SIZE> hashKey;
  public:
    HashGenTable();
    void addHashKey(HashKey& hk,Position pos,PtypeO ptypeo) const{
      assert(isValid(pos) && isValid(ptypeo));
      hk+=hashKey[pos][ptypeo-PTYPEO_MIN];
    }
    void subHashKey(HashKey& hk,Position pos,PtypeO ptypeo) const{
      assert(isValid(pos) && isValid(ptypeo));
      hk-=hashKey[pos][ptypeo-PTYPEO_MIN];
    }
  };
  extern const HashGenTable hashGenTable;

  inline 
  PtypeO HashKey::doSimpleMove(const PawnMaskState& state,
			       Position from, Position to, int promoteMask) 
  {
    const PtypeO ptypeo=getPtypeO(state.getPieceAt(from));
    const PtypeO newPtypeo=promoteWithMask(ptypeo,promoteMask);
    hashGenTable.subHashKey(*this,from,ptypeo);
    hashGenTable.addHashKey(*this,to,newPtypeo);
    xorPlayer();
    return newPtypeo;
  }
  inline
  void HashKey::doDropMove(const PawnMaskState& state, Position to,Ptype ptype)
  {
    const Player p=state.getTurn();
    hashGenTable.subHashKey(*this,OFFBOARD,newPtypeO(p,ptype));
    hashGenTable.addHashKey(*this,to,newPtypeO(p,ptype));
    xorPlayer();
  }
  inline
  void HashKey::doCaptureMove(const PawnMaskState& state,
			      Position from, Position to, Piece p1, int promoteMask)
  {
    const Ptype capturePtype=getPtype(p1);
    const Piece p0=state.getPieceAt(from);
    const Piece newPiece=promoteWithMask(p0,promoteMask);
    const Player p=state.getTurn();
    // ä
    hashGenTable.subHashKey(*this,to,newPtypeO(alt(p),capturePtype));
    hashGenTable.addHashKey(*this,OFFBOARD,newPtypeO(p,unpromote(capturePtype)));
    // ư
    hashGenTable.subHashKey(*this,from,getPtypeO(p0));
    hashGenTable.addHashKey(*this,to,getPtypeO(newPiece));
    xorPlayer();
  }
} // namespace osl


#if __GNUG__ >= 3
using namespace __gnu_cxx;
namespace __gnu_cxx
#else
using namespace std;
namespace std
#endif
{
  template <typename T> struct hash;
  struct hash<osl::HashKey>{
    unsigned long operator()(const osl::HashKey& h) const{
      return h.value(0);
    }
  };
  // for icc
  static inline size_t hash_value(const osl::HashKey& val){
    return val.value(0);
  }
}

#endif /* _HASH_KEY_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
