#ifndef _MOVE_H
#define _MOVE_H
#include "board.h"
#include "player.h"
#include "ptype.h"
#include <iosfwd>

/** move طǤĤޤΤʤ顼ǤȤ */
// #define MOVE_DEBUG
#ifdef MOVE_DEBUG
#  include <cassert>
#  define move_assert(x) assert(x)
#else
#  define move_assert(x) 
#endif

namespace osl{
/**
 * ̤Ƥʤ moveɽ 
 * Piecepromotep򤽤 
 * ̤ 
 * - from     : 8 bit 
 * - to       : 8 bit 
 * - dummy    : 3 bit 
 * - promote? : 1 bit  
 * - capture ptype    : 4 bit 
 * - ptype    : 4 bit 
 * - owner    : signed 
 */
  enum Move{
    MOVE_INVALID=0,
    /** 32bit ݤ뤿Τޤʤ */
    MOVE_MIN=-0x7fffffff,
    /** 32bit ݤ뤿Τޤʤ */
    MOVE_MAX=0x7fffffff,
  };

/** դιˡκ 
 * ʣƼ뤳Ȥϡ600Ǥ­⤷ʤ
 */
  const size_t MoveMax=600;

  inline Move newMove(Position from, Position to, Ptype ptype,
		      Ptype capturePtype, bool isPromoteMove,
		      Player player){
    move_assert(isValid(from));
    move_assert(isValid(to));
    move_assert(isValid(ptype));
    move_assert(isValid(capturePtype));
    move_assert(isValid(player));
    return static_cast<Move>(static_cast<unsigned int>(from)|
			     (static_cast<unsigned int>(to)<<8)|
			     (static_cast<unsigned int>(isPromoteMove)<<19)|
			     (static_cast<unsigned int>(capturePtype)<<20)|
			     (static_cast<unsigned int>(ptype)<<24)|
			     (static_cast<int>(player)<<28));
  }

  inline Move newDrop(Position to,Ptype ptype,Player player){
    move_assert(isValid(to));
    move_assert(isValid(ptype));
    move_assert(isValid(player));
    return newMove(OFFBOARD,to,ptype,PTYPE_EMPTY,false,player);
  }

  inline Position getFrom(Move move){
    const Position result = static_cast<Position>(static_cast<int>(move)&0xff);
    move_assert(isValid(result));
    return result;
  }

  inline Position getTo(Move move){
    const Position result = static_cast<Position>((static_cast<int>(move)>>8)&0xff);
    move_assert(isValid(result));
    return result;
  }

  inline int getPromoteMask(Move move){
    return (static_cast<int>(move)&(1<<19));
  }

  inline int promoteMask(bool isPromote){
    return isPromote<<19;
  }
  inline bool isPromote(Move move){
    return (static_cast<int>(move)&(1<<19))!=0;
  }
  inline Ptype getPtype(Move move){
    const Ptype result = static_cast<Ptype>((static_cast<int>(move)>>24)&0xf);
    move_assert(isValid(result));
    return result;
  }

/** ưPtype, i.e., ää */
  inline PtypeO getPtypeO(Move move){
    const PtypeO result = static_cast<PtypeO>(static_cast<int>(move)>>24);
    move_assert(isValid(result));
    return result;
  }

/** ưPtypeO, i.e., ä */
  inline PtypeO getOldPtypeO(Move move){
    const PtypeO result = static_cast<PtypeO>((static_cast<int>(move)>>24)+((static_cast<int>(move)>>(19-3))&8));
    move_assert(isValid(result));
    return result;
  }

/** ưPtype, i.e., ä */
  inline Ptype getOldPtype(Move move){
    return getPtype(getOldPtypeO(move));
  }

  inline Ptype getCapturePtype(Move move){
    const Ptype result = static_cast<Ptype>((static_cast<int>(move)>>20)&0xf);
    move_assert(isValid(result));
    return result;
  }

  inline Player getPlayer(Move move){
    const Player result = static_cast<Player>(static_cast<int>(move)>>28);
    move_assert(isValid(result));
    return result;
  }

  inline Move noPromote(Move move){
    move_assert(isPromote(move) && isPromoted(getPtype(move)));
    return static_cast<Move>(static_cast<int>(move)^((1<<19)|(1<<27)));
  }

  inline Move addTo(Move move,Offset offset){
    return static_cast<Move>(static_cast<int>(move)+(static_cast<int>(offset)<<8));
  }

  inline bool isValid(Move m)
  {
    const Position from = getFrom(m);
    if (! (isOffBoard(from) || isOnBoard(from)))
      return false;
    const Position to = getFrom(m);
    if (! (isOffBoard(to) || isOnBoard(to)))
      return false;
    return isValid(getPtype(m))
      && isValid(getCapturePtype(m))
      && getCapturePtype(m)!=KING
      && isValid(getPlayer(m));
  }

  inline bool operator<(Move lhs,Move rhs){
    return static_cast<int>(lhs) < static_cast<int>(rhs);
  }

  std::ostream& operator<<(std::ostream& os,const Move move);
  std::ostream& csaShow(std::ostream& os,const Move move);
}
#endif /* _MOVE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
