#ifndef _GENERATE_PIECE_MOVES_H
#define _GENERATE_PIECE_MOVES_H
#include "osl/loki.h"
#include "osl/player.h"
#include "osl/ptype.h"
#include "osl/position.h"
#include "osl/direction.h"
#include "osl/piece.h"
#include "osl/promote.h"
#include "osl/bitXmask.h"
#include "osl/move_action/concept.h"
#include <boost/static_assert.hpp>

namespace osl
{
  namespace move_generator
  {

    /**
     * pieceư
     */
    template<class State,class Action>
    class PieceOnBoard
    {
      BOOST_CLASS_REQUIRE(Action,osl::move_action,Concept);
      BOOST_STATIC_ASSERT(sizeof(typename State::simple_board_t));
    private:
      template<Player P,Ptype T,PromoteType CanP,Direction Dir>
      static void generateLongMovesDir(const State& state,Piece p,int num,Position pos, Action& action);
      template<Player P,Ptype T,PromoteType CanP,Direction Dir>
      static void generateShortMovesDir(const State& state,Piece p,int num,Position pos, Action& action){
	const Offset offset=DirectionPlayerTraits<Dir,P>::offset;
	Position toPos=pos+offset;
	Piece p1=state.getPieceAt(toPos);
	if(canMoveOn<P>(p1)){
	  if(CanP==CanPromoteType 
	     || (CanP==CheckPromoteType && DirectionTraits<Dir>::canPromoteTo && canPromotePosition<P>(toPos))){
	    action.unknownMove(pos,toPos,p1,PtypeFuns<T>::promotePtype,true,P);
	    if (PtypePlayerTraits<T,P>::canDropTo(toPos))
	      action.unknownMove(pos,toPos,p1,T,false,P);
	  }
	  else{
	    action.unknownMove(pos,toPos,p1,T,false,P);
	  }
	}
      }

      template<Player P,Ptype T,Ptype MoveT,PromoteType CanP,Direction Dir>
      static void generateMovesDir(const State& state,Piece p,int num,Position pos, Action& action,Int2Type<true>,Int2Type<true>){
	generateMovesDirSub<P,T,NoPromoteType,Dir>(state,p,num,pos,action);
      }

      template<Player P,Ptype T,Ptype MoveT,PromoteType CanP,Direction Dir>
      static void generateMovesDir(const State& state,Piece p,int num,Position pos, Action& action,Int2Type<true>,Int2Type<false>){
	generateMovesDirSub<P,T,CanP,Dir>(state,p,num,pos,action);
      }

      template<Player P,Ptype T,Ptype MoveT,PromoteType CanP,Direction Dir>
      static void generateMovesDir(const State& state,Piece p,int num,Position pos, Action& action,Int2Type<false>,Int2Type<true>){
      }

      template<Player P,Ptype T,Ptype MoveT,PromoteType CanP,Direction Dir>
      static void generateMovesDir(const State& state,Piece p,int num,Position pos, Action& action,Int2Type<false>,Int2Type<false>){
      }

      template<Player P,Ptype T,Ptype MoveT,PromoteType CanP,Direction Dir>
      static void generateMovesDir(const State& state,Piece p,int num,Position pos, Action& action){
	generateMovesDir<P,T,MoveT,CanP,Dir>(state,p,num,pos,action,
					     Int2Type<(PtypeTraits<MoveT>::moveMask & DirectionTraits<Dir>::mask) !=0>(),
					     Int2Type<CanP == CheckPromoteType && !DirectionTraits<Dir>::canPromoteTo>());
      }

    public:

      template<Player P,Ptype T,Ptype MoveT,PromoteType CanP,bool isLong>
      static void generateMovesSub(const State& state,Piece p,int num,Position pos, Action& action,Int2Type<isLong>);

      template<Player P,Ptype T,PromoteType CanP,Direction Dir>
      static inline void generateMovesDirSub(const State& state,Piece p,int num,Position pos, Action& action,Int2Type<true>){
	generateLongMovesDir<P,T,CanP,Dir>(state,p,num,pos,action);
      }

      template<Player P,Ptype T,PromoteType CanP,Direction Dir>
      static inline void generateMovesDirSub(const State& state,Piece p,int num,Position pos, Action& action,Int2Type<false>){
	generateShortMovesDir<P,T,CanP,Dir>(state,p,num,pos,action);
      }


      template<Player P,Ptype T,PromoteType CanP,Direction Dir>
      static inline void generateMovesDirSub(const State& state,Piece p,int num,Position pos, Action& action){
	generateMovesDirSub<P,T,CanP,Dir>(state,p,num,pos,action,
					  Int2Type<DirectionTraits<Dir>::isLong>());
      }

      /**
       * Ptypepromoteǽɤˤä dispatch
       */
      template<Player P,Ptype T,bool canPromote>
      static void generateMoves(const State& state,Piece p, Action& action,Int2Type<canPromote>);

      template<Player P,Ptype T,Ptype MoveT,PromoteType CanP>
      static void generateMovesSub(const State& state,Piece p,int num,Position pos, Action& action);

      template<Player P,Ptype T>
      static void generateMoves(const State& state,Piece p, Action& action,Int2Type<true>){
	Position pos=getPosition(p);
	int num=getNum(p);
	if(isPromoted(p)){
	  generateMovesSub<P,PtypeFuns<T>::promotePtype,PtypeTraits<PtypeFuns<T>::promotePtype>::moveType,NoPromoteType>(state,p,num,pos,action);
	}
	else{
	  if(canPromotePosition<P>(pos))
	    generateMovesSub<P,T,T,CanPromoteType>(state,p,num,pos,action);
	  else
	    generateMovesSub<P,T,T,CheckPromoteType>(state,p,num,pos,action);
	}
      }
      template<Player P,Ptype T>
      static void generateMoves(const State& state,Piece p, Action& action,Int2Type<false>){
	Position pos=getPosition(p);
	int num=getNum(p);
	generateMovesSub<P,T,T,NoPromoteType>(state,p,num,pos,action);
      }

      template<Player P>
      static void generateMovesPawn(const State& state,Piece p, Action& action,BitXmask& pawnMask){
	Position pos=getPosition(p);
	int num=getNum(p);
	if(isPromoted(p)){
	  generateMovesSub<P,PPAWN,GOLD,NoPromoteType>(state,p,num,pos,action);
	}
	else{
	  set(pawnMask,getX(pos));
	  if(canPromotePosition<P>(pos))
	    generateMovesSub<P,PAWN,PAWN,CanPromoteType>(state,p,num,pos,action);
	  else
	    generateMovesSub<P,PAWN,PAWN,CheckPromoteType>(state,p,num,pos,action);
	}
      }

      template<Player P,Ptype T>
      static void generateMoves(const State& state,Piece p, Action& action){
	generateMoves<P,T>(state,p,action,Int2Type<PtypeTraits<T>::canPromote>());
      }



    };
  } // namespace move_generator
} // namespace osl

#endif /* _GENERATE_PIECE_MOVES_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
