#ifndef _NUM_SIMPLE_EFFECT_H
#define _NUM_SIMPLE_EFFECT_H

#include "osl/ptype.h"
#include "osl/numBitmapEffect.h"
#include "osl/carray.h"
#include "osl/doUndoMoveLock.h"
#include "osl/applyMove.h"

#include <boost/static_assert.hpp>

// Note: no-inline (and no-template-member) functions are moved into numSimpleEffect.tcc

namespace osl 
{
  template<class State>
  class NumSimpleEffectTable;
  template<class State>
  bool operator==(const NumSimpleEffectTable<State>& et1,const NumSimpleEffectTable<State>& et2);
  /**
   * 0-31 : 0-31ľܸ
   * 32-39 : 32-39û32-39ûܸ
   * 40-47 : 32-39Ĺ32-39Ĺܸ
   */
  template<class State>
  class NumSimpleEffectTable{
  protected:
    CArray<NumBitmapEffect, POSITION_SIZE> effects;
  public:
    template<Player P,Ptype T,Direction Dir>
    inline void xorEffectShort(Position pos,int num){
      if ((PtypeTraits<T>::moveMask & DirectionTraits<Dir>::mask)!=0){
	// setNum0Ȥ
	effects[pos+DirectionPlayerTraits<Dir,P>::offset].xorNum(num);
      }
    }
    // DirectionϹˤȤäƤdirection
    template<Player P,Ptype T,Direction Dir>
    inline void xorEffectLong(const State& state,Position pos,int num){
      if ((PtypeTraits<T>::moveMask & DirectionTraits<DirectionPlayerTraits<Dir,P>::directionByBlack>::mask)!=0){

	Offset offset=DirectionPlayerTraits<Dir,BLACK>::offset;
	assert(offset!=0);

	for (;;){
	  pos=pos+offset;
	  effects[pos].xorNum1(num+8);
	  if (!isEmpty(state.getPieceAt(pos))) break;
	}
      }
    }
    template<Player P,Ptype T>
    void xorEffectBy(const State& state,Position pos,int num);
    void xorEffect(const State& state,PtypeO ptypeo,Position pos,int num);
    void xorEffect(const State& state,Piece p){
      xorEffect(state,getPtypeO(p),getPosition(p),getNum(p));
    }
    void init(const State& state);
    NumSimpleEffectTable(const State& state){
      init(state);
    }
    NumBitmapEffect getEffect(Position pos) const{
      return effects[static_cast<int>(pos)];
    }
    void xorBlockAt(const State& state,Position pos,PtypeO ptypeo);
    friend bool operator== <>(const NumSimpleEffectTable& et1,const NumSimpleEffectTable& et2);
  };

  template<class State>
  inline bool operator!=(const NumSimpleEffectTable<State>& et1,const NumSimpleEffectTable<State>& et2){
    return !(et1==et2);
  }

  template <Player P,class State>
  struct ApplyDoUndoSimpleMove;
  template <Player P,class State>
  struct ApplyDoUndoCaptureMove;
  template <Player P,class State>
  struct ApplyDoUndoDropMove;

  template<class State>
  class NumSimpleEffect;
  template<class State>
  bool operator==(const NumSimpleEffect<State>& st1, const NumSimpleEffect<State>& st2);
  
  template<class State>
  class NumSimpleEffect: public State{
  public:
    /** Ѿ饹 effect ĤФ˻Ѥ */
    typedef NumSimpleEffect<State> effect_state_t;
  private:
    //    State state;
    NumSimpleEffectTable<State> effects;
    friend bool operator==<>(const NumSimpleEffect& st1,
			     const NumSimpleEffect& st2);
    typedef NumSimpleEffect state_t;
    friend class ApplyDoUndoSimpleMove<BLACK,NumSimpleEffect>;
    friend class ApplyDoUndoCaptureMove<BLACK,NumSimpleEffect>;
    friend class ApplyDoUndoDropMove<BLACK,NumSimpleEffect>;
    friend class ApplyDoUndoSimpleMove<WHITE,NumSimpleEffect>;
    friend class ApplyDoUndoCaptureMove<WHITE,NumSimpleEffect>;
    friend class ApplyDoUndoDropMove<WHITE,NumSimpleEffect>;
    CArray<PieceMask,2> onBoardMask;
  public:
    PieceMask& getOnBoardMask(Player p) {
      return onBoardMask[playerToIndex(p)];
    }
    const PieceMask& getOnBoardMask(Player p) const {
      return onBoardMask[playerToIndex(p)];
    }
    bool isOnBoardNum(int num) const{
      return getOnBoardMask(BLACK).isSet(num) ||
	getOnBoardMask(WHITE).isSet(num);
    }

    /** T  isBasic ǤʤƤư unpromote(T) η̤Ʊ */ 
    template<Player P,Ptype T,typename F>
    void forEachOnBoard(F & func) const {
      int onMask=getOnBoardMask(P).getMask(PtypeFuns<T>::indexNum) & PtypeFuns<T>::indexMask;
    
      while (onMask){
	int numLow=bsf(onMask);
	onMask ^= (1<<numLow);
	int num=numLow+((PtypeFuns<T>::indexNum)<<5);
	Piece p=getPieceOf(num);
	func(p);
      }
    }

    explicit NumSimpleEffect(const State& st);
#if 1
    NumBitmapEffect getEffect(Position pos) const{
      return effects.getEffect(pos);
    }
#endif
    //
    // move methods
    // effect
    bool isConsistent(bool showError) const;
    // 
    void doSimpleMove(Position from, Position to, int promoteMask);
    void doDropMove(Position to,Ptype ptype);
    /**
     * SimpleStatedoCaptureMoveȤƤ⡤
     * capturenumʤΤ
     */
    void doCaptureMove(Position from, Position to, Piece target,int promoteMask);
    void doMove(Move move){ ApplyMoveOfTurn::doMove(*this,move); }
    /**
     * ޥˤDirectionǤĹ뤫ɤ
     */
    template<Direction Dir,Player P>
    bool hasEffectDir(Position to) const {
      BOOST_STATIC_ASSERT( (DirectionTraits<Dir>::isLong) );
      const PieceMask& onBoardMask=getOnBoardMask(P);
      unsigned int mask1=onBoardMask.getMask(1);
      mask1&=(PtypeDirectionTraits<LANCE,Dir>::canMove ? 
	      PtypeFuns<LANCE>::indexMask : 0) |
	(PtypeDirectionTraits<BISHOP,Dir>::canMove ? 
	 PtypeFuns<BISHOP>::indexMask : 0) |
	(PtypeDirectionTraits<ROOK,Dir>::canMove ? 
	 PtypeFuns<ROOK>::indexMask : 0);
      // ûӽ
      mask1&=effects.getEffect(to).getMask(1)>>8;
      while (mask1!=0){
	int num32=bsf(mask1);
	mask1&= ~(1<<num32);
	int num=num32+32;
	Position from=getPosition(getPieceOf(num));
	Direction dir=directionTable.getLongDirection<BLACK>(positionDiff(to,from));
	if (dir==DirectionPlayerTraits<Dir,P>::directionByBlack)
	  return true;
      }
      return false;
    }
    /** 
     * @param player ¦
     * @param target оݤΥޥ
     */
    bool hasEffectBy(Player player,Position target) const {
      const PieceMask& onBoardMask=getOnBoardMask(player);
      unsigned int mask0=onBoardMask.getMask(0);
      unsigned int mask1=onBoardMask.getMask(1);
      mask1|=(mask1<<8);
      NumBitmapEffect effect=effects.getEffect(target);
      mask0&=effect.getMask(0);
      mask1&=effect.getMask(1);
      return (mask0|mask1)!=0;
    }


    /** 
     * @param player ¦
     * @param piece ¦ζ
     * @param target оݤΥޥ
     */
    bool hasEffectNotBy(Player player,Piece piece,Position target) const {
      assert(getOwner(piece)==player);
      PieceMask onBoardMask=getOnBoardMask(player);
      int num=getNum(piece);
      onBoardMask.clearNum(num);
      unsigned int mask0=onBoardMask.getMask(0);
      unsigned int mask1=onBoardMask.getMask(1);
      mask1|=(mask1<<8);
      NumBitmapEffect effect=effects.getEffect(target);
      mask0&=effect.getMask(0);
      mask1&=effect.getMask(1);
      return (mask0|mask1)!=0;
    }
    /** 
     * İʾζ𤫤
     * @param player ¦
     * @param target оݤΥޥ
     */
    bool hasMultipleEffectBy(Player player,Position target) const {
      NumBitmapEffect effect=effects.getEffect(target);
      unsigned int mask1=effect.getMask(1);
      mask1|=mask1>>8;
      const PieceMask& onBoardMask=getOnBoardMask(player);
      mask1&=onBoardMask.getMask(1);
#if 0
      /** ûʣ뤳ȤϤǤʤ */
      if (mask1==0) return false;
#else
      /** ȤΤ, λˤΤ߸뤳 */
      if (mask1==0){
	unsigned int mask0=effect.getMask(0);
	mask0&=onBoardMask.getMask(0);
	if (mask0==0) return false;
	int num=bsf(mask0);
	return mask0!=(1u<<num);
      }
#endif
      int num=bsf(mask1);
      if (mask1!=(1u<<num)) return true;
      unsigned int mask0=effect.getMask(0);
      mask0&=onBoardMask.getMask(0);
      return mask0!=0;
    }

    /**

    * ! ߻ͤɤƤ

|  NumSimpleEffectäϡλͤǽ񤤤ɤ¾ effectǤ
| 뤳ȤꤷäΤ⤷ޤ(뤳Ȥꤹ
| ȡNumSimpleEffectǤifʸĸäƤ줷ΤϤǤ)
| templateˤ뤫뤤PIECE_EMPTY֤Ȥ褦ˤ
| Τɤ餫ˤޤ礦

     * 1Ĥpiece֤
     * ʤPIECE_EMPTY֤
     * @param P(template) ¦
     * @param p åȤΰ
     */
    template<Player P> 
    Piece hasEffectPiece(Position p) const{
      assert(hasEffectBy(P,p));
      const PieceMask& onBoardMask=getOnBoardMask(P);
      unsigned int mask0=onBoardMask.getMask(0);
      NumBitmapEffect effect=effects.getEffect(p);
      mask0&=effect.getMask(0);
      if (mask0!=0){
	return getPieceOf(bsf(mask0));
      }
      else{
	unsigned int mask1=effect.getMask(1);
	mask1|=(mask1>>8);
	mask1&=onBoardMask.getMask(1);
	if (mask1!=0){
	  int num=bsf(mask1)+32;
	  assert(32<=num && num<=39);
	  return getPieceOf(num);
	}
	else{
	  return PIECE_EMPTY;
	}
      }
    }
    Piece hasEffectPiece(Player pl,Position pos) const{
      if (pl==BLACK) return hasEffectPiece<BLACK>(pos);
      else return hasEffectPiece<WHITE>(pos);
    }
    /**
     * @param P ¦
     * @param Type 
     */
    template<Player P,Ptype Type,class Action>
    void forEachEffectPtype(Position pos,Action & action) const {
      unsigned int mask=getEffect(pos).getMask(PtypeFuns<Type>::indexNum);
      if (PtypeFuns<Type>::indexNum==1){
	mask|=mask>>8;
      }
      mask&=PtypeFuns<Type>::indexMask;
      const PieceMask& onBoardMask=getOnBoardMask(P);
      mask&=onBoardMask.getMask(PtypeFuns<Type>::indexNum);
      while (mask!=0){
	int num=bsf(mask);
	mask &= ~(1<<num);
	num+=32*PtypeFuns<Type>::indexNum;
	action.template doActionPtype<P,Type>(getPieceOf(num),pos);
      }
    }

    /** ®ٽŻǤϤʤֽŻ 
     * pos ˤ move  action  member ƤӽФ
     * @param action Ȥ AlwayMoveAction
     */
    template<Player P,class Action>
    void forEachEffect(Position pos,Action & action) const {
      const PieceMask& onBoardMask=getOnBoardMask(P);
      unsigned int mask0=getEffect(pos).getMask(0);
      mask0&=onBoardMask.getMask(0);
      while (mask0!=0){
	int num=bsf(mask0);
	mask0 &= ~(1<<num);
	action.template doAction<P>(getPieceOf(num),pos);
      }
      unsigned int mask1=getEffect(pos).getMask(1);
      mask1|=(mask1>>8);
      mask1&=onBoardMask.getMask(1);
      while (mask1!=0){
	int numLow=bsf(mask1);
	mask1 &= ~(1<<numLow);
	int num=numLow+32;
	action.template doAction<P>(getPieceOf(num),pos);
      }
    }

    /** 
     * pos ˤ move  action  member ƤӽФ
     * @param action Ȥ AlwayMoveAction
     * @param piece  ʳζȤ
     */
    template<Player P,class Action>
    void forEachEffectNotBy(Position pos,Piece piece,Action & action) const {
      PieceMask onBoardMask=getOnBoardMask(P);
      onBoardMask.clearNum(getNum(piece));
      unsigned int mask0=getEffect(pos).getMask(0);
      mask0&=onBoardMask.getMask(0);
      while (mask0!=0){
	int num=bsf(mask0);
	mask0 &= ~(1<<num);
	action.template doAction<P>(getPieceOf(num),pos);
      }
      unsigned int mask1=getEffect(pos).getMask(1);
      mask1|=(mask1>>8);
      mask1&=onBoardMask.getMask(1);
      while (mask1!=0){
	int numLow=bsf(mask1);
	mask1 &= ~(1<<numLow);
	int num=numLow+32;
	action.template doAction<P>(getPieceOf(num),pos);
      }
    }
    

    /** ®ٽŻǤϤʤֽŻ 
     * pos ˤ move  action  member ƤӽФ
     * @param action Ȥ AlwayMoveAction
     */
    template<Player P,class Action>
    void forSortedEffect(Position pos,Action & action) const ;
    /**
     * attackerPiecetargetPositionĤɤ
     * targetPosition϶Ǥɤ
     * @param attackerPiece 
     * @param targetPosition
     */
    bool hasEffectTo(Piece attackerPiece,Position targetPosition) const
    {
      assert(isPiece(attackerPiece) && isOnBoard(targetPosition));
      NumBitmapEffect effect=effects.getEffect(targetPosition);
      int num=getNum(attackerPiece);
      int index=numToIndex(num);
      int mask=numToMask(num);
#if 1
      if (index!=0)mask|=mask<<8;
#else
      /** ΥɤϾʬʤƤӽФǤ٤ʤΤ
       *  褵ʤ¸Ƥ
       */
      mask|=((mask<<8)&(-index));
#endif
      return (effect.getMask(index) & mask)!=0;
    }
    /**
     * @param P(template) - ĤƤ¦Υץ쥤
     * @param target - Ĥ줿
     * @param attackerPiece - multiple attackξPIECE_EMPTY
     *        ǤʤʤĤƤ֤
     */
    template<Player P>
    bool hasEffectBy(Position target,Piece& attackerPiece) const {
      attackerPiece=PIECE_EMPTY;
      const PieceMask& onBoardMask=getOnBoardMask(P);
      unsigned int mask0=onBoardMask.getMask(0);
      NumBitmapEffect effect=effects.getEffect(target);
      mask0&=effect.getMask(0);
      unsigned int mask1=effect.getMask(1);
      mask1|=(mask1>>8);
      mask1&=onBoardMask.getMask(1);
      if (mask0!=0){
	if(mask1!=0){
	  return true;
	}
	else{
	  int num=bsf(mask0);
	  if((mask0 & ~(1<<num))==0){
	    attackerPiece=getPieceOf(num);
	    return true;
	  }
	  else{
	    return true;
	  }
	}
      }
      else if (mask1!=0){
	int num=bsf(mask1);
	if((mask1 & ~(1<<num))==0){
	  attackerPiece=getPieceOf(num+32);
	  return true;
	}
	return true;
      }
      else return false;
    }
    /**
     * μ꤬꤫ɤ
     * Ĥζˤ벦ξattckPieceˤζ
     * ʣζˤ벦ξPIECE_EMPTY
     * @param P(template) - ߤμ(ä, movealt(P))
     */
    template<Player P>
    bool lastMoveIsCheck(Move lastMove,Piece &attackPiece) const{
      return hasEffectBy<PlayerTraits<P>::opponent>
	(getPosition(getPieceOf(KingTraits<P>::index)),attackPiece);
    }
  };

  template<class State>
  template<Player P,class Action>
  void NumSimpleEffect<State>::forSortedEffect(Position pos,Action & action) const {
    forEachEffectPtype<P,PAWN,Action>(pos,action);
    forEachEffectPtype<P,LANCE,Action>(pos,action);
    forEachEffectPtype<P,KNIGHT,Action>(pos,action);
    forEachEffectPtype<P,SILVER,Action>(pos,action);
    forEachEffectPtype<P,GOLD,Action>(pos,action);
    forEachEffectPtype<P,BISHOP,Action>(pos,action);
    forEachEffectPtype<P,ROOK,Action>(pos,action);
    forEachEffectPtype<P,KING,Action>(pos,action);
  }

  template<class State>
  inline bool operator!=(const NumSimpleEffect<State>& s1,
			 const NumSimpleEffect<State>& s2){
    return !(s1==s2);
  }

  template<Player P,typename BaseState>
  struct ApplyDoUndoSimpleMove<P,NumSimpleEffect<BaseState> >
  {
    typedef NumSimpleEffect<BaseState> state_t;
    template <typename F>
    static void doUndoSimpleMove(state_t& s, 
				 Position from, Position to, int promoteMask,F& func);

    static 
    void prologue(state_t& s, Position from, Position to, int promoteMask,
		  Piece& oldPiece, int& num, 
		  PtypeO& oldPtypeO, PtypeO& newPtypeO)
    {
      oldPiece=s.getPieceAt(from);
      Piece newPiece=promoteWithMask(oldPiece,promoteMask);
      newPiece+=(to-from);
      num=getNum(oldPiece);

      oldPtypeO=getPtypeO(oldPiece);
      newPtypeO=getPtypeO(newPiece);
      // ʬȤθ򳰤
      s.setPieceOf(num,newPiece);
      s.effects.xorEffect(s,oldPtypeO,from,num);
      // ʬȤ֥åƤpromoteαĹ
      // 뤤ϼʬȤΥ֥å
      s.setBoard(to,newPiece);
      s.effects.xorBlockAt(s,to,newPtypeO);
      s.setBoard(from,PIECE_EMPTY);
      s.effects.xorBlockAt(s,from,oldPtypeO);
      // 
      s.effects.xorEffect(s,newPtypeO,to,num);
    }
    static 
    void epilogue(state_t& s, Position from, Position to, Piece oldPiece, 
		  int num, PtypeO oldPtypeO, PtypeO newPtypeO)
    {
      s.setPieceOf(num,oldPiece);
      s.effects.xorEffect(s,newPtypeO,to,num);
      s.setBoard(from,oldPiece);
      s.effects.xorBlockAt(s,from,oldPtypeO);
      s.setBoard(to,PIECE_EMPTY);
      s.effects.xorBlockAt(s,to,newPtypeO);
      // 
      s.effects.xorEffect(s,oldPtypeO,from,num);
    }
  };
  
  
  template<Player P,typename BaseState>
  template <typename F>
  void ApplyDoUndoSimpleMove<P,NumSimpleEffect<BaseState> >::
  doUndoSimpleMove(state_t& s, 
		   Position from, Position to, int promoteMask,F& func)
  {
    Piece oldPiece;
    int num;
    PtypeO oldPtypeO, newPtypeO;
    prologue(s, from, to, promoteMask, oldPiece, num, oldPtypeO, newPtypeO);

    if (promoteMask!=0 && num < PtypeTraits<PAWN>::indexLimit)
      {
	s.clearPawn(P,from);
	s.changeTurn();
	func(to);
	s.changeTurn();
	s.setPawn(P,from);
      }
    else
      {
	s.changeTurn();
	func(to);
	s.changeTurn();
      }

    epilogue(s, from, to, oldPiece, num, oldPtypeO, newPtypeO);
  }

  template<Player P,typename BaseState>
  struct ApplyDoUndoDropMove<P,NumSimpleEffect<BaseState> >
  {
    typedef NumSimpleEffect<BaseState> state_t;
    template <typename F>
    static void doUndoDropMove(state_t& s, 
			       Position to, Ptype ptype,F func);
    static 
    void prologue(state_t& s, Position to, Ptype ptype,
		  Piece& oldPiece, int& num, PtypeO& ptypeO, 
		  int& numIndex, int& numLow)
    {
      numIndex=ptypeTable.getIndex(ptype);
      const unsigned int ownMochigoma=
	s.getOffBoardMask(P).getMask(numIndex) & ptypeTable.getMaskLow(ptype);
      assert(ownMochigoma!=0);
      numLow = bsf(ownMochigoma);
      num = numLow|(numIndex<<5);
      oldPiece=s.getPieceOf(num);
      Piece newPiece=oldPiece;
      newPiece+=to-OFFBOARD;
      ptypeO=getPtypeO(newPiece);
      s.setPieceOf(num,newPiece);
      s.effects.xorEffect(s,ptypeO,to,num);
      s.setBoard(to,newPiece);
      s.effects.xorBlockAt(s,to,ptypeO);
      s.getOffBoardMask(P).xorMask(numIndex,1<<numLow);
      s.getOnBoardMask(P).xorMask(numIndex,1<<numLow);
    }

    static 
    void epilogue(state_t& s, Position to, Piece oldPiece, 
		  int num, PtypeO ptypeO, int numIndex, int numLow)
    {
      s.getOffBoardMask(P).xorMask(numIndex,1<<numLow);
      s.getOnBoardMask(P).xorMask(numIndex,1<<numLow);
      s.setBoard(to,PIECE_EMPTY);
      s.effects.xorEffect(s,ptypeO,to,num);
      s.effects.xorBlockAt(s,to,ptypeO);
      s.setPieceOf(num,oldPiece);
    }
  };
  
  template<Player P,typename BaseState>
  template <typename F>
  void ApplyDoUndoDropMove<P,NumSimpleEffect<BaseState> >::
  doUndoDropMove(state_t& s, 
		 Position to, Ptype ptype,F func)
  {
    Piece oldPiece;
    PtypeO ptypeO;
    int num, numIndex, numLow;
    prologue(s, to, ptype, oldPiece, num, ptypeO, numIndex, numLow);

    if (ptype==PAWN)
      {
	s.setPawn(P,to);
	s.changeTurn();
	func(to);
	s.changeTurn();
	s.clearPawn(P,to);
      }
    else
      {
	s.changeTurn();
	func(to);
	s.changeTurn();
      }
    epilogue(s, to, oldPiece, num, ptypeO, numIndex, numLow);
    
  }

  template<Player P,typename BaseState>
  struct ApplyDoUndoCaptureMove<P,NumSimpleEffect<BaseState> >
  {
    typedef NumSimpleEffect<BaseState> state_t;
    template <typename F>
    static void doUndoCaptureMove(state_t& s, Position from,Position to, 
				  Piece target, int promoteMask,F func);
    static
    void prologue(state_t& s, Position from, Position to, Piece target, 
		  int promoteMask,
		  Piece& oldPiece, PtypeO& oldPtypeO, PtypeO& capturePtypeO, 
		  PtypeO& newPtypeO, int& num0, int& num1, 
		  int& num1Index, int& num1Mask)
    {
      num1=getNum(target);
      num1Index=numToIndex(num1);
      num1Mask=numToMask(num1);
      s.getOnBoardMask(PlayerTraits<P>::opponent).xorMask(num1Index,num1Mask);
      s.getOffBoardMask(P).xorMask(num1Index,num1Mask);
      oldPiece=s.getPieceAt(from);
      Piece newPiece=promoteWithMask(oldPiece,promoteMask);
      newPiece+=(to-from);
      num0=getNum(oldPiece);
      s.setPieceOf(num0,newPiece);
      s.setPieceOf(num1,captured(target));
      
      oldPtypeO=getPtypeO(oldPiece);
      newPtypeO=getPtypeO(newPiece);
      capturePtypeO=getPtypeO(target);
      s.effects.xorEffect(s,capturePtypeO,to,num1);
      s.effects.xorEffect(s,oldPtypeO,from,num0);
      s.setBoard(to,newPiece);
      s.effects.xorBlockAt(s,to,capturePtypeO);
      s.effects.xorBlockAt(s,to,newPtypeO);
      s.setBoard(from,PIECE_EMPTY);
      s.effects.xorBlockAt(s,from,oldPtypeO);
      s.effects.xorEffect(s,newPtypeO,to,num0);
    }
    static
    void epilogue(state_t& s, Position from, Position to, Piece target, 
		  Piece oldPiece, PtypeO oldPtypeO, PtypeO capturePtypeO, 
		  PtypeO newPtypeO, int num0, int num1, 
		  int num1Index, int num1Mask)
    {
      s.getOffBoardMask(P).xorMask(num1Index,num1Mask);
      s.getOnBoardMask(PlayerTraits<P>::opponent).xorMask(num1Index,num1Mask);
      s.setPieceOf(num0,oldPiece);
      s.setPieceOf(num1,target);
      s.effects.xorEffect(s,newPtypeO,to,num0);
      s.effects.xorEffect(s,capturePtypeO,to,num1);
      s.setBoard(from,oldPiece);
      s.effects.xorBlockAt(s,from,oldPtypeO);
      s.setBoard(to,target);
      s.effects.xorBlockAt(s,to,newPtypeO);
      s.effects.xorBlockAt(s,to,capturePtypeO);
      s.effects.xorEffect(s,oldPtypeO,from,num0);
    }
    
  };
  
  template<Player P,typename BaseState>
  template <typename F>
  void ApplyDoUndoCaptureMove<P,NumSimpleEffect<BaseState> >::
  doUndoCaptureMove(state_t& s, Position from,Position to, Piece target, 
		    int promoteMask,F func)
  {
    Piece oldPiece;
    PtypeO oldPtypeO, capturePtypeO, newPtypeO;
    int num0, num1, num1Index, num1Mask;
    prologue(s, from, to, target, promoteMask, oldPiece, oldPtypeO, 
	     capturePtypeO, newPtypeO, num0, num1, num1Index,num1Mask);

    s.changeTurn();
    const Ptype capturePtype=getPtype(target);
    if (capturePtype==PAWN)
      {
	s.clearPawn(PlayerTraits<P>::opponent,to);
	if (promoteMask!=0 && num0<PtypeTraits<PAWN>::indexLimit)
	  {
	    s.clearPawn(P,from);
	    func(to);
	    s.setPawn(P,from);
	  }
	else
	  {
	    func(to);
	  }
	s.setPawn(PlayerTraits<P>::opponent,to);
      }
    else if (promoteMask!=0 && num0<PtypeTraits<PAWN>::indexLimit)
      {
	s.clearPawn(P,from);
	func(to);
	s.setPawn(P,from);
      }
    else
      {
	func(to);
      }
    s.changeTurn();

    epilogue(s, from, to, target, oldPiece, oldPtypeO, capturePtypeO, newPtypeO, 
	     num0, num1, num1Index,num1Mask);
  }

}
#endif // _NUM_SIMPLE_EFFECT_H
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
