#ifndef _NUM_SIMPLE_EFFECT_TCC
#define _NUM_SIMPLE_EFFECT_TCC

#include "numSimpleEffect.h"
namespace osl
{
  template<class State>
  void NumSimpleEffectTable<State>::xorBlockAt(const State& state,Position pos,PtypeO ptypeo){
    //EffectPass effectPass=ptypeOToEffectPass.getEffectPass(ptypeo);
    unsigned int mask1=(effects[pos].getMask(1))>>8;
    while (mask1!=0){
      int num32=bsf(mask1);
      int mask3= 1<<num32;
      int num=num32+32;
      mask1&= ~mask3;
      Piece p1=state.getPieceOf(num);
      //	Player pl1=getOwner(p1);
      assert(getPtype(p1)!=PPAWN);
      Position pos1=getPosition(p1);
      Direction dir0=directionTable.getLongDirection<BLACK>(positionDiff(pos,pos1));
      Offset offset0=directionTable.getOffsetForBlack(dir0);
      assert(offset0!=0);
      int mask4Long=(1<<(num32+8));
      Position pos2=pos+offset0;
      effects[pos2].xorMask(1,mask4Long);
      for(;;){
	if (!isEmpty(state.getPieceAt(pos2))) break;
	pos2+=offset0;
	effects[pos2].xorMask(1,mask4Long);
      }
    }
  }
  template<class State>
  void NumSimpleEffectTable<State>::xorEffect(const State& state,PtypeO ptypeo,Position pos,int num){
    switch(ptypeo){
    case NEW_PTYPEO(WHITE,PAWN): xorEffectBy<WHITE,PAWN>(state,pos,num); break;
    case NEW_PTYPEO(WHITE,LANCE): xorEffectBy<WHITE,LANCE>(state,pos,num); break;
    case NEW_PTYPEO(WHITE,KNIGHT): xorEffectBy<WHITE,KNIGHT>(state,pos,num); break;
    case NEW_PTYPEO(WHITE,SILVER): xorEffectBy<WHITE,SILVER>(state,pos,num); break;
    case NEW_PTYPEO(WHITE,PPAWN):
    case NEW_PTYPEO(WHITE,PLANCE):
    case NEW_PTYPEO(WHITE,PKNIGHT):
    case NEW_PTYPEO(WHITE,PSILVER):
    case NEW_PTYPEO(WHITE,GOLD): xorEffectBy<WHITE,GOLD>(state,pos,num); break;
    case NEW_PTYPEO(WHITE,BISHOP): xorEffectBy<WHITE,BISHOP>(state,pos,num); break;
    case NEW_PTYPEO(WHITE,PBISHOP): xorEffectBy<WHITE,PBISHOP>(state,pos,num); break;
    case NEW_PTYPEO(WHITE,ROOK): xorEffectBy<WHITE,ROOK>(state,pos,num); break;
    case NEW_PTYPEO(WHITE,PROOK): xorEffectBy<WHITE,PROOK>(state,pos,num); break;
    case NEW_PTYPEO(WHITE,KING): xorEffectBy<WHITE,KING>(state,pos,num); break;
    case NEW_PTYPEO(BLACK,PAWN): xorEffectBy<BLACK,PAWN>(state,pos,num); break;
    case NEW_PTYPEO(BLACK,LANCE): xorEffectBy<BLACK,LANCE>(state,pos,num); break;
    case NEW_PTYPEO(BLACK,KNIGHT): xorEffectBy<BLACK,KNIGHT>(state,pos,num); break;
    case NEW_PTYPEO(BLACK,SILVER): xorEffectBy<BLACK,SILVER>(state,pos,num); break;
    case NEW_PTYPEO(BLACK,PPAWN):
    case NEW_PTYPEO(BLACK,PLANCE):
    case NEW_PTYPEO(BLACK,PKNIGHT):
    case NEW_PTYPEO(BLACK,PSILVER):
    case NEW_PTYPEO(BLACK,GOLD): xorEffectBy<BLACK,GOLD>(state,pos,num); break;
    case NEW_PTYPEO(BLACK,BISHOP): xorEffectBy<BLACK,BISHOP>(state,pos,num); break;
    case NEW_PTYPEO(BLACK,PBISHOP): xorEffectBy<BLACK,PBISHOP>(state,pos,num); break;
    case NEW_PTYPEO(BLACK,ROOK): xorEffectBy<BLACK,ROOK>(state,pos,num); break;
    case NEW_PTYPEO(BLACK,PROOK): xorEffectBy<BLACK,PROOK>(state,pos,num); break;
    case NEW_PTYPEO(BLACK,KING): xorEffectBy<BLACK,KING>(state,pos,num); break;
    default: assert(0);
    }
    return;
  }
  template<class State>
  template<Player P,Ptype T>
  void NumSimpleEffectTable<State>::xorEffectBy(const State& state,Position pos,int num){
    xorEffectShort<P,T,UL>(pos,num);
    xorEffectShort<P,T,U>(pos,num);
    xorEffectShort<P,T,UR>(pos,num);
    xorEffectShort<P,T,L>(pos,num);
    xorEffectShort<P,T,R>(pos,num);
    xorEffectShort<P,T,DL>(pos,num);
    xorEffectShort<P,T,D>(pos,num);
    xorEffectShort<P,T,DR>(pos,num);
    xorEffectShort<P,T,UUL>(pos,num);
    xorEffectShort<P,T,UUR>(pos,num);
    xorEffectLong<P,T,LONG_UL>(state,pos,num);
    xorEffectLong<P,T,LONG_U>(state,pos,num);
    xorEffectLong<P,T,LONG_UR>(state,pos,num);
    xorEffectLong<P,T,LONG_L>(state,pos,num);
    xorEffectLong<P,T,LONG_R>(state,pos,num);
    xorEffectLong<P,T,LONG_DL>(state,pos,num);
    xorEffectLong<P,T,LONG_D>(state,pos,num);
    xorEffectLong<P,T,LONG_DR>(state,pos,num);
  }
  template<class State>
  void NumSimpleEffectTable<State>::init(const State& state){
    std::fill(effects.begin(), effects.end(), NumBitmapEffect());
    for(int num=0;num<40;num++){
      if (state.isOnBoardNum(num)){
	xorEffect(state,state.getPieceOf(num));
      }
    }
  }
  template<class State>
  bool operator==(const NumSimpleEffectTable<State>& et1,const NumSimpleEffectTable<State>& et2){
    for(int y=1;y<=9;y++)
      for(int x=9;x>0;x--){
	Position pos=newPosition(x,y);
	if (!(et1.getEffect(pos)==et2.getEffect(pos))) return false;
      }
    return true;
  }

  template<class State>
  std::ostream& operator<<(std::ostream& os,const NumSimpleEffectTable<State> & effectTable){
    os << "EffectPass" << std::endl;
    for(int y=1;y<=9;y++){
      for(int x=9;x>0;x--){
	Position pos=newPosition(x,y);
	os << static_cast<unsigned int>(effectTable.getEffectPass(pos)) << " ";
      }
      os << std::endl;
    }
    os << "Effect" << std::endl;
    for(int y=1;y<=9;y++){
      for(int x=9;x>0;x--){
	Position pos=newPosition(x,y);
	os << effectTable.getEffect(pos) << " ";
      }
      os << std::endl;
    }
    return os;
  }

  template<class State>
  bool operator==(const NumSimpleEffect<State>& st1,
		  const NumSimpleEffect<State>& st2){
    //      if (!(s1.effects == s2.effects)) return false;
    // return true;
    assert(st1.isConsistent(false));
    assert(st2.isConsistent(false));
    if (st1.getTurn()!=st2.getTurn()) return false;
#if 0
    if (st1.getPawnMask(0)!=st2.getPawnMask(0)) return false;
    if (st1.getPawnMask(1)!=st2.getPawnMask(1)) return false;
#endif
    for(int y=1;y<=9;y++)
      for(int x=9;x>0;x--){
	Piece p1=st1.getPieceAt(newPosition(x,y));
	Piece p2=st2.getPieceAt(newPosition(x,y));
	if (getPtypeO(p1)!=getPtypeO(p2)) return false;
      }
    return true;
  }

  template<class State>
  NumSimpleEffect<State>::NumSimpleEffect(const State& st) 
    : State(st),effects(st){
    onBoardMask[0].clearAll();
    onBoardMask[1].clearAll();
    for(int num=0;num<40;num++){
      Piece p=getPieceOf(num);
      if (isOnBoard(p)){
	getOnBoardMask(getOwner(p)).setNum(num);
      }
    }
  }

  template<class State>
  void NumSimpleEffect<State>::
  doSimpleMove(Position from, Position to, int promoteMask)
  {
    Piece oldPiece;
    int num;
    PtypeO oldPtypeO, newPtypeO;
    if(getTurn()==BLACK){
      ApplyDoUndoSimpleMove<BLACK,NumSimpleEffect<State> >
	::prologue(*this, from, to, promoteMask, 
		   oldPiece, num, oldPtypeO, newPtypeO);
    }
    else{
      ApplyDoUndoSimpleMove<WHITE,NumSimpleEffect<State> >
	::prologue(*this, from, to, promoteMask, 
		   oldPiece, num, oldPtypeO, newPtypeO);
    }
    if (promoteMask!=0 && num < PtypeTraits<PAWN>::indexLimit)
      clearPawn(getTurn(),from);
  }
  template<class State>
  void NumSimpleEffect<State>::
  doCaptureMove(Position from, Position to, Piece target, int promoteMask)
  {
    Piece oldPiece;
    PtypeO oldPtypeO, capturePtypeO, newPtypeO;
    int num0, num1, num1Index, num1Mask;
    if(getTurn()==BLACK){
      ApplyDoUndoCaptureMove<BLACK,NumSimpleEffect<State> >
	::prologue(*this, from, to, target, promoteMask, oldPiece, oldPtypeO, 
		   capturePtypeO, newPtypeO, num0, num1, num1Index,num1Mask);
    }
    else{
      ApplyDoUndoCaptureMove<WHITE,NumSimpleEffect<State> >
	::prologue(*this, from, to, target, promoteMask, oldPiece, oldPtypeO, 
		   capturePtypeO, newPtypeO, num0, num1, num1Index,num1Mask);
    }
    const Ptype capturePtype=getPtype(target);
    if (capturePtype==PAWN)
      clearPawn(alt(getTurn()),to);
    if (promoteMask!=0 && num0<PtypeTraits<PAWN>::indexLimit)
      clearPawn(getTurn(),from);
  }
  template <class State>
  void NumSimpleEffect<State>::
  doDropMove(Position to,Ptype ptype)
  {
    Piece oldPiece;
    PtypeO ptypeO;
    int num, numIndex, numLow;
    if(getTurn()==BLACK){
      ApplyDoUndoDropMove<BLACK,NumSimpleEffect<State> >
	::prologue(*this, to, ptype, oldPiece, num, ptypeO, numIndex, numLow);
    }
    else{
      ApplyDoUndoDropMove<WHITE,NumSimpleEffect<State> >
	::prologue(*this, to, ptype, oldPiece, num, ptypeO, numIndex, numLow);
    }
    if (ptype==PAWN)
      setPawn(getTurn(),to);
  }
  template<class State>
  bool NumSimpleEffect<State>::isConsistent(bool showError) const
  {
    if (!State::isConsistent(showError)) 
    {
      if (showError)
	std::cerr << "error before effect\n";
      return false;
    }
    NumSimpleEffectTable<State> effects1(*this);
    if (!(effects1==effects))
    {
      if (showError)
      {
	std::cerr << "Effect error 1" << std::endl;
	std::cerr << *this;
	for(int y=1;y<=9;y++)
	  for(int x=9;x>0;x--)
	  {
	    Position pos=newPosition(x,y);
	    if (!(effects1.getEffect(pos)==effects.getEffect(pos)))
	    {
	      std::cerr << pos << ",real=" << effects.getEffect(pos) << ",ideal=" << effects1.getEffect(pos) << std::endl;
	    }
	  }
      }
      return false;
    }
    return true;
  }

  template<class State>
  std::ostream& operator<<(std::ostream& os,const NumSimpleEffect<State>& state)
  {
    os<< static_cast<State const&>(state);
    for(int y=1;y<=9;y++){
      os << 'P' << y;  
      for(int x=9;x>0;x--){
	Position pos=newPosition(x,y);
	csaShow(os,state.getPieceAt(pos));
	os<<state.getEffect(pos);
      }
      os << std::endl;
    }
    // ɽ
    for(int num=0;num<PIECE_SIZE;num++){
      if (state.getOffBoardMask(BLACK).isSet(num)){
	os << "P+00";
	csaShow(os,numTable.getPtype(num));
	os << std::endl;
      }
      else if (state.getOffBoardMask(WHITE).isSet(num)){
	os << "P-00";
	csaShow(os,numTable.getPtype(num));
	os << std::endl;
      }
    }
    return os;
  }

  template <class BaseState>
  class DoUndoMoveLockSimple<NumSimpleEffect<BaseState> >
    : public DoUndoMoveLockCleanUp
  {
    typedef NumSimpleEffect<BaseState> state_t;
    state_t& state;
    Position from, to;
    Piece oldPiece;
    PtypeO oldPtypeO, newPtypeO;
    int num;
    bool pawnPromotion;
  public:
    DoUndoMoveLockSimple(state_t& s, 
			 Position f, Position t, int promoteMask)
      : state(s), from(f), to(t)
    {
      if(state.getTurn()==BLACK){
	ApplyDoUndoSimpleMove<BLACK,NumSimpleEffect<BaseState> >::
	  prologue(s, from, to, promoteMask, oldPiece, num, oldPtypeO, newPtypeO);
      }
      else{
	ApplyDoUndoSimpleMove<WHITE,NumSimpleEffect<BaseState> >::
	  prologue(s, from, to, promoteMask, oldPiece, num, oldPtypeO, newPtypeO);
      }
      pawnPromotion = (promoteMask!=0) && (num < PtypeTraits<PAWN>::indexLimit);
      if (pawnPromotion)
	s.clearPawn(s.getTurn(),from);
    }
    ~DoUndoMoveLockSimple()
    {
      if (pawnPromotion)
	state.setPawn(state.getTurn(),from);
      if(state.getTurn()==BLACK){
	ApplyDoUndoSimpleMove<BLACK,NumSimpleEffect<BaseState> >::
	  epilogue(state, from, to, oldPiece, num, oldPtypeO, newPtypeO);
      }
      else{
	ApplyDoUndoSimpleMove<WHITE,NumSimpleEffect<BaseState> >::
	  epilogue(state, from, to, oldPiece, num, oldPtypeO, newPtypeO);
      }
    }

  };

  template <class BaseState>
  class DoUndoMoveLockDrop<NumSimpleEffect<BaseState> >
    : public DoUndoMoveLockCleanUp
  {
    typedef NumSimpleEffect<BaseState> state_t;
    state_t& state;
    Position to;
    Piece oldPiece;
    PtypeO ptypeO;
    int num, numIndex, numLow;
    bool pawnDrop;
  public:
    DoUndoMoveLockDrop(state_t& s, Position t, Ptype ptype)
      : state(s), to(t)
    {
      if(state.getTurn()==BLACK){
	ApplyDoUndoDropMove<BLACK,NumSimpleEffect<BaseState> >::
	  prologue(s, to, ptype, oldPiece, num, ptypeO, numIndex, numLow);
      }
      else{
	ApplyDoUndoDropMove<WHITE,NumSimpleEffect<BaseState> >::
	  prologue(s, to, ptype, oldPiece, num, ptypeO, numIndex, numLow);
      }
      pawnDrop = (ptype==PAWN);
      if (pawnDrop)
	s.setPawn(s.getTurn(),to);
    }
    ~DoUndoMoveLockDrop()
    {
      if (pawnDrop)
	state.clearPawn(state.getTurn(),to);
      if(state.getTurn()==BLACK){
	ApplyDoUndoDropMove<BLACK,NumSimpleEffect<BaseState> >::
	  epilogue(state, to, oldPiece, num, ptypeO, numIndex, numLow);
      }
      else{
	ApplyDoUndoDropMove<WHITE,NumSimpleEffect<BaseState> >::
	  epilogue(state, to, oldPiece, num, ptypeO, numIndex, numLow);
      }
    }
  };

  template <class BaseState>
  class DoUndoMoveLockCapture<NumSimpleEffect<BaseState> >
    : public DoUndoMoveLockCleanUp
  {
    typedef NumSimpleEffect<BaseState> state_t;
    state_t& state;
    Position from, to;
    Piece target;
    Piece oldPiece;
    PtypeO oldPtypeO, capturePtypeO, newPtypeO;
    int num0, num1;
    int num1Index, num1Mask;
    Ptype capturePtype;
    bool pawnPromotion, pawnCapture;
  public:
    DoUndoMoveLockCapture(state_t& s, Position f, Position t, Piece tg,
			  int promoteMask)
      : state(s), from(f), to(t), target(tg)
    {
      if(state.getTurn()==BLACK){
	ApplyDoUndoCaptureMove<BLACK,NumSimpleEffect<BaseState> >::
	  prologue(s, from, to, target, promoteMask, oldPiece, oldPtypeO, 
		   capturePtypeO, newPtypeO, num0, num1, num1Index,num1Mask);
      }
      else{
	ApplyDoUndoCaptureMove<WHITE,NumSimpleEffect<BaseState> >::
	  prologue(s, from, to, target, promoteMask, oldPiece, oldPtypeO, 
		   capturePtypeO, newPtypeO, num0, num1, num1Index,num1Mask);
      }
      capturePtype=getPtype(target);
      pawnCapture = (capturePtype==PAWN);
      if (pawnCapture)
	s.clearPawn(alt(s.getTurn()),to);
      pawnPromotion = (promoteMask!=0) && (num0<PtypeTraits<PAWN>::indexLimit);
      if (pawnPromotion)
	s.clearPawn(s.getTurn(),from);
    }
    ~DoUndoMoveLockCapture()
    {
      if (pawnPromotion)
	state.setPawn(state.getTurn(),from);
      if (pawnCapture)
	state.setPawn(alt(state.getTurn()),to);
      if(state.getTurn()==BLACK){
	ApplyDoUndoCaptureMove<BLACK,NumSimpleEffect<BaseState> >::
	  epilogue(state, from, to, target, oldPiece, 
		   oldPtypeO, capturePtypeO, newPtypeO, 
		   num0, num1, num1Index,num1Mask);
      }
      else{
	ApplyDoUndoCaptureMove<WHITE,NumSimpleEffect<BaseState> >::
	  epilogue(state, from, to, target, oldPiece, 
		   oldPtypeO, capturePtypeO, newPtypeO, 
		   num0, num1, num1Index,num1Mask);
      }
    }
  };
} // namespace osl

#endif /* _NUM_SIMPLE_EFFECT_TCC */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
