/* PawnMaskState.h
 */
#ifndef _PAWN_MASK_STATE_H
#define _PAWN_MASK_STATE_H

#include "osl/bitXmask.h"
#include "osl/direction.h"
#include "osl/board.h"
#include "osl/ptype.h"
#include "osl/piece.h"
#include "osl/pieceMask.h"
#include "osl/move.h"
#include "osl/player.h"
#include "osl/state.h"
#include "osl/loki.h"
#include "osl/handicap.h"
#include "osl/carray.h"
#include "osl/doUndoMoveLock.h"
#include "osl/simpleState.h"

#include <iosfwd>

namespace osl
{
  template <class State>  struct DoUndoMoveLockSimple;
  template <class State>  struct DoUndoMoveLockCapture;
  template <class State>  struct DoUndoMoveLockDrop;
  class PawnMaskState;
  std::ostream& operator<<(std::ostream& os,const PawnMaskState& state);
  bool operator==(const PawnMaskState& st1,const PawnMaskState& st2);


  /**
   * PawnMask State
   */
  class PawnMaskState : public SimpleState
  {
  public:
    /** Ѿ饹 ĤФ˻Ѥ */
    typedef PawnMaskState pawn_mask_state_t;
    typedef SimpleState simple_state_t;
    typedef SimpleState effect_state_t;
    static const bool hasPawnMask=true;
  private:
    friend std::ostream& operator<<(std::ostream& os,const PawnMaskState& state);
    friend bool operator==(const PawnMaskState& st1,const PawnMaskState& st2);
    typedef PawnMaskState state_t;
    CArray<BitXmask,2> pawnMask;
    
    friend class ApplyDoUndoSimpleMove<BLACK,PawnMaskState>;
    friend class ApplyDoUndoCaptureMove<BLACK,PawnMaskState>;
    friend class ApplyDoUndoDropMove<BLACK,PawnMaskState>;
    friend class ApplyDoUndoSimpleMove<WHITE,PawnMaskState>;
    friend class ApplyDoUndoCaptureMove<WHITE,PawnMaskState>;
    friend class ApplyDoUndoDropMove<WHITE,PawnMaskState>;
    friend class DoUndoMoveLockSimple<PawnMaskState>;
    friend class DoUndoMoveLockCapture<PawnMaskState>;
    friend class DoUndoMoveLockDrop<PawnMaskState>;
  public:
    // ˴ؤ
    explicit PawnMaskState();
    explicit PawnMaskState(Handicap h);
    // publicѾˤϡvirtual destructor롥
    virtual ~PawnMaskState();
    
    /** ̤ξ֤˽ */
    void init();
    /** ϥǥ˱֤˽ */
    void init(Handicap h);

    bool isPawnMaskSet(Player player, int x)const {return isSet(pawnMask[playerToIndex(player)],x);}

    template<Player P>
    bool isPawnMaskSet(int x)const {return isPawnMaskSet(P,x); }

    void clearPawn(Player pl,Position pos){
      clearPos(pawnMask[playerToIndex(pl)],pos);
    }
    void setPawn(Player pl,Position pos){
      setPos(pawnMask[playerToIndex(pl)],pos);
    }

    void setPiece(Player player,Position pos,Ptype ptype);
    void setPieceAll(Player player);

    // check
    bool isConsistent(bool showError=true) const;
    // move methods
    void doSimpleMove(Position from, Position to, int promoteMask);
    void doDropMove(Position to,Ptype ptype);
    void doCaptureMove(Position from, Position to, Piece target,int promoteMask);
    void doMove(Move move);

  };


  template<Player P>
  struct ApplyDoUndoSimpleMove<P,PawnMaskState>
  {
    static void prologue(PawnMaskState& s, 
			 Position from, Position to, int promoteMask,
			 Piece& oldPiece, int& num)
    {
      oldPiece=s.getPieceOnBoard(from);
      Piece newPiece=promoteWithMask(oldPiece,promoteMask);
      newPiece+=(to-from);
      num=getNum(oldPiece);
      s.setPieceOf(num,newPiece);
      s.setBoard(to,newPiece);
      s.setBoard(from,PIECE_EMPTY);
    }
    static void epilogue(PawnMaskState& s, Position from, Position to,
			 Piece oldPiece, int num)
    {
      // ʬȤθ򳰤
      s.setPieceOf(num,oldPiece);
      s.setBoard(from,oldPiece);
      s.setBoard(to,PIECE_EMPTY);
    }
    template <typename F>
    static void doUndoSimpleMove(PawnMaskState& s, 
				 Position from, Position to, int promoteMask,F& func);
  };
  
  template<Player P>
  template <typename F>
  void ApplyDoUndoSimpleMove<P,PawnMaskState>::
  doUndoSimpleMove(PawnMaskState& s, 
		   Position from, Position to, int promoteMask,F& func)
  {
    Piece oldPiece;
    int num;
    prologue(s, from, to, promoteMask, oldPiece, num);
    if (promoteMask!=0 && num < PtypeTraits<PAWN>::indexLimit)
    {
      s.clearPawn(P,from);
      s.changeTurn();
      func(to);
      s.changeTurn();
      s.setPawn(P,from);
    }
    else{
      s.changeTurn();
      func(to);
      s.changeTurn();
    }
    epilogue(s, from, to, oldPiece, num);
  }

  template<Player P>
  struct ApplyDoUndoDropMove<P,PawnMaskState>
  {
    typedef PawnMaskState state_t;

    static void prologue(PawnMaskState& s, Ptype ptype, Position to,
			 Piece& oldPiece, int& num, int& numIndex, int& numLow)
    {
      numIndex=ptypeTable.getIndex(ptype);
      const unsigned int ownMochigoma=
	s.getOffBoardMask(P).getMask(numIndex) & ptypeTable.getMaskLow(ptype);
      assert(ownMochigoma!=0 || (s.dump(), 0));
      numLow=bsf(ownMochigoma);
      num=numLow|(numIndex<<5);
      oldPiece=s.getPieceOf(num);
      Piece p=oldPiece;
      p+=to-OFFBOARD;

      s.setPieceOf(num,p);
      s.setBoard(to,p);
      s.getOffBoardMask(P).xorMask(numIndex,1<<numLow);
    }
    static void epilogue(PawnMaskState& s, Position to,
			 Piece oldPiece, int num,
			 int numIndex, int numLow)
    {
      s.getOffBoardMask(P).xorMask(numIndex,1<<numLow);
      s.setBoard(to,PIECE_EMPTY);
      s.setPieceOf(num,oldPiece);
    }
    template <typename F>
    static void doUndoDropMove(PawnMaskState& s, 
			       Position to, Ptype ptype,F& func);
  };
  
  
  template<Player P>
  template <typename F>
  void ApplyDoUndoDropMove<P,PawnMaskState>::
  doUndoDropMove(PawnMaskState& s, 
		 Position to, Ptype ptype,F& func)
  {
    Piece oldPiece;
    int num, numIndex, numLow;
    prologue(s, ptype, to, oldPiece, num, numIndex, numLow);

    if (ptype==PAWN)
    {
      s.setPawn(P,to);
      s.changeTurn();
      func(to);
      s.changeTurn();
      s.clearPawn(P,to);
    }
    else
    {
      s.changeTurn();
      func(to);
      s.changeTurn();
    }
    epilogue(s, to, oldPiece, num, numIndex, numLow);
  }

  
  template<Player P>
  struct ApplyDoUndoCaptureMove<P,PawnMaskState>
  {
    typedef PawnMaskState state_t;
    template <typename F>
    static void doUndoCaptureMove(state_t& s,
				  Position from,Position to, Piece p1, int promoteMask,F& func);

    static
    void prologue(PawnMaskState& s, 
		  Position from, Position to, Piece target, int promoteMask,
		  Ptype& capturePtype, Piece& oldPiece, int& num0, 
		  int& num1, int& num1Index, int& num1Mask)
    {
      capturePtype=getPtype(target);
      num1=getNum(target);
      num1Index=numToIndex(num1);
      num1Mask=numToMask(num1);

      s.getOffBoardMask(P).xorMask(num1Index,num1Mask);
      oldPiece=s.getPieceOnBoard(from);
      Piece newPiece=promoteWithMask(oldPiece,promoteMask);
      newPiece+=(to-from);
      num0=getNum(oldPiece);
      s.setPieceOf(num0,newPiece);
      s.setPieceOf(num1,captured(target));
      s.setBoard(to,newPiece);
      s.setBoard(from,PIECE_EMPTY);
    }
    static 
    void epilogue(PawnMaskState& s, Position from, Position to, Piece target, 
		  Piece oldPiece, int num0, 
		  int num1, int num1Index, int num1Mask)
    {
      s.getOffBoardMask(P).xorMask(num1Index,num1Mask);
      s.setPieceOf(num0,oldPiece);
      s.setPieceOf(num1,target);
      s.setBoard(from,oldPiece);
      s.setBoard(to,target);
    }
  };
  
  template<Player P>
  template <typename F>
  void ApplyDoUndoCaptureMove<P,PawnMaskState>::
  doUndoCaptureMove(state_t& s,
		    Position from,Position to, Piece target, int promoteMask,F& func)
  {
    Piece oldPiece;
    int num0, num1;
    int num1Index, num1Mask;
    Ptype capturePtype;
    prologue(s, from, to, target, promoteMask,
	     capturePtype, oldPiece, num0, num1, num1Index, num1Mask);

    s.changeTurn();
    if (capturePtype==PAWN){
      s.clearPawn(PlayerTraits<P>::opponent,to);
      if (promoteMask!=0 && num0<PtypeTraits<PAWN>::indexLimit){
	s.clearPawn(P,from);
	func(to);
	s.setPawn(P,from);
      }
      else{
	func(to);
      }
      s.setPawn(PlayerTraits<P>::opponent,to);
    }
    else if (promoteMask!=0 && num0<PtypeTraits<PAWN>::indexLimit){
      s.clearPawn(P,from);
      func(to);
      s.setPawn(P,from);
    }
    else{
      func(to);
    }
    s.changeTurn();

    epilogue(s, from, to, target, oldPiece, num0, num1, num1Index, num1Mask);
  }
  class DoUndoMoveLockSimple<PawnMaskState> : public DoUndoMoveLockCleanUp
  {
    PawnMaskState& state;
    Position from, to;
    Piece oldPiece;
    int num;
    bool clearPawn;
  public:
    DoUndoMoveLockSimple(PawnMaskState& s, 
			 Position f, Position t, int promoteMask);
    ~DoUndoMoveLockSimple();
  };

  class DoUndoMoveLockDrop<PawnMaskState> : public DoUndoMoveLockCleanUp
  {
    PawnMaskState& state;
    Position to;
    Piece oldPiece;
    int num, numIndex, numLow;
    bool pawnDrop;
  public:
    DoUndoMoveLockDrop(PawnMaskState& state, 
		       Position to, Ptype ptype);
    ~DoUndoMoveLockDrop();
  };
  
  class DoUndoMoveLockCapture<PawnMaskState> : public DoUndoMoveLockCleanUp
  {
    PawnMaskState& state;
    Position from, to;
    Piece target;
    Piece oldPiece;
    int num0, num1;
    int num1Index, num1Mask;
    Ptype capturePtype;
    bool pawnPromotion, pawnCapture;
  public:
    DoUndoMoveLockCapture(PawnMaskState& state, 
			  Position from, Position to, Piece target,
			  int promoteMask);
    ~DoUndoMoveLockCapture();
  };



} // namespace osl

#endif /* _PAWN_MASK_STATE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
