#ifndef _PIECE_H
#define _PIECE_H
#include "osl/bitOp.h"
#include "osl/board.h"
#include "osl/player.h"
#include "osl/ptype.h"

namespace osl{

#define NEW_PIECE(owner,ptype,num,position)\
  ((static_cast<int>(owner)<<20)|(static_cast<int>(ptype)<<16)|((num)<<8)|\
   static_cast<int>(position))

  /**
   * ptypeo(-15 - 15), ֹ(0-39), ݥ(0-0xff)ʤ 
   * 16 bitptypeo, 8bitֹ, 8bitǥݥȤƤߤ?/
   * ɤˤ 16 bitǤϼޤʤ
   */
  enum Piece{
    PIECE_EMPTY=NEW_PIECE(BLACK,PTYPE_EMPTY,0,OFFBOARD),
    PIECE_EDGE=NEW_PIECE(WHITE,PTYPE_EDGE,0,OFFBOARD),
    /** 32bit ݤ뤿Τޤʤ */
    PIECE_MIN= -0x7fffffff,
    /** 32bit ݤ뤿Τޤʤ */
    PIECE_MAX=  0x7fffffff,
  };

  static inline Piece newPiece(Player owner,Ptype ptype,int num,Position position){
    return static_cast<Piece>(NEW_PIECE(owner,ptype,num,position));
  }

  static inline Ptype getPtype(Piece piece){
    return static_cast<Ptype>((static_cast<int>(piece)>>16)&0xf);
  }

  static inline PtypeO getPtypeO(Piece piece){
    return static_cast<PtypeO>(static_cast<int>(piece)>>16);
  }

  /**
   * pgetPtypeOptypeOȰפƤ0
   * פƤʤä֤ͤ
   */
  static inline unsigned int isPtypeOMask(Piece p,PtypeO ptypeO){
    int ret= -((static_cast<int>(p)&0x00ff0000)^
	       ((static_cast<int>(ptypeO)<<16)&0x00ff0000));
    assert( (getPtypeO(p)==ptypeO && ret==0) ||
	    (getPtypeO(p)!=ptypeO && ret<0));
    return static_cast<unsigned int>(ret);
  }

  static inline int getNum(Piece piece){
    return (static_cast<int>(piece)>>8)&0xff;
  }

  static inline Position getPosition(Piece piece){
    return static_cast<Position>(static_cast<int>(piece)&0xff);
  }

  // must be offset rather than position
  inline Piece& operator+=(Piece& piece,Offset offset){
    piece=static_cast<Piece>(static_cast<int>(piece)+static_cast<int>(offset));
    return piece;
  }

  static inline Piece setPosition(Piece piece,int position){
    return static_cast<Piece>((static_cast<int>(piece)&0xfffffe00)+position);
  }

  static inline bool isOnBoard(Piece piece){
    return !isOffBoard(getPosition(piece));
  }

  //
  template<Player P>
  static inline bool isOnBoardBy(Piece piece);


  template<> inline
  bool isOnBoardBy<BLACK>(Piece piece){
    return static_cast<int>(static_cast<unsigned int>(piece)&0xfff000ff)>0;
  }

  template<> inline
  bool isOnBoardBy<WHITE>(Piece piece){
    return (static_cast<unsigned int>(piece)&0xfff000ff)>0xfff00000;
  }


  static inline bool validNum(int num){
    return 0<=num && num<=39;
  }

  /* PROMOTEԲĤpieceŬѤȥХФ */
  static inline Piece promote(Piece piece){
    return static_cast<Piece>(static_cast<int>(piece)-0x80000);
  }

  /* PROMOTEԲĤpieceŬѤƤOK  */
  static inline Piece unpromote(Piece piece){
    return (Piece)((int)piece|0x80000);
  }

  /**
   * 줿piece unpromoteơPosition0
   * λΤˤʤ
   */
  static inline Piece captured(Piece piece){
    // return (Piece)((((int)piece|0x80000)&0xffffff00)^0xfff00000);
    // optimize
    return static_cast<Piece>((static_cast<int>(piece)&0xfff7ff00)^0xfff80000);
  }

  static inline Piece unCaptured(Piece piece,Position to,Ptype ptype){
    piece+=to-OFFBOARD;
    return static_cast<Piece>((static_cast<int>(piece)^0xfff00000)&((static_cast<int>(ptype)<<(19-3))|0xfff7ffff));
  }


  static inline Piece promoteWithMask(Piece piece,int promoteMask){
    return static_cast<Piece>(static_cast<int>(piece)-promoteMask);
  }
  static inline Piece unpromoteWithMask(Piece piece,int promoteMask){
    return static_cast<Piece>(static_cast<int>(piece)+promoteMask);
  }

  static inline Piece checkPromote(Piece piece,int promotep){
    return promoteWithMask(piece,promotep<<19);
  }

  static inline Piece checkUnpromote(Piece piece,int promotep){
    return unpromoteWithMask(piece,promotep<<19);
  }

  /**
   * promoteǽʼζpromoteƤ뤫ɤå
   * @param piece - å(, ʳ)
   */
  static inline bool isPromoted(Piece piece){
    assert(getPtype(piece)!=KING && getPtype(piece)!=GOLD);
    return (static_cast<int>(piece)&(1<<19))==0;
  }



  static inline bool isEmpty(Piece piece){
    return piece==PIECE_EMPTY;
  }

  static inline bool isEdge(Piece piece){
    return piece==PIECE_EDGE;
  }

  static inline bool isPiece(Piece piece){
    return !isEmpty(piece) && !isEdge(piece);
  }

  static inline Player getOwner(Piece piece){
    assert(isPiece(piece));
    return static_cast<Player>(static_cast<int>(piece)>>20);
  }

  /** Player Pζ𤬡board[]p˰ưǤ뤫? */
  template<Player P>
  bool canMoveOn(Piece p);

  template<>
  inline bool canMoveOn<BLACK>(Piece p){
    // std::cerr << "canMoveOn<BLACK>" << static_cast<unsigned int>(PIECE_EDGE) 
    //	  << "," <<  (static_cast<unsigned int>(p)-1) << std::endl;
    return static_cast<unsigned int>(PIECE_EDGE)<(static_cast<unsigned int>(p)-1);
  }

  template<>
  inline bool canMoveOn<WHITE>(Piece p){
    //std::cerr << "canMoveOn<WHITE>" << static_cast<int>(PIECE_EMPTY) 
    //	  << "," <<  static_cast<int>(p) << std::endl;
    return static_cast<int>(PIECE_EMPTY)<=static_cast<int>(p);
  }

  std::ostream& operator<<(std::ostream& os,const Piece piece);
  std::ostream& csaShow(std::ostream& os,const Piece piece);

  static const int PIECE_SIZE=40;

  class NumTable{
  private:
    CArray<Ptype, PIECE_SIZE> ptypes;
    template<Ptype T>
    void initPtype();
  public:
    NumTable();
    Ptype getPtype(int num) const{
      assert(validNum(num));
      return ptypes[num];
    }
  };

  extern const NumTable numTable;

  // Note: NumToPtypeTable and NumToMaskLowTable
  // were moved into piece.cc because they are in #if 0 ... #endif

  extern const CArray<unsigned int, 16> ptypeToNum;
}

#endif /* _PIECE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
