/* pieceEval.h
 */
#ifndef _PIECEEVAL_H
#define _PIECEEVAL_H

#include "osl/piece.h"
#include "osl/ptype.h"
#include "osl/move.h"
#include "osl/evalTraits.h"
#include "osl/loki.h"
#include "osl/carray.h"
#include "osl/simpleState.h"
#include <cassert>

// see also pieceEval.tcc
namespace osl
{
  template<Ptype T>
  struct PtypeEvalTraits;
  
  template<>
  struct PtypeEvalTraits<PAWN>{
    static const int val=100;
  };

  template<>
  struct PtypeEvalTraits<PPAWN>{
    static const int val=600;
  };

  template<>
  struct PtypeEvalTraits<LANCE>{
    static const int val=400;
  };

  template<>
  struct PtypeEvalTraits<PLANCE>{
    static const int val=600;
  };

  template<>
  struct PtypeEvalTraits<KNIGHT>{
    static const int val=400;
  };

  template<>
  struct PtypeEvalTraits<PKNIGHT>{
    static const int val=600;
  };

  template<>
  struct PtypeEvalTraits<SILVER>{
    static const int val=550;
  };

  template<>
  struct PtypeEvalTraits<PSILVER>{
    static const int val=600;
  };

  template<>
  struct PtypeEvalTraits<GOLD>{
    static const int val=600;
  };

  template<>
  struct PtypeEvalTraits<BISHOP>{
    static const int val=800;
  };

  template<>
  struct PtypeEvalTraits<PBISHOP>{
    static const int val=1150;
  };

  template<>
  struct PtypeEvalTraits<ROOK>{
    static const int val=950;
  };

  template<>
  struct PtypeEvalTraits<PROOK>{
    static const int val=1300;
  };

  template<>
  struct PtypeEvalTraits<KING>{
    static const int val=10000;
  };
  
  class PtypeEvalTable
  {
  private:
    CArray<int, PTYPEO_SIZE> ptypeO2Val;
    CArray<int, PTYPEO_SIZE> promoteVal;
    CArray<int, PTYPEO_SIZE> captureVal;
    CArray<int, 2> delta;
  public:
    PtypeEvalTable();
    ~PtypeEvalTable();
    void init();
    /**
     * ꤫鸫ptypeOζβ
     */
    int getVal(PtypeO ptypeO) const{
      assert(isValid(ptypeO));
      return ptypeO2Val[ptypeO-PTYPEO_MIN];
    }
    /**
     * ptypeζβ
     */
    int getVal(Ptype ptype) const{
      assert(isValid(ptype));
      return ptypeO2Val[ptype-PTYPEO_MIN];
    }
    /**
     * playerˤȤäƤäȹ⤤
     */
    int getDelta(Player player) const{
      assert(isValid(player));
      return delta[playerToIndex(player)];
    }
    /**
     * ptypeOpromoteɾͤ
     */
    int getPromoteVal(PtypeO ptypeO) const{
      assert(isPromoted(ptypeO));
      return promoteVal[ptypeO-PTYPEO_MIN];
    }
    /**
     * ownerptypeOcapture줿ɾͤ
     */
    int getCaptureVal(PtypeO ptypeO) const{
      assert(isValid(ptypeO));
      return captureVal[ptypeO-PTYPEO_MIN];
    }
  };
  extern const PtypeEvalTable ptypeEvalTable;

  /**
   * βͥ١ɾؿ
   * ɬ
   * ͭ +, ͭ -
   * 100
   */
  class PieceEval
  {
    int val;
  public:
    explicit PieceEval(const SimpleState& state);
    explicit PieceEval(int v) : val(v) {}
    int getVal() const
    { 
      assert(isEvalValueConsistent(val)); 
      return val; 
    }
    template <class State>
    static int diffWithMove(const State&, Move move)
    {
      const PtypeO capturePtypeO=newPtypeO(alt(getPlayer(move)),
				     getCapturePtype(move));
      int retVal=ptypeEvalTable.getCaptureVal(capturePtypeO);
      if (isPromote(move))
      {
	retVal+=ptypeEvalTable.getPromoteVal(getPtypeO(move));
      }
      return retVal;
    }
  private:
    static int minusInfty(Int2Type<BLACK>)
    {
      return -45300;
    }
    static int minusInfty(Int2Type<WHITE>)
    {
      return 45300;
    }
  public:
    template <Player P>
    static int minusInfty()
    {
      return minusInfty(Int2Type<P>());
    }
    static bool isEvalValueConsistent(int v)
    {
      // 礤ϤǶ̤ޤ pieceEval  45300 ˤϤʤǽ?
      return ((v % 2) == 0)
	&& EvalTraits<BLACK>::betterThan(v, minusInfty<BLACK>())
	&& EvalTraits<WHITE>::betterThan(v, minusInfty<WHITE>());
    }
    /** 1β MTDF Ȥ */
    int pawnValue() const
    {
      return PtypeEvalTraits<PAWN>::val;
    }
    
    /**
     * move ɾ֤̣ͤƵ
     */
    template<class State,Player P>
    int computeValWithMove(const State& state,Move move) const
    {
      typedef typename State::effect_state_t effect_state_t;
      return computeDiffAfterMove<effect_state_t,P>(state,move) + getVal();
    }
    template<class State>
    int computeValWithMove(Player P, const State& state,Move move) const
    {
      if (P == BLACK)
	return computeValWithMove<State,BLACK>(state,move);
      else
	return computeValWithMove<State,WHITE>(state,move);
    }

    /**
     * EvalState ͳõüǸƤФɾؿ
     *
     */
    template<class State,Player P>
    int computeValWithTakeBack(const State& state, Position focus) const
    {
      // return computeValWithTakeBackLimited<State,P>(state,focus);
      return computeValWithTakeBackMulti<State,P>(state);
    }

    /**
     * ɾͤܤ֤˴ؤ֤̣Ƶ
     *
     * getTo(move) == focus ä computeValWithMove(state,move)
     * η̤ state.doMove(move), computeValWithTakeBack(focus)
     * η̤ϰ
     * @param P focus ˤνͭ alt(P)֤ǼϤ
     *   Ȥߤʤ(ºݤstate.getTurn() ȰפƤɬפϤʤ)
     */
    template<class State,Player P>
    int computeValWithTakeBackLimited(const State& state, Position focus) const
    {
      typedef typename State::effect_state_t effect_state_t;
      return computeDiffAfterTakeBackLimited<effect_state_t,P>(state,focus) 
	+ getVal();
    }
    template<class State>
    int computeValWithTakeBackLimited(Player P, const State& state, Position focus) const
    {
      if (P == BLACK)
	return computeValWithTakeBackLimited<State,BLACK>(state,focus);
      else
	return computeValWithTakeBackLimited<State,WHITE>(state,focus);
    }

    template<class State,Player P>
    static int computeDiffAfterTakeBackLimited(const State& state, Position focus);
    template<class State>
    static int computeDiffAfterTakeBackLimited(Player P, const State& state, Position focus)
    {
      if (P == BLACK)
	return PieceEval::template 
	  computeDiffAfterTakeBackLimited<State,BLACK>(state, focus);
      else
	return PieceEval::template 
	  computeDiffAfterTakeBackLimited<State,WHITE>(state, focus);
    }

    /**
     * alt(P)ˤȤäưʾǤμ礤
     *
     * computeValWithTakeBack 㴳ΤΤϤ
     * @param P focus ˤνͭ alt(P)֤ǼϤ
     *   (ºݤstate.getTurn() ȰפƤɬפϤʤ)
     */
    template<class State,Player P>
    int computeValWithTakeBackMulti(const State& state) const
    {
      typedef typename State::effect_state_t effect_state_t;
      return computeDiffAfterTakeBackMulti<effect_state_t,P>(state) + getVal();
    }
    template<class State>
    int computeValWithTakeBackMulti(Player P, const State& state) const
    {
      if (P == BLACK)
	return computeValWithTakeBackMulti<State,BLACK>(state);
      else
	return computeValWithTakeBackMulti<State,WHITE>(state);
    }
    
    /** 
     * move ˤ֤ͤѲ (SOMA)
     *
     * - move Υޥͤ
     * - ܤϲͤξĴ٤
     * - ֤˴ؤơPROMOTE ̵ͭϹͤƤʤ
     * - ߡ֤move θˤΤӤʤ
     * - ?? ̣ζӱۤͤ
     * - ͤβͤ Player ˤȤäơ
     * - ʤɤϵˤʤ
     */
    template<class State,Player P>
    static int computeDiffAfterMove(const State& state,Move move);
    template<class State>
    static int computeDiffAfterMove(Player P, const State& state,Move move)
    {
      if (P == BLACK)
	return computeDiffAfterMove<State,BLACK>(state,move);
      else
	return computeDiffAfterMove<State,WHITE>(state,move);
    }
    template<class State,Player P>
    static int computeDiffAfterTakeBackMulti(const State& state);

    /** 
     * move ˤ֤ͤѲ 
     *
     * - move ΥޥǤʤ¾Υޥͤưʾ
     *   getTo(move)˴ؤƤΤ߿Ӥ
     * - ʤɤϵˤʤ
     * - move θĹϼ礤˹θʤȤ
     */
    template<class State,Player P>
    static int computeDiffAfterMoveMulti(const State& state,Move move);
    template<class State>
    static int computeDiffAfterMoveMulti(Player P, const State& state,Move move)
    {
      if (P == BLACK)
	return computeDiffAfterMoveMulti<State,BLACK>(state,move);
      else
	return computeDiffAfterMoveMulti<State,WHITE>(state,move);
    }
    
    /**
     * ¸ΨõѼ֤
     * 
     * ߤζ̤ɾͤ move ζ̤κʬ(P礬Ȥ
     * 褦)֤
     */
    template<Player P,class State>
    static int computeDiffAfterMoveForRP(const State& state,Move move)
    {
      assert(getPlayer(move)==P);
      typedef typename State::effect_state_t effect_state_t;
      const int diff = computeDiffAfterMove<effect_state_t,P>(state,move);
      return (P==BLACK) ? diff : -diff;
    }
    template<class State>
    static int computeDiffAfterMoveForRP(const State& state, Move move)
    {
      if (getPlayer(move)==BLACK)
	return computeDiffAfterMoveForRP<BLACK,State>(state,move);
      else
	return computeDiffAfterMoveForRP<WHITE,State>(state,move);
    }


    void addVal(int d) { val+=d; }
    // Ϥʤ
    // promoteѲ
    void doSimpleMove(const SimpleState& state,Position from, Position to, 
		      int promoteMask)
    {
      const Piece p0=state.getPieceOnBoard(from);
      const Piece newPiece=promoteWithMask(p0,promoteMask);
      addVal(ptypeEvalTable.getVal(getPtypeO(newPiece))-ptypeEvalTable.getVal(getPtypeO(p0)));
    }
    void doDropMove(const SimpleState& state,Position to,Ptype ptype)
    {
    }
    void doCaptureMove(const SimpleState& state,Position from, Position to, 
		       Piece target,int promoteMask)
    {
      const Piece p0=state.getPieceOnBoard(from);
      const Piece newPiece=promoteWithMask(p0,promoteMask);
      const PtypeO capturePtypeO=getPtypeO(target);
      addVal(ptypeEvalTable.getVal(getPtypeO(newPiece))-
	     ptypeEvalTable.getVal(getPtypeO(p0))+
	     ptypeEvalTable.getVal(captured(capturePtypeO))-
	     ptypeEvalTable.getVal(capturePtypeO));
    }

    /**
     * QuiescenceSearch λ޴ǻ
     */
    static int getCaptureVal(PtypeO ptypeO)
    {
      return ptypeEvalTable.getCaptureVal(ptypeO);
    }
  };

} // namespace osl

#endif /* _PIECEEVAL_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
