#ifndef _PIECE_MASK_H
#define _PIECE_MASK_H
#include "osl/carray.h"

namespace osl{

  static inline int numToIndex(int num){ return num>>5; }
  static inline int numToOffset(int num){ return num&31; }
  static inline int numToMask(int num){ return 1<<(num&31); }

  class PieceMask{
    // private:
    CArray<unsigned int,2> mask;
  public:
    PieceMask(){
      clearAll();
    }
    inline bool isSet(int num) const{
      return (mask[numToIndex(num)]&numToMask(num))!=0;
    }
    inline void setNum(int num){
      mask[numToIndex(num)]|=numToMask(num);
    }
    inline void setNum0(int num){
      mask[0]|=(1<<num);
    }
    inline void setNum1(int num){
      mask[1]|=numToMask(num);
    }
    inline void xorNum(int num){
      mask[numToIndex(num)]^=numToMask(num);
    }
    inline void xorNum0(int num){
      mask[0]^=(1<<num);
    }
    inline void xorNum1(int num){
      mask[1]^=numToMask(num);
    }
    inline void clearNum(int num){
      mask[numToIndex(num)]&= ~numToMask(num);
    }
    inline void xorMask(int numIndex,int numMask){
      mask[numIndex]^= numMask;
    }
    inline void clearAll(){
      mask[0]=mask[1]=0;
    }
    inline void setAll(){
      mask[0]=0xffffffff; mask[1]=0xff;
    }
    inline unsigned int getMask(int index) const{
      return mask[index];
    }
    inline void setMask(int index,unsigned int val){
      mask[index]=val;
    }
  };
  bool operator==(const PieceMask &m1, const PieceMask &m2);
  std::ostream& operator<<(std::ostream& os,const PieceMask pieceMask);
}


#endif /* _PIECE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
