/* piecePairEval.h
 */
#ifndef _PIECE_PAIR_EVAL_H
#define _PIECE_PAIR_EVAL_H

#include "osl/piece.h"
#include "osl/board.h"
#include "osl/move.h"
#include "osl/evalTraits.h"
#include "osl/positionCompressor.h"
#include "osl/carray.h"
#include "osl/pawnMaskState.h"
#include <cassert>

#ifdef PIECE_PAIR_EXTRA_DEBUG
#  define piece_pair_assert(x) assert(x)
#else
#  define piece_pair_assert(x)
#endif

namespace osl
{
  /**
   * [Position*(Player*Ptype==PtypeO)] ^2
   */
  class PiecePairEvalTable
  {
  public:
    static const unsigned int maxPositionIndex = 82;
    static const unsigned int maxPtypeOIndex = PTYPEO_MAX - PTYPEO_MIN + 1;
    static const unsigned int maxPieceIndex = maxPositionIndex*maxPtypeOIndex;
    static const unsigned int maxPairIndex = maxPieceIndex*maxPieceIndex;
  private:
    /** 
     * const object ФơǡѹԲġɤ߹ߤϵĤ뤿
     * mutable ˤ롥
     */
    mutable CArray<signed char, maxPairIndex> values;
  public:
    PiecePairEvalTable();
    ~PiecePairEvalTable();

    /**
     *  log 줿ǡ򰷤
     * @return successful load
     */
    bool loadFromBinaryFile(const char *filename) const;
    void writeInBinaryFile(const char *filename) const;
    
    static unsigned int indexOf(unsigned int i1, unsigned int i2)
    {
      assert(i1 < maxPieceIndex);
      assert(i2 < maxPieceIndex);
      return i1*maxPieceIndex + i2;
    }
    static unsigned int canonicalIndexOf(unsigned int i1, unsigned int i2)
    {
      if (i1 > i2)
	std::swap(i1,i2);
      return indexOf(i1,i2);
    }
    /** Ѵ */
    static void meltIndex(unsigned int index, 
			  unsigned int& i1, unsigned int& i2)
    {
      i1 = index / maxPieceIndex;
      i2 = index % maxPieceIndex;
    }
    static unsigned int positionIndexOf(Position pos)
    {
      unsigned int result = PositionCompressor::compress(pos);
      assert(result < maxPositionIndex);
      return result;
    }
    static unsigned int ptypeOIndexOf(PtypeO ptypeo)
    {
      return ptypeo - PTYPEO_MIN;
    }
    static unsigned int indexOf(Position pos, PtypeO ptypeo)
    {
      const int result = ptypeOIndexOf(ptypeo)
	+ maxPtypeOIndex*positionIndexOf(pos);
      return result;
    }
    /** Ѵ */
    static void meltIndex(unsigned int index, Position& pos, PtypeO& ptypeo)
    {
      ptypeo = static_cast<PtypeO>((index % maxPtypeOIndex)+PTYPEO_MIN);
      pos    = PositionCompressor::melt(index / maxPtypeOIndex);
    }

    static unsigned int indexOf(Piece piece)
    {
      return indexOf(getPosition(piece), getPtypeO(piece));
    }
    static unsigned int indexOf(Piece p1, Piece p2)
    {
      return indexOf(indexOf(p1), indexOf(p2));
    }
    const signed char& value(unsigned int i) const
    {
      return values[i];
    }
    signed char& valueOf(unsigned int i1, unsigned int i2)
    {
      const unsigned int index = indexOf(i1,i2);
      return values[index];
    }
    signed char& valueOf(Piece p1, Piece p2)
    {
      const unsigned int index = indexOf(p1,p2);
      return values[index];
    }
    int valueOf(unsigned int i1, unsigned int i2) const
    {
      const unsigned int index = indexOf(i1,i2);
      return values[index];
    }
    int valueOf(Piece p1, Piece p2) const
    {
      const unsigned int index = indexOf(p1,p2);
      return values[index];
    }

    /**
     * Note: SimpleState ¸
     *  SimpleState  id  onboard ˤʤȡgetPieceOf(id) ̵ä
     * Ͼ˻Ȥ褦ˤʤä
     */
    static unsigned int getIndexOfPieceNum(const SimpleState& s, int id)
    {
#ifdef STATE_REQUIRE_OFFBOARD_CHECKING
      const PieceMask& blackPieces = s.getOffBoardMask(BLACK);
      if (blackPieces.isSet(id))
	return indexOf(BLACK, OFFBOARD, numTable.getPtype(id));

      const PieceMask& whitePieces = s.getOffBoardMask(WHITE);
      if (whitePieces.isSet(id))
	return indexOf(WHITE, OFFBOARD, numTable.getPtype(id));
#endif
      return indexOf(s.getPieceOf(id));
    }

    /** ƤδطˤĤf¹Ԥ롥ʣطˬʤ */
    template <class F>
    static void forEachRelation(F f);
  };

  /** user must initialize this befor use */
  extern const PiecePairEvalTable piecePairEvalTable;

  /**
   * Υڥ׾򸵤ˤɾؿ
   * ɬ
   * ͭ +, ͭ -
   */
  class PiecePairEval 
  {
    int val;
  public:
    PiecePairEval(const PawnMaskState& state);

    /** roundup  2^n Ǥ뤳 */
    static const int ROUND_UP = 64;
    static int roundUp(int v)
    {
      return v & (~(ROUND_UP-1)); 
    }
    int getVal() const	{ return roundUp(val); }
    int getRawVal() const { return val; }
    /**
     * TODO: SOMA Υɤ񤯤ȽŤ?
     */
    template<class State,Player P>
    int computeValWithMove(const State& state,Move move) const
    {
      return computeValWithMove(P, state, move);
    }
    int computeValWithMove(Player P, const PawnMaskState& state,Move move) const;
    template<class State,Player P>
    int computeValWithTakeBack(const State&, Position) const
    {
      return getVal();
    }
    /**
     *  indexOld  indexNew ưȤͤκʬ
     * @param state ưζ
     * @param indexOld +ư
     * @param indexNew +ư
     */
    static int adjustPairs(const SimpleState& state,
			   unsigned int indexOld, unsigned int indexNew);    
    static int adjustPairs(const SimpleState& state, 
			   unsigned int indexOld, unsigned int indexOld2, 
			   unsigned int indexNew, unsigned int indexNew2);
    /** λ state  move 
     */
    void doSimpleMove(const PawnMaskState& state,Position from, Position to, int promoteMask);
    template <class State>
    static int diffAfterSimpleMove(const State& state,
				   Position from, Position to, int promoteMask)
    {
      const Piece oldPiece=state.getPieceAt(from);
      const PtypeO newPtypeO = promoteWithMask(getPtypeO(oldPiece), promoteMask);
      const unsigned int indexOld = PiecePairEvalTable::indexOf(oldPiece);
      const unsigned int indexNew = PiecePairEvalTable::indexOf(to, newPtypeO);
      return adjustPairs(state, indexOld, indexNew);
    }
    void doDropMove(const PawnMaskState& state,Position to,Ptype ptype);
    template <class State>
    static int diffAfterDropMove(const State& state,Position to,PtypeO ptypeo)
    {
      const unsigned int indexOld = PiecePairEvalTable::indexOf(OFFBOARD, ptypeo);
      const unsigned int indexNew = PiecePairEvalTable::indexOf(to, ptypeo);
      return adjustPairs(state, indexOld, indexNew);
    }
    void doCaptureMove(const PawnMaskState& state,Position from, Position to, Piece victim,int promoteMask);
    template <class State>
    static int diffAfterCaptureMove(const State& state,
				    Position from, Position to, 
				    PtypeO victim,int promoteMask)
    {
      const Piece oldPiece=state.getPieceAt(from);
      const PtypeO newPtypeO = promoteWithMask(getPtypeO(oldPiece), promoteMask);
      const unsigned int indexOld = PiecePairEvalTable::indexOf(oldPiece);
      const unsigned int indexNew = PiecePairEvalTable::indexOf(to, newPtypeO);

      const unsigned int indexVictim = PiecePairEvalTable::indexOf(to, victim);
      const unsigned int indexCaptured = 
	PiecePairEvalTable::indexOf(OFFBOARD, captured(victim));
      return adjustPairs(state, indexOld, indexVictim, indexNew, indexCaptured);
    }
    template <class State>
    static int diffWithMove(const State& state, Move move)
    {
      const Position from=getFrom(move);
      const Position to=getTo(move);
      if (isOffBoard(from))
	return diffAfterDropMove(state, to,getPtypeO(move));

      const Ptype ptypeCaptured=getCapturePtype(move);
      if (ptypeCaptured != PTYPE_EMPTY)
	return diffAfterCaptureMove(state, from, to,
				    newPtypeO(alt(getPlayer(move)),
					      ptypeCaptured),
				    getPromoteMask(move));

      return diffAfterSimpleMove(state, from,to,getPromoteMask(move));
    }
  private:
    static int minusInfty(Int2Type<BLACK>)
    {
      /** CompositEval ­Ƥoverflowʤͤˤ */
      return -3000000;
    }
    static int minusInfty(Int2Type<WHITE>)
    {
      return -minusInfty(Int2Type<BLACK>());
    }
  public:
    template <Player P>
    static int minusInfty()
    {
      return minusInfty(Int2Type<P>());
    }

    /** 1β MTDF Ȥ */
    int pawnValue() const
    {
      return 100; // ??
    }

    static bool isEvalValueConsistent(int v)
    {
      return ((v % 2) == 0)
	&& EvalTraits<BLACK>::betterThan(v, minusInfty<BLACK>())
	&& EvalTraits<WHITE>::betterThan(v, minusInfty<WHITE>());
    }
  };
} // namespace osl

template <class F>
void osl::PiecePairEvalTable::forEachRelation(F f)
{
  for (int x=1; x<=9; ++x)
  {
    for (int y=1; y<=9; ++y)
    {
      const Position pos1 = newPosition(x,y);
      for (int ptype=PPAWN; ptype<=PTYPE_MAX; ++ptype)
      {
	const Ptype p1 = static_cast<Ptype>(ptype);
	const unsigned int i1 = 
	  indexOf(pos1, newPtypeO(BLACK, p1));
	const unsigned int i1w = 
	  indexOf(pos1, newPtypeO(WHITE, p1));
	f(indexOf(i1,i1));
	f(indexOf(i1w,i1w));
	
	for (int x2=x; x2<=9; ++x2)
	{
	  for (int y2=((x2 == x) ? y+1 : 1); y2<=9; ++y2)
	  {
	    const Position pos2 = newPosition(x2,y2);
	    
	    for (int ptype2=PPAWN; ptype2<=PTYPE_MAX; ++ptype2)
	    {
	      const Ptype p2 = static_cast<Ptype>(ptype2);
	      const unsigned int i2 = 
		indexOf(pos2, newPtypeO(BLACK, p2));
	      const unsigned int i2w = 
		indexOf(pos2, newPtypeO(WHITE, p2));

	      f(indexOf(i1, i2));
	      f(indexOf(i1, i2w));
	      f(indexOf(i1w, i2));
	      f(indexOf(i1w, i2w));
	    }
	  }
	}

	for (int ptype2=KING; ptype2<=PTYPE_MAX; ++ptype2)
	{
	  const Ptype p2 = static_cast<Ptype>(ptype2);
	  const unsigned int i2 = 
	    indexOf(OFFBOARD, newPtypeO(BLACK, p2));
	  const unsigned int i2w = 
	    indexOf(OFFBOARD, newPtypeO(WHITE, p2));

	  f(indexOf(i1, i2));
	  f(indexOf(i1, i2w));
	  f(indexOf(i1w, i2));
	  f(indexOf(i1w, i2w));
	}
      }	// ptype
    } // x
  } // y

  // ƱΤϺǸ
  for (int ptype=KING; ptype<=PTYPE_MAX; ++ptype)
  {
    const Ptype p1 = static_cast<Ptype>(ptype);
    const unsigned int i1 = 
      indexOf(OFFBOARD, newPtypeO(BLACK, p1));
    const unsigned int i1w = 
      indexOf(OFFBOARD, newPtypeO(WHITE, p1));
    f(indexOf(i1, i1));
    f(indexOf(i1, i1w));
    f(indexOf(i1w, i1w));
    for (int ptype2=ptype+1; ptype2<=PTYPE_MAX; ++ptype2)
    {
      const Ptype p2 = static_cast<Ptype>(ptype2);
      const unsigned int i2 = 
	indexOf(OFFBOARD, newPtypeO(BLACK, p2));
      const unsigned int i2w = 
	indexOf(OFFBOARD, newPtypeO(WHITE, p2));

      f(indexOf(i1, i2));
      f(indexOf(i1, i2w));
      f(indexOf(i1w, i2));
      f(indexOf(i1w, i2w));
    }
  }
}

#endif /* _PIECE_PAIR_EVAL_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
