/* pieceStand.h
 */
#ifndef _PIECESTAND_H
#define _PIECESTAND_H
#include "osl/ptype.h"
#include "osl/carray.h"
#include <iosfwd>
#include <cassert>
namespace osl
{
  /**
   * μ֤λϿ륯饹
   * - bit field Ȥ٤¤
   * -  king ˤɤȤˤƤ
   * 쥤 Ĺ:index
   * -  reserved : 1;31
   * -  carry    : 1;
   * -  KING     : 2;28
   * -  carry    : 1;
   * -  GOLD     : 3;24
   * -  carry    : 1;
   * -  PAWN     : 5;18
   * -  carry    : 1;
   * -  LANCE    : 3;14
   * -  carry    : 1;
   * -  KNIGHT   : 3;10
   * -  carry    : 1;
   * -  SILVER   : 3; 6
   * -  carry    : 1;
   * -  BISHOP   : 2; 3
   * -  carry    : 1; 
   * -  ROOK     : 2; 0
   *
   * == ڤ뤿 carry off ξ֤ܤȤ
   */
  class PieceStand
  {
  public:
    static const CArray<unsigned char,PTYPE_MAX+1> shift;
    static const CArray<unsigned char,PTYPE_MAX+1> mask;
    static const unsigned int carryMask = 0x48822224;
  private:
    mutable unsigned int flags;
  public:
    explicit PieceStand(unsigned int value=0) : flags(value)
    {
    }

    void add(Ptype type, unsigned int num=1)
    {
      assert(isBasic(type));
      assert(num == (num & mask[type]));
      flags += (num << (shift[type]));
      assert(testCarries() == 0);	// overflow 
    }
    void sub(Ptype type, unsigned int num=1)
    {
      assert(isBasic(type));
      assert(num == (num & mask[type]));
      assert(get(type) >= num);
      flags -= (num << (shift[type]));
    }

    unsigned int get(Ptype type) const
    {
      return (flags >> (shift[type])) & mask[type];
    }
    void carriesOff() const { flags &= (~carryMask); }
    void carriesOn()  const { flags |= carryMask; }
    unsigned int testCarries() const { return (flags & carryMask); }
    bool isSuperiorOrEqualTo(PieceStand other) const
    {
      carriesOn();
      other.carriesOff();
      const bool result = (((flags - other.flags) & carryMask) == carryMask);
      carriesOff();
      return result;
    }
    unsigned int getFlags() const { return flags; }
  };

  inline bool operator==(PieceStand l, PieceStand r)
  {
    assert(! l.testCarries());
    assert(! r.testCarries());
    return l.getFlags() == r.getFlags();
  }
  inline bool operator!=(PieceStand l, PieceStand r)
  {
    return ! (l == r);
  }
  inline bool operator<(PieceStand l, PieceStand r)
  {
    assert(! l.testCarries());
    assert(! r.testCarries());
    return l.getFlags() < r.getFlags();
  }
  std::ostream& operator<<(std::ostream&, PieceStand l);
} // namespace osl

#endif /* _PIECESTAND_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
