#ifndef _PTYPE_H
#define _PTYPE_H

#include "player.h"
#include "board.h"
#include "direction.h"
#include "loki.h"
#include "carray.h"
#include "carray2d.h"
#include <algorithm>
#include <cassert>
#include <iosfwd>
#include <boost/static_assert.hpp>

namespace osl{

  /** μ4ӥåȤǥɲ */
  enum Ptype{
    PTYPE_EMPTY=0,
    PTYPE_EDGE=1,
    PPAWN=2,
    PLANCE=3,
    PKNIGHT=4,
    PSILVER=5,
    PBISHOP=6,
    PROOK=7,
    KING=8,
    GOLD=9,
    PAWN=10,
    LANCE=11,
    KNIGHT=12,
    SILVER=13,
    BISHOP=14,
    ROOK=15,
    PTYPE_MIN=0,
    PTYPE_PIECE_MIN=2,
    PTYPE_MAX=15,
  };
  const int PTYPE_SIZE=PTYPE_MAX-PTYPE_MIN+1;
  
  std::ostream& operator<<(std::ostream& os,const Ptype ptype);
  std::ostream& csaShow(std::ostream& os,const Ptype ptype);
  
  inline bool isValid(Ptype ptype) {
    return static_cast<int>(ptype)>=PTYPE_MIN &&
      static_cast<int>(ptype)<=PTYPE_MAX;
  }
  inline bool isPiece(Ptype ptype) {
    assert(isValid(ptype));
    return static_cast<int>(ptype)>=PTYPE_PIECE_MIN;
  }
  inline bool isBasic(Ptype ptype) {
    assert(isValid(ptype));
    return static_cast<int>(ptype)>PROOK;
  }
  
  inline bool isPromoted(Ptype ptype) {
    assert(isPiece(ptype));
    return static_cast<int>(ptype)<KING; 
  }
  inline bool canPromote(Ptype ptype) {
    assert(isPiece(ptype));
    return static_cast<int>(ptype)>GOLD; 
  }
  
  /** promoteƤʤptypeͿƤ褤 */
  inline Ptype unpromote(Ptype ptype) {
    assert(isPiece(ptype));
    Ptype ret=static_cast<Ptype>(static_cast<int>(ptype)|8); 
    assert(isPiece(ret));
    return ret;
  }
  
  /** promoteԲĤptypeͿƤϤʤ*/
  inline Ptype promote(Ptype ptype) {
    assert(canPromote(ptype));
    Ptype ret=static_cast<Ptype>(static_cast<int>(ptype)-8); 
    assert(isPiece(ret));
    return ret;
  }
  
  template<Ptype T>
  struct PtypeTraits;
  
  template <>
  struct PtypeTraits<PTYPE_EMPTY>{
    static const bool isBasic=false;
    static const bool canPromote=false;
    /** Ǥͤƹͤ promoteϾ */
    static const bool betterToPromote=false;
    static const char *name(){ return "PTYPE_EMPTY";}
    static const char *csaName(){ return "..";}
    static const int moveMask=0;
  };
  
  template <>
  struct PtypeTraits<PTYPE_EDGE>{
    static const bool isBasic=false;
    static const bool canPromote=false;
    static const bool betterToPromote=false;
    static const char *name(){ return "PTYPE_EDGE";}
    static const char *csaName(){ return "XX";}
    static const int moveMask=0;
  };
  
  template <>
  struct PtypeTraits<GOLD>{
    static const bool isBasic=true;
    static const bool canPromote=false;
    static const bool betterToPromote=false;
    static const Ptype moveType=GOLD;
    static const char *name(){ return "GOLD";}
    static const char *csaName(){ return "KI";}
    static const int indexMin=26;
    static const int indexLimit=30;
    static const int dropBlackFromY=1;
    static const int dropBlackToY=9;
    static const int moveMask=
    DirectionTraits<UL>::mask|DirectionTraits<U>::mask
    |DirectionTraits<UR>::mask|DirectionTraits<L>::mask
    |DirectionTraits<R>::mask|DirectionTraits<D>::mask;
  };
  
  template <>
  struct PtypeTraits<PAWN>{
    static const bool isBasic=true;
    static const bool canPromote=true;
    static const bool betterToPromote=true;
    static const Ptype moveType=PAWN;
    static const char *name(){ return "PAWN";}
    static const char *csaName(){ return "FU";}
    static const int indexMin=0;
    static const int indexLimit=18;
    static const int dropBlackFromY=2;
    static const int dropBlackToY=9;
    static const int moveMask=DirectionTraits<U>::mask;
  };
  
  template <>
  struct PtypeTraits<PPAWN>{
    static const bool isBasic=false;
    static const bool canPromote=false;
    static const bool betterToPromote=true;
    static const Ptype moveType=GOLD;
    static const char *name(){ return "PPAWN";}
    static const char *csaName(){ return "TO";}
    static const int moveMask=PtypeTraits<GOLD>::moveMask;
    static const Ptype basicType=PAWN;
    static const int indexMin=PtypeTraits<basicType>::indexMin;
  };
  
  template <>
  struct PtypeTraits<LANCE>{
    static const bool isBasic=true;
    static const bool canPromote=true;
    static const bool betterToPromote=false;
    static const Ptype moveType=LANCE;
    static const char *name(){ return "LANCE";}
    static const char *csaName(){ return "KY";}
    static const int indexMin=32;
    static const int indexLimit=36;
    static const int dropBlackFromY=2;
    static const int dropBlackToY=9;
    static const int moveMask=DirectionTraits<LONG_U>::mask;
  };
  
  template <>
  struct PtypeTraits<PLANCE>{
    static const bool isBasic=false;
    static const bool canPromote=false;
    static const bool betterToPromote=false;
    static const Ptype moveType=GOLD;
    static const char *name(){ return "PLANCE";}
    static const char *csaName(){ return "NY";}
    static const int moveMask=PtypeTraits<GOLD>::moveMask;
    static const Ptype basicType=LANCE;
    static const int indexMin=PtypeTraits<basicType>::indexMin;
  };
  
  template <>
  struct PtypeTraits<KNIGHT>{
    static const bool isBasic=true;
    static const bool canPromote=true;
    static const bool betterToPromote=false;
    static const Ptype moveType=KNIGHT;
    static const char *name(){ return "KNIGHT";}
    static const char *csaName(){ return "KE";}
    static const int indexMin=18;
    static const int indexLimit=22;
    static const int dropBlackFromY=3;
    static const int dropBlackToY=9;
    static const int moveMask=DirectionTraits<UUL>::mask|DirectionTraits<UUR>::mask;
  };
  
  template <>
  struct PtypeTraits<PKNIGHT>{
    static const bool isBasic=false;
    static const bool canPromote=false;
    static const bool betterToPromote=false;
    static const Ptype moveType=GOLD;
    static const char *name(){ return "PKNIGHT";}
    static const char *csaName(){ return "NK";}
    static const int moveMask=PtypeTraits<GOLD>::moveMask;
    static const Ptype basicType=KNIGHT;
    static const int indexMin=PtypeTraits<basicType>::indexMin;
  };
  
  template <>
  struct PtypeTraits<SILVER>{
    static const bool isBasic=true;
    static const bool canPromote=true;
    static const bool betterToPromote=false;
    static const Ptype moveType=SILVER;
    static const char *name(){ return "SILVER";}
    static const char *csaName(){ return "GI";}
    static const int indexMin=22;
    static const int indexLimit=26;
    static const int dropBlackFromY=1;
    static const int dropBlackToY=9;
    static const int moveMask=
    DirectionTraits<UL>::mask|DirectionTraits<U>::mask
    |DirectionTraits<UR>::mask|DirectionTraits<DL>::mask
    |DirectionTraits<DR>::mask;
  };
  
  template <>
  struct PtypeTraits<PSILVER>{
    static const bool isBasic=false;
    static const bool canPromote=false;
    static const bool betterToPromote=false;
    static const Ptype moveType=GOLD;
    static const char *name(){ return "PSILVER";}
    static const char *csaName(){ return "NG";}
    static const int moveMask=PtypeTraits<GOLD>::moveMask;
    static const Ptype basicType=SILVER;
    static const int indexMin=PtypeTraits<basicType>::indexMin;
  };
  
  template <>
  struct PtypeTraits<BISHOP>{
    static const bool isBasic=true;
    static const bool canPromote=true;
    static const bool betterToPromote=true;
    static const Ptype moveType=BISHOP;
    static const char *name(){ return "BISHOP";}
    static const char *csaName(){ return "KA";}
    static const int indexMin=36;
    static const int indexLimit=38;
    static const int dropBlackFromY=1;
    static const int dropBlackToY=9;
    static const int moveMask=
    DirectionTraits<LONG_UL>::mask|DirectionTraits<LONG_UR>::mask
    |DirectionTraits<LONG_DL>::mask|DirectionTraits<LONG_DR>::mask;
  };
  
  template <>
  struct PtypeTraits<PBISHOP>{
    static const bool isBasic=false;
    static const bool canPromote=false;
    static const bool betterToPromote=true;
    static const Ptype moveType=PBISHOP;
    static const char *name(){ return "PBISHOP";}
    static const char *csaName(){ return "UM";}
    static const int moveMask=
    DirectionTraits<LONG_UL>::mask|DirectionTraits<LONG_UR>::mask
    |DirectionTraits<LONG_DL>::mask|DirectionTraits<LONG_DR>::mask
    |DirectionTraits<U>::mask|DirectionTraits<L>::mask
    |DirectionTraits<R>::mask|DirectionTraits<D>::mask;
    static const Ptype basicType=BISHOP;
    static const int indexMin=PtypeTraits<basicType>::indexMin;
  };
  
  template <>
  struct PtypeTraits<ROOK>{
    static const bool isBasic=true;
    static const bool canPromote=true;
    static const bool betterToPromote=true;
    static const Ptype moveType=ROOK;
    static const char *name(){ return "ROOK";}
    static const char *csaName(){ return "HI";}
    static const int indexMin=38;
    static const int indexLimit=40;
    static const int dropBlackFromY=1;
    static const int dropBlackToY=9;
    static const int moveMask=
    DirectionTraits<LONG_U>::mask|DirectionTraits<LONG_L>::mask
    |DirectionTraits<LONG_R>::mask|DirectionTraits<LONG_D>::mask;
  };
  
  template <>
  struct PtypeTraits<PROOK>{
    static const bool isBasic=false;
    static const bool canPromote=false;
    static const bool betterToPromote=true;
    static const Ptype moveType=PROOK;
    static const char *name(){ return "PROOK";}
    static const char *csaName(){ return "RY";}
    static const int moveMask=
    DirectionTraits<LONG_U>::mask|DirectionTraits<LONG_L>::mask
    |DirectionTraits<LONG_R>::mask|DirectionTraits<LONG_D>::mask
    |DirectionTraits<UL>::mask|DirectionTraits<UR>::mask
    |DirectionTraits<DL>::mask|DirectionTraits<DR>::mask;
    static const Ptype basicType=ROOK;
    static const int indexMin=PtypeTraits<basicType>::indexMin;
  };
  
  
  template <>
  struct PtypeTraits<KING>{
    static const bool isBasic=true;
    static const bool canPromote=false;
    static const bool betterToPromote=false;
    static const Ptype moveType=KING;
    static const char *name(){ return "KING";}
    static const char *csaName(){ return "OU";}
    static const int indexMin=30;
    static const int indexLimit=32;
    static const int dropBlackFromY=1;
    static const int dropBlackToY=9;
    static const int moveMask=
    DirectionTraits<U>::mask|DirectionTraits<L>::mask
    |DirectionTraits<R>::mask|DirectionTraits<D>::mask
    |DirectionTraits<UL>::mask|DirectionTraits<UR>::mask
    |DirectionTraits<DL>::mask|DirectionTraits<DR>::mask;
  };

#if 0
  template<Ptype T,bool IsBasic>
  struct BasicPtypeSub{};

  template<Ptype T>
  struct BasicPtypeSub<T,true>{ 
    static const Ptype type=T;
  };

  template<Ptype T>
  struct BasicPtypeSub<T,false>{ 
    static const Ptype type=PtypeTraits<T>::basicPtype;
  };
  template<Ptype T>
  struct BasicPtype{
    static const Ptype type=BasicPtypeSub<T,PtypeTraits<T>::isBasic>::type;
  };

#endif

  template<Ptype T,bool IsBasic>
  struct PtypeFunsSub;
  
  template<Ptype T>
  struct PtypeFunsSub<T,true>{
    static const unsigned int indexMask=static_cast<unsigned int>((-1<<(PtypeTraits<T>::indexMin&31))^((-1<<(((PtypeTraits<T>::indexLimit-1)&31)))<<1));
    static const Ptype promotePtype=static_cast<Ptype>(static_cast<int>(T)-8);
  };
  
  template<Ptype T>
  struct PtypeFunsSub<T,false>{
    static const unsigned int indexMask=0;
    //    static const Ptype promotePtype=PTYPE_EMPTY;
    static const Ptype promotePtype=T;
  };
  
  template<Ptype T>
  struct PtypeFuns{
    static const unsigned int indexNum=(PtypeTraits<T>::indexMin >> 5);
    static const unsigned int indexMask=PtypeFunsSub<T,PtypeTraits<T>::isBasic>::indexMask;
    static const Ptype promotePtype=PtypeFunsSub<T,PtypeTraits<T>::canPromote>::promotePtype;
  };

  enum MoveType{ noMove, promoteMove, noPromoteMove, alwaysMove};

  template<Ptype T,Direction D>
  struct PtypeDirectionTraits{
    static const bool hasMove=(PtypeTraits<T>::moveMask & DirectionTraits<D>::mask)!=0;
    static const bool canMove=
         (PtypeTraits<T>::moveMask & DirectionTraits<D>::mask)!=0 ||
         (PtypeTraits<T>::moveMask & 
          DirectionTraits<DirectionTraits<D>::longDir>::mask)!=0;
#if 1
    static const MoveType moveType=
         (PtypeDirectionTraits<T,D>::canMove ? 
          (PtypeDirectionTraits<PtypeFuns<T>::promotePtype,D>::canMove ?
           alwaysMove : noPromoteMove) :
          (PtypeDirectionTraits<PtypeFuns<T>::promotePtype,D>::canMove ?
           promoteMove : noMove));
#endif
  };
  
  
  template<Player T>
  struct KingTraits{
    static const int index=PtypeTraits<KING>::indexMin+PlayerTraits<T>::index;
  };
  
  // ؿݥ󥿤
  inline void for_each_piece_ptype(void (*func)(Ptype)){
    for(int i=static_cast<int>(PTYPE_PIECE_MIN);i<=static_cast<int>(PTYPE_MAX);i++){
      func(static_cast<Ptype>(i));
    }
  }
  
  inline void for_each_ptype(void (*func)(Ptype)){
    for(int i=static_cast<int>(PTYPE_MIN);i<=static_cast<int>(PTYPE_MAX);i++){
      func(static_cast<Ptype>(i));
    }
  }
  
  // inline bool newNEW_PTYPE(Ptype ptype,bool promotep){
  //   return (Ptype)(((int)ptype)-((int)promotep)*8);
  // }
  
  template<Ptype T,Player P>
  struct PtypePlayerTraits{
    static inline bool canDropTo(Position pos,Int2Type<false>){
#if 0
      /**
       * Ūˤ
       */
     BOOST_STATIC_ASSERT(0);
#endif
      assert(0);
      return false;
    }
    static inline bool canDropTo(Position pos,Int2Type<true>){
      if(PtypeTraits<T>::dropBlackFromY == 1)
	return true; 
      else if(P==BLACK){
	return getY(pos)>=PtypeTraits<T>::dropBlackFromY;
      }
      else
	return getY(pos)<=reverseY(PtypeTraits<T>::dropBlackFromY);
    }
    static inline bool canDropTo(Position pos){
      return canDropTo(pos,Int2Type<PtypeTraits<T>::isBasic>());
    }
  };
  
  /**
   * Player + Ptype [-15, 15] 
   * PtypeO  O  Owner  O
   */
  enum PtypeO{
    PTYPEO_MIN= PTYPE_EMPTY-16,
    PTYPEO_MAX= 15,
  };
  
#define NEW_PTYPEO(player,ptype) static_cast<PtypeO>(static_cast<int>(ptype)-(16&static_cast<int>(player)))
  
  static inline PtypeO newPtypeO(Player player,Ptype ptype){
    return static_cast<PtypeO>(static_cast<int>(ptype)-(16&static_cast<int>(player)));
  }
  
  
  static inline Ptype getPtype(PtypeO ptypeO){
    return static_cast<Ptype>(static_cast<int>(ptypeO)& 15);
  }
  
  /** promoteԲĤptypeͿƤϤʤ*/
  static inline PtypeO promote(PtypeO ptypeO) {
    assert(canPromote(getPtype(ptypeO)));
    PtypeO ret=static_cast<PtypeO>(static_cast<int>(ptypeO)-8); 
    assert(isPiece(getPtype(ret)));
    return ret;
  }
  
  /** piecepromote뤿Τ
   * promoteMask0x80000 0
   */
  static inline PtypeO promoteWithMask(PtypeO ptypeO,int promoteMask) {
    PtypeO ret=static_cast<PtypeO>(static_cast<int>(ptypeO)-(promoteMask>>16)); 
    return ret;
  }
  
  /** promoteƤʤptypeͿƤ褤 */
  inline PtypeO unpromote(PtypeO ptypeO) {
    return static_cast<PtypeO>(static_cast<int>(ptypeO)|8); 
  }

  /**
   * piecepromote뤿Τ
   * promoteMask0x80000 0
   */
  static inline PtypeO unpromoteWithMask(PtypeO ptypeO,int promoteMask) {
    PtypeO ret=static_cast<PtypeO>(static_cast<int>(ptypeO)+(promoteMask>>16)); 
    return ret;
  }

  /** unpromoteȶˡownerȿž롥 */
  static inline PtypeO captured(PtypeO ptypeO){
    return static_cast<PtypeO>((static_cast<int>(ptypeO)|8)^(~15));
  }
  
  inline bool canPromote(PtypeO ptypeO){
    return canPromote(getPtype(ptypeO));
  }

  inline bool isValid(PtypeO ptypeO) {
    return static_cast<int>(ptypeO)>=PTYPEO_MIN &&
      static_cast<int>(ptypeO)<=PTYPEO_MAX;
  }
  
  /**
   * EMPTY, EDGEǤϤʤ
   */
  inline bool isPiece(PtypeO ptypeO) {
    assert(isValid(ptypeO));
    return isPiece(getPtype(ptypeO));
  }

  static inline Player getOwner(PtypeO ptypeO){
    assert(isPiece(ptypeO));
    return static_cast<Player>(static_cast<int>(ptypeO)>>31);
  }
  

  /**
   * ptypeO promoteѤߤɤ
   */
  inline bool isPromoted(PtypeO ptypeO) {
    assert(isValid(ptypeO));
    return isPromoted(getPtype(ptypeO));
  }


  static const PtypeO PTYPEO_EMPTY=newPtypeO(BLACK,PTYPE_EMPTY);
  static const PtypeO PTYPEO_EDGE=newPtypeO(WHITE,PTYPE_EDGE);
  
  std::ostream& operator<<(std::ostream& os,const PtypeO ptypeO);
  
  static const int PTYPEO_SIZE=PTYPEO_MAX-PTYPEO_MIN+1;
  
  enum EffectContent{
  };
  
  /**
   * û뤫֤emptyʤĹ
   */
  inline bool hasEffect(EffectContent eval){
    return static_cast<int>(eval)!=0;
  }
  /**
   * û롥Ĺ٤ޤ
   */
  inline bool hasDirectEffect(EffectContent eval){
    return (static_cast<int>(eval)&1)!=0;
  }
  /**
   * ֤ͤ0ʤĹʤ, 뤤Ĺ
   * 0ʳʤéΤɬפoffset
   */
  inline Offset effectOffset(EffectContent eval){
    return static_cast<Offset>(static_cast<int>(eval)>>1);
  }
  inline bool hasIndirectEffect(EffectContent eval){
    return effectOffset(eval)!=0;
  }
  static const EffectContent directEffectVal=static_cast<EffectContent>(1);
  inline EffectContent offsetEffectVal(int offset){
    return static_cast<EffectContent>(offset<<1);
  }
  
  
  
  class PtypeTable{
  private:
    CArray<unsigned int, PTYPE_SIZE> numMaskLows;
    CArray<int, PTYPE_SIZE> numIndices;
    CArray<const char *, PTYPE_SIZE> names;
    CArray<const char *, PTYPE_SIZE> csaNames;
    CArray<bool, PTYPE_SIZE> betterToPromote;
    CArray<int, PTYPE_SIZE> moveMasks;
    CArray<int, PTYPE_SIZE> indexMins;
    CArray<int, PTYPE_SIZE> indexLimits;

    CArray2d<int, 2, PTYPE_SIZE> canDropLimit;
    // 22^nˤƤ®
    CArray2d<EffectContent,PTYPEO_SIZE,OFFSET32_SIZE> effectTable;
    CArray2d<EffectContent,PTYPEO_SIZE,OFFSET32_SIZE> effectTableNotLongU;
    CArray2d<unsigned int, 2, SHORT_DIRECTION_SIZE> shortMoveMask;
    
    template<Ptype T> void initPtypeSub(Int2Type<false> isBasic);
    template<Ptype T> void initPtypeSub(Int2Type<true> isBasic);
    template<Ptype T> void initPtype();
  public:
    PtypeTable();
    void init();
    unsigned int getShortMoveMask(Player p,PtypeO ptypeo,Direction dir) const{
      return shortMoveMask[playerToIndex(p)][static_cast<int>(dir)] &
	(1<<(ptypeo-PTYPEO_MIN));
    }
    unsigned int getMaskLow(Ptype ptype) const{
      return numMaskLows[ptype];
    }
    int getIndex(Ptype ptype) const{
      return numIndices[ptype];
    }
    bool hasLongMove(Ptype ptype) const{
      return getIndex(ptype)==1;
    }
    bool isBetterToPromote(Ptype ptype) const{
      return betterToPromote[ptype];
    }
    int getCanDropLimit(Player player,Ptype ptype) const{
      assert(isValid(ptype) && !isPromoted(ptype));
      return canDropLimit[playerToIndex(player)][ptype];
    }
    const char *getName(Ptype ptype) const{
      return names[ptype];
    }
    const char *getCsaName(Ptype ptype) const{
      return csaNames[ptype];
    }
    int getMoveMask(Ptype ptype) const{
      return moveMasks[ptype];
    }
    int getIndexMin(Ptype ptype) const{
      assert(isBasic(ptype));
      return indexMins[ptype];
    }
    int getIndexLimit(Ptype ptype) const{
      assert(isBasic(ptype));
      return indexLimits[ptype];
    }
    int getKingIndex(Player p) const{
      assert(isValid(p));
      if(p==BLACK){return KingTraits<BLACK>::index;}
      else{return KingTraits<WHITE>::index; }
    }
    /** 
     * fromˤptypeotoĤ?
     * @param ptypeo - μ
     * @param from - ΰ
     * @param to - åޥΰ
     */
    EffectContent getEffect(PtypeO ptypeo,Position from, Position to) const{
      assert(isOnBoard(from) && isOnBoard(to));
      return getEffect(ptypeo,positionDiff(to,from));
    }
    EffectContent getEffect(PtypeO ptypeo,Offset32 offset32) const{
      assert(PTYPEO_MIN<=ptypeo && ptypeo<=PTYPEO_MAX);
      return effectTable[ptypeo-PTYPEO_MIN][offset32-OFFSET32_MIN];
    }

    /** ptypeo ʬ offset ΤȤ˸Ĥ? U */
    EffectContent getEffectNotLongU(PtypeO ptypeo, Position from, Position to) const{
      assert(PTYPEO_MIN<=ptypeo && ptypeo<=PTYPEO_MAX);
      assert(isOnBoard(from) && isOnBoard(to));
      Offset32 offset32=positionDiff(to,from);
      return effectTableNotLongU[ptypeo-PTYPEO_MIN][offset32-OFFSET32_MIN];
    }
  };
  
  extern PtypeTable ptypeTable;

  template<Player P> 
  static inline bool canDropTo(Ptype ptype,Position pos);
  template<>
  static inline bool canDropTo<BLACK>(Ptype ptype,Position pos){
    return getY(pos)>=ptypeTable.getCanDropLimit(BLACK,ptype);
  }
  template<>
  static inline bool canDropTo<WHITE>(Ptype ptype,Position pos){
    return getY(pos)<=ptypeTable.getCanDropLimit(WHITE,ptype);
  }

  inline bool canDropTo(Player pl,Ptype ptype,Position pos)
  {
    if(pl==BLACK)
      return canDropTo<BLACK>(ptype,pos);
    else
      return canDropTo<WHITE>(ptype,pos);
  }
  
}

#endif /* _PTYPE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
