#ifndef _RECORD_H
#define _RECORD_H
#include "osl/player.h"
#include "osl/move.h"
#include "osl/pawnMaskState.h"
#include "osl/carray.h"
#include <iosfwd>
#include <string>
#include <vector>
namespace osl
{
  namespace record
  {
    class Record;
    class IRecordStream{
    public:
      virtual void load(Record*)=0;
      virtual ~IRecordStream();
    private:
    };
    class ORecordStream{
    public:
      virtual void save(Record*)=0;
      virtual ~ORecordStream();
    private:
    };
  
    enum NodeType{
      MOVE,
      TORYO,
      MATTA,
      CHUDAN,
      SENNICHITE,
      JISHOGI,
      TSUMI,
      FUZUMI,
      ERROR,
      KACHI,
      HIKIWAKE,
    };
    /**
     * Ȥꤢ tree ɽǤ褦ˤ롥
     * ήϹͤʤ
     * ФΤ CSA GCF
     */
    class MoveRecord{
    private:
      Move move;
      int nodeIndex;
      int time;
      std::string comment;
    public:
      MoveRecord(const Move& mv, int ni):move(mv),nodeIndex(ni){} 
      Move getMove() const;
      int getNodeIndex() const;
      void setTime(int t);
      int getTime() const{ return time; }
      std::string getComment() const{ return comment; }
    };
  
    class NodeRecord{
    private:
      NodeType type;
      std::vector<int> moves;
      std::string comment;
    public:
      NodeRecord():type(MOVE){}
      NodeType getType() const{ return type; }
      int size() const { return moves.size(); }
      int at(int index) const{ return moves.at(index); }
      void setComment(std::string com){ comment=com; }
      std::string getComment() const{ return comment; }
      void addMoveRecord(int moveIndex);
    };
  
    class Record{
    private:
      std::string version;
      CArray<std::string,2> playerNames;
      std::vector<NodeRecord> nrs;
      std::vector<MoveRecord> mrs;
      PawnMaskState initialState;
    public:
      Record();
      void init();
      void setVersion(const std::string& str);
      std::string getVersion() const;
      void setPlayer(Player player,const std::string& str);
      std::string getPlayer(Player player) const;
      void setInitialState(const PawnMaskState& state);
      PawnMaskState const& getInitialState() const;
      int addNodeRecord();
      int addMoveRecord(const MoveRecord& moveRecord);
      NodeRecord* nodeOf(int index);
      NodeRecord& operator[](int index);
      MoveRecord* moveOf(int index);
      void load(IRecordStream&); 
      void save(ORecordStream&); 
      std::vector<Move> getMoves() const;
      const NodeRecord* nodeOf(int index) const;
      const MoveRecord* moveOf(int index) const;
    };
  
    class RecordVisitor{
    private:
      Record* rec;
      PawnMaskState* state;
      int lastMoveIndex;
      int nodeIndex;
    public:
      RecordVisitor():rec(NULL),state(NULL),lastMoveIndex(0),nodeIndex(0){}
      RecordVisitor(Record& r);
      PawnMaskState *getState() const{ return state; }
      void setState(PawnMaskState *s){ state=s;}
      Record *getRecord() { return rec; }
      void setRecord(Record *r){ rec=r;}
      MoveRecord *getLastMove(){ return rec->moveOf(lastMoveIndex); }
      void addMoveAndAdvance(Move move);
      NodeRecord *getNode(){ return rec->nodeOf(nodeIndex); }
    };
  
    std::ostream& operator<<(std::ostream&,const MoveRecord &);
    std::ostream& operator<<(std::ostream&,Record &);
  } // namespace record
  using record::Record;
} // namespace osl
#endif /* _RECORD_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
