/* alphaBeta.h
 */
#ifndef _ALPHA_BETA_H
#define _ALPHA_BETA_H

#include "osl/search/realizationProbability.h"
#include "osl/search/searchBase.h"
#include "osl/search/searchWindow.h"
#include "osl/search/killerMoveTable.h"
#include "osl/checkmate/dualCheckMateSearcher.h"
#include "osl/evalTraits.h"

#ifndef USE_CHECKMATE
#  define USE_CHECKMATE 1
#endif

namespace osl
{
  class SimpleHashRecord;
  class MoveStack;
  class MoveLogProbSet;
  class MoveLogProbVector;
  namespace search
  {
    /**
     * ¸ΨѤ alpha beta õ.
     *
     * ֤ state, ߤ limit, root Ǥ depth, limit, ġ
     * 
     * null window search  alpha beta search 롥
     * ̤ι¤:
     * - searchAllMoves:
     *   - Ƥλؼäɾ򤹤롥
     *   - ƥ examineMoves ƤӽФ
     * - examineMoves: ƻؼ˰Ϳ줿 policy Ƥ֡
     * - policy ϥإѥ饹ǡΤΥ᥽å fooSearch Ƥ֡
     *   fooSearch ϥåȽ뤫ɤ֤
     * - fooSearch  searchWithMove ͳ ApplyMove::doUndoMove Ԥʤ
     *   ˥إѤͳ fooSearchAfterMove ƤФ롥
     *   fooSearchAfterMove ϺƵŪ searchAllMoves Ƥӡɾ֤ͤ
     *
     * 1. null window search ξ
     *
     *   mtdf ƤִطǡmoveؤͤĴ٤ؿ
     *   
     *   nullWindowSearchAfterMove ϰʲεǽġ
     *   - null window search
     *   - õ
     *   - ơõ (mtdf  null window ͭ)
     *   
     * 2. ̾ alpha beta
     *   
     *   ƤmoveĴ٤֤ͤؿ
     *
     *   alphaBetaSearchAfterMove:
     *   - null window search (alpha ͤ򹹿뤫)
     *   - alpha beta search (alpha beta õ + õĹ)
     *   
     */
    template<typename State, typename MoveGenerator, typename Table=void, 
	     typename Recorder=NullRecorder, 
	     typename Probabilities=RealizationProbability>
    class AlphaBeta
      : public SearchBase<State,MoveGenerator,Table,Recorder,RealizationProbability>
    {
      typedef SearchBase<State,MoveGenerator,Table,Recorder,RealizationProbability> base_t;
    public:
      typedef State state_t;
      typedef typename State::effect_state_t effect_state_t;
      typedef typename State::eval_t eval_t;
    
      typedef MoveGenerator MoveGenerator_t; // ̾ΤĤɤ?
      typedef Probabilities Probabilities_t;
      typedef AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities> 
      alpha_beta_t;
      /** õ䡤ؼǤΨι⤤꤬äȤ̵뤹ϰ */
      static const int reSearchLimitMargin = 80;
#if 0
      /** comment for debug */
    private:
#endif
      State state;
      int rootDepth;
      int rootLimit;
      int curLimit;

      typedef DualCheckMateSearcher<typename State::hash_state_t> checkmate_t;
      checkmate_t& checkmateSearcher;
      MoveStack& history;
      KillerMoveTable killerMoves;
      /** NullWindow search Ǥ window stack Ѥɬפʤ */
      NullWindow nullWindow;
    public:
      NullWindow getNullWindow() const { return nullWindow; }
      int curDepth() const;
      int getCurLimit() const { return curLimit; }
    public:
      explicit AlphaBeta(const State& s, checkmate_t& checker,
			 MoveStack& moves,
			 const Recorder& r=Recorder(), Table *t=0);
      ~AlphaBeta();

      /**
       * state ΥԡԤ.
       *
       * this->state õλ¸뤬õ exception 
       * ˲Ƥ
       */
      void setState(const State& s);
      /** 
       * entrance of null window search.
       *
       * rootDepth, curLimit are initialized here.
       */
      template <Player P>
      int nullWindowSearch(int threshold, const MoveLogProb& move, 
			   int limit);
      /** 
       * entrance of alpha beta window search.
       *
       * see http://www.logos.t.u-tokyo.ac.jp/~gekisashi/algorithm/re_search.html.
       * rootDepth, curLimit are initialized here.
       */
      template <Player P>
      int alphaBetaSearch(AlphaBetaWindow window, MoveLogProb& bestMove,
			  int limit);
      int alphaBetaSearch(MoveLogProb& bestMove, int limit)
      {
	const Player P = state.getTurn();
	AlphaBetaWindow fullWindow(winValue(alt(P)), winValue(P));
	if (P == BLACK)
	  return alphaBetaSearch<BLACK>(fullWindow, bestMove, limit);
	else
	  return alphaBetaSearch<WHITE>(fullWindow, bestMove, limit);
      }
    
      /**
       * entrance of alpha beta iterative search.
       *
       * MTDF::computeBestMoveIteratively  lastMove ˼뤬
       * lastMove ѤϤ餫 history ˤƤ
       */
      Move computeBestMoveIteratively(int limit,
				      const int step=100, 
				      const int initialLimit=600,
				      int nodeLimit=1600000);

      /**
       * (private) 
       * õޤ move õmaxVal  bestMove (ǽʤ).
       *
       * @return cut Ǥ뤫
       * @param maxVal max(val,maxVal) ˹
       * @param bestMoveExtension true ξ cut ʤǤϺõ
       *   false ʤ cut õ
       */
      template<Player P, bool bestMoveExtension>
      bool nullWindowSearch(const MoveLogProb& move, 
			    int& maxVal, MoveLogProb& bestMove);
      /**
       * (private) 
       * õޤ move õmaxVal  bestMove (ǽʤ).
       *
       * Υ᥽åɤPμ֤ doMove Ԥơ alt(P) μ
       * Υ᥽åɤȴ undoMove  Pμ֤
       * @return P λǺɾ
       * @param baseVal ɾͤ򹹿Ȥߤʤƺõ򤹤
       * @param bestMoveExtension true ξ cut ʤǤϺõ
       *   false ʤ cut õ
       */
      template<Player P, bool bestMoveExtension>
      int nullWindowSearchAfterMove(const MoveLogProb& move, 
				    int baseVal);

      /** (private) */
      template <Player P>
      bool alphaBetaSearch(const MoveLogProb& move, AlphaBetaWindow& window, 
			   int& baseVal, MoveLogProb& bestMove);
      /** (private) 
       * : nullWindow ˲
       */
      template <Player P>
      int alphaBetaSearchAfterMove(const MoveLogProb& move, 
				   AlphaBetaWindow window, 
				   MoveLogProb& bestMove);

    private:
      /**
       * move ؤɾͤ
       *
       * move Ф doUndoMove Ԥʤ Helper Υɤ¹Ԥ롣
       * 
       * @param P 
       * @param Helper ˼¹Ԥ륳
       * @param baseVal ɾͤ򹹿Ȥߤʤƺõ򤹤
       *   (null window search ξ)
       * @return ɾ
       */
      template<Player P, class Helper>
      void searchWithMove(const MoveLogProb& move, Helper& helper);
    public:
      /**
       * (private)ؼäɾ
       * 
       * friend Ĺˤʤ뤿publicˤƤ뤬̤֤˰¸
       * ¼private
       * @param maxval P μǤɾ
       * @param P 
       * @param Policy examineMoves ǤɤΥ᥽åɤƤ֤ݲΡ
       * @param bestMove  ȿ§λʤɤǤʤ
       * @return ɾ
       */
      template<Player P, class Policy>
      int searchAllMoves(const Policy&, MoveLogProb& bestMove);

      /**
       * (private) moves ơ̤ bestmove  maxVal .
       *
       * @param moves β¿ξĤΥƥ꤬ä
       * @param tried ˤμɤɤΤ
       * @param helper ƻؼ˼¹Ԥ륳
       *   null window search ξ nullWindowSearch ƤФ
       * @param isKingEscape 򤱤
       * @param P turn
       * @return Υƥɬפ뤫
       */
      template<Player P, class Helper, bool isKingEscape>
      bool examineMoves(const char *categoryName, const MoveLogProbVector& moves,
			const Helper&,
			MoveLogProb& bestMove, int& maxVal, MoveLogProbSet& tried);
      template<Player P, class Helper>
      bool examineMoves(const char *categoryName, const MoveLogProbVector& moves,
			const Helper&, MoveLogProb& bestMove, int& maxVal);
    }; // class AlphaBeta
  } // namespace search
  using search::AlphaBeta;
} // namespace osl

#endif /* _ALPHA_BETA_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
