#include "osl/search/searchTable.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/search/quiescenceSearch.h"
#include "osl/search/shouldPromoteCut.h"
#include "osl/moveStack.h"
#include "osl/moveLogProbSet.h"
#include "osl/moveLogProbVector.h"
#include "osl/effectUtil.h"
#include "osl/category/categoryListUtil.h"

#ifndef _ALPHABETA_TCC
#define _ALPHABETA_TCC

/** ơõ򤵤ʤȤ */
// #define NO_BEST_MOVE_EXTENSION


/* ------------------------------------------------------------------------- */
/*	Constructors							     */
/* ------------------------------------------------------------------------- */

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
AlphaBeta(const State& s, checkmate_t& c, MoveStack& h,
	  const Recorder& r, Table *t)
  : base_t(r, t), state(s), 
    rootDepth(0), rootLimit(0), curLimit(0),
    checkmateSearcher(c), history(h), nullWindow(0)
{
}
  
template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
~AlphaBeta()
{
}

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
void osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
setState(const State& s)
{
  state = s;
  killerMoves.clear();
}

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
int osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
curDepth() const
{
  return history.size() - rootDepth;
}

/* ------------------------------------------------------------------------- */
/*	Common methods					     		     */
/* ------------------------------------------------------------------------- */

namespace osl
{
  namespace search
  {
    namespace alpha_beta
    {
      /**
       * category ˼
       * examineMoves Ƥ֤helper
       * @param Policy examineMoves Ǽ¹Ԥ륳
       */
      template <typename AlphaBeta, Player P, class Policy>
      struct CategoryHelper
      {
	const Policy& policy;
	MoveLogProb *bestMove;
	int *maxVal;
	MoveLogProbSet *tried;
	CategoryHelper(const Policy& h, MoveLogProb *b, int *m,
		       MoveLogProbSet *t=0)
	  : policy(h), bestMove(b), maxVal(m), tried(t)
	{
	}
	template <bool isKingEscape>
	bool operator()(const char *categoryName, MoveLogProbVector& moves)
	{
	  if (tried)
	    return (*policy.searcher).template examineMoves<P,Policy,isKingEscape>
	      (categoryName, moves, policy, *bestMove, *maxVal, *tried);
	  else
	    return (*policy.searcher).template examineMoves<P,Policy>
	      (categoryName, moves, policy, *bestMove, *maxVal);
	}
      };
    } // namespace alpha_beta
  } // namespace search
} // namespace osl

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P, class Helper>
void osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
searchWithMove(const MoveLogProb& ml, Helper& helper)
{
  const Move move = ml.getMove();
  recorder.countNode(curLimit);

  history.push(move);
  ApplyMove<P>::doUndoMove(state, move, helper);
  history.pop();
}

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P, class Policy, bool isKingEscape>
bool osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
examineMoves(const char *categoryName, const MoveLogProbVector& moves, 
	     const Policy& policy,
	     MoveLogProb& bestMove, int& maxVal, MoveLogProbSet& triedMoves)
{
  assert(P == state.getTurn());
  recorder.newCategory(categoryName, curLimit);
  for(MoveLogProbVector::const_iterator p=moves.begin(); p!=moves.end(); p++)
  {
    if (! isKingEscape)
    {
      if (! triedMoves.assignIfBetter(*p, reSearchLimitMargin))
	continue;
    }
    else
    {
      // ꤫ƨμʤȾԤ
      if (! triedMoves.assignIfBetter(*p, 0))
	continue;
    }
    if (ShouldPromoteCut::canIgnore(p->getMove()))
      continue;
      
    if (policy.searchWithMove(*p, maxVal, bestMove))
      return false;
  }
  return true;
}

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P, class Policy>
inline 
bool osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
examineMoves(const char *categoryName, const MoveLogProbVector& moves, 
	     const Policy& policy, MoveLogProb& bestMove, int& maxVal)
{
  assert(P == state.getTurn());
  recorder.newCategory(categoryName, curLimit);
  for(MoveLogProbVector::const_iterator p=moves.begin(); p!=moves.end(); p++)
  {
    if (ShouldPromoteCut::canIgnore(p->getMove()))
      continue;
    if (policy.searchWithMove(*p, maxVal, bestMove))
      return false;
  }
  return true;
}

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P, class Policy>
int osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
searchAllMoves(const Policy& policy, MoveLogProb& bestMove)
{
  int maxVal=minusInfty(P);
  const Player Opponent = PlayerTraits<P>::opponent;
  // ʬ֤β => ľμ꤬ˡ => 
  if (EffectUtil::isKingInCheck(Opponent,state)) {
    maxVal = winValue(P);
    recorder.recordLeaf(maxVal, curLimit);
    return maxVal;
  }

  // ơ֥ˤͤĴ٤
  const MoveLogProb *bestMoveInTable=0;
  SimpleHashRecord *record = findRecord(state.getHash());
  if (record)
  {
    if (isOutOfWindow(P, *record, curLimit, policy.window(), maxVal, 
		      Policy::isBestMoveExtension))
      return maxVal;
      
    assert(maxVal == minusInfty(P));
    bestMoveInTable = &(record->bestMove);
  }

  /** ;Ϥ root β礭ˡޤ꿼oracle  */

  int checkMateMax
    = (record && (curDepth() == 1))
    ? checkmateSearcher.limitToCheckCount(3500)
    : (((curDepth() <= 2) 
	|| (/*Policy::isBestMoveExtension
	      &&*/ ((rootLimit - curLimit) <= 500)
	    && (curDepth() <= 5)))
       ? checkmateSearcher.limitToCheckCount(curLimit)
       : 1);

  bestMove = MoveLogProb();	// invalidate

  if (curLimit < 100)
  {
    // Ǥˤʤȥơ֥ΥåϤ
    goto move_generation_failure;
  }

  // ơ֥ˤ
  if (bestMoveInTable && validTableMove(state, *bestMoveInTable, curLimit))
  {
    MoveLogProb tryingMove=*bestMoveInTable;
    assert(tryingMove.getLogProb() >= Probabilities_t::tableMove);
    // tryingMove.setLogProbAtMost(Probabilities_t::tableMove);
    if (Policy::isBestMoveExtension)
      tryingMove.setLogProbAtMost(Probabilities_t::reSearch);

    MoveLogProbVector moves;
    moves.push_back(tryingMove);
    const bool tryNext = examineMoves<P,Policy>
      ("table", moves, policy, bestMove, maxVal);
    if ((! tryNext)
	|| (Policy::isBestMoveExtension && bestMove.validMove()))
    {
      assert(bestMove.validMove());
      goto register_table;
    }
  }

  // killer move 
  {
    MoveLogProb killerMove = killerMoves[curDepth()];
    if (killerMove.validMove())
    {
      assert(getPlayer(killerMove.getMove()) == state.getTurn());
      assert(getPlayer(killerMove.getMove()) == P);
      assert(killerMove.getLogProb() >= Probabilities_t::killerMove);
      if (Policy::isBestMoveExtension && (! bestMove.validMove()))
	killerMove.setLogProbAtMost(Probabilities_t::reSearch);
      if (state.template isAlmostValidMove<false>(killerMove.getMove())
	  && (curLimit>=killerMove.getLogProb()))
      {
	MoveLogProbVector moves;
	moves.push_back(killerMove);
	const bool tryNext = examineMoves<P,Policy>
	  ("killer move", moves, policy, bestMove, maxVal);
	if (! tryNext)
	{
	  assert(bestMove.validMove());
	  goto register_table;
	}
      }
    }
  }
    
#if USE_CHECKMATE
  if (record)
  {
    // ;ϤٿȤƤ
    // oracle ϻƤɤ⤷ʤǤϤʤΤǻƤ
    checkMateMax -= record->checkMateNodes;
    if (checkMateMax <= 0)
      goto check_mate_done;
    record->checkMateNodes += checkMateMax;
  }
  else
  {
    checkMateMax /= 2;		// for safety against multiple calls
  }
  
  // ͤޤȤͤ 
  {
    Move checkMateMove=MOVE_INVALID;
    recorder.gotoCheckMateSearch(state, checkMateMax);
    const bool win = checkmateSearcher.
      template isWinningState<P>(checkMateMax, state, checkMateMove);
    recorder.backFromCheckMateSearch();
    if (win)
    {
      assert(checkMateMove);
      assert(isValid(checkMateMove));
      maxVal= winValue(P);
      const MoveLogProb bestMove = MoveLogProb(checkMateMove,100);
      recordToTable(state, SearchTable::CheckMateSpecialDepth, 
		    policy.window(), bestMove, maxVal);
      killerMoves.setMove(curDepth(), bestMove, Probabilities_t::killerMove);
      return maxVal;
    }
  }
check_mate_done:
#endif

  // ˥ơ֥Ͽ줿
  if (record)
  {
    MoveLogProbVector moves;
    for (MoveLogProbSet::const_iterator p=record->moves.begin(); 
	 p!=record->moves.end(); ++p)
    {
      if (p->getLogProb() <= curLimit)
	moves.push_back(*p);
    }
    // ɾͽ¤٤Ȥ
    moves.sortByProbability();
    const bool tryNext = examineMoves<P,Policy>
      ("recorded", moves, policy, bestMove, maxVal);
    if ((! tryNext)
	|| (Policy::isBestMoveExtension && bestMove.validMove()))
    {
      assert(bestMove.validMove());
      goto register_table;
    }
  }

  // ƥ˻
  {
    if (! record)		// Фζ̤ξ礳ǳݤ
      record = allocateRecord(state.getHash(), curLimit);

    const category::CategoryEnv env(&state, curLimit, &history);
    bool cutOccured;
    if (record) 
    {
      if (record->limitOfMoves + reSearchLimitMargin < curLimit)
	record->usedCategories.clear();
      cutOccured = CategoryListUtil::forEachCategory
	(MoveGenerator(), env,
	 alpha_beta::CategoryHelper<AlphaBeta,P,Policy>
	 (policy, &bestMove, &maxVal, (record ? &record->moves : 0)),
	 record->usedCategories);
      record->limitOfMoves = curLimit;
    }
    else
    {
      cutOccured = CategoryListUtil::forEachCategory
	(MoveGenerator(), env,
	 alpha_beta::CategoryHelper<AlphaBeta,P,Policy>
	 (policy, &bestMove, &maxVal, (record ? &record->moves : 0)));
    }
      
    if (! bestMove.validMove())
    {
      assert(! cutOccured);
      // cutOccured ǤʲξϤ
      // (1) ꤬äƤơ
      // (2) 褽Ǻä kingEscapeMoves ¸
      // (3) reSearchLimitMargin ǺäΤ̵̤ˤʤä
      //     9/26 к
      goto move_generation_failure;
    }
  }
register_table:
  assert(bestMove.validMove());
  assert(eval_t::isEvalValueConsistent(maxVal));

  if (! record) // killer move Ǥ뤳ȤΤǡȤ
    record = allocateRecord(state.getHash(), curLimit);
  if (record)
    recordToTable(P, record, curLimit, policy.window(), bestMove, maxVal);

  killerMoves.setMove(curDepth(), bestMove, Probabilities_t::killerMove);
  return maxVal;
move_generation_failure:
  // Ǥʤä
  // limit  100 ̤äȡʾä move generator  limit ʲμǤʤä礳
#if DONT_USE_QUIESCENCESEARCH
  maxVal = state.template computeValWithTakeBack<Opponent>
    (getTo(history.lastMove())); 
#else
  typedef QuiescenceSearch<effect_state_t,eval_t> qsearcher_t;
  qsearcher_t qs(state);
  maxVal = qs.template search<P>(policy.window(), 
				 state.getEval(), getTo(history.lastMove()));
#endif
  recorder.recordTerminal(maxVal, curLimit);
  assert(eval_t::isEvalValueConsistent(maxVal)
	 || (std::cerr << maxVal << " " << policy.window().alpha() << "\n",0));
  return maxVal;
}


/* ------------------------------------------------------------------------- */
/*	Methods for null window search					     */
/* ------------------------------------------------------------------------- */

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
int osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
nullWindowSearch(int threshold, const MoveLogProb& move, int limit)
{
  assert(threshold % 2);
  nullWindow = NullWindow(threshold);
  rootDepth = history.size();
  rootLimit = limit;
  assert(curDepth() == 0);
  curLimit = limit;
  int maxVal = minusInfty(P);
  MoveLogProb bestMove;
  nullWindowSearch<P,false>(move, maxVal, bestMove);
  return maxVal;
}

namespace osl
{
  namespace search
  {
    namespace alpha_beta
    {
      /**
       * examineMoves ͳǼ¹Ԥ륳
       * (null window search ξ)
       */
      template <typename AlphaBeta, Player P,
		bool bestMoveExtension>
      struct NullWindowSearchPolicy
      {
	AlphaBeta *searcher;
	NullWindow nullWindow;
	NullWindowSearchPolicy(AlphaBeta *s, NullWindow window) 
	  : searcher(s), nullWindow(window)
	{
	}
	/** @return cut Ǥ뤫 */
	bool searchWithMove(const MoveLogProb& ml, int& maxVal, 
			    MoveLogProb& bestMove) const
	{
	  return (*searcher).template
	    nullWindowSearch<P,bestMoveExtension>
	    (ml, maxVal, bestMove);
	}
	static const bool isBestMoveExtension = bestMoveExtension;
	NullWindow window() const { return nullWindow; }
      };

      /**
       * searchWithMove ƤФ
       * @param P move ؤ
       */
      template <typename AlphaBeta, Player P, bool bestMoveExtension>
      struct DoUndoNullWindow
      {
	AlphaBeta *searcher;
	const MoveLogProb& moved;
	int *maxVal;
	DoUndoNullWindow(AlphaBeta *s, const MoveLogProb& md, int *m)
	  : searcher(s), moved(md), maxVal(m)
	{
	  assert(P == getPlayer(md.getMove()));
	}
	void operator()(Position /*lastTo*/)
	{
#ifndef NDEBUG
	  const int curLimit = searcher->getCurLimit();
#endif
	  *maxVal = (*searcher).template 
	    nullWindowSearchAfterMove<P,bestMoveExtension>
	    (moved, *maxVal);
	  assert(curLimit == searcher->getCurLimit() ||
		 (std::cout << searcher->state << std::endl <<
		  moved.getMove() << std::endl,0)
	    );
	}
      };
    } // namespace search
  } // namespace alpha_beta
} // namespace osl

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P, bool bestMoveExtension>
bool osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
nullWindowSearch(const MoveLogProb& move, int& maxVal, MoveLogProb& bestMove)
{
  assert(P == getPlayer(move.getMove()));
  assert(P == state.getTurn());
  assert(betterThan(P, nullWindow.beta(), maxVal));
    
  typedef alpha_beta::DoUndoNullWindow<AlphaBeta,P,bestMoveExtension> 
    helper_t;
  int val = maxVal;
  helper_t helper(this, move, &val);
  searchWithMove<P,helper_t>(move, helper);
  if (betterThan(P, val, maxVal))
  {
    maxVal = val;
    bestMove = move;
    if (betterThan(P, val, nullWindow.alpha()))
      return true;
  }
  return false;
}

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P, bool bestMoveExtension>
int osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
nullWindowSearchAfterMove(const MoveLogProb& moved, int baseVal)
{
  assert(alt(P) == state.getTurn());
  assert(P == getPlayer(moved.getMove()));
  assert(betterThan(P, nullWindow.alpha(), baseVal));

  const Player Turn = PlayerTraits<P>::opponent;
  typedef alpha_beta::NullWindowSearchPolicy<AlphaBeta,Turn,false>
    normal_policy_t;
  typedef alpha_beta::NullWindowSearchPolicy<AlphaBeta,Turn,true>
    extend_policy_t;
  MoveLogProb bestMove;

  
  // normal search
  recorder.tryMove(moved,nullWindow.alpha(),curLimit);
  curLimit -= moved.getLogProb();
  assert(0 <= curLimit);

  normal_policy_t normalPolicy(this, getNullWindow());
  int val = searchAllMoves<Turn,normal_policy_t>(normalPolicy, bestMove);

  curLimit += moved.getLogProb();
  recorder.recordValue(moved,val,betterThan(P, val,baseVal),
		       curLimit);

  if (! betterThan(P, val, baseVal))
    return val;
  if ((! bestMoveExtension) && (! betterThan(P, val, nullWindow.alpha())))
    return val;

  // first extension
  if (moved.getLogProb() > Probabilities::reSearch + reSearchLimitMargin) 
  {
    // ᤫ鿼ɤߤäϱĹʤ

    recorder.retryMove(moved,nullWindow.alpha(),curLimit,1);
    curLimit -= Probabilities::reSearch;
    assert(0 <= curLimit);

    val = searchAllMoves<Turn,normal_policy_t>(normalPolicy, bestMove);

    curLimit += Probabilities::reSearch;
    recorder.recordValue(moved,val,betterThan(P, val,baseVal),
			 curLimit);
  }
  
  if (bestMoveExtension)
    return val;
  if (! betterThan(P, val, nullWindow.alpha()))
    return val;

  // second extension
#ifndef NO_BEST_MOVE_EXTENSION
  // ơõ κϥåȤʤƤĹ⡼ɤ
  extend_policy_t extendPolicy(this, getNullWindow());

  recorder.retryMove(moved,nullWindow.alpha(),curLimit,2);
  curLimit -= Probabilities::reSearch;

  val = searchAllMoves<Turn,extend_policy_t>(extendPolicy, bestMove);

  curLimit += Probabilities::reSearch;
  recorder.recordValue(moved, val, betterThan(P, val,baseVal),
		       curLimit);
#endif
  return val;
}


/* ------------------------------------------------------------------------- */
/*	Methods for alpha beta search					     */
/* ------------------------------------------------------------------------- */


template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::Move osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
computeBestMoveIteratively(int limit, const int step, 
			   const int initialLimit, int nodeLimit)
{
  recorder.resetNodeCount();
  int limitIterative = initialLimit;
  MoveLogProb bestMove;
  while (limitIterative < limit)
  {
    recorder.startSearch(limitIterative);
    const int bestValue = alphaBetaSearch(bestMove, limitIterative);
    recorder.finishSearch(bestMove, bestValue);
#ifdef USE_CHECKMATE
    recorder.checkMateStatus(checkmateSearcher.black().getTotalNodeCount());
    recorder.checkMateStatus(checkmateSearcher.white().getTotalNodeCount());
#endif
    if (recorder.currentNodeCount() *8 > nodeLimit)
    {
      std::cerr << "iteration stop\n";
      return bestMove.getMove();
    }
    limitIterative += step;
  }
  std::cerr << "go to final iteration\n";
  recorder.startSearch(limitIterative);
  const int bestValue = alphaBetaSearch(bestMove, limitIterative);
  recorder.finishSearch(bestMove, bestValue);
  return bestMove.getMove();
}

namespace osl
{
  namespace search
  {
    namespace alpha_beta
    {
      /**
       * examineMoves ͳǼ¹Ԥ륳
       * (alpha beta window search ξ)
       */
      template <typename AlphaBeta, Player P>
      struct AlphaBetaSearchPolicy
      {
	AlphaBeta *searcher;
	mutable AlphaBetaWindow abWindow;	
	AlphaBetaSearchPolicy(AlphaBeta *s, AlphaBetaWindow w) 
	  : searcher(s), abWindow(w) 
	{
	}
	/** @return cut Ǥ뤫 */
	bool searchWithMove(const MoveLogProb& ml, int& maxVal, 
			    MoveLogProb& bestMove) const
	{
	  return (*searcher).template
	    alphaBetaSearch<P>(ml, abWindow, maxVal, bestMove);
	}
	static const bool isBestMoveExtension = false;
	AlphaBetaWindow& window() const { return abWindow; }
      };

      /**
       * searchWithMove ƤФ
       * @param P move ؤ(ؤ)
       */
      template <typename AlphaBeta, Player P>
      struct DoUndoAlphaBeta
      {
	AlphaBeta *searcher;
	const MoveLogProb& moved;
	AlphaBetaWindow window;
	int *result;
	DoUndoAlphaBeta(AlphaBeta *s, const MoveLogProb& md, AlphaBetaWindow w,
			int *r)
	  : searcher(s), moved(md), window(w), result(r)
	{
	  assert(P == getPlayer(md.getMove()));
	}
	void operator()(Position /*lastTo*/)
	{
#ifndef NDEBUG
	  const int curLimit = searcher->getCurLimit();
#endif
	  MoveLogProb bestMove;
	  *result = (*searcher).template 
	    alphaBetaSearchAfterMove<P>(moved, window, bestMove);
	  assert(curLimit == searcher->getCurLimit());
	}
      };
    } // namespace search
  } // namespace alpha_beta
} // namespace osl

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
int osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
alphaBetaSearch(AlphaBetaWindow window, MoveLogProb& bestMove, int limit)
{
  assert(window.alpha() % 2 == 0);
  assert(window.beta() % 2 == 0);
  rootDepth = history.size()-1;
 	// nullWindowSearch ȵ;ο·뤿κٹ
  rootLimit = limit;
  assert(curDepth() == 1);
  curLimit = limit;

  // ޤơ֥򸣤
  SimpleHashRecord *record = findRecord(state.getHash());
  if (record)
  {
    int tableVal;
    if (isOutOfWindow(P, *record, limit, window, tableVal, false))
    {
      bestMove = record->bestMove;
      return tableVal;
    }
  }
#if USE_CHECKMATE
  // ;ƤǤߤ root ǤƤǤʤ
  if (record)
  {
    const int rootCheckMateMax = checkmateSearcher.limitToCheckCount(3800);
    int rootCheckMateLimit = rootCheckMateMax - record->checkMateNodes;
    if (rootCheckMateLimit >= 0)
    {
      record->checkMateNodes = rootCheckMateMax;
      if(state.hasEffectBy(alt(P),state.template getKingPosition<P>()))
      {
	// ꤫鲦꤬äƤ
	recorder.gotoCheckMateSearch(state, rootCheckMateLimit);
	const bool lose = checkmateSearcher.
	  template isLosingState<P>(rootCheckMateLimit, state);
	recorder.backFromCheckMateSearch();
	if (lose)
	{
	  const int val=winValue(alt(P));
	  bestMove = MoveLogProb(MOVE_INVALID,100);
	  recordToTable(state, SearchTable::CheckMateSpecialDepth, window,
			bestMove,val); 
	  return val;
	}
      }
      // ͤޤʤСͤޤȤͤ 
      {
	Move checkMateMove;
	recorder.gotoCheckMateSearch(state, rootCheckMateLimit);
	const bool win = checkmateSearcher.
	  template isWinningState<P>(rootCheckMateLimit, state, checkMateMove);
	recorder.backFromCheckMateSearch();
	if (win)
	{
	  const int val = winValue(P);
	  bestMove = MoveLogProb(checkMateMove,100);
	  recordToTable(state, SearchTable::CheckMateSpecialDepth, window,
			bestMove, val);
	  return val;
	}
      }
    }
  }
#endif

  typedef alpha_beta::AlphaBetaSearchPolicy<AlphaBeta,P>
    ab_policy_t;
  ab_policy_t abPolicy(this, window);
  return searchAllMoves<P,ab_policy_t>(abPolicy, bestMove);
}

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
bool osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
alphaBetaSearch(const MoveLogProb& move, 
		AlphaBetaWindow& w, int& baseVal, MoveLogProb& bestMove)
{
  // baseVal ϡlow fail ξǤ⤽餷뤿ɬ
  assert(w.alpha() % 2 == 0);
  assert(w.beta() % 2 == 0);
  assert(P == getPlayer(move.getMove()));
  assert(P == state.getTurn());
  // std::cerr << P << " " << w.beta() << " " << w.alpha() << "\n";
  assert(betterThan(P, w.beta(), w.alpha()));
  assert(! betterThan(P, baseVal, w.alpha()));

  typedef alpha_beta::DoUndoAlphaBeta<AlphaBeta,P> helper_t;
  int val;
  helper_t helper(this, move, w, &val);
  searchWithMove<P,helper_t>(move, helper);
  if (betterThan(P, val, baseVal))
  {
    baseVal = val;
    bestMove = move;
    if (betterThan(P, val, w.alpha()))
    {
      w.alpha() = val;
      if (notLessThan(P, val, w.beta()))
	return true;
    }
  }
  return false;
}

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
int osl::search::AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities>::
alphaBetaSearchAfterMove(const MoveLogProb& moved, AlphaBetaWindow w,
			 MoveLogProb& bestMove)
{
  assert(w.alpha() % 2 == 0);
  assert(w.beta() % 2 == 0);
  assert(alt(P) == state.getTurn());
  assert((! moved.validMove()) || (P == getPlayer(moved.getMove())));
  assert(betterThan(P, w.beta(), w.alpha()));

  const Player Turn = PlayerTraits<P>::opponent;
  typedef alpha_beta::NullWindowSearchPolicy<AlphaBeta,Turn,false>
    null_policy_t;
  typedef alpha_beta::AlphaBetaSearchPolicy<AlphaBeta,Turn>
    ab_policy_t;

  nullWindow.alpha() = w.alpha() + delta(P);
  assert(nullWindow.alpha() % 2);
  // null window search
  recorder.tryMove(moved, nullWindow.alpha(), curLimit);
  curLimit -= moved.getLogProb();
  assert(0 <= curLimit);

  null_policy_t nullPolicy(this, getNullWindow());
  int val = searchAllMoves<Turn,null_policy_t>(nullPolicy, bestMove);

  curLimit += moved.getLogProb();
  recorder.recordValue(moved,val,betterThan(P, val, w.alpha()),
		       curLimit);

  if (! betterThan(P, val, w.alpha()))
    return val;

  // 򹹿õ
  recorder.retryMove(moved, w.alpha(), curLimit, 1);
  curLimit -= Probabilities::reSearch;
  assert(0 <= curLimit);

  ab_policy_t abPolicy(this, w.flipPlayer());
  val = searchAllMoves<Turn,ab_policy_t>(abPolicy, bestMove);

  curLimit += Probabilities::reSearch;
  recorder.recordValue(moved,val,betterThan(P, val, w.alpha()),
		       curLimit);
  
  return val;
}

/* ------------------------------------------------------------------------- */

#endif /* _ALPHABETA_TCC */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
