#ifndef _MTDF_H
#define _MTDF_H

/** ;롼Ƥ֤ɤ */
#define USE_CHECKMATE 1

#include "osl/search/searchTable.h"
#include "osl/search/searchBase.h"
#include "osl/search/alphaBeta.h"
#include "osl/checkmate/dualCheckMateSearcher.h"
#include "osl/moveLogProbVector.h"
#include "osl/evalTraits.h"
#include "osl/loki.h"
#include "osl/applyMove.h"
#include "osl/moveSorter.h"
#include "osl/category/categoryList.h"
#include "osl/moveLogProbSet.h"
#include "osl/moveStack.h"

// for ptypeEvalTable in setInitialGuess()
#include "osl/pieceEval.h"


namespace osl
{
  namespace search
  {
    /**
     * MTDF
     *
     * @param MoveGenerator  CategoryList 
     *   㤨õʤ FWMoveGenerator 
     *   ʣ generator  CategoryList<A,CategoryList<B,C> > ʤɤȥꥹȤǽ
     * @param State can be EvalState
     * @param Probabilities fullwidth ξ FullWidthProbability Ȥ
     * @param Recorder 㤨 NullMtdfRecorder
     *
     * null window õ AlphaBeta Ǥ
     * ˤroot ΡնΥ᥽åɤ
     */
    template<typename State, typename MoveGenerator, typename Table=void, 
	     typename Recorder=NullRecorder, 
	     typename Probabilities=RealizationProbability>
    class MTDF 
      : public SearchBase<State,MoveGenerator,Table,Recorder,RealizationProbability>
    {
      typedef SearchBase<State,MoveGenerator,Table,Recorder,RealizationProbability> base_t;
    public:
      typedef MTDF<State,MoveGenerator,Table,Recorder,Probabilities> mtdf_t;
      typedef AlphaBeta<State,MoveGenerator,Table,Recorder,Probabilities> alpha_beta_t;
    private:
      State state;
      int lastF;
      DualCheckMateSearcher<typename State::hash_state_t> checkmateSearcher;
      MoveStack history;
      alpha_beta_t alphaBetaSearcher;
    public:
      explicit MTDF(State const& s, const Recorder& r=Recorder(), Table *t=0);
      ~MTDF();

      /**
       * ͤ
       */
      void setInitialGuess(int guess) 
      {
	lastF = (guess & (~1)); 
	lastF -= delta(state.getTurn());
	assert(lastF % 2);
      }
      /** f νͤ */
      void setInitialGuess() { 
	setInitialGuess(state.getVal()-ptypeEvalTable.getDelta(state.getTurn()));
      }

      // õ
      /**
       * ̾ƤФ᥽å
       * @param limit 1400Ǻ14ޤɤ
       */
      Move computeBestMove(int limit,Move lastMove=MOVE_INVALID)
      {
	if (lastMove)
	  history.push(lastMove);
	Move bestMove;
	if (state.getTurn()==BLACK)
	  bestMove = computeBestMoveOfPlayer<BLACK>(limit,lastMove);
	else
	  bestMove = computeBestMoveOfPlayer<WHITE>(limit,lastMove);
	if (lastMove)
	  history.pop();
	return bestMove;
      }

      /**
       * computeBestMove  iterative deepening version
       * @param limit controls search depth
       * @param nodeLimit maximum number of nodes
       * @param step iteration step. 
       */
      Move computeBestMoveIteratively(int limit,
				      Move lastMove=MOVE_INVALID,
				      const int step=100, 
				      const int initialLimit=600,
				      int nodeLimit=1600000);

    private:
      // ȥåץ٥礭ʴؿ
      /** ᥤõ */
      template<Player P>
      Move computeBestMoveOfPlayer(int limit,Move lastMove=MOVE_INVALID);
      /** ٤٤Ƥμį᤿ */
      template<Player P>
      Move retryByLowerF(int limit, MtdfMoveSorter<P>& lowerMoves, MtdfMoveSorter<P>& higherMoves);
      /** 
       * F ۤʣμǺ
       *
       * ͤι⤤˸¤١夫2ܰʹߤõ
       * 1ɤβ+delta ۤʤȤǧ롣
       * ۤʤ1ܤμ꤬ꡢۤΤʤ¤ľƤľ
       *
       * @param lowerMoves F ⤤֤õΤǡߤ lastF ۤǽν
       * @param higherMoves ȤF ۤν
       */
      template<Player P>
      Move selectBestMoveByHigherF(int limit, MtdfMoveSorter<P>& lowerMoves, MtdfMoveSorter<P>& higherMoves);
    public:
      /**
       * ȥåץ٥  moves μĴ٤ 
       * lowerMoves  higherMoves ƥʥ󥹤
       * ޤ lastF Ŭڤѹ
       *
       * @param currentVal ߤκ betterThan<P>(lastF, currentVal) 
       * @param lowerMoves lastF ꤬ʤäȤƤμɤ¸
       * @param higherMoves lastF Τߤɤ¸
       */
      template<Player P>
      void examineMoves(const char *categoryName, int& currentVal, 
			int limit, const MoveLogProbVector& moves, 
			MtdfMoveSorter<P>& lowerMoves, 
			MtdfMoveSorter<P>& higherMoves);
      /** 
       * 줿 moves  out ˽
       */
      void gatherMoves(const MoveLogProbVector& moves,
		       MoveLogProbSet& out);
    private:
      /**
       * move õmaxVal (ǽʤ)
       * @return cut Ǥ뤫
       * @param val  move  val
       * @param maxVal max(val,maxVal) ˹
       */
      template<Player P>
      bool testMove(const MoveLogProb& move, int& val,
		    int& maxVal, int limit);
    private:    
      // ʪ
      template <Player P>
      bool lookUpTable(const State& state, int limit, 
		       const MoveLogProb **bestMove, int& val) const
      {
	const SimpleHashRecord *record = TableUtil::find(table, state.getHash());
	if (! record)
	  return false;
	NullWindow w(lastF);
	*bestMove = &(record->bestMove);
	return
	  base_t::template isOutOfWindow<P,NullWindow,false>
	  (*record, limit, w, val);
      }
    
      void recordToTable(const State& state, int limit, 
			 const MoveLogProb& bestMove, int val) const
      {
	base_t::recordToTable
	  (state, limit, NullWindow(lastF), bestMove, val);
      }
    public:
      const DualCheckMateSearcher<typename State::hash_state_t>&
      getCheckMateSearcher() const { return checkmateSearcher; }
    }; // class MTDF
  } // namespace search
  using search::MTDF;
} // namespace osl

#endif /* _MTDF_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
