#ifndef _MTDF_TCC
#define _MTDF_TCC
#include "osl/search/alphaBeta.tcc"
#include "osl/search/shouldPromoteCut.h"
#include "osl/category/categoryListUtil.h"

/* ------------------------------------------------------------------------- */
/*	Constructors							     */
/* ------------------------------------------------------------------------- */

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::search::MTDF<State,MoveGenerator,Table,Recorder,Probabilities>::
MTDF(State const& s, const Recorder& r, Table *t) 
  : base_t(r, t), state(s), 
    alphaBetaSearcher(s, checkmateSearcher, history, r, t)
{
  setInitialGuess();
}
template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::search::MTDF<State,MoveGenerator,Table,Recorder,Probabilities>::
~MTDF() 
{
}

/* ------------------------------------------------------------------------- */
/*	Methods for the toplevel loop					     */
/* ------------------------------------------------------------------------- */

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::Move osl::search::MTDF<State,MoveGenerator,Table,Recorder,Probabilities>::
computeBestMoveIteratively(int limit, Move lastMove, const int step, 
			   const int initialLimit, int nodeLimit)
{
  recorder.resetNodeCount();
  int limitIterative=initialLimit;
  while (limitIterative < limit)
  {
    std::cerr << "iteration " << limitIterative << "\n";
    const Move bestMove = computeBestMove(limitIterative,lastMove);
#if USE_CHECKMATE
    recorder.checkMateStatus(checkmateSearcher.black().getTotalNodeCount());
    recorder.checkMateStatus(checkmateSearcher.white().getTotalNodeCount());
#endif
    // depth  200 䤹8ܤ餤ˤʤȤɤ
    // פ20ܤˤʤ뤳Ȥ⤢
    if (recorder.currentNodeCount() *8 > nodeLimit)
    {
      std::cerr << "iteration stop\n";
      return bestMove;
    }
    limitIterative += step;
  }
  std::cerr << "go to final iteration\n";
  return computeBestMove(limit,lastMove);
}

/**
 ** ᥤõ
 * ǽ low/high fail ȽꤷѴؿ˸Ǥ
 */
template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
osl::Move osl::search::MTDF<State,MoveGenerator,Table,Recorder,Probabilities>::
computeBestMoveOfPlayer(int limit, Move lastMove)
{
  assert(lastF % 2);
  recorder.startSearch(limit);
  alphaBetaSearcher.setState(state);

  // ޤơ֥򸣤
  const MoveLogProb *bestMoveInTable=0;
  SimpleHashRecord *record = findRecord(state.getHash());
  if (record)
  {
    bestMoveInTable = &(record->bestMove);
    int tableVal;
    if (isOutOfWindow(P, *record, limit, NullWindow(lastF), tableVal, false))
      return bestMoveInTable->getMove();
  }

#if USE_CHECKMATE
  // ;ƤǤߤ root ǤƤǤʤ
  if (record)
  {
    const int rootCheckMateMax = checkmateSearcher.limitToCheckCount(3800);
    int rootCheckMateLimit = rootCheckMateMax - record->checkMateNodes;
    if (rootCheckMateLimit >= 0)
    {
      record->checkMateNodes = rootCheckMateMax;
      if(state.hasEffectBy(alt(P),state.template getKingPosition<P>()))
      {
	// ꤫鲦꤬äƤ
	recorder.gotoCheckMateSearch(state, rootCheckMateLimit);
	const bool lose = checkmateSearcher.
	  template isLosingState<P>(rootCheckMateLimit, state);
	recorder.backFromCheckMateSearch();
	if (lose)
	{
	  const int val=winValue(alt(P));
	  recordToTable(state,SearchTable::CheckMateSpecialDepth, 
			MoveLogProb(MOVE_INVALID,100),val); 
	  return MOVE_INVALID;
	}
      }
      // ͤޤʤСͤޤȤͤ 
      {
	Move checkMateMove;
	recorder.gotoCheckMateSearch(state, rootCheckMateLimit);
	const bool win = checkmateSearcher.
	  template isWinningState<P>(rootCheckMateLimit, state, checkMateMove);
	recorder.backFromCheckMateSearch();
	if (win)
	{
	  const int val = winValue(P);
	  const MoveLogProb bestMove = MoveLogProb(checkMateMove,100);
	  recordToTable(state, SearchTable::CheckMateSpecialDepth, 
			bestMove, val);
	  return checkMateMove;
	}
      }
    }
  }
#endif
  /**
   * f ۤȱۤʤä
   */
  MtdfMoveSorter<P> lowerMoves, higherMoves;

  /** 
   * ߤκɾ lastF ˤä upperbound ä
   * lower bound äꤹΤ lastF ѹ 
   */
  int currentVal = minusInfty(P);
  // ǽκ
  recorder.newCategory("table",limit);
  if (bestMoveInTable && validTableMove(state, *bestMoveInTable, limit))
  {
    MoveLogProb tryingMove=*bestMoveInTable;
    tryingMove.setLogProbAtMost(Probabilities::tableMove);

#ifndef NDEBUG
    int val = minusInfty(P);
    // 줳ͤʤäǸФǤ褦
    assert(! eval_t::isEvalValueConsistent(val));
#else
    int val;
#endif
    if (testMove<P>(tryingMove, val, currentVal, limit))
    {
      lastF = currentVal + delta(P);
      higherMoves.insert(std::make_pair(currentVal, tryingMove));
    }
    else if (! isWinValue(alt(P), val))
    {
      lastF = currentVal - delta(P);
      currentVal = minusInfty(P);
    }
  }

  // ƤΥƥμ򽸤
  typedef typename State::effect_state_t effect_state_t;
  MoveLogProbSet moveSet;
  const category::CategoryEnv env(&state, limit, &history);
  CategoryListUtil::gatherAllMoves<MoveGenerator>(env, moveSet);
  // ٤Ƥμ
  MoveLogProbVector allMoves;
  for (MoveLogProbSet::const_iterator p=moveSet.begin(); p!=moveSet.end(); ++p)
  {
    if (higherMoves.empty() 
	|| (p->getMove() != higherMoves.begin()->second.getMove()))
    allMoves.push_back(*p);
  }
  allMoves.sortByProbability();
#if 1
  for (MoveLogProbVector::iterator p=allMoves.begin(); p!=allMoves.end(); ++p)
  {
    p->setLogProbAtMost(1100);	// ϳ̵褦
  }
#endif
  examineMoves<P>("uniqued moves", currentVal, limit, allMoves,
		  lowerMoves, higherMoves);

  if (higherMoves.empty()) {
    // low fail
    return retryByLowerF<P>(limit, lowerMoves, higherMoves);
  }
  // high fail: loop ƬǤ lastF Ͱʾμʣǧ
  return selectBestMoveByHigherF<P>(limit, lowerMoves, higherMoves);
}

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
void osl::search::MTDF<State,MoveGenerator,Table,Recorder,Probabilities>::
examineMoves(const char *categoryName, 
	     int& currentVal, int limit, const MoveLogProbVector& moves, 
	     MtdfMoveSorter<P>& lowerMoves, MtdfMoveSorter<P>& higherMoves) 
{
  recorder.newCategory(categoryName, limit);
  for(MoveLogProbVector::const_iterator p=moves.begin(); p!=moves.end(); ++p)
  {
    if (isWinValue(P, currentVal))
    {
      assert(! higherMoves.empty());
      return;
    }
    if (ShouldPromoteCut::canIgnore(p->getMove()))
      continue;
    
    int val;
    if (testMove<P>(*p, val, currentVal, limit))
    {
      lastF=currentVal + delta(P);
      higherMoves.insert(std::make_pair(currentVal, *p));
    }
    if (higherMoves.empty() 
	&& (! isWinValue(alt(P), val)))
    {
      // low fail ʤΤǤƤ
      lowerMoves.insert(std::make_pair(val, *p));
      lastF = val - delta(P); // ɤɤ f 򲼤Ƹǻ
      currentVal = lastF - delta(P);
    }
  }
#if USE_CHECKMATE
  recorder.checkMateStatus(checkmateSearcher.black().getTotalNodeCount());
  recorder.checkMateStatus(checkmateSearcher.white().getTotalNodeCount());
#endif
}
  
/**
 * ȥåץ٥ 롼 ( low fail )
 * 
 * lastF 򲼤ʤõlastF ꤬ۤǤ selectBestMoveByHigherF 
 */
template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
osl::Move osl::search::MTDF<State,MoveGenerator,Table,Recorder,Probabilities>::
retryByLowerF(int limit, MtdfMoveSorter<P>& lowerMoves, MtdfMoveSorter<P>& higherMoves)
{
  assert(higherMoves.empty());
  MoveLogProbVector moves;
  while (higherMoves.empty() && (lowerMoves.size() > 1))
  {
    recorder.recordTopLevelLowFail(lowerMoves.begin()->second, lastF);
    std::cerr << "low fail " << lastF << "\n";
    for (typename MtdfMoveSorter<P>::const_iterator p=lowerMoves.begin(); p!=lowerMoves.end(); ++p)
      moves.push_back(p->second);
    lowerMoves.clear();
    // 餱λؤϿʤΤ lower moves Ϥ󸺤äƤ椯ǽ
    int currentVal = minusInfty(P);
    examineMoves<P>("low fail", currentVal, limit, 
		    moves, lowerMoves, higherMoves);
    moves.clear();
    lastF -= state.pawnValue()*delta(P)*8; // 8礵˲
  }

  if (! higherMoves.empty())
    return selectBestMoveByHigherF<P>(limit, lowerMoves, higherMoves);

  if (lowerMoves.size() == 0)
    return MOVE_INVALID;

  assert(lowerMoves.size() == 1);
  const MoveLogProb& bestMove = lowerMoves.begin()->second;
  const int val = lowerMoves.begin()->first;
  recordToTable(state, limit, bestMove, val);
  recorder.finishSearch(bestMove, val);
  lastF = val - delta(P);
  return bestMove.getMove();
}

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
osl::Move osl::search::MTDF<State,MoveGenerator,Table,Recorder,Probabilities>::
selectBestMoveByHigherF(int limit, MtdfMoveSorter<P>& lowerMoves, 
			MtdfMoveSorter<P>& higherMoves)
{
  recorder.recordTopLevelHighFail(higherMoves.begin()->second, lastF);
  // lowerMoves ǽΤ򽦤
  {
    assert(! higherMoves.empty());
    int currentVal = higherMoves.begin()->first;
    lastF = currentVal + delta(P);
    MoveLogProbVector movesMayBeBeyondF;
    for (typename MtdfMoveSorter<P>::const_iterator p=lowerMoves.begin(); p!=lowerMoves.end(); ++p)
    {
      // p->first ˤ ¤äƤ롤 lowerMoves Ͼ¤ι⤤¤Ǥ
      if (! betterThan(P, p->first, lastF))
	break;
      movesMayBeBeyondF.push_back(p->second);
    }
    examineMoves<P>("high fail", currentVal, limit,
		    movesMayBeBeyondF, lowerMoves, higherMoves);
  }
    
  while (higherMoves.size() > 1)
  {
    higherMoves.dump(5);
    typename MtdfMoveSorter<P>::iterator p = higherMoves.begin();
    int bestValue = p->first;
    if (isWinValue(P, bestValue))
    {
      // ʬ⤱Ф⤦ɤμǤɤ(ߤϵͤߤ򸫤ĤȤ)
      break;
    }

    MtdfMoveSorter<P> nextJobs;
    nextJobs.insert(*p);
    while (++p != higherMoves.end())
    {
      lastF = bestValue + delta(P);

      MoveLogProb moveToTry = p->second; // copy
      if (higherMoves.size() <= 3)
	moveToTry.setLogProbAtMost(Probabilities::reSearch);

      int val,maxVal=lastF-delta(P);
      if (testMove<P>(moveToTry, val, maxVal, limit))
      {
	bestValue = val;
	nextJobs.insert(std::make_pair(val, moveToTry)); // ȥå׸
      }
    }
    higherMoves.swap(nextJobs);
  }
  higherMoves.dump(1);
  const MoveLogProb& bestMove = higherMoves.begin()->second;
  const int bestValue = higherMoves.begin()->first;
  lastF = bestValue - delta(P);
  recordToTable(state, limit, bestMove, bestValue);
  recorder.finishSearch(bestMove, bestValue);
  return bestMove.getMove();
}

template <typename State,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
bool osl::search::MTDF<State,MoveGenerator,Table,Recorder,Probabilities>::
testMove(const MoveLogProb& move, int& val,
	 int& maxVal, int limit)
{
  assert(P == getPlayer(move.getMove()));
  assert(P == state.getTurn());
  assert(betterThan(P, lastF,maxVal));
  assert(lastF % 2);

  val = alphaBetaSearcher.template 
    nullWindowSearch<P>(lastF, move, limit);
  if (betterThan(P, val,maxVal))
  {
    maxVal = val;
    if (betterThan(P, val,lastF))
      return true;
    // fall thru
  }
  return false;
}

#endif /* _MTDF_TCC */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
