/* quiescenceSearch.tcc
 */
#ifndef _QUIESCENCESEARCH_TCC
#define _QUIESCENCESEARCH_TCC

#include "osl/search/quiescenceSearch.h"
#include "osl/search/shouldPromoteCut.h"
#include "osl/move_generator/capture.h"
#include "osl/move_generator/escape.h"
#include "osl/move_action/store.h"
#include "osl/moveVector.h"
#include "osl/applyMove.h"
#include "osl/effectUtil.h"
#include "osl/neighboring8.h"

namespace osl
{
  namespace search
  {
    template <class QSearch,Player P>
    struct QSearchNextMove
    {
      typedef typename QSearch::eval_t eval_t;
      int& result;
      QSearch *searcher;
      int alpha, beta, depth;
      eval_t ev;
      QSearchNextMove(int& r, QSearch *s, int a, int b, int d, eval_t e)
	: result(r), searcher(s), alpha(a), beta(b), depth(d), ev(e)
      {
      }
      void operator()(Position lastTo)
      {
	result = (*searcher).template search<PlayerTraits<P>::opponent>
	  (beta, alpha, depth+1, ev, lastTo);
      }
    };

    template <class QSearch, Player P>
    struct QSearchGenerate
    {
      const QSearch *searcher;
      MoveVector& out;
      QSearchGenerate(const QSearch *s, MoveVector& o)
	: searcher(s), out(o)
      {
      }
      void operator()(Piece targetPiece)
      {
	(*searcher).template generateCaptureMoves<P>(targetPiece, out);
      }
    };

    template <class QSearch, Player P>
    struct QSearchGenerateWithCheck
    {
      const QSearch *searcher;
      MoveVector& out;
      Piece exceptFor;
      QSearchGenerateWithCheck(const QSearch *s, MoveVector& o, Piece e)
	: searcher(s), out(o), exceptFor(e)
      {
      }
      void operator()(Piece targetPiece)
      {
	if (targetPiece != exceptFor)
	  (*searcher).template generateCaptureMoves<P>(targetPiece, out);
      }
    };
  } // namespace search
} // namespace osl

template <class State, class EvalT>
template <osl::Player P, osl::Ptype PTYPE> 
inline
void osl::search::QuiescenceSearch<State,EvalT>::
generateCaptureMoves(MoveVector& moves, Piece exceptFor) const
{
  if (isPiece(exceptFor) && (getPtype(exceptFor) == PTYPE))
  {
    typedef QSearchGenerateWithCheck<QuiescenceSearch,P> generator_t;
    generator_t generator(this, moves, exceptFor);
    state.template forEachOnBoard<PlayerTraits<P>::opponent,PTYPE,generator_t>
      (generator);
  }
  else
  {
    typedef QSearchGenerate<QuiescenceSearch,P> generator_t;
    generator_t generator(this, moves);
    state.template forEachOnBoard<PlayerTraits<P>::opponent,PTYPE,generator_t>
      (generator);
  }
}

template <class State, class EvalT>
template <osl::Player P> 
inline
void osl::search::QuiescenceSearch<State,EvalT>::
generateCaptureMoves(Piece target, MoveVector& moves) const
{
  move_action::Store store(moves);
  move_generator::Capture<State,move_action::Store>::template
    generateMoves<P>(state, getPosition(target), store);
}

template <class State, class EvalT>
template <osl::Player P, bool allowCut> 
inline
bool osl::search::QuiescenceSearch<State,EvalT>::
examineMoves(int& curVal, const MoveVector& moves, int& alpha, int beta, int depth, eval_t ev)
{
  assert(EvalTraits<P>::notLessThan(alpha, curVal));
  const Player Opponent = PlayerTraits<P>::opponent;
  const Position myKing = state.template getKingPosition<P>();
  const Position opKing = state.template getKingPosition<Opponent>();
  
  for (size_t i=0; i<moves.size(); ++i)
  {
    eval_t newEv = ev;
    const Move move = moves[i];
    ApplyMoveOfTurn::doMove(newEv, state, move);

    if (allowCut)
    {
      // ջ޴δबޤäѤμɤ񤯤®

      // ꤬ߤǤʤϢ³Ƽǽ㤤Τǡͤ㤤ϻ޴
      const Position to = getTo(move);
      if (EvalTraits<P>::notLessThan(curVal, newEv.getVal())
	  && (! Neighboring8::isNeighboring8(to, myKing))
	  && (! state.hasEffectFromTo(getPtypeO(move), to, opKing)))
	continue;

      // »ۤܳμ¤ʶϻ޴ (Ȥȴ)
      // TODO: computeValWithMove Ƥ֤ɤƤ
      if (state.hasEffectBy(Opponent, to)
	  && (ptypeEvalTable.getVal(getOldPtype(move))
	      > ptypeEvalTable.getVal(getCapturePtype(move)))
	  && (! state.hasEffectNotBy(P, state.
				     getPieceOnBoard(getFrom(move)), to)))
	continue;
      const Ptype ptype = getPtype(move);
      // 
      if (ShouldPromoteCut::canIgnore<P>(ptype, to))
	continue;
    }
    
    // std::cerr << "+ " << depth << " " << move << " " << alpha << " " << beta << "\n";
    int result;
    QSearchNextMove<QuiescenceSearch,P> 
      helper(result, this, alpha, beta, depth, newEv);
    ApplyMove<P>::doUndoMove(state, move, helper);
    // std::cerr << "- " << depth << " " << move << " " << result << "\n";
    if (EvalTraits<P>::betterThan(result, curVal))
    {
      curVal = result;
      if (EvalTraits<P>::betterThan(result, alpha))
      {
	alpha = result;
	if (EvalTraits<P>::notLessThan(result, beta))
	  return true;
      }
    }
  }
  return false;
}

template <class State, class EvalT>
template <osl::Player P>
int osl::search::QuiescenceSearch<State,EvalT>::
search(int alpha, int beta, int depth, eval_t ev, Position lastTo)
{
  assert((alpha % 2) ==0);
  assert((beta % 2) == 0);
  assert(EvalTraits<P>::notLessThan(beta, alpha));
  
  if (EffectUtil::isKingInCheck(alt(P), state))
    return winValue(P);

  MoveVector moves;
  if (EffectUtil::isKingInCheck(P, state))
  {
    int curVal = winValue(alt(P));
    // ꤬äƤϤƨ
    move_action::Store store(moves);
    move_generator::Escape<P,State,move_action::Store>::
      generateMoves(state, state.template getKingPiece<P>(), store);
    examineMoves<P,false>(curVal, moves, alpha, beta, depth, ev);
    // ͤdrop β̵꤬еˤʤɤ
    return curVal;
  }

  // ǤʤХѥǧΤǺǽ getVal() 
  int curVal = ev.getVal();
  if (depth >= maxDepth)
    return curVal;
  alpha = EvalTraits<P>::max(alpha, curVal);
  if (EvalTraits<P>::notLessThan(alpha, beta))
    return curVal;

  // TakeBack ͥ
  const Piece lastPiece = state.getPieceOnBoard(lastTo);
  generateCaptureMoves<P>(lastPiece, moves);
  if (examineMoves<P,true>(curVal, moves, alpha, beta, depth, ev))
      return curVal;
  
  // ϼ
  moves.clear();
  generateCaptureMoves<P,ROOK>(moves, lastPiece);
  if (examineMoves<P,true>(curVal, moves, alpha, beta, depth, ev))
    return curVal;

  moves.clear();
  generateCaptureMoves<P,BISHOP>(moves, lastPiece);
  if (examineMoves<P,true>(curVal, moves, alpha, beta, depth, ev))
    return curVal;

  moves.clear();
  generateCaptureMoves<P,GOLD>(moves, lastPiece);
  if (examineMoves<P,true>(curVal, moves, alpha, beta, depth, ev))
    return curVal;

  moves.clear();
  generateCaptureMoves<P,SILVER>(moves, lastPiece);
  if (examineMoves<P,true>(curVal, moves, alpha, beta, depth, ev))
    return curVal;

  moves.clear();
  generateCaptureMoves<P,KNIGHT>(moves, lastPiece);
  if (examineMoves<P,true>(curVal, moves, alpha, beta, depth, ev))
    return curVal;

  moves.clear();
  generateCaptureMoves<P,LANCE>(moves, lastPiece);
  if (examineMoves<P,true>(curVal, moves, alpha, beta, depth, ev))
    return curVal;

  moves.clear();
  generateCaptureMoves<P,PAWN>(moves, lastPiece);
  examineMoves<P,true>(curVal, moves, alpha, beta, depth, ev);
  return curVal;
}


#endif /* _QUIESCENCESEARCH_TCC */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
