/* searchBase.h
 */
#ifndef _SEARCHBASE_H
#define _SEARCHBASE_H

#include "osl/search/searchEvalBase.h"
#include "osl/search/searchRecorder.h"
#include "osl/search/realizationProbability.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/search/tableUtil.h"
#include "osl/search/searchWindow.h"

namespace osl
{
  namespace search
  {
    /**
     * MTDF  AlphaBeta ˶̤
     * ʪΥ᥽åɤ
     */
    template<typename State, typename MoveGenerator, typename Table=void, 
	     typename Recorder=NullRecorder,
	     typename Probabilities=RealizationProbability>
    struct SearchBase : protected SearchEvalBase<typename State::eval_t>
    {
      // types
      typedef SearchEvalBase<typename State::eval_t> base_t;
      typedef State state_t;
      typedef typename State::effect_state_t effect_state_t;
      typedef typename State::hash_state_t hash_state_t;
      typedef typename State::HashType HashKey;
    
      typedef MoveGenerator MoveGenerator_t; // ̾ΤĤɤ?
      typedef Probabilities Probabilities_t;
    protected:
      const Recorder& recorder;
      Table *table;		// acquaintance
    public:
      SearchBase(const Recorder& r, Table *t) : recorder(r), table(t)
      {
      }
      virtual ~SearchBase() {}
    
      /** 
       * ơ֥λؼΤ
       *
       * ĤǤ˼¸ΨΥå⤹
       * HashKey conflict table 鸣 move Ǥ뤳Ȥ
       * conflict  turn ΰ㤤 isAlmostValidMove ǤϥåǤʤ
       * Ȥ
       */
      bool validTableMove(const State& state,
			  const MoveLogProb& bestMoveInTable, int limit) const
      {
	if ((limit < bestMoveInTable.getLogProb())
	    || (limit < Probabilities::tableMove)
	    || (! bestMoveInTable.validMove())) 
	  return false;
	assert(bestMoveInTable.getMove());
	const bool valid = 
	  state.template isAlmostValidMove<false>(bestMoveInTable.getMove());
	if (! valid)
	  recorder.recordInvalidMoveInTable(bestMoveInTable, limit);
	return valid;
      }

      /** 
       * Record Window ٤ cut Ǥ뤫ɤȽꤹ
       *
       * @return cut Ǥ뤫ɤ
       * @param val cut  upper/lower bound 
       * @param bestMoveExtension ξ window ۤƤƤ
       *   ͤǤʤ cut ʤƱõ
       *   줫ɤγΨ⤯Ȥ뤳ȤǱĹ˻Ѥ뤿
       */
      bool isOutOfWindow(Player P, const SimpleHashRecord& record, int limit, 
			 NullWindow w, int& val, bool bestMoveExtension) const
      {
	int tableVal;
	if (record.hasGreaterLowerBound(P, limit, w.beta(), tableVal))
	{
	  assert(eval_t::isEvalValueConsistent(tableVal));
	  recorder.tableHitLowerBound(P, tableVal, w.beta(), limit);
	  if ((! bestMoveExtension)
	      || isWinValue(P, tableVal))
	  {
	    val = tableVal;
	    return true;
	  }
	} 
	if (record.hasLesserUpperBound(P, limit, w.alpha(), tableVal))
	{
	  assert(eval_t::isEvalValueConsistent(tableVal));
	  recorder.tableHitUpperBound(P, tableVal, w.alpha(), limit);
	  if ((! bestMoveExtension)
	      || isWinValue(alt(P), tableVal))
	  {
	    val = tableVal;
	    return true;
	  }
	}
	return false;
      }
      /**
       * NullWindow ξȰä window 򶹤뤳Ȥ롥
       * bestMoveExtension ϵˤʤ(󥿡եߥޥåʤΤǤĤ)
       */
      bool isOutOfWindow(Player P, const SimpleHashRecord& record, int limit, 
			 AlphaBetaWindow& w, int& val, bool) const
      {
	int tableVal;
	if (record.hasGreaterLowerBound(P, limit, w.alpha(), tableVal))
	{
	  assert(eval_t::isEvalValueConsistent(tableVal));
	  w.alpha() = tableVal;
	  if (notLessThan(P, tableVal, w.beta()))
	  {
	    recorder.tableHitLowerBound(P, tableVal, w.beta(), limit);
	    val = tableVal;
	    return true;
	  }
	} 
	if (record.hasLesserUpperBound(P, limit, w.beta(), tableVal))
	{
	  assert(eval_t::isEvalValueConsistent(tableVal));
	  w.beta() = tableVal;
	  if (notLessThan(P, w.alpha(), tableVal))
	  {
	    recorder.tableHitUpperBound(P, tableVal, w.alpha(), limit);
	    val = tableVal;
	    return true;
	  }
	}
	return false;
      }

      SimpleHashRecord* findRecord(const HashKey& key) const
      {
	return TableUtil::find(table, key);
      }
      SimpleHashRecord* allocateRecord(const HashKey& key, int limit) const
      {
	return TableUtil::allocate(table, key, limit);
      }

      /** SimpleHashRecord* Τʤinterface */
      template <class Window>
      void recordToTable(const State& state, int limit, const Window& w,
			 const MoveLogProb& bestMove, int val) const
      {
	assert(eval_t::isEvalValueConsistent(val));
	SimpleHashRecord *record = 
	  TableUtil::allocate(table, state.getHash(), limit);
	if (record)
	  recordToTable(state.getTurn(), record, limit, w, bestMove, val);
      }

      /** 󥿡ե */
      void recordToTable(Player P,
			 SimpleHashRecord *record, int limit, NullWindow w,
			 const MoveLogProb& bestMove, int val) const
      {
	MoveLogProb storeMove = bestMove;
	storeMove.setLogProbAtLeast(Probabilities::tableMove);
	if (isWinValue(P,val) || isWinValue(alt(P),val))
	{
	  recorder.tableStoreLowerBound(P, bestMove, val, w.alpha(), limit);
	  record->setLowerBound(P, limit, storeMove, val);
	  recorder.tableStoreUpperBound(P, bestMove, val, w.alpha(), limit);
	  record->setUpperBound(P, limit, storeMove, val);
	}
	else if (betterThan(P, val, w.alpha()))
	{
	  recorder.tableStoreLowerBound(P, bestMove, val, w.alpha(), limit);
	  record->setLowerBound(P, limit, storeMove, val);
	}
	else
	{
	  recorder.tableStoreUpperBound(P, bestMove, val, w.alpha(), limit);
	  record->setUpperBound(P, limit, storeMove, val);
	}
      }
      void recordToTable(Player P, SimpleHashRecord *record, int limit, 
			 const AlphaBetaWindow& w,
			 const MoveLogProb& bestMove, int val) const
      {
	MoveLogProb storeMove = bestMove;
	storeMove.setLogProbAtLeast(Probabilities::tableMove);
	if (isWinValue(P,val) || isWinValue(alt(P),val))
	{
	  recorder.tableStoreUpperBound(P, bestMove, val, w.beta(), limit);
	  record->setUpperBound(P, limit, storeMove, val);
	  recorder.tableStoreLowerBound(P, bestMove, val, w.alpha(), limit);
	  record->setLowerBound(P, limit, storeMove, val);
	}
	else if (betterThan(P, w.alpha(), val))
	{
	  recorder.tableStoreUpperBound(P, bestMove, val, w.alpha(), limit);
	  record->setUpperBound(P, limit, storeMove, val);
	}
	else if (betterThan(P, w.beta(), val))
	{
	  recorder.tableStoreUpperBound(P, bestMove, val, w.beta(), limit);
	  record->setUpperBound(P, limit, storeMove, val);
	  recorder.tableStoreLowerBound(P, bestMove, val, w.alpha(), limit);
	  record->setLowerBound(P, limit, storeMove, val);
	}
	else
	{
	  recorder.tableStoreLowerBound(P, bestMove, val, w.beta(), limit);
	  record->setLowerBound(P, limit, storeMove, val);
	}
      }

      using base_t::isWinValue;
      using base_t::winValue;
      using base_t::delta;
      using base_t::minusInfty;
    
    };
  } // namespace search
} // namespace osl


#endif /* _SEARCHBASE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
