/* searchRecorder.h
 */
#ifndef _MTDF_RECORDER_H
#define _MTDF_RECORDER_H

#include "osl/player.h"
#include <boost/scoped_ptr.hpp>
#include <iosfwd>
namespace osl
{
  class MoveLogProb;
  class PawnMaskState;
  namespace search
  {
    /**
     * recorder of MTDF/AlphaBeta
     *
     * TODO: mutable 
     */
    class NullRecorder
    {
      mutable int nodeCount;
    public:
      virtual ~NullRecorder();

      /** õ潪λǻȤ褦ɬ */
      void countNode(int /*limit*/) const { ++nodeCount; }
      void resetNodeCount() const { nodeCount = 0; }
      int currentNodeCount() const { return nodeCount; }
    
      /** recordValue ȥåȤǸƤ */
      void tryMove(const MoveLogProb& m, int lastF, int limit) const {}
      /** recordValue ȥåȤǸƤ */
      void retryMove(const MoveLogProb& m, int lastF, int limit,
		     int retryCount) const {}
      /** tryMove ȥåȤǸƤ */
      void recordValue(const MoveLogProb& m, int val, bool betterMove, int limit) const {}

      /** ˿εϿ */
      void recordLeaf(int val, int limit) const {}
      /** ˿εϿ */
      void recordTerminal(int val, int limit) const {}

      /** ˿εϿ */
      void recordTopLevelLowFail(const MoveLogProb& /* best */, int lastF) const {}
      void recordTopLevelHighFail(const MoveLogProb& bestMove, int lastF) const {}

      void tableHitLowerBound(Player p, int val, int lastF, int limit) const {}
      void tableHitUpperBound(Player p, int val, int lastF, int limit) const {}

      void tableStoreLowerBound(Player p, const MoveLogProb&, int val, int lastF, int limit) const {}
      void tableStoreUpperBound(Player p, const MoveLogProb&, int val, int lastF, int limit) const {}
    
      void checkMateStatus(int nodeCount) const {}

      void startSearch(int limit) const {}
      /** ٤Ƥⵤˤʤ */
      virtual void finishSearch(const MoveLogProb& /* best */, int lastF) const {}

      void recordInvalidMoveInTable(const MoveLogProb& move, int limit) const {}
      void newCategory(const char *name, int limit) const {}

      /** ;̵¥롼ȯ */
      void gotoCheckMateSearch(const PawnMaskState&, int nodeLimit) const {}
      void backFromCheckMateSearch() const {}
    };

    class CountRecorder : public NullRecorder
    {
    protected:
      mutable int numTerminalNodes, numLeafNodes, numTotalNodes;
    public:
      CountRecorder();
      ~CountRecorder();

      void recordValue(const MoveLogProb& m, int val, bool betterMove, int limit) const
      {
	++numTotalNodes;
      }
      void recordLeaf(int val, int limit) const
      {
	++numLeafNodes;
      }
      void recordTerminal(int val, int limit) const
      {
	++numTerminalNodes;
      }
      void startSearch(int limit) const
      {
	numTerminalNodes = numLeafNodes = numTotalNodes = 0;
      }
      void finishSearch(const MoveLogProb& bestMove, int lastF) const;
    };

    class SearchRecorder : public CountRecorder
    {
      class Recorder;
      /** hide implementation */
      boost::scoped_ptr<Recorder> recorder;
    public:
      explicit SearchRecorder(const char *filename="mtdf.log");
      ~SearchRecorder();

      /** ɤٿϿ뤫ؼ */
      void setLogMargin(int margin=500);

      void tryMove(const MoveLogProb& m, int lastF, int limit) const;
      void retryMove(const MoveLogProb& m, int lastF, int limit,
		     int retryCount) const;

      void recordValue(const MoveLogProb& m, int val, bool betterMove, int limit) const;
      void recordLeaf(int val, int limit) const;

      void tableHitLowerBound(Player p, int val, int lastF, int limit) const;
      void tableHitUpperBound(Player p, int val, int lastF, int limit) const;

      void tableStoreLowerBound(Player p, const MoveLogProb& bestMove, int val, int lastF, int limit) const;
      void tableStoreUpperBound(Player p, const MoveLogProb& bestMove, int val, int lastF, int limit) const;

      void recordTopLevelLowFail(const MoveLogProb& /* best */, int lastF) const;
      void recordTopLevelHighFail(const MoveLogProb& bestMove, int lastF) const;

      void startSearch(int limit) const;
      void finishSearch(const MoveLogProb& bestMove, int lastF) const;
      void checkMateStatus(int nodeCount) const;

      void recordInvalidMoveInTable(const MoveLogProb& move, int limit) const;
      void newCategory(const char *name, int limit) const;

      void gotoCheckMateSearch(const PawnMaskState&, int nodeLimit) const;
      void backFromCheckMateSearch() const;

      /** ˥å򺮤Ȥ˻Ȥ */
      std::ostream& stream() const;
    };
  } // namespace search

  using search::NullRecorder;
  using search::CountRecorder;
  using search::SearchRecorder;
} // namespace osl


#endif /* _MTDF_RECORDER_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
