#ifndef _SIGNATURE_EFFECT_H
#define _SIGNATURE_EFFECT_H

#include "osl/board.h"
#include "osl/piece.h"
#include "osl/carray.h"
#include "osl/moveSignature.h"

namespace osl 
{

  /**
   *
   */
  class SignatureTable{
    CArray<MoveSignature,16*(9+4)> signatures;
  public:
    template<typename State>
    SignatureTable(State& state){
      std::fill(signatures.begin(),signatures.end(),signature_EDGE);
      for(int y=1;y<=9;y++)
	for(int x=1;x<=9;x++){
	  Position pos=newPosition(x,y);
	  PtypeO ptypeO=getPtypeO(state.getPieceAt(pos));
	  set(pos,moveSignatureTable.getSignature(ptypeO));
	}
    }
    MoveSignature get(Position pos) const{
      return signatures[pos];
    }
    void set(Position pos,MoveSignature signature){ 
      signatures[pos]=signature;
    }
  };

  template <Player P,class State>
  struct ApplyDoUndoSimpleMove;
  template <Player P,class State>
  struct ApplyDoUndoCaptureMove;
  template <Player P,class State>
  struct ApplyDoUndoDropMove;

  template<class State>
  class SignatureEffect: public State{
  public:
    SignatureTable signatureTable;
    /** Ѿ饹 effect ĤФ˻Ѥ */
    typedef SignatureEffect<State> effect_state_t;
  public:
    template<typename OrigState>
    explicit SignatureEffect(OrigState const& st) :State(st),signatureTable(st){}
    MoveSignature getSignature(Position pos) const{
      return signatureTable.get(pos);
    }
    void setSignature(Position pos,MoveSignature signature){ 
      signatureTable.set(pos,signature);
    }

  };

  template<Player P,typename BaseState>
  struct ApplyDoUndoSimpleMove<P,SignatureEffect<BaseState> >
  {
    typedef SignatureEffect<BaseState> state_t;
    template <typename F>
    static void doUndoSimpleMove(state_t& s, 
				 Position from, Position to, int promoteMask,F& func){
      /* signatureset */
      Piece oldPiece=s.getPieceAt(from);
      PtypeO ptypeO=getPtypeO(promoteWithMask(oldPiece,promoteMask));
      MoveSignature oldSignature=s.getSignature(from);
      s.setSignature(from,signature_EMPTY);
      s.setSignature(to,moveSignatureTable.getSignature(ptypeO));

      ApplyDoUndoSimpleMove<P,BaseState>::doUndoSimpleMove(s,from,to,promoteMask,func);

      /* signatureνᤷ */
      s.setSignature(from,oldSignature);
      s.setSignature(to,signature_EMPTY);
    }
  };
  
  template<Player P,typename BaseState>
  struct ApplyDoUndoDropMove<P,SignatureEffect<BaseState> >
  {
    typedef SignatureEffect<BaseState> state_t;
    template <typename F>
    static void doUndoDropMove(state_t& s, 
			       Position to, Ptype ptype,F func){
      /* signatureset */
      PtypeO ptypeO=newPtypeO(P,ptype);
      s.setSignature(to,moveSignatureTable.getSignature(ptypeO));

      ApplyDoUndoDropMove<P,BaseState>::doUndoDropMove(s,to,ptype,func);
      /* signatureundo */
      s.setSignature(to,signature_EMPTY);

    }
  };

  template<Player P,typename BaseState>
  struct ApplyDoUndoCaptureMove<P,SignatureEffect<BaseState> >
  {
    typedef SignatureEffect<BaseState> state_t;
    template <typename F>
    static void doUndoCaptureMove(state_t& s,
				  Position from,Position to, Piece p1, int promoteMask,F func){
      /* signatureset */
      Piece oldPiece=s.getPieceAt(from);
      PtypeO ptypeO=getPtypeO(promoteWithMask(oldPiece,promoteMask));
      MoveSignature oldSignatureFrom=s.getSignature(from);
      s.setSignature(from,signature_EMPTY);
      MoveSignature oldSignatureTo=s.getSignature(to);
      s.setSignature(to,moveSignatureTable.getSignature(ptypeO));
      /* base classθƤӽФ */
      ApplyDoUndoCaptureMove<P,BaseState>::doUndoCaptureMove(s,from,to,p1,promoteMask,func);
      /* signatureνᤷ */
      s.setSignature(from,oldSignatureFrom);
      s.setSignature(to,oldSignatureTo);
    }
  };
}
#endif // _SIGNATURE_EFFECT_H
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
