/* simpleState.h
 */
#ifndef _SIMPLE_STATE_H
#define _SIMPLE_STATE_H

#include "osl/direction.h"
#include "osl/board.h"
#include "osl/ptype.h"
#include "osl/piece.h"
#include "osl/pieceMask.h"
#include "osl/move.h"
#include "osl/player.h"
#include "osl/state.h"
#include "osl/loki.h"
#include "osl/handicap.h"
#include "osl/carray.h"
#include "osl/doUndoMoveLock.h"
#include "osl/effect_action/pieceFilter.h"

#include <iosfwd>

namespace osl
{

  /**
   * ˵Ȥʤ functor
   */
  class NotEmpty{
  public:
    bool operator()(Position pos) const{ return false; }
  };

  class SimpleState;
  std::ostream& operator<<(std::ostream& os,const SimpleState& state);
  bool operator==(const SimpleState& st1,const SimpleState& st2);

  class SimpleState
  {
  public:
    /** Ѿ饹 ĤФ˻Ѥ */
    typedef SimpleState simple_board_t;
  private:
    friend std::ostream& operator<<(std::ostream& os,const SimpleState& state);
    friend bool operator==(const SimpleState& st1,const SimpleState& st2);
    typedef SimpleState state_t;
  public:
    static const bool hasPawnMask=false;
    PieceMask usedMask;
    CArray<PieceMask,2> offBoardMask;
    CArray<Piece,POSITION_SIZE> board;
    /**
     * onboard pieceΤϿƤä
     */
    CArray<Piece,PIECE_SIZE> pieces;

    /** . ĹŪˤϾä⤷ʤʬɬ */
    Player turn;

    friend class ApplyDoUndoSimpleMove<BLACK,SimpleState>;
    friend class ApplyDoUndoSimpleMove<WHITE,SimpleState>;
    friend class ApplyDoUndoCaptureMove<BLACK,SimpleState>;
    friend class ApplyDoUndoCaptureMove<WHITE,SimpleState>;
    friend class ApplyDoUndoDropMove<BLACK,SimpleState>;
    friend class ApplyDoUndoDropMove<WHITE,SimpleState>;
  public:
    // ˴ؤ
    explicit SimpleState();
    explicit SimpleState(Handicap h);
    // publicѾˤϡvirtual destructor롥
    virtual ~SimpleState();
    
    /** ̤ξ֤˽ */
    void init();
    /** ϥǥ˱֤˽ */
    void init(Handicap h);
    // setter getter
    /** 
     * onBoardǤʤ
     */
    Piece SimpleState::getPieceOf(int num) const{
      assert(validNum(num));
      return pieces[num];
    }
    void setPieceOf(int num,Piece p){
      pieces[num]=p;
    }
    template<Player P>
    Piece getKingPiece() const{
      return getPieceOf(KingTraits<P>::index);
    }
    Piece getKingPiece(Player P) const{
      assert(isValid(P));
      if (P==BLACK)
	return getKingPiece<BLACK>();
      else
	return getKingPiece<WHITE>();
    }
    template<Player P>
    Position getKingPosition() const{
      return getPosition(getKingPiece<P>());
    }
    Position getKingPosition(Player player) const{
      assert(isValid(player));
      if (player==BLACK)
	return getKingPosition<BLACK>();
      else
	return getKingPosition<WHITE>();
    }

    void SimpleState::setBoard(Position pos,Piece piece)
    {
      //assert(isValid(pos));
      board[static_cast<int>(pos)]=piece;
    }
    PieceMask& getOffBoardMask(Player p) {
      return offBoardMask[playerToIndex(p)];
    }
    PieceMask& getUsedMask() {return usedMask;}
    const PieceMask& getOffBoardMask(Player p) const {
      return offBoardMask[playerToIndex(p)];
    }
    const PieceMask& getUsedMask() const {return usedMask;}
    bool isOffBoardNum(int num) const{
      return getOffBoardMask(BLACK).isSet(num) ||
	getOffBoardMask(WHITE).isSet(num);
    }
    
    template<Player P>
    bool isPawnMaskSet(int x) const{
      Position pos=newPosition(x,(P==BLACK ? 2 : 1));
      Piece p1=getPieceAt(pos);
      Piece p2=getPieceAt(pos+DirectionTraits<D>::blackOffset);
      Piece p3=getPieceAt(pos+DirectionTraits<D>::blackOffset*2);
      Piece p4=getPieceAt(pos+DirectionTraits<D>::blackOffset*3);
      Piece p5=getPieceAt(pos+DirectionTraits<D>::blackOffset*4);
      Piece p6=getPieceAt(pos+DirectionTraits<D>::blackOffset*5);
      Piece p7=getPieceAt(pos+DirectionTraits<D>::blackOffset*6);
      Piece p8=getPieceAt(pos+DirectionTraits<D>::blackOffset*7);
      return (isPtypeOMask(p1,newPtypeO(P,PAWN)) &
	      isPtypeOMask(p2,newPtypeO(P,PAWN)) &
	      isPtypeOMask(p3,newPtypeO(P,PAWN)) &
	      isPtypeOMask(p4,newPtypeO(P,PAWN)) &
	      isPtypeOMask(p5,newPtypeO(P,PAWN)) &
	      isPtypeOMask(p6,newPtypeO(P,PAWN)) &
	      isPtypeOMask(p7,newPtypeO(P,PAWN)) &
	      isPtypeOMask(p8,newPtypeO(P,PAWN)))==0;
    }
    bool isPawnMaskSet(Player player, int x)const {
      if(player==BLACK)
	return isPawnMaskSet<BLACK>(x);
      else
	return isPawnMaskSet<WHITE>(x);
    }


    void setPiece(Player player,Position pos,Ptype ptype);
    void setPieceAll(Player player);

    /**
     * @param pos  isOnboard Position 12˵(8˵+Ϥ)
     * isOffBoard(pos) ξ PIECE_EDGE ֤
     */
    Piece SimpleState::getPieceAt(Position pos) const
    {
      return board[static_cast<int>(pos)];
    }
    Piece SimpleState::getPieceOnBoard(Position pos) const
    {
      assert(isOnBoard(pos));
      return getPieceAt(pos);
    }

    bool isOnBoardNum(int num) const{
      return isOnBoard(getPieceOf(num));
    }
    template<Player P,Ptype T>
    bool hasOffBoard() const {
      return (getOffBoardMask(P).getMask(PtypeFuns<T>::indexNum) & PtypeFuns<T>::indexMask)!=0;
    }
    bool hasOffBoard(Player player,Ptype ptype) const{
      return (getOffBoardMask(player).getMask(ptypeTable.getIndex(ptype))
	      & ptypeTable.getMaskLow(ptype))!=0;
    }
    /**
     * ǡѡ
     */
    int countOffBoard(Player pl,Ptype ptype) const{
      return countBit(getOffBoardMask(pl).getMask(ptypeTable.getIndex(ptype))
		      & ptypeTable.getMaskLow(ptype));
    }
    
    void setTurn(Player player){
      turn=player;
    }
    Player getTurn() const{
      return turn;
    }
    /**
     * ֤ѹ
     */
    void changeTurn(){
      osl::changeTurn(turn);
    }

    // check
    bool isConsistent(bool showError=true) const;
    /** 顼ɽ򤹤뤫ɤtemplateѥ᡼ˤ® */
    template <bool showError>
    bool isAlmostValidMove(Move move) const;
    /**
     * ˡ꤫ɤñ˸롥̤˰¸åΤߡ
     * 롼ؤʤǤǽϡisValidMove Ѥ롥
     *
     * ̤˰¸븡Ǥ⡤̤ȴ䲦ɤǤ뤫
     * ꡤͤɤϸʤ
     */
    bool isAlmostValidMove(Move move,bool showError=true) const;
    /**
     * ˡ꤫ɤ̩˸롥isAlmostValidMove θޤࡥ
     */
    bool isValidMove(Move move,bool showError=true) const;

    /**
     * ̰ʳʬȿ§Υå
     *
     * static ˤǤ뤬ޤǤΤǸ
     */
    bool isValidMoveByRule(Move move,bool showError) const;

    /**
     * @param from - ޥΰ
     * @param to - ޥΰ
     * @param offset - fromtoؤshort offset
     * fromtoִطˤȤ
     * ǡ֤򤫤å
     */
    bool isEmptyBetween(Position from, Position to,Offset offset) const{
      assert(offset!=OFFSET_ZERO);
      assert(offset==directionTable.getShortOffset(positionDiff(to,from)));
      Position pos=from+offset;
      for(;isEmpty(getPieceAt(pos));pos+=offset){
	assert(isOnBoard(pos));
	if(pos==to) return true;
      }
      return pos==to;
      
    }
    /**
     * @param from - ޥΰ
     * @param to - ޥΰ
     * fromtoִطˤȤ
     * ǡ֤򤫤å
     */
    bool isEmptyBetween(Position from, Position to) const{
      assert(isOnBoard(from) && isOnBoard(to));
      Offset offset=directionTable.getShortOffset(positionDiff(to,from));
      assert(offset!=OFFSET_ZERO);
      return isEmptyBetween(from,to,offset);
    }
    /**
     * attackerptypeoζ𤬤Ȳꤷtarget뤫ɤ
     *  stateupdateʤǳΤ
     * targetPosition϶Ǥɤ
     * ׾ζưƤƤȤϤˡ
     * ʬȤαäʤȸʤ뤳Ȥ
     */
    bool SimpleState::hasEffectFromTo(PtypeO ptypeo,Position attacker,
				      Position target) const
    {
      Offset32 offset32=positionDiff(target,attacker);
      EffectContent effect=ptypeTable.getEffect(ptypeo,offset32);
      if (!hasEffect(effect)) return false;
      if (hasDirectEffect(effect))return true;
      assert(directionTable.getShortOffset(offset32) == effectOffset(effect));
      return isEmptyBetween(attacker,target,effectOffset(effect));
    }
    /**
     * attackerPiecetargetPositionĤɤ
     * targetPosition϶Ǥɤ
     * @param attackerPiece
     * @param targetPosition
     */
    bool hasEffectTo(Piece attackerPiece,Position targetPosition) const{
      assert(isPiece(attackerPiece) && isOnBoard(targetPosition));
      return hasEffectFromTo(getPtypeO(attackerPiece),
			     getPosition(attackerPiece),
			     targetPosition);
    }

    /**
     * μ꤬꤫ɤ
     * Ĥζˤ벦ξattckPieceˤζ
     * ʣζˤ벦ξPIECE_EMPTY
     * @param P(template) - ߤμ(ä, movealt(P))
     * @param setPiece(template) - attackPieceͤ뤫ɤ
     */
    template<Player P,bool setPiece>
    bool lastMoveIsOpenCheck(Position from,Position kingPosition,Piece &attackPiece) const{
      Offset offset=directionTable.getShortOffset(positionDiff(from,kingPosition));
      if(!isEmptyBetween(kingPosition,from,offset)) return false;
      Piece p;
      Position pos;
      for(pos=from+offset;isEmpty(p=getPieceAt(pos));pos+=offset);
      if(!isEdge(p) && hasEffectFromTo(getPtypeO(p),pos,kingPosition)){
	if(setPiece) attackPiece=p;
	return true;
      }
      return false;
    }

    /**
     * μ꤬꤫ɤ
     * Ĥζˤ벦ξattckPieceˤζ
     * ʣζˤ벦ξPIECE_EMPTY
     * @param P(template) - ߤμ(ä, movealt(P))
     */
    template<Player P>
    bool lastMoveIsCheck(Move lastMove,Piece &attackPiece) const{
      assert(getPlayer(lastMove)==alt(P));
      attackPiece=PIECE_EMPTY;
      Position kingPosition=getPosition(getPieceOf(KingTraits<P>::index));
      Position lastTo=getTo(lastMove);
      Position lastFrom=getFrom(lastMove);
      Piece lastPiece=getPieceAt(lastTo);
      if(hasEffectTo(lastPiece,kingPosition)){
	if(!lastMoveIsOpenCheck<P,false>(getFrom(lastMove),kingPosition,attackPiece))
	  attackPiece=lastPiece;
	return true;
      }
      else if(!isOffBoard(lastFrom))
	return lastMoveIsOpenCheck<P,true>(lastFrom,kingPosition,attackPiece);
      else
	return false;
    }

    // ֤
    //
    /** T  isBasic ǤʤƤư unpromote(T) η̤Ʊ */ 
    template<Player P,Ptype T,typename F>
    void forEachOnBoard(F & func) const {
      for(int num=PtypeTraits<T>::indexMin;num<PtypeTraits<T>::indexLimit;num++){
	Piece p=getPieceOf(num);
	if (isOnBoardBy<P>(p)){
	  func(p);
	}
      }
    }
    /**
     *
     */
    template<Player P,Ptype Type,class Action,Direction Dir>
    void forEachEffectOfPieceDir(Position piecePosition,Action & action,Int2Type<false>) const {}
    template<Player P,Ptype Type,class Action,Direction Dir>
    void forEachEffectOfPieceDir(Position piecePosition,Action & action,Int2Type<true>) const {
      const Offset offset=DirectionPlayerTraits<Dir,P>::offset;
#if 0
      assert(offset!=35);
      std::cout << "P=" << P << ",Dir=" << Dir << ",Type=" << Type << ",piecePosition=" << piecePosition << ",offset=" << offset << std::endl;
#endif
      action.template doAction<P>(getPieceAt(piecePosition),piecePosition+offset);
    }

    template<Player P,Ptype Type,class Action,Direction Dir>
    void forEachEffectOfPieceDir(Position piecePosition,Action & action) const {
      forEachEffectOfPieceDir<P,Type,Action,Dir>(piecePosition,action,Int2Type<(PtypeTraits<Type>::moveMask & DirectionTraits<Dir>::mask)!=0>());
    }
    /**
     *
     */
    template<Player P,Ptype Type,class Action,Direction Dir>
    void forEachEffectOfPieceLongDir(Position piecePosition,Action & action,Int2Type<false>) const {}
    template<Player P,Ptype Type,class Action,Direction Dir>
    void forEachEffectOfPieceLongDir(Position piecePosition,Action & action,Int2Type<true>) const {
      Piece piece=getPieceAt(piecePosition);
      const Offset offset=DirectionPlayerTraits<Dir,P>::offset;
      assert(offset!=35);
      Position pos=piecePosition+offset;
#if 0
      std::cout << pos << ",offset=" << offset << std::endl;
#endif
      for(;isEmpty(getPieceAt(pos));pos+=offset)
	action.template doAction<P>(piece,pos);
      action.template doAction<P>(piece,pos);
    }

    template<Player P,Ptype Type,class Action,Direction Dir>
    void forEachEffectOfPieceLongDir(Position piecePosition,Action & action) const {
      forEachEffectOfPieceLongDir<P,Type,Action,Dir>(piecePosition,action,Int2Type<(PtypeTraits<Type>::moveMask & DirectionTraits<Dir>::mask)!=0>());
    }
    /**
     * piecePositionˤˤä뤹٤Ƥposition
     * (ޤ)ˤĤ
     * action¹Ԥ
     */
    template<Player P,Ptype Type,class Action>
    void forEachEffectOfPiece(Position piecePosition,Action & action) const;
    /**
     * pieceˤä뤹٤Ƥposition(ޤˤĤ
     * action¹Ԥ
     */
    template<class Action>
    void forEachEffectOfPiece(Piece piece,Action & action) const;

    /**
     * @param P(template) - ¦
     * @param pos - оݤΥޥ
     * @param action - ¹Ԥ
     */
    template<Player P,class Action,bool checkDone,class PositionCheck>
    void forEachEffectWithCheck(Position pos,Action & action,PositionCheck const& positionCheck) const;
    
    template<Player P,class Action>
    void forEachEffect(Position pos,Action & action) const {
      forEachEffectWithCheck<P,Action,false,NotEmpty>(pos,action,NotEmpty());
    }
    /**
     * @param P(template) - ¦
     * @param target - ĤɸΥޥ
     */
    template<Player P>
    Piece hasEffectPiece(Position target) const;
    /**
     * @param pl - ¦
     * @param target - ĤɸΥޥ
     */
    Piece hasEffectPiece(Player pl,Position target) const;

    template<Player P>
    bool hasEffectBy(Position pos) const{
      return hasEffectPiece<P>(pos)!=PIECE_EMPTY;
    }
    bool hasEffectBy(Player player,Position pos) const{
      if(player==BLACK)
	return hasEffectBy<BLACK>(pos)!=PIECE_EMPTY;
      else
	return hasEffectBy<WHITE>(pos)!=PIECE_EMPTY;
    }
    /**
     *
     */
    template<Player P>
    bool hasEffectByWithRemove(Position target,Position removed) const;
    
    /**
     * @param P(template) - ĤƤ¦Υץ쥤
     * @param target - Ĥ줿
     * @param attackerPiece - multiple attackξPIECE_EMPTY
     *        ǤʤʤĤƤ֤
     */
    template<Player P>
    bool hasEffectBy(Position target,Piece& attackerPiece) const;

    /**
     * ٤ƤɤȤǸƤ
     */
    template<Player P>
    bool hasMultipleEffectBy(Position target) const;
    bool hasMultipleEffectBy(Player player,Position target) const;
    
    template<Player P,class Action>
    void forEachEffectNotBy(Position pos,Piece p,Action & action) const {
      typedef PieceFilter<Action> action_t;
      action_t notByAction(action,p);
      forEachEffect<P,action_t>(pos,notByAction);
    }
    /** 
     * @param P(template) - ¦
     * @param piece - ¦ζ, ζоݳȤ
     * @param target - оݤΥޥ
     */
    template<Player P>
    bool hasEffectNotBy(Piece piece,Position target) const;

    /** 
     * @param player - ¦
     * @param piece - ¦ζ
     * @param target - оݤΥޥ
     */
    bool hasEffectNotBy(Player player,Piece piece,Position target) const {
      if(player==BLACK)
	return hasEffectNotBy<BLACK>(piece,target);
      else
	return hasEffectNotBy<WHITE>(piece,target);
    }


    template<Player P,class Action>
    void forEachEffectPtypeDirection(Position pos,Action & action,Direction dir,Ptype ptype,Int2Type<false> canMove) const{}

    template<Player P,class Action>
    void forEachEffectPtypeDirection(Position pos,Action & action,Direction dir,Ptype ptype,Int2Type<true> canMove) const{
#if 0
            std::cerr << "forEachEffectPtypeDirection<" << P << ">(" <<
      	pos << "," << dir << "," << ptype << ")"  << std::endl;
#endif
      Position pos1=pos-directionTable.template getOffset<P>(dir);
      Piece p=getPieceAt(pos1);
      PtypeO ptypeo=getPtypeO(p);
#if 0
      std::cerr << "p=" << p << ",ptypeo=" << ptypeo << "shortMoveMask=" << ptypeTable.getShortMoveMask(P,ptypeo,dir) << std::endl;
#endif      
      if(ptypeTable.getShortMoveMask(P,ptypeo,dir)!=0 && unpromote(getPtype(ptypeo))==ptype){
	action.template doAction<P>(p,pos);
      }
    }

    template<Player P,Ptype Type,Direction Dir,class Action>
    void forEachEffectPtypeDirection(Position pos,Action & action) const{
      BOOST_STATIC_ASSERT((PtypeTraits<Type>::isBasic));
      forEachEffectPtypeDirection<P,Action>(pos,action,Dir,Type,Int2Type<PtypeDirectionTraits<Type,Dir>::canMove || PtypeDirectionTraits<PtypeFuns<Type>::promotePtype,Dir>::canMove>());
    }

    template<Player P,Ptype Type,class Action>
    void forEachEffectPtype(Position pos,Action & action) const;
    /**
     * ޥˤDirectionǤĹ뤫ɤ
     */
    template<Direction Dir,Player P>
    bool hasEffectDir(Position to) const{
      BOOST_STATIC_ASSERT((DirectionTraits<Dir>::isLong));
      const Offset offset=DirectionPlayerTraits<Dir,P>::offset;
      Piece p;
      for(Position pos=to-offset;isEmpty(p=getPieceAt(pos));pos-=offset);
      /**
       * PTYPE_EDGEƤפassertäƤޤäΤ
       */
      if(isOnBoardBy<P>(p) &&
	 (ptypeTable.getMoveMask(getPtype(p))&DirectionTraits<Dir>::mask)!=0)
	return true;
      return false;
    }


    /** dump: ʬ cerr ɽ롣abort ʤɤ˥ǥХå˻Ȥ */
    bool dump();
  };

  template<class Action>
  void SimpleState::forEachEffectOfPiece(Piece piece,Action & action) const 
  {
    Position piecePosition=getPosition(piece);
    switch(getPtypeO(piece)){
    case NEW_PTYPEO(WHITE,PAWN): forEachEffectOfPiece<WHITE,PAWN,Action>(piecePosition,action); break;
    case NEW_PTYPEO(WHITE,LANCE): forEachEffectOfPiece<WHITE,LANCE,Action>(piecePosition,action); break;
    case NEW_PTYPEO(WHITE,KNIGHT): forEachEffectOfPiece<WHITE,KNIGHT,Action>(piecePosition,action); break;
    case NEW_PTYPEO(WHITE,SILVER): forEachEffectOfPiece<WHITE,SILVER,Action>(piecePosition,action); break;
    case NEW_PTYPEO(WHITE,PPAWN): forEachEffectOfPiece<WHITE,PPAWN,Action>(piecePosition,action); break;
    case NEW_PTYPEO(WHITE,PLANCE): forEachEffectOfPiece<WHITE,PLANCE,Action>(piecePosition,action); break;
    case NEW_PTYPEO(WHITE,PKNIGHT): forEachEffectOfPiece<WHITE,PKNIGHT,Action>(piecePosition,action); break;
    case NEW_PTYPEO(WHITE,PSILVER): forEachEffectOfPiece<WHITE,PSILVER,Action>(piecePosition,action); break;
    case NEW_PTYPEO(WHITE,GOLD): forEachEffectOfPiece<WHITE,GOLD,Action>(piecePosition,action); break;
    case NEW_PTYPEO(WHITE,BISHOP): forEachEffectOfPiece<WHITE,BISHOP,Action>(piecePosition,action); break;
    case NEW_PTYPEO(WHITE,PBISHOP): forEachEffectOfPiece<WHITE,PBISHOP,Action>(piecePosition,action); break;
    case NEW_PTYPEO(WHITE,ROOK): forEachEffectOfPiece<WHITE,ROOK,Action>(piecePosition,action); break;
    case NEW_PTYPEO(WHITE,PROOK): forEachEffectOfPiece<WHITE,PROOK,Action>(piecePosition,action); break;
    case NEW_PTYPEO(WHITE,KING): forEachEffectOfPiece<WHITE,KING,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,PAWN): forEachEffectOfPiece<BLACK,PAWN,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,LANCE): forEachEffectOfPiece<BLACK,LANCE,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,KNIGHT): forEachEffectOfPiece<BLACK,KNIGHT,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,SILVER): forEachEffectOfPiece<BLACK,SILVER,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,PPAWN): forEachEffectOfPiece<BLACK,PPAWN,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,PLANCE): forEachEffectOfPiece<BLACK,PLANCE,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,PKNIGHT): forEachEffectOfPiece<BLACK,PKNIGHT,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,PSILVER): forEachEffectOfPiece<BLACK,PSILVER,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,GOLD): forEachEffectOfPiece<BLACK,GOLD,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,BISHOP): forEachEffectOfPiece<BLACK,BISHOP,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,PBISHOP): forEachEffectOfPiece<BLACK,PBISHOP,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,ROOK): forEachEffectOfPiece<BLACK,ROOK,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,PROOK): forEachEffectOfPiece<BLACK,PROOK,Action>(piecePosition,action); break;
    case NEW_PTYPEO(BLACK,KING): forEachEffectOfPiece<BLACK,KING,Action>(piecePosition,action); break;
    default: assert(0);
    }
  }

  template<Player P,Ptype Type,class Action>
  void SimpleState::forEachEffectPtype(Position pos,Action & action) const
  {
#if 0
    std::cerr << "forEachEffectPtype<" << P << "," << Type << ">(" <<
      pos << ")"  << std::endl;
#endif
    forEachEffectPtypeDirection<P,Type,UL,Action>(pos,action);
    forEachEffectPtypeDirection<P,Type,UR,Action>(pos,action);
    forEachEffectPtypeDirection<P,Type,L,Action>(pos,action);
    forEachEffectPtypeDirection<P,Type,R,Action>(pos,action);
    forEachEffectPtypeDirection<P,Type,DL,Action>(pos,action);
    forEachEffectPtypeDirection<P,Type,D,Action>(pos,action);
    forEachEffectPtypeDirection<P,Type,DR,Action>(pos,action);
    forEachEffectPtypeDirection<P,Type,UUL,Action>(pos,action);
    forEachEffectPtypeDirection<P,Type,UUR,Action>(pos,action);
    // U
    if(Type!=LANCE && Type!=ROOK){
      forEachEffectPtypeDirection<P,Type,U,Action>(pos,action);
    }
    else{
      Position pos1=pos-DirectionPlayerTraits<U,P>::offset;
      Piece p=getPieceAt(pos1);
      if(isEmpty(p)){
	pos1-=DirectionPlayerTraits<U,P>::offset;
	while(isEmpty(p=getPieceAt(pos1))){
	  pos1-=DirectionPlayerTraits<U,P>::offset;
	}
	PtypeO ptypeo=getPtypeO(p);
	int moveMask=ptypeTable.getMoveMask(getPtype(ptypeo))&
	  DirectionTraits<LONG_U>::mask;
	int playerMask= (~(static_cast<signed int>(ptypeo)>>31)) ^
	  PlayerTraits<P>::mask;
	if((moveMask&playerMask)!=0 && unpromote(getPtype(ptypeo))==Type ){
	  action.template doAction<P>(p,pos);
	}
      }
      else{
	PtypeO ptypeo=getPtypeO(p);
	int moveMask=ptypeTable.getMoveMask(getPtype(ptypeo))&
	  (DirectionTraits<U>::mask|DirectionTraits<LONG_U>::mask);
	int playerMask= (~(static_cast<signed int>(ptypeo)>>31)) ^
	  PlayerTraits<P>::mask;
	if((moveMask&playerMask)!=0 && unpromote(getPtype(ptypeo))==Type){
	  action.template doAction<P>(p,pos);
	}
      }
    }
    if(Type==BISHOP){
      for(int num=36;num<=37;num++){
	Piece p=getPieceOf(num);
	if(isOnBoardBy<P>(p)){
	  Position pos1=getPosition(p);
	  EffectContent effect=ptypeTable.getEffectNotLongU(getPtypeO(p),pos,pos1);
	  if(hasEffect(effect)){
	    Offset offset=effectOffset(effect);
	    Position pos2=pos1+offset;
	    do{
	      if(!isEmpty(getPieceAt(pos2))) goto retryBIshop;
	      pos2+=offset;
	    } while(pos2!=pos);
	    action.template doAction<P>(p,pos);
	  }
	retryBIshop:;
	}
      }
    }
    else if(Type==ROOK){
      for(int num=38;num<=39;num++){
	Piece p=getPieceOf(num);
	if(isOnBoardBy<P>(p)){
	  Position pos1=getPosition(p);
	  EffectContent effect=ptypeTable.getEffectNotLongU(getPtypeO(p),pos,pos1);
	  if(hasEffect(effect)){
	    Offset offset=effectOffset(effect);
	    Position pos2=pos1+offset;
	    do{
	      if(!isEmpty(getPieceAt(pos2))) goto retryRook;
	      pos2+=offset;
	    } while(pos2!=pos);
	    action.template doAction<P>(p,pos);
	  }
	retryRook:;
	}
      }
    }
  }


  namespace
  {
    template <bool checkDone> struct CheckDoneIf;
    template <> struct CheckDoneIf<true>
    {
      template <class Action>
      static bool done(Action& a) { return a.done(); }
    };
    template <> struct CheckDoneIf<false>
    {
      template <class Action>
      static bool done(Action&) { return false; }
    };
  } // anonymous namespace

  template<Player P,class Action,bool checkDone,class PositionCheck>
  inline
  void SimpleState::
  forEachEffectWithCheck(Position pos,Action & action,PositionCheck const& positionCheck) const 
  {
    for(int i=0;i<U;i++){
      Direction dir=static_cast<Direction>(i);
      Position pos1=pos-directionTable.template getOffset<P>(dir);
      Piece p=getPieceAt(pos1);
      PtypeO ptypeo=getPtypeO(p);
      if(ptypeTable.getShortMoveMask(P,ptypeo,dir)!=0){
	action.template doAction<P>(p,pos);
	if (CheckDoneIf<checkDone>::done(action)) return;
      }
    }
    for(int i=U+1;i<=SHORT_DIRECTION_MAX;i++){
      Direction dir=static_cast<Direction>(i);
      Position pos1=pos-directionTable.template getOffset<P>(dir);
      Piece p=getPieceAt(pos1);
      PtypeO ptypeo=getPtypeO(p);
      if(ptypeTable.getShortMoveMask(P,ptypeo,dir)!=0){
	action.template doAction<P>(p,pos);
	if (CheckDoneIf<checkDone>::done(action)) return;
      }
    }
    // U
    {
      Position pos1=pos-DirectionPlayerTraits<U,P>::offset;
      Piece p=getPieceAt(pos1);
      if(isEmpty(p) || positionCheck(pos1)){
	pos1-=DirectionPlayerTraits<U,P>::offset;
	while(isEmpty(p=getPieceAt(pos1)) || positionCheck(pos1)){
	  pos1-=DirectionPlayerTraits<U,P>::offset;
	}
	PtypeO ptypeo=getPtypeO(p);
	int moveMask=ptypeTable.getMoveMask(getPtype(ptypeo))&
	  DirectionTraits<LONG_U>::mask;
	int playerMask= (~(static_cast<signed int>(ptypeo)>>31)) ^
	  PlayerTraits<P>::mask;
	if((moveMask&playerMask)!=0){
	  action.template doAction<P>(p,pos);
	  if (CheckDoneIf<checkDone>::done(action)) return;
	}
      }
      else{
	PtypeO ptypeo=getPtypeO(p);
	int moveMask=ptypeTable.getMoveMask(getPtype(ptypeo))&
	  (DirectionTraits<U>::mask|DirectionTraits<LONG_U>::mask);
	int playerMask= (~(static_cast<signed int>(ptypeo)>>31)) ^
	  PlayerTraits<P>::mask;
	if((moveMask&playerMask)!=0){
	  action.template doAction<P>(p,pos);
	  if (CheckDoneIf<checkDone>::done(action)) return;
	}
      }
    }
    for(int num=36;num<=39;num++){
      Piece p=getPieceOf(num);
      if(isOnBoardBy<P>(p)){
	Position pos1=getPosition(p);
	EffectContent effect=ptypeTable.getEffectNotLongU(getPtypeO(p),pos1,pos);
	if(hasEffect(effect)){
	  Offset offset=effectOffset(effect);
	  Position pos2=pos1+offset;
	  do{
	    if(!isEmpty(getPieceAt(pos2)) || positionCheck(pos2)) goto retry;
	    pos2+=offset;
	  } while(pos2!=pos);
	  action.template doAction<P>(p,pos);
	  if (CheckDoneIf<checkDone>::done(action)) return;
	}
      retry:;
      }
    }
  }

  template<Player P,Ptype Type,class Action>
  void SimpleState::
  forEachEffectOfPiece(Position piecePosition,Action & action) const 
  {
    forEachEffectOfPieceDir<P,Type,Action,UL>(piecePosition,action);
    forEachEffectOfPieceDir<P,Type,Action,U>(piecePosition,action);
    forEachEffectOfPieceDir<P,Type,Action,UR>(piecePosition,action);
    forEachEffectOfPieceDir<P,Type,Action,L>(piecePosition,action);
    forEachEffectOfPieceDir<P,Type,Action,R>(piecePosition,action);
    forEachEffectOfPieceDir<P,Type,Action,DL>(piecePosition,action);
    forEachEffectOfPieceDir<P,Type,Action,D>(piecePosition,action);
    forEachEffectOfPieceDir<P,Type,Action,DR>(piecePosition,action);
    forEachEffectOfPieceDir<P,Type,Action,UUL>(piecePosition,action);
    forEachEffectOfPieceDir<P,Type,Action,UUR>(piecePosition,action);
    forEachEffectOfPieceLongDir<P,Type,Action,LONG_UL>(piecePosition,action);
    forEachEffectOfPieceLongDir<P,Type,Action,LONG_U>(piecePosition,action);
    forEachEffectOfPieceLongDir<P,Type,Action,LONG_UR>(piecePosition,action);
    forEachEffectOfPieceLongDir<P,Type,Action,LONG_L>(piecePosition,action);
    forEachEffectOfPieceLongDir<P,Type,Action,LONG_R>(piecePosition,action);
    forEachEffectOfPieceLongDir<P,Type,Action,LONG_DL>(piecePosition,action);
    forEachEffectOfPieceLongDir<P,Type,Action,LONG_D>(piecePosition,action);
    forEachEffectOfPieceLongDir<P,Type,Action,LONG_DR>(piecePosition,action);
  }

  template<Player P>
  struct ApplyDoUndoSimpleMove<P,SimpleState>
  {
    static void prologue(SimpleState& s, 
			 Position from, Position to, int promoteMask,
			 Piece& oldPiece, int& num)
    {
      oldPiece=s.getPieceAt(from);
      Piece newPiece=promoteWithMask(oldPiece,promoteMask);
      newPiece+=(to-from);
      num=getNum(oldPiece);
      s.setPieceOf(num,newPiece);
      s.setBoard(to,newPiece);
      s.setBoard(from,PIECE_EMPTY);
    }
    static void epilogue(SimpleState& s, Position from, Position to,
			 Piece oldPiece, int num)
    {
      // ʬȤθ򳰤
      s.setPieceOf(num,oldPiece);
      s.setBoard(from,oldPiece);
      s.setBoard(to,PIECE_EMPTY);
    }
    template <typename F>
    static void doUndoSimpleMove(SimpleState& s, 
				 Position from, Position to, int promoteMask,F& func);
  };
  
  template<Player P>
  template <typename F>
  void ApplyDoUndoSimpleMove<P,SimpleState>::
  doUndoSimpleMove(SimpleState& s, 
		   Position from, Position to, int promoteMask,F& func)
  {
    Piece oldPiece;
    int num;
    prologue(s, from, to, promoteMask, oldPiece, num);
    s.changeTurn();
    func(to);
    s.changeTurn();
    epilogue(s, from, to, oldPiece, num);
  }

  template<Player P>
  struct ApplyDoUndoDropMove<P,SimpleState>
  {
    typedef SimpleState state_t;

    static void prologue(SimpleState& s, Ptype ptype, Position to,
			 Piece& oldPiece, int& num, int& numIndex, int& numLow)
    {
      numIndex=ptypeTable.getIndex(ptype);
      const unsigned int ownMochigoma=
	s.getOffBoardMask(P).getMask(numIndex) & ptypeTable.getMaskLow(ptype);
      assert(ownMochigoma!=0 || (s.dump(), 0));
      numLow=bsf(ownMochigoma);
      num=numLow|(numIndex<<5);
      oldPiece=s.getPieceOf(num);
      Piece p=oldPiece;
      p+=to-OFFBOARD;

      s.setPieceOf(num,p);
      s.setBoard(to,p);
      s.getOffBoardMask(P).xorMask(numIndex,1<<numLow);
    }

    static void epilogue(SimpleState& s, Position to,
			 Piece oldPiece, int num,
			 int numIndex, int numLow)
    {
      s.getOffBoardMask(P).xorMask(numIndex,1<<numLow);
      s.setBoard(to,PIECE_EMPTY);
      s.setPieceOf(num,oldPiece);
    }
    template <typename F>
    static void doUndoDropMove(SimpleState& s, 
			       Position to, Ptype ptype,F func);
  };
  
  
  template<Player P>
  template <typename F>
  void ApplyDoUndoDropMove<P,SimpleState>::
  doUndoDropMove(SimpleState& s, 
		 Position to, Ptype ptype,F func)
  {
    Piece oldPiece;
    int num, numIndex, numLow;
    prologue(s, ptype, to, oldPiece, num, numIndex, numLow);
    s.changeTurn();
    func(to);
    s.changeTurn();
    epilogue(s, to, oldPiece, num, numIndex, numLow);
  }

  template<Player P>
  struct ApplyDoUndoCaptureMove<P,SimpleState>
  {
    typedef SimpleState state_t;
    template <typename F>
    static void doUndoCaptureMove(state_t& s,
				  Position from,Position to, Piece p1, int promoteMask,F func);

    static
    void prologue(SimpleState& s, 
		  Position from, Position to, Piece target, int promoteMask,
		  Ptype& capturePtype, Piece& oldPiece, int& num0, 
		  int& num1, int& num1Index, int& num1Mask)
    {
      capturePtype=getPtype(target);
      num1=getNum(target);
      num1Index=numToIndex(num1);
      num1Mask=numToMask(num1);

      s.getOffBoardMask(P).xorMask(num1Index,num1Mask);
      oldPiece=s.getPieceAt(from);
      Piece newPiece=promoteWithMask(oldPiece,promoteMask);
      newPiece+=(to-from);
      num0=getNum(oldPiece);
      s.setPieceOf(num0,newPiece);
      s.setPieceOf(num1,captured(target));
      s.setBoard(to,newPiece);
      s.setBoard(from,PIECE_EMPTY);
    }

    static 
    void epilogue(SimpleState& s, Position from, Position to, Piece target, 
		  Piece oldPiece, int num0, 
		  int num1, int num1Index, int num1Mask)
    {
      s.getOffBoardMask(P).xorMask(num1Index,num1Mask);
      s.setPieceOf(num0,oldPiece);
      s.setPieceOf(num1,target);
      s.setBoard(from,oldPiece);
      s.setBoard(to,target);
    }
  };
  
  
  template<Player P>
  template <typename F>
  void ApplyDoUndoCaptureMove<P,SimpleState>::
  doUndoCaptureMove(state_t& s,
		    Position from,Position to, Piece target, int promoteMask,F func)
  {
    Piece oldPiece;
    int num0, num1;
    int num1Index, num1Mask;
    Ptype capturePtype;
    prologue(s, from, to, target, promoteMask,
	     capturePtype, oldPiece, num0, num1, num1Index, num1Mask);
    s.changeTurn();
    func(to);
    s.changeTurn();

    epilogue(s, from, to, target, oldPiece, num0, num1, num1Index, num1Mask);
  }

  

} // namespace osl

#endif /* _SIMPLE_STATE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
