#include "osl/boardBitMask.h"
#include "osl/ptype.h"
#include <iostream>

namespace osl{

  std::ostream& operator<<(std::ostream& os,BoardBitMask const& boardBitMask){
    os << "[";
    for(int i=15;i>=0;i--){
      unsigned char uc=boardBitMask.bMask[i];
      for(int j=7;j>=0;j--){
	if((uc&(1<<j))!=0) os << "1";
	else os<<"0";
      }
      os<<" ";
    }
    return os << "]";
  }
  
  BoardBitMaskTable boardBitMaskTable;

  void BoardBitMaskTable::initMaskOfPosition(){
    for(int i=0;i<POSITION_SIZE;i++){
      maskOfPosition[i].clearAll();
    }
    for(int y=1;y<=9;y++)
      for(int x=1;x<=9;x++){
	Position pos=newPosition(x,y);
	maskOfPosition[pos].setBit(BoardBitMask::positionToOffset(pos));
      }
  }

  static void setBetweenMask(BoardBitMask& mask,Position from,Position to,
			     Ptype ptype){
    Offset offset=
      effectOffset(ptypeTable.getEffect(newPtypeO(BLACK,ptype),positionDiff(to,from)));
    if(offset==OFFSET_ZERO) return;
    mask.clearAll();
    for(Position pos=from+offset;pos!=to;pos+=offset){
      mask.setBit(pos);
    }
  }
  void BoardBitMaskTable::initBetweenMask(){
    for(int j=0;j<POSITION_SIZE;j++){
      for(int i=0;i<POSITION_SIZE;i++){
	rookBetweenMask[i][j].setAll();
	lanceBetweenMask[i][j].setAll();
	bishopBetweenMask[i][j].setAll();
      }
    }
    for(int y1=1;y1<=9;y1++)
      for(int x1=1;x1<=9;x1++){
	Position from=newPosition(x1,y1);
	for(int y2=1;y2<=9;y2++)
	  for(int x2=1;x2<=9;x2++){
	    Position to=newPosition(x2,y2);
	    /**
	     * äбӥåȤΩƤ
	     */
	    setBetweenMask(lanceBetweenMask[from][to],from,to,LANCE);
	    setBetweenMask(bishopBetweenMask[from][to],from,to,BISHOP);
	    setBetweenMask(rookBetweenMask[from][to],from,to,ROOK);
	  }
      }
  }

  BoardBitMaskTable::BoardBitMaskTable(){
    initMaskOfPosition();
    initBetweenMask();
  }
}
