/* addEffect.cc
 */
#include "osl/category/addEffect.h"
#include "osl/category/probTable.tcc"
#include "osl/move_generator/addEffect.h"
#include "osl/move_action/store.h"
#include "osl/pieceEval.h"
#include <string>
#include <fstream>
#include <stdexcept>

namespace osl
{
  namespace category
  {
    DEF_CATEGORY_TABLE(PtypePtypeProbTable, AddEffect);
    DEF_CATEGORY_TABLE(PtypePtypeProbTable, AddEffectToLastMove);
  }
}

const int osl::category::AddEffectToLastMove::pieceValueLimit
= -PtypeEvalTraits<PAWN>::val*2;

void osl::category::AddEffectToLastMove::
generate(const CategoryEnv& env, MoveLogProbVector& out, int *indices)
{
  if (! env.history->hasLastMove())
    return;
  const Move lastMove = env.history->lastMove();
  const Ptype target = getPtype(env.history->lastMove());
  MoveVector moves;
  move_action::Store store(moves);
  move_generator::AddEffect<NumEffectState,move_action::Store>::generateMoves
    (env.state->getTurn(), *env.state, getTo(lastMove), store);

  for (size_t i=0; i<moves.size(); ++i)
  {
    const int pieceValue = PieceEval::
      computeDiffAfterMoveForRP(*env.state, moves[i]);
    if (pieceValue < pieceValueLimit)
      continue;
    const bool supported 
      = env.state->hasEffectBy(alt(getPlayer(lastMove)),getTo(lastMove));
    const size_t index = 
      probTable.getIndex(target,supported, getPtype(moves[i]),pieceValue);
    const int prob = probTable[index];
    if (prob <= env.limit)
    {
      out.push_back(moves[i], prob);
      if (indices)		// FIXME: ٤ɤޤ
	*indices++ = index;
    }
  }
}

void osl::category::AddEffectToLastMove::
generate(const CategoryEnv& env, MoveLogProbVector& out)
{
  generate(env, out, 0);
}

const char * osl::category::AddEffectToLastMove::
getName() 
{
  return "AddEffect!";
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
