/* capture.cc
 */
#include "osl/category/capture.h"
#include "osl/move_generator/capture.h"
#include "osl/move_action/store.h"

namespace osl
{
  namespace category
  {
    DEF_CATEGORY_TABLE(PieceValueTable, ContinueCapture);
    template <osl::Ptype PTYPE> 
    DEF_CATEGORY_TABLE(PieceValueTable, Capture<PTYPE>);
  }
}


template <osl::Ptype PTYPE>
void osl::category::Capture<PTYPE>::
generate(const CategoryEnv& env, MoveVector& out)
{
  move_action::Store a(out);
  for (int i=PtypeTraits<PTYPE>::indexMin; 
       i < PtypeTraits<PTYPE>::indexLimit; i++)
  {
    const Piece target = env.state->getPieceOf(i);
    if (getOwner(target) == alt(env.state->getTurn()))
    {
      move_generator::Capture<NumEffectState,move_action::Store>
	::generateMoves(env.state->getTurn(), *env.state, 
			getPosition(target), a);
    }
  }
}

template <osl::Ptype PTYPE>
const char * osl::category::Capture<PTYPE>::getName()
{
  static std::string name = std::string("Capture-")
    + PtypeTraits<PTYPE>::name();
  return name.c_str(); 
}

template struct osl::category::Capture<osl::ROOK>;
template struct osl::category::Capture<osl::BISHOP>;
template struct osl::category::Capture<osl::GOLD>;
template struct osl::category::Capture<osl::SILVER>;
template struct osl::category::Capture<osl::KNIGHT>;
template struct osl::category::Capture<osl::LANCE>;
template struct osl::category::Capture<osl::PAWN>;


struct osl::category::ContinueCapture::PositionAction
{
  const CategoryEnv::effect_state_t *state;
  Position attackFrom;
  Ptype attackerPtype;
  MoveVector *out;
  PositionAction(const CategoryEnv::effect_state_t *s, 
		 Position from, Ptype ptype,
		 MoveVector *o)
    : state(s), attackFrom(from), attackerPtype(ptype), out(o)
  {
  }
  template<Player P>
  void doAction(Piece, Position pos)
  {
    const Piece target = state->getPieceAt(pos);
    if (isPiece(target) && (getOwner(target) == alt(state->getTurn())))
    {
      assert(getPosition(target)==pos);
      if (isBasic(attackerPtype) && canPromote(attackerPtype)
	  && (canPromotePosition(state->getTurn(), pos)
	      || canPromotePosition(state->getTurn(), attackFrom)))
      {
	// promotion move
	const Move m1 = 
	  newMove(attackFrom, pos, promote(attackerPtype),
		  getPtype(target), true, state->getTurn());
	out->push_back(m1);
      }
      if ((! isBasic(attackerPtype))
	  || canDropTo(state->getTurn(), attackerPtype, pos))
      {
	// no promotion move
	const Move m2 = 
	  newMove(attackFrom, pos, attackerPtype,
		  getPtype(target), false, state->getTurn());
	out->push_back(m2);
      }
    }
  }
};
void osl::category::
ContinueCapture::generate(const CategoryEnv& env, MoveVector& out)
{
  if (! isEffective(env))
    return;
  const Move lastLastMove=env.history->lastMove(2);
  const Position attackFrom = getTo(lastLastMove);
  const Piece attacker = env.state->getPieceOnBoard(attackFrom);
  assert(isPiece(attacker));
  if (getOwner(attacker) != env.state->getTurn())
    return;
	
  PositionAction a(env.state, attackFrom, getPtype(attacker), &out);
  env.state->forEachEffectOfPiece(attacker,a);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
