/* defense.cc
 */
#include "osl/category/defense.h"
#include "osl/category/categoryEnv.h"
#include "osl/pieceEval.h"
#include "osl/effectUtil.h"
#include "osl/effectUtil.tcc"
#include "osl/moveStack.h"

namespace osl
{
  namespace category
  {
DEF_CATEGORY_TABLE(PtypePtypeProbTable, AddSupport);
DEF_CATEGORY_TABLE(PieceValueTable, AntiPromotion);
DEF_CATEGORY_TABLE(PieceValueTable, AddEffectToLastEffect);
  }
}

struct osl::category::AddSupport::Test
{
  const CategoryEnv::effect_state_t *state;
  Player turn;

  Piece result;
  bool supported;
  int value;
  Test(const CategoryEnv::effect_state_t *s, Player t)
    : state(s), turn(t),
      result(PIECE_EMPTY), supported(1), value(0)
  {
  }
  void operator()(Position pos)
  {
    const Piece target = state->getPieceAt(pos);
    assert(isPiece(target));
    if ((getOwner(target) == turn)
	&& state->hasEffectBy(alt(turn), pos))
    {
      update(pos, target);
    }
  }
  /**
   * ʣζĤ:
   * - ʬʤͥ
   * - ͤ⤤ͥ
   */
  void update(Position pos, Piece target)
  {
    const bool supported = state->hasEffectBy(turn, pos);
    if ((! this->supported) && supported)
      return;
    const int targetVal = ptypeEvalTable.getVal(getPtype(target));
    if (((! supported) && this->supported)
	|| (targetVal > value))
    {
      result = target;
      this->supported = supported;
      value = targetVal;
    }
  }
};

/**
 * move ŨΤ̣ζĤ
 * fixme: open attack ˤʤȤʤ
 */
std::pair<bool,size_t> osl::category::
AddSupport::getIndexIf(const CategoryEnv& env, Move move, int pieceValue)
{
  const Player turn = getPlayer(move);
  assert(env.state->getTurn() == turn);
  Test selector(env.state, turn);
  EffectUtil::forEachEffectPieceOfMove(*env.state, move, selector);
  if (! isPiece(selector.result))
    return std::make_pair(false,0u);
    
  const Piece attacker 
    = EffectUtil::cheapestPiece(alt(turn), *env.state, 
				getPosition(selector.result));
  const size_t index
    = probTable.getIndex(getPtype(selector.result), selector.supported, 
			 getPtype(attacker), pieceValue);
  return std::make_pair(true,index);
}

struct osl::category::AddEffectToLastEffect::CheckEffect
{
  bool *result;
  const PawnMaskState& state;
  Move move;
  CheckEffect(bool *r, const PawnMaskState& s, Move m)
    : result(r), state(s), move(m)
  {
  }
  template <Player P>
  void doAction(Piece piece, Position lastAttack)
  {
    if (! isOnBoard(lastAttack))
      return;
    const Piece attacked = state.getPieceOnBoard(lastAttack);
    if (((attacked == PIECE_EMPTY) // Ũζޤ⤢ޤѤʤΤǸɵ
	 || (getOwner(attacked) == state.getTurn()))
	&& state.hasEffectFromTo(getPtypeO(move), getTo(move), lastAttack))
      *result = true;
  }
};
bool osl::category::AddEffectToLastEffect::
isInCategory(const CategoryEnv& env, Move move)
{
  if (! env.history->hasLastMove()) 
    return false;
  const Move lastMove=env.history->lastMove();
  const Position lastTo = getTo(lastMove);
  if (lastTo==getTo(move))
    return false;		// TakeBack Ͻ
  bool result = false;
  CheckEffect action(&result, *env.state, move);
  env.state->forEachEffectOfPiece(env.state->getPieceOnBoard(lastTo), action);
  return result;
}

struct osl::category::AntiPromotion::Test
{
  const CategoryEnv::effect_state_t& state;
  const Player opponent;
  bool result;

  Test(const CategoryEnv::effect_state_t& s, Player o)
    : state(s), opponent(o), result(false)
  {
  }
  void operator()(Position focus)
  {
    // focus ؤɤ
    if (! canPromotePosition(opponent, focus))
      return;
      
    if ((state.getPieceOnBoard(focus) == PIECE_EMPTY)
	&& (! state.hasEffectBy(alt(opponent), focus)))
    {
      const Piece promoter = 
	EffectUtil::promotablePiece(opponent, state, focus); 
      if (isPiece(promoter))
      {
	assert(canPromote(getPtype(promoter)));
	result = true;
      }
    }
  }
};

bool osl::category::AntiPromotion::
isInCategory(const CategoryEnv& env, Move move)
{
  const Player opponent = alt(env.state->getTurn());
  int ymin, ymax;
  if (opponent == BLACK)
  {
    ymin = 1;
    ymax = 4;
  }
  else 
  {
    ymin = 7;
    ymax = 10;
  }
  const Ptype ptype = unpromote(getPtype(move));
  const Position to = getTo(move);
  if ((ptype != BISHOP) && (ptype != ROOK))
  {
    const int toY = getY(to);
    if (opponent == BLACK)
    {
      if (toY > ymax)		// 5-9
	return false;
    }
    else
    {
      if (toY < ymin-1)		// 1-5
	return false;
    }
  }

  Test checker(*env.state, opponent);
  EffectUtil::forEachEffectPositionOfMove(*env.state, move, checker);
  return false;
};

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
