/* checkHashRecord.cc
 */
#include "osl/checkmate/checkHashRecord.h"
#include <iostream>

int osl::checkmate::CheckHashRecord::currentDagFindingID = 0;

void osl::checkmate::CheckHashRecord::propagateCheckMateRecursive()
{
#ifdef CHECK_PROPAGATE_BY_DOMINANCE
  // std::cerr << "prop check " << this << "\n";
  for (CheckHashRecordList::iterator p=moreProvableStates.begin();
       p!=moreProvableStates.end(); ++p)
  {
    if (getProof((*p)->proofDisProof))
    {
      (*p)->finalByDominance = this;
      (*p)->bestMove = bestMove;
      (*p)->propagateCheckMate(proofDisProof);
    }
  }
  // std::cerr << "prop check done\n";
#endif
}

void osl::checkmate::CheckHashRecord::propagateNoCheckMateRecursive()
{
#ifdef CHECK_PROPAGATE_BY_DOMINANCE
  for (CheckHashRecordList::iterator p=lessProvableStates.begin();
       p!=lessProvableStates.end(); ++p)
  {
    if (getDisProof((*p)->proofDisProof))
    {
      (*p)->finalByDominance = this;
      (*p)->propagateNoCheckMate(proofDisProof);
    }
  }
#endif
}

void osl::checkmate::CheckHashRecord::propagateProof(unsigned int proof)
{
#ifdef PROOF_NUMBER_PROPAGATE_BY_DOMINANCE
  for (CheckHashRecordList::iterator p=lessProvableStates.begin();
       p!=lessProvableStates.end(); ++p)
  {
    const unsigned int pp = getProof((*p)->proofDisProof);
    if (pp && (pp < proof))
    {
      if (! (*p)->isVisited)
	(*p)->proofDisProof 
	  = makeProofDisProof(proof, getDisProof((*p)->proofDisProof));
      (*p)->propagateProof(proof);
    }
  }
#endif
}

void osl::checkmate::CheckHashRecord::propagateDisProof(unsigned int disProof)
{
#ifdef PROOF_NUMBER_PROPAGATE_BY_DOMINANCE
  for (CheckHashRecordList::iterator p=moreProvableStates.begin();
       p!=moreProvableStates.end(); ++p)
  {
    const unsigned int pd = getDisProof((*p)->proofDisProof);
    if (pd && (pd < disProof))
    {
      if (! (*p)->isVisited)
	(*p)->proofDisProof 
	  = makeProofDisProof(getProof((*p)->proofDisProof), disProof);
      (*p)->propagateDisProof(disProof);
    }
  }
#endif
}

inline 
osl::checkmate::CheckHashRecord * 
osl::checkmate::CheckHashRecord::grandParentOfDag(bool isAttackNode, int ID)
{
  if ((! parent) || (! parent->parent))
    return 0;
  if (parent->parent->dagFindingID == ID)
    return 0;
  // loop ε
  if (parent->useMaxInsteadOfSum || parent->parent->useMaxInsteadOfSum)
    return 0;
  if ((!isAttackNode)
      ? (getProof(parent->proofDisProof)    != getProof(proofDisProof))
      : (getDisProof(parent->proofDisProof) != getDisProof(proofDisProof)))
    return 0;
  parent->parent->dagFindingID = ID;
  return parent->parent;
}

osl::checkmate::CheckHashRecord *
osl::checkmate::CheckHashRecord::findDagOrigin(CheckHashRecord *other, bool isAttackNode)
{
  if (! other)
    return 0;
  assert(this != other);
  int curID = ++currentDagFindingID;
  // std::cerr << "--- " << this << " " << other << "\n";
  CheckHashRecord *cur = this;    
  this->dagFindingID = curID;
  while (cur)
  {
    int otherID = ++currentDagFindingID;
    CheckHashRecord *work = other;
    work->dagFindingID = otherID;
    while (work)
    {
      if (cur == work)
      {
	// std::cerr << "dag " << this << " " << parent << " ==> " << cur << "\n";
	return cur;
      }
      work = work->grandParentOfDag(isAttackNode, otherID);
    } 
    cur = cur->grandParentOfDag(isAttackNode, curID);
  }
  return 0;
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
