/* checkMateRecorder.cc
 */
#include "osl/checkmate/checkMateRecorder.h"
#include "osl/checkmate/checkHashRecord.h"
#include "osl/checkmate/checkMoveList.h"
#include "osl/pawnMaskState.h"
#include <string>
#include <iostream>
#include <fstream>

/** ͤɬפ˱Խ */
unsigned int osl::checkmate::CheckMateRecorder::DepthTracer::maxLogDepth=1;
/** depth νͤ 0 ˸ */
unsigned int osl::checkmate::CheckMateRecorder::DepthTracer::depth=0;
osl::checkmate::CheckMove *osl::checkmate::CheckMateRecorder::nextMove;
const char * osl::checkmate::CheckMateRecorder::leaveReason=0;

std::ostream& osl::checkmate::CheckMateRecorder::stream()
{
#if 0
  return std::cerr;
#endif
  static std::ofstream os("checkmate.log");
  return os;
}

void osl::checkmate::CheckMateRecorder::dumpRecord(const PawnMaskState& state,
				   const CheckHashRecord *record, 
				   unsigned int proofLimit,
				   unsigned int disProofLimit,
				   unsigned int currentProofNumber,
				   unsigned int currentDisProofNumber)
{
  stream() << "dumpRecord\n";
  stream() << "proofLimit " << proofLimit << "\n"
	   << "disProofLimit " << disProofLimit << "\n"
	   << "cur proofNumber " << currentProofNumber << "\n"
	   << "cur disProofNumber " << currentDisProofNumber << "\n";
  stream() << state;
  stream() << record << "\n";
  if (record)
  {
    stream() << record->proofDisProof << "\n"
	     << "bestMove " << record->bestMove << "\n"
	     << "moves.size()=" << record->moves.size() << "\n";
    for (CheckMoveList::const_iterator p=record->moves.begin();
	 p!=record->moves.end(); ++p)
    {
      stream() << "moves = " << p->move;
      if (p->record)
      {
	stream() << " " << p->record->proofDisProof
		 << " " << p->record << " parent " << p->record->parent;
      }
      stream() << "\n";
    }
  }
  stream() << std::flush;
}

void osl::checkmate::CheckMateRecorder::writeRootLog(const char *msg, unsigned int tableSize,
				     unsigned int continuousNoExpandLoop)
{
#ifdef CHECKMATE_DEBUG
  stream() << msg << " table " << tableSize
	   << " continuousNoExpandLoop "<< continuousNoExpandLoop << "\n"
	   << std::flush;
#endif
}

void osl::checkmate::CheckMateRecorder::stat(const char *msg, Player P,
			     unsigned int tableSize,
			     unsigned int totalNodeCount, 
			     unsigned int totalNodeLimit)
{
  std::cerr << msg << " (" << P << ") "
	    << tableSize << "/"
	    << totalNodeCount << "/" << totalNodeLimit << std::endl;
}

void osl::checkmate::CheckMateRecorder::enter(int depth, const char *name, 
			      const CheckHashRecord *record, 
			      unsigned int proofLimit, unsigned int disProofLimit)
{
  const std::string spacer(depth,' ');
  stream() << spacer << "==> " << name << " ";
  if (nextMove)
  {
    csaShow(stream(), nextMove->move);
    if (nextMove->record != record)
      std::cerr << "nextMove->record " << nextMove->record
		<< " != record " << record << "\n";
  }
  else
    stream() << "root";
  stream() << " p " << proofLimit << " d " << disProofLimit;
  if (record)
  {
    stream() << " " << record->proofDisProof;
    if (record->isConfluence)
      stream() << " (c)";
    if (record->useMaxInsteadOfSum)
      stream() << " (max)";
  }
  stream() << ' ' << record << "\n" << std::flush;
}
  
void osl::checkmate::CheckMateRecorder::leave(int depth, const char *name, 
			      const CheckHashRecord *record)
{
  const std::string spacer(depth,' ');
  stream() << spacer << "<== " << name << " "
	   << record->proofDisProof;
  if (leaveReason)
  {
    stream() << " " << leaveReason;
    leaveReason = 0;
  }
  stream() << " " << record << "\n" << std::flush;
}

void osl::checkmate::CheckMateRecorder::allocate(int depth, const CheckHashRecord *record)
{
  const std::string spacer(depth,' ');
  stream() << spacer << "alloc " << record << "\n";
}
    
/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
