/* checkTableAnalyzer.cc
 */
#include "osl/checkmate/checkTableAnalyzer.h"
#include "osl/checkmate/checkHashRecord.h"
#include "osl/checkmate/checkMoveList.h"
#include "osl/moveVector.h"
// isPawnDropFoul ΤʤΤʬΥۤɤ
#include "osl/checkmate/checkmateSearcher.h"
#include <set>
#include <map>
#include <string>
#include <iostream>
#include <stdexcept>

namespace osl
{
  namespace checkmate
  {
    struct CorruptCheckTable : std::runtime_error
    {
      const CheckHashRecord *record;
      CorruptCheckTable(const CheckHashRecord *r, 
			const char *msg="CorruptCheckTable") 
	: std::runtime_error(msg), record(r)
      {
      }
    };
  
    void fancyShow(std::ostream& os, const ProofDisProof& pdp)
    {
      switch (pdp)
      {
      case CheckMate:
	os << "CheckMate";
	break;
      case NoEscape:
	os << "NoEscape";
	break;
      case NoCheckMate:
	os << "NoCheckMate";
	break;
      case PawnCheckMate:
	os << "PawnCheckMate";
	break;
      default:
	os << "(" << getProof(pdp) << "," << getDisProof(pdp) << ")";
      }
    }

    typedef std::set<const CheckHashRecord*> RecordSet;
    typedef std::multimap<int, const CheckMove*> CheckMoveMap;
    namespace CTA
    {
      struct TreeWriter
      {
	std::ostream& os;
	int curDepth;
	explicit TreeWriter(std::ostream& o)
	  : os(o), curDepth(0)
	{
	}
	void incDepth() { ++curDepth; }
	void decDepth() { --curDepth; }
	const std::string spacer() { return std::string(curDepth*2, ' '); }
	void showRecord(const CheckHashRecord *record)
	{
	  os << spacer();
	  fancyShow(os, record->proofDisProof);
	  if (record->bestMove)
	  {
	    os << " best move ";
	    csaShow(os, record->bestMove->move);
	  }
	  os << "\n";
	}
	void showMove(Move m)
	{
	  os << spacer();
	  csaShow(os, m);
	  os << "\n";
	}
	void showMoveAndRecord(Move m, const CheckHashRecord *record)
	{
	  os << spacer();
	  csaShow(os, m);
	  os << ' ';
	  fancyShow(os, record->proofDisProof);
	  if (record->bestMove && record->bestMove->move)
	  {
	    os << " best move ";
	    csaShow(os, record->bestMove->move);
	  }
	  if (record->isConfluence)
	    os << " (c)";
	  if (record->isLoopOrigin)
	    os << " (l)";
	  if (record->useMaxInsteadOfSum)
	    os << " (sum/max)";
	  os << ' ' << record << " up " << record->parent << "\n";
	}
	void showMoves(const MoveVector& moves)
	{
	  if (moves.size() == 0)
	    return;
	  os << spacer() << 'T';	// terminal
	  for (MoveVector::const_iterator p=moves.begin(); p!=moves.end(); ++p)
	  {
	    os << ' ';
	    csaShow(os, *p);
	  }
	  os << "\n";
	}
	void writeln(const char *msg)
	{
	  os << spacer() << msg << "\n";
	}
      };
      struct TreeTraverser
      {
	TreeWriter writer;
	RecordSet visited;
	int maxDepth;
	bool expandFinalState, showTerminalMoves;
	TreeTraverser(std::ostream& o, int d, bool e, bool t)
	  : writer(o), maxDepth(d), expandFinalState(e), showTerminalMoves(t)
	{
	}
	void traverse(const CheckHashRecord *record);
      };
      void TreeTraverser::traverse(const CheckHashRecord *record)
      {
	if (! visited.insert(record).second)
	{
	  writer.writeln("(confluence)");
	  if (! record->isConfluence)
	    writer.writeln(" not recognized !!");
	  return;
	}
	if ((! expandFinalState) && isFinal(record->proofDisProof))
	  return;
	writer.incDepth();
	if (writer.curDepth < maxDepth)
	{
	  // üΡɤޤȤɽ
	  if (showTerminalMoves)
	  {
	    MoveVector moves;
	    for (CheckMoveList::const_iterator p=record->moves.begin(); 
		 p!=record->moves.end(); ++p)
	    {
	      if (! p->record)
		moves.push_back(p->move);
	    }
	    writer.showMoves(moves);
	  }
	  // ŸΡɤ proof number ɽ (Ϲθ)
	  CheckMoveMap m;
	  for (CheckMoveList::const_iterator p=record->moves.begin(); 
	       p!=record->moves.end(); ++p)
	  {
	    if (p->record)
	      m.insert(std::make_pair(getProof(p->record->proofDisProof), &*p));
	  }
	  for (CheckMoveMap::const_iterator p=m.begin(); p!=m.end(); ++p)
	  {
	    writer.showMoveAndRecord(p->second->move, p->second->record);
	    traverse(p->second->record);
	  }
	}
	writer.decDepth();
      }
    }
    using namespace CTA;
  
    CheckTableAnalyzer::
    CheckTableAnalyzer()
    {
    }
  
    CheckTableAnalyzer::~CheckTableAnalyzer()
    {
    }

    void CheckTableAnalyzer::
    showTree(const CheckHashRecord *record, std::ostream& os,
	     int maxDepth, bool expandFinalState, bool showTerminalMoves) const
    {
      if (! record)
	return;
      TreeTraverser traverser(os, maxDepth, expandFinalState, showTerminalMoves);
      traverser.traverse(record);
    }

    static void traverse(const CheckHashRecord *record,
			 RecordSet& out)
    {
      if (! record)
	return;
      if (! out.insert(record).second)
	return;
    
      for (CheckMoveList::const_iterator p=record->moves.begin();
	   p!=record->moves.end(); ++p)
      {
	traverse(p->record, out);
      }
    }

    size_t CheckTableAnalyzer::treeSize(const CheckHashRecord *record) const
    {
      RecordSet nodes;
      traverse(record, nodes);
      return nodes.size();
    }

    template <bool isProof>
    static void traverseAndNode(Move, const CheckHashRecord *record,
				RecordSet& out, TreeWriter *writer=0);
    template <bool isProof>
    static void traverseOrNode(Move m, const CheckHashRecord *record,
			       RecordSet& out, TreeWriter *writer=0)
    {
      if (isProof && (! isCheckmateSuccess(record->proofDisProof)))
	throw CorruptCheckTable(record, "! isCheckmateSuccess in proof");
      if ((! isProof) 
	  && (! isCheckmateFail(record->proofDisProof))
	  && (! isPawnDropFoul(record->proofDisProof, m)))
	throw CorruptCheckTable(record, "! isCheckmateFail in disproof");

      if (record->moves.empty())
	return;	// isProof ǡμ꤬ƨˤʤäƤʤäȤ
      assert((!isProof) || record->bestMove);

      if (! out.insert(record).second)
      {
	if (writer)
	{
	  assert(record->isConfluence);
	  writer->writeln("confluence");
	}
	return;
      }
      if (record->finalByDominance)
      {
	if (writer)
	  writer->writeln("finalByDominance");
	return;
      }
#if 0
      if (writer)
	writer->showRecord(record);
#endif
      for (CheckMoveList::const_iterator p=record->moves.begin();
	   p!=record->moves.end(); ++p)
      {
	if ((isProof && (p->move == record->bestMove->move))
	    || ((!isProof) && p->record 
		&& isCheckmateFail(p->record->proofDisProof)))
	{
	  if (writer)
	  {
	    writer->showMove(p->move);
	    writer->incDepth();
	  }
	  traverseAndNode<isProof>(p->move, p->record, out, writer);
	  if (writer)
	    writer->decDepth();
	  return;
	}
      }
      std::cerr << "corrupt proof tree \n";
      abort();
    }

    template <bool isProof>
    static void traverseAndNode(Move, const CheckHashRecord *record,
				RecordSet& out, TreeWriter *writer=0)
    {
      assert((isProof && isCheckmateSuccess(record->proofDisProof))
	     || ((!isProof) && isCheckmateFail(record->proofDisProof)));
      if (! out.insert(record).second)
      {
	if (writer)
	{
	  assert(record->isConfluence);
	  writer->writeln("confluence");
	}
	return;
      }
      if (record->finalByDominance)
      {
	if (writer)
	  writer->writeln("finalByDominance");
	return;
      }
#if 0
      if (writer)
	writer->showRecord(record);
#endif
      for (CheckMoveList::const_iterator p=record->moves.begin();
	   p!=record->moves.end(); ++p)
      {
	if (! p->record)
	  continue;
	if (writer)
	{
	  writer->showMove(p->move);
	  writer->incDepth();
	}
	traverseOrNode<isProof>(p->move, p->record, out, writer);
	if (writer)
	  writer->decDepth();
      }
    }

    static size_t examineProofTree(const CheckHashRecord *record,
				   bool win, TreeWriter *writer)
    {
      if (! record)
	return 0;
      if (! isFinal(record->proofDisProof))
	return 0;

      RecordSet nodes;
      if (isCheckmateSuccess(record->proofDisProof))
      {
	if (win)			// ¦ͤ᤿
	  traverseOrNode<true>(MOVE_INVALID, record, nodes, writer);
	else			// ¦ͤޤ줿
	  traverseAndNode<true>(MOVE_INVALID, record, nodes, writer);
      }
      else
      {
	if (win)			// ¦ͤƨ줿
	  traverseOrNode<false>(MOVE_INVALID, record, nodes, writer);
	else			// ¦ƨ줿
	  traverseAndNode<false>(MOVE_INVALID, record, nodes, writer);
      }
      return nodes.size();
    }

    size_t CheckTableAnalyzer::proofTreeSize(const CheckHashRecord *record,
					     bool win) const
    {
      try
      {
	return examineProofTree(record, win, 0);
      }
      catch (CorruptCheckTable& t)
      {
	std::cerr << "corrupt tree is " << t.record << "\n";
	CheckTableAnalyzer a;
	std::cerr << "entire tree is \n";
	a.showTree(record, std::cerr, 100, true, true);
	throw;
      }
    }
    size_t CheckTableAnalyzer::showProofTree(const CheckHashRecord *record,
					     bool win, std::ostream& os) const
    {
      TreeWriter writer(os);
      return examineProofTree(record, win, &writer);
    }
  } // namespace checkmate
} // namespace osl

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
