#include "osl/hash.h"
#include <boost/random/mersenne_twister.hpp>
#include <iomanip>

namespace osl
{
  static boost::mt11213b mtRandom;
  const HashGenTable hashGenTable;

  std::ostream& operator<<(std::ostream& os,const osl::HashKey& h){
    os << std::setfill('0') << std::setbase(16) << std::setw(8);
    for (int i=0; i<3; ++i)
      os << ':' << h.value(i);
#ifdef HASH_KEY_128_BIT
    os << ':' << h.value(3);
#endif
    return os << ':' << std::setbase(10);
  }
} // namespace osl

osl::HashKey osl::HashKey::calcHash(const PawnMaskState& state)
{
  HashKey v;
  for(int num=0;num<40;num++){
    Piece p=state.getPieceOf(num);
    hashGenTable.addHashKey(v,getPosition(p),getPtypeO(p));
  }
  v.setPlayer(state.getTurn());
  return v;
}

osl::HashKey osl::HashKey::newHashWithMove(Move move) const
{
  HashKey ret(*this);
  Position from=getFrom(move);
  Position to=getTo(move);
  Ptype capturePtype=getCapturePtype(move);
  PtypeO ptypeO=getPtypeO(move);
  PtypeO oldPtypeO=getOldPtypeO(move) ;
  if (capturePtype!=PTYPE_EMPTY)
  {
    PtypeO capturePtypeO=newPtypeO(alt(getPlayer(move)),capturePtype);
    PtypeO capturedPtypeO=captured(capturePtypeO);

    hashGenTable.subHashKey(ret,to,capturePtypeO);
    hashGenTable.addHashKey(ret,OFFBOARD,capturedPtypeO);

    hashGenTable.subHashKey(ret,from,oldPtypeO);
    hashGenTable.addHashKey(ret,to,ptypeO);
  }
  else{
    hashGenTable.subHashKey(ret,from,oldPtypeO);
    hashGenTable.addHashKey(ret,to,ptypeO);
  }
  ret.xorPlayer();
  return ret;
}
  

void osl::HashKey::setRandom()
{
  v[0]=((unsigned int)mtRandom() & (~1)); // ̰ӥåȤϼӤ˳Ƥ
  v[1]=(unsigned int)mtRandom(); 
  v[2]=0;			// reserved for PieceStand
#ifdef HASH_KEY_128_BIT
  v[3]=(unsigned int)mtRandom();
#endif
}

osl::HashGenTable::HashGenTable()
{
  for(int j=0;j<PTYPEO_SIZE;j++)
  {
    const PtypeO pjo = (PtypeO)(j+PTYPEO_MIN);
    for(int i=0;i<POSITION_SIZE;i++)
    {
      if ((Position)i == OFFBOARD)
      {
	const Ptype pj = getPtype(pjo);
	if (isBasic(pj) && (getOwner(pjo) == BLACK))
	{
	  PieceStand stand;
	  stand.add(pj);
	  hashKey[i][j] = HashKey(0,0,stand.getFlags(),0);
	}
      }
      else
      {
	hashKey[i][j].setRandom();
      }
    }
  }
}


// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
