/* kanjiPrint.cc
 */
#include "osl/kanjiPrint.h"
#include "osl/pawnMaskState.h"
#include "osl/ptype.h"
#include <iostream>
#include <map>

osl::KanjiPrint::
KanjiPrint(std::ostream& o) : os(o)
{
}

osl::KanjiPrint::
~KanjiPrint()
{
}

namespace osl
{
  namespace
  {
    const size_t NumKanjis = PTYPEO_MAX - PTYPEO_MIN+2;
    const char kanjiPieces[NumKanjis][3] = 
    {
      // 
      "", "-E", 
      "", ""/*""*/, ""/*""*/, "" /*""*/, "", "", "", "", "", "", "", "", "", "", 
      // 
      "", "+E",
      "", "", "", "", "", "ζ", "", "", "", "", "", "", "", "", 
    };
  } // anonymous namespace
} // namespace osl

const char *osl::KanjiPrint::kanji(PtypeO ptypeo)
{
  const size_t index = ptypeo -PTYPEO_MIN;
  assert(index < NumKanjis);
  return kanjiPieces[index];
}

void osl::KanjiPrint::
print(const PawnMaskState& state)
{
  for(int y=1;y<=9;y++){
    os << 'P' << y;  
    for(int x=9;x>0;x--){
      const PtypeO ptypeo = getPtypeO(state.getPieceAt(newPosition(x,y)));
      os << kanji(ptypeo);
    }
    os << std::endl;
  }
  // ɽ
  typedef std::map<PtypeO,int> map_t;
  map_t blackStand, whiteStand;
  for(int num=0;num<PIECE_SIZE;num++){
    if(state.getOffBoardMask(BLACK).isSet(num)){
      blackStand[newPtypeO(BLACK, numTable.getPtype(num))]++;
    }
    else if(state.getOffBoardMask(WHITE).isSet(num)){
      whiteStand[newPtypeO(BLACK, numTable.getPtype(num))]++;
    }
  }
  os << " ";
  for (map_t::const_iterator p=blackStand.begin(); p!=blackStand.end(); ++p)
    os << kanji(p->first) << p->second << " ";
  os << std::endl;  
  os << " ";
  for (map_t::const_iterator p=whiteStand.begin(); p!=whiteStand.end(); ++p)
    os << kanji(p->first) << p->second << " ";
  os << std::endl;  
  os << state.getTurn() << std::endl;
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
