/* pawnMaskState.cc
 */

#include "osl/pawnMaskState.h"
#include "osl/applyMove.h"
#include "osl/doUndoMoveLock.h"
#include "osl/doUndoMoveLock.tcc"
#include <iostream>

namespace osl 
{
  // explicit instantiation
  template DoUndoMoveLock::DoUndoMoveLock(PawnMaskState&, Move);

} // namespace osl

osl::PawnMaskState::PawnMaskState(){
  init();
}

osl::PawnMaskState::PawnMaskState(Handicap h){
  init(h);
}

void osl::PawnMaskState::init()
{
  SimpleState::init();
  clearAll(pawnMask[0]);
  clearAll(pawnMask[1]);
}
  

void osl::PawnMaskState::init(Handicap h){
  SimpleState::init(h);
  clearAll(pawnMask[0]);
  clearAll(pawnMask[1]);
  for(int num=PtypeTraits<PAWN>::indexMin;
      num< PtypeTraits<PAWN>::indexLimit; num++){
    Piece p=getPieceOf(num);
    Player player=getOwner(p);
    Position pos=getPosition(p);
    if(!isOffBoard(pos)){
      assert(!isPawnMaskSet(player,getX(pos)));
      setPos(pawnMask[playerToIndex(player)],pos);
    }
  }
}
  

osl::PawnMaskState::~PawnMaskState(){}

void osl::PawnMaskState::setPiece(Player player,Position pos,Ptype ptype){
  int num;
  for (num=0;num<40;num++){
    if (!usedMask.isSet(num) && numTable.getPtype(num)==unpromote(ptype)
	&& (ptype!=KING || 
	    num==PtypeTraits<KING>::indexMin+playerToIndex(player))){
      usedMask.setNum(num);
	
      Piece p=newPiece(player,ptype,num,pos);
      setPieceOf(num,p);
      setBoard(pos,p);
      if (pos==OFFBOARD)
	offBoardMask[playerToIndex(player)].setNum(num);
      else{
	if (ptype==PAWN)
	  setPos(pawnMask[playerToIndex(player)],pos);

      }
      return;
    }
  }
  std::cerr << "osl::PawnMaskState::setPiece! maybe too many pieces " 
	    << ptype << " " << pos << " " << player << "\n";
  abort();
}

void osl::PawnMaskState::setPieceAll(Player player){
  for (int num=0;num<40;num++){
    if (!usedMask.isSet(num)){
      usedMask.setNum(num);
      offBoardMask[playerToIndex(player)].setNum(num);
      Piece p=newPiece(player,numTable.getPtype(num),num,OFFBOARD);
      setPieceOf(num,p);
    }
  }
}
// check

bool osl::PawnMaskState::isConsistent(bool showError) const
{
  if(!SimpleState::isConsistent(showError)) 
    return false;
  // pawnMask;
  {
    BitXmask pawnMask1[2];
    clearAll(pawnMask1[0]);
    clearAll(pawnMask1[1]);
    for (int num=PtypeTraits<PAWN>::indexMin;
	 num<PtypeTraits<PAWN>::indexLimit;num++){
      if (isOnBoardNum(num)){
	Piece p=getPieceOf(num);
	if (!isPromoted(p)){
	  setPos(pawnMask1[playerToIndex(getOwner(p))],
		 getPosition(p));
	}
      }
    }
    if ((pawnMask[0]!=pawnMask1[0])
	|| (pawnMask[1]!=pawnMask1[1]))
    {
      if (showError) 
	std::cerr << "pawnMask "
		  << pawnMask[0] << "!=" << pawnMask1[0] 
		  << " || " <<  pawnMask[1] << "!=" << pawnMask1[1]
		  << std::endl;
      return false;
    }
  }
  return true;
}

bool osl::operator==(const PawnMaskState& st1,const PawnMaskState& st2){
  assert(st1.isConsistent(false));
  assert(st2.isConsistent(false));
  if (st1.pawnMask[0]!=st2.pawnMask[0]) return false;
  if (st1.pawnMask[1]!=st2.pawnMask[1]) return false;
  for (int y=1;y<=9;y++)
    for (int x=9;x>0;x--){
      Piece p1=st1.getPieceOnBoard(newPosition(x,y));
      Piece p2=st2.getPieceOnBoard(newPosition(x,y));
      if (getPtypeO(p1)!=getPtypeO(p2)) return false;
    }
  return true;
      
}

std::ostream& osl::operator<<(std::ostream& os,const PawnMaskState& state){
  os << static_cast<SimpleState>(state);
  return os;
}

void osl::PawnMaskState::doMove(Move move)
{
  ApplyMoveOfTurn::doMove(*this,move);
}
  
void osl::PawnMaskState::doSimpleMove(Position from, Position to, int promoteMask)
{
  Piece oldPiece;
  int num;
  if(getTurn()==BLACK){
    ApplyDoUndoSimpleMove<BLACK,PawnMaskState>::
      prologue(*this, from, to, promoteMask, oldPiece, num);
  }
  else{
    ApplyDoUndoSimpleMove<WHITE,PawnMaskState>::
      prologue(*this, from, to, promoteMask, oldPiece, num);
  }

  if (promoteMask!=0 && num<PtypeTraits<PAWN>::indexLimit)
    clearPawn(getTurn(), from);
}

void osl::PawnMaskState::doDropMove(Position to,Ptype ptype)
{
  Piece oldPiece;
  int num, numIndex, numLow;
  if(getTurn()==BLACK){
    ApplyDoUndoDropMove<BLACK,PawnMaskState>::
      prologue(*this, ptype, to, oldPiece, num, numIndex, numLow);
  }
  else{
    ApplyDoUndoDropMove<WHITE,PawnMaskState>::
      prologue(*this, ptype, to, oldPiece, num, numIndex, numLow);
  }
  if (ptype==PAWN)
    setPawn(getTurn(),to);
}

void osl::PawnMaskState::doCaptureMove(Position from, Position to, Piece target,int promoteMask)
{
  Piece oldPiece;
  int num0, num1;
  int num1Index, num1Mask;
  Ptype capturePtype;
  if(getTurn()==BLACK){
    ApplyDoUndoCaptureMove<BLACK,PawnMaskState>::
      prologue(*this, from, to, target, promoteMask,
	       capturePtype, oldPiece, num0, num1, num1Index, num1Mask);
  }
  else{
    ApplyDoUndoCaptureMove<WHITE,PawnMaskState>::
      prologue(*this, from, to, target, promoteMask,
	       capturePtype, oldPiece, num0, num1, num1Index, num1Mask);
  }

  if (capturePtype==PAWN)
    clearPawn(alt(getTurn()),to);
  if (promoteMask!=0 && num0<PtypeTraits<PAWN>::indexLimit)
    clearPawn(getTurn(),from);
}

osl::DoUndoMoveLockSimple<osl::PawnMaskState>::
DoUndoMoveLockSimple(PawnMaskState& s, 
		     Position f, Position t, int promoteMask)
  : state(s), from(f), to(t)
{
  if(s.getTurn()==BLACK){
    ApplyDoUndoSimpleMove<BLACK,PawnMaskState>::
      prologue(state, from, to, promoteMask, oldPiece, num);
  }
  else{
    ApplyDoUndoSimpleMove<WHITE,PawnMaskState>::
      prologue(state, from, to, promoteMask, oldPiece, num);
  }
  clearPawn = (promoteMask!=0) && (num < PtypeTraits<PAWN>::indexLimit);
  if (clearPawn)
    s.clearPawn(s.getTurn(),from);
}

osl::DoUndoMoveLockSimple<osl::PawnMaskState>::
~DoUndoMoveLockSimple()
{
  if (clearPawn)
    state.setPawn(state.getTurn(),from);
  if(state.getTurn()==BLACK){
    ApplyDoUndoSimpleMove<BLACK,PawnMaskState>::
      epilogue(state, from, to, oldPiece, num);
  }
  else{
    ApplyDoUndoSimpleMove<WHITE,PawnMaskState>::
      epilogue(state, from, to, oldPiece, num);
  }
}

osl::DoUndoMoveLockDrop<osl::PawnMaskState>::
DoUndoMoveLockDrop(PawnMaskState& s, Position t, Ptype ptype)
  : state(s), to(t)
{
  if(state.getTurn()==BLACK){
    ApplyDoUndoDropMove<BLACK,PawnMaskState>::
      prologue(s, ptype, to, oldPiece, num, numIndex, numLow);
  }
  else{
    ApplyDoUndoDropMove<WHITE,PawnMaskState>::
      prologue(s, ptype, to, oldPiece, num, numIndex, numLow);
  }
  pawnDrop = (ptype==PAWN);
  if (pawnDrop)
    s.setPawn(state.getTurn(),to);
}

osl::DoUndoMoveLockDrop<osl::PawnMaskState>::
~DoUndoMoveLockDrop()
{
  if (pawnDrop)
    state.clearPawn(state.getTurn(),to);
  if(state.getTurn()==BLACK){
    ApplyDoUndoDropMove<BLACK,PawnMaskState>::
      epilogue(state, to, oldPiece, num, numIndex, numLow);
  }
  else{
    ApplyDoUndoDropMove<WHITE,PawnMaskState>::
      epilogue(state, to, oldPiece, num, numIndex, numLow);
  }
}

osl::DoUndoMoveLockCapture<osl::PawnMaskState>::
DoUndoMoveLockCapture(PawnMaskState& s, 
		      Position f, Position t, Piece tg,
		      int promoteMask)
  : state(s), from(f), to(t), target(tg)
{
  if(state.getTurn()==BLACK){
    ApplyDoUndoCaptureMove<BLACK,PawnMaskState>::
      prologue(s, from, to, target, promoteMask,
	       capturePtype, oldPiece, num0, num1, num1Index, num1Mask);
  }
  else{
    ApplyDoUndoCaptureMove<WHITE,PawnMaskState>::
      prologue(s, from, to, target, promoteMask,
	       capturePtype, oldPiece, num0, num1, num1Index, num1Mask);
  }
  pawnCapture = (capturePtype==PAWN);
  if (pawnCapture)
    s.clearPawn(alt(s.getTurn()),to);
  pawnPromotion = (promoteMask!=0) && (num0<PtypeTraits<PAWN>::indexLimit);
  if (pawnPromotion)
    s.clearPawn(s.getTurn(),from);
}
osl::DoUndoMoveLockCapture<osl::PawnMaskState>::
~DoUndoMoveLockCapture()
{
  if (pawnPromotion)
    state.setPawn(state.getTurn(), from);
  if (pawnCapture)
    state.setPawn(alt(state.getTurn()), to);
  if(state.getTurn()==BLACK){
    ApplyDoUndoCaptureMove<BLACK,PawnMaskState>::
      epilogue(state, from, to, target, oldPiece, num0, num1, num1Index, num1Mask);
  }
  else{
    ApplyDoUndoCaptureMove<WHITE,PawnMaskState>::
      epilogue(state, from, to, target, oldPiece, num0, num1, num1Index, num1Mask);
  }
}

    

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
