#include "osl/pieceEval.h"
#include "osl/pieceEval.tcc"
#include "osl/evalTraits.h"
#include "osl/numEffectState.h"

namespace osl
{
  // explicit template instantiation
  template int PieceEval::computeDiffAfterMove<NumEffectState, BLACK>
  (const NumEffectState&, Move);
  template int PieceEval::computeDiffAfterMove<NumEffectState, WHITE>
  (const NumEffectState&, Move);
  template int PieceEval::computeDiffAfterMoveMulti<NumEffectState, BLACK>
  (const NumEffectState&, Move);
  template int PieceEval::computeDiffAfterMoveMulti<NumEffectState, WHITE>
  (const NumEffectState&, Move);
  template int PieceEval::computeDiffAfterTakeBackLimited<NumEffectState, BLACK>
  (const NumEffectState&, Position);
  template int PieceEval::computeDiffAfterTakeBackLimited<NumEffectState, WHITE>
  (const NumEffectState&, Position);
  template int PieceEval::computeDiffAfterTakeBackMulti<NumEffectState, BLACK>
  (const NumEffectState&);
  template int PieceEval::computeDiffAfterTakeBackMulti<NumEffectState, WHITE>
  (const NumEffectState&);
}

osl::PtypeEvalTable const osl::ptypeEvalTable;

osl::PieceEval::PieceEval(const SimpleState& state)
{
  int ret=0;
  for (int num=0;num<PIECE_SIZE;num++){
    if (state.getOffBoardMask(BLACK).isSet(num)){
      ret+=ptypeEvalTable.getVal(newPtypeO(BLACK,numTable.getPtype(num)));
    }
    else if (state.getOffBoardMask(WHITE).isSet(num)){
      ret+=ptypeEvalTable.getVal(newPtypeO(WHITE,numTable.getPtype(num)));
    }
    else{
      assert(state.isOnBoardNum(num));
      Piece p=state.getPieceOf(num);
      ret+=ptypeEvalTable.getVal(getPtypeO(p));
    }
  }
  val=ret;
}

osl::PtypeEvalTable::PtypeEvalTable(){ 
  init();
}
osl::PtypeEvalTable::~PtypeEvalTable(){ 
}

void osl::PtypeEvalTable::init(){
  delta[playerToIndex(BLACK)]=EvalTraits<BLACK>::delta;
  delta[playerToIndex(WHITE)]=EvalTraits<WHITE>::delta;
  ptypeO2Val[PTYPE_EMPTY-PTYPEO_MIN]=0;
  ptypeO2Val[PTYPE_EDGE-PTYPEO_MIN]=0;
  ptypeO2Val[PAWN-PTYPEO_MIN]=PtypeEvalTraits<PAWN>::val;
  ptypeO2Val[PPAWN-PTYPEO_MIN]=PtypeEvalTraits<PPAWN>::val;
  ptypeO2Val[LANCE-PTYPEO_MIN]=PtypeEvalTraits<LANCE>::val;
  ptypeO2Val[PLANCE-PTYPEO_MIN]=PtypeEvalTraits<PLANCE>::val;
  ptypeO2Val[KNIGHT-PTYPEO_MIN]=PtypeEvalTraits<KNIGHT>::val;
  ptypeO2Val[PKNIGHT-PTYPEO_MIN]=PtypeEvalTraits<PKNIGHT>::val;
  ptypeO2Val[SILVER-PTYPEO_MIN]=PtypeEvalTraits<SILVER>::val;
  ptypeO2Val[PSILVER-PTYPEO_MIN]=PtypeEvalTraits<PSILVER>::val;
  ptypeO2Val[GOLD-PTYPEO_MIN]=PtypeEvalTraits<GOLD>::val;
  ptypeO2Val[KING-PTYPEO_MIN]=PtypeEvalTraits<KING>::val;
  ptypeO2Val[BISHOP-PTYPEO_MIN]=PtypeEvalTraits<BISHOP>::val;
  ptypeO2Val[PBISHOP-PTYPEO_MIN]=PtypeEvalTraits<PBISHOP>::val;
  ptypeO2Val[ROOK-PTYPEO_MIN]=PtypeEvalTraits<ROOK>::val;
  ptypeO2Val[PROOK-PTYPEO_MIN]=PtypeEvalTraits<PROOK>::val;
  for(int i=PTYPE_MIN;i<=PTYPE_MAX;i++){
    Ptype ptype=static_cast<Ptype>(i);
    ptypeO2Val[newPtypeO(WHITE,ptype)-PTYPEO_MIN]=
      -ptypeO2Val[newPtypeO(BLACK,ptype)-PTYPEO_MIN];
  }
  for(int i=PTYPEO_MIN;i<=PTYPEO_MAX;i++){
    PtypeO ptypeO=static_cast<PtypeO>(i);
    PtypeO basicPtypeO=unpromote(ptypeO);
    promoteVal[i-PTYPEO_MIN]=ptypeEvalTable.getVal(ptypeO)-
      ptypeEvalTable.getVal(basicPtypeO);
  }
  // EMPTYcapture
  captureVal[0]=0;
  for(int i=PTYPEO_MIN;i<=PTYPEO_MAX;i++){
    PtypeO ptypeO=static_cast<PtypeO>(i);
    if(isPiece(ptypeO))
      captureVal[i-PTYPEO_MIN]=ptypeEvalTable.getVal(captured(ptypeO))-
	ptypeEvalTable.getVal(ptypeO);
    else
      captureVal[i-PTYPEO_MIN]=0;
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
