/* piecePairEval.cc
 */
#include "osl/piecePairEval.h"
#include <boost/static_assert.hpp>
#include <algorithm>
#include <cmath>
#include <fstream>
#include <iostream>

namespace osl
{
  const PiecePairEvalTable piecePairEvalTable;

  BOOST_STATIC_ASSERT(PiecePairEvalTable::maxPtypeOIndex == 32);
  // roundup  2^n  2 ʾǤ뤳Ȥγǧ
  BOOST_STATIC_ASSERT((PiecePairEval::ROUND_UP & (PiecePairEval::ROUND_UP-1))
		      == 0);
  BOOST_STATIC_ASSERT(PiecePairEval::ROUND_UP >= 2);
}

osl::PiecePairEval::PiecePairEval(const PawnMaskState& state) 
  : val(0)
{
  for (int i=0;i<PIECE_SIZE;i++){
    for (int j=i;j<PIECE_SIZE;j++){
      val += piecePairEvalTable.valueOf
	(PiecePairEvalTable::getIndexOfPieceNum(state, i), 
	 PiecePairEvalTable::getIndexOfPieceNum(state, j));
    }
  }
}

void osl::PiecePairEval::
doSimpleMove(const PawnMaskState& state,Position from, Position to, 
	     int promoteMask)
{
  piece_pair_assert(state.getPieceAt(from) != PIECE_EMPTY);
  piece_pair_assert(state.getPieceAt(to) == PIECE_EMPTY);
  val += diffAfterSimpleMove(state, from, to, promoteMask);
}

int osl::PiecePairEval::    
computeValWithMove(Player, const PawnMaskState& state, Move move) const
{
  return roundUp(val + diffWithMove(state, move));
}

void osl::PiecePairEval::
doDropMove(const PawnMaskState& state,Position to,Ptype ptype)
{
  piece_pair_assert(state.getPieceAt(to) == PIECE_EMPTY);
  const Player player = state.getTurn();
  val += diffAfterDropMove(state, to, newPtypeO(player, ptype));
}

void osl::PiecePairEval::
doCaptureMove(const PawnMaskState& state,
	      Position from, Position to, Piece victim,int promoteMask)
{
  piece_pair_assert(state.getPieceAt(from) != PIECE_EMPTY);
  piece_pair_assert(state.getPieceAt(to) != PIECE_EMPTY);
  val += diffAfterCaptureMove(state, from, to, getPtypeO(victim), promoteMask);
}

int osl::PiecePairEval::
adjustPairs(const SimpleState& state,
	    unsigned int indexOld, unsigned int indexNew)
{
  int diff = 0;
  for (int i=0;i<PIECE_SIZE;i++) 
  {
    const unsigned int target = 
      PiecePairEvalTable::getIndexOfPieceNum(state, i);
    diff -= piecePairEvalTable.valueOf(target, indexOld);
    // ʲ target==indexOld 1ʾ()ޥå
    diff += piecePairEvalTable.valueOf(target, indexNew);
  }
  diff -= piecePairEvalTable.valueOf(indexOld, indexNew); // ­ʬ
  diff += piecePairEvalTable.valueOf(indexNew, indexNew);
  return diff;
}

int osl::PiecePairEval::
adjustPairs(const SimpleState& state, 
	    unsigned int indexOld, unsigned int indexOld2, 
	    unsigned int indexNew, unsigned int indexNew2)
{
  int diff = 0;
  for (int i=0;i<PIECE_SIZE;i++) 
  {
    const unsigned int target = 
      PiecePairEvalTable::getIndexOfPieceNum(state, i);
    diff += piecePairEvalTable.valueOf(target, indexNew);
    diff += piecePairEvalTable.valueOf(target, indexNew2);
    diff -= piecePairEvalTable.valueOf(target, indexOld);
    diff -= piecePairEvalTable.valueOf(target, indexOld2);
  }

  diff -= piecePairEvalTable.valueOf(indexOld, indexNew);
  diff -= piecePairEvalTable.valueOf(indexOld, indexNew2);
  diff -= piecePairEvalTable.valueOf(indexOld2, indexNew);
  diff -= piecePairEvalTable.valueOf(indexOld2, indexNew2);
  diff += piecePairEvalTable.valueOf(indexOld, indexOld2);

  diff += piecePairEvalTable.valueOf(indexNew, indexNew);
  diff += piecePairEvalTable.valueOf(indexNew, indexNew2);
  diff += piecePairEvalTable.valueOf(indexNew2, indexNew2);
  return diff;
}
    

osl::PiecePairEvalTable::PiecePairEvalTable()
{
  values.clear();
}

osl::PiecePairEvalTable::~PiecePairEvalTable()
{
}

void osl::PiecePairEvalTable::writeInBinaryFile(const char *filename) const
{
  std::ofstream os(filename);
  os.write(reinterpret_cast<const char*>(values.begin()), sizeof(values));
}

bool osl::PiecePairEvalTable::
loadFromBinaryFile(const char *filename) const
{
  std::cerr << "PiecePairEvalTable loading  " << filename << "\n";
  std::ifstream is(filename);
  is.read(reinterpret_cast<char*>(values.begin()), sizeof(values));
  if (! is)
    return false;

  return true;
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
