#include "osl/ptype.h"
#include <iostream>

namespace osl{

  PtypeTable ptypeTable;

  std::ostream& operator<<(std::ostream& os,const Ptype ptype){
    return os << ptypeTable.getName(ptype);
  }
  std::ostream& csaShow(std::ostream& os,const Ptype ptype){
    return os << ptypeTable.getCsaName(ptype);
  }
  
  std::ostream& operator<<(std::ostream& os,const PtypeO ptypeO)
  {
    if (isPiece(ptypeO))
      return os << "PtypeO(" << getOwner(ptypeO) << "," 
		<< getPtype(ptypeO) << ")";
    return os << "PtypeO(" << (int)ptypeO << "," << getPtype(ptypeO) << ")";
  }
  
} // namespace osl

osl::PtypeTable::PtypeTable(){
  init();
}

template<osl::Ptype T> inline
void osl::PtypeTable::initPtypeSub(Int2Type<true> isBasic)
{
  initPtypeSub<T>(Int2Type<false>());
  numMaskLows[static_cast<int>(T)]=PtypeFuns<T>::indexMask;
  numIndices[static_cast<int>(T)]=PtypeFuns<T>::indexNum;
  if(canPromote(T)){
    numMaskLows[static_cast<int>(promote(T))]=PtypeFuns<T>::indexMask;
    numIndices[static_cast<int>(promote(T))]=PtypeFuns<T>::indexNum;
  }
  /**  */
  canDropLimit[0][static_cast<int>(T)]=PtypeTraits<T>::dropBlackFromY;
  /**  */
  canDropLimit[1][static_cast<int>(T)]=reverseY(PtypeTraits<T>::dropBlackFromY);
  indexMins[static_cast<int>(T)]=PtypeTraits<T>::indexMin;
  indexLimits[static_cast<int>(T)]=PtypeTraits<T>::indexLimit;
}

template<osl::Ptype T> inline
void osl::PtypeTable::initPtypeSub(Int2Type<false> isBasic)
{
  names[static_cast<int>(T)]=PtypeTraits<T>::name();
  csaNames[static_cast<int>(T)]=PtypeTraits<T>::csaName();
  moveMasks[static_cast<int>(T)]=PtypeTraits<T>::moveMask;
  betterToPromote[static_cast<int>(T)]=PtypeTraits<T>::betterToPromote;
}

template<osl::Ptype T> inline
void osl::PtypeTable::initPtype()
{
  initPtypeSub<T>(Int2Type<PtypeTraits<T>::isBasic>());
}

void osl::PtypeTable::init(){
#if 0
  std::fill(numMaskLows.begin(), numMaskLows.end(), 0);
  std::fill(numIndices.begin(), numIndices.end(), 0);
#endif
  numMaskLows.clear();
  numIndices.clear();

  initPtype<PTYPE_EMPTY>(); 
  initPtype<PTYPE_EDGE>(); 
  initPtype<PPAWN>(); 
  initPtype<PLANCE>(); 
  initPtype<PKNIGHT>(); 
  initPtype<PSILVER>(); 
  initPtype<PBISHOP>(); 
  initPtype<PROOK>(); 
  initPtype<GOLD>(); 
  initPtype<KING>(); 
  initPtype<PAWN>(); 
  initPtype<LANCE>(); 
  initPtype<KNIGHT>(); 
  initPtype<SILVER>(); 
  initPtype<BISHOP>(); 
  initPtype<ROOK>(); 
  effectTable.clear();
  effectTableNotLongU.clear();
  shortMoveMask.clear();
  //    std::cout << "moveMasks[ROOK]=" << moveMasks[ROOK] << std::endl;
  for(int ptype=PTYPE_MIN;ptype<=PTYPE_MAX;ptype++){
    for(int j=DIRECTION_MIN;j<=DIRECTION_MAX;j++){
      Direction dir=static_cast<Direction>(j);
      if((moveMasks[ptype]&(1<<dir))!=0){
	int dx=directionTable.getDxForBlack(dir);
	int dy=directionTable.getDyForBlack(dir);
	Offset32 offset32=newOffset32(dx,dy);
	Offset offset=newOffset(dx,dy);
	if(isLong(dir)){
	  shortMoveMask[0][dir-10]|=1<<(ptype-PTYPEO_MIN);
	  shortMoveMask[1][dir-10]|=1<<(ptype-16-PTYPEO_MIN);
	  effectTable[ptype-PTYPEO_MIN][offset32-OFFSET32_MIN]=directEffectVal;
	  effectTable[ptype-16-PTYPEO_MIN][-offset32-OFFSET32_MIN]=directEffectVal;
	  for(int i=2;i<9;i++){
	    offset32=newOffset32(dx*i,dy*i);
	    assert(offset32-OFFSET32_MIN>=0);
	    assert(-offset32-OFFSET32_MIN>=0);
	    effectTable[ptype-PTYPEO_MIN][offset32-OFFSET32_MIN]=offsetEffectVal(offset);
	    effectTable[ptype-16-PTYPEO_MIN][-offset32-OFFSET32_MIN]=offsetEffectVal(-offset);
	    if(static_cast<int>(dir)!=LONG_U){
	      effectTableNotLongU[ptype-PTYPEO_MIN][offset32-OFFSET32_MIN]=effectTable[ptype-PTYPEO_MIN][offset32-OFFSET32_MIN];
	      effectTableNotLongU[ptype-16-PTYPEO_MIN][-offset32-OFFSET32_MIN]=effectTable[ptype-16-PTYPEO_MIN][-offset32-OFFSET32_MIN];
	    }
	  }
	}
	else{
	  shortMoveMask[0][dir]|=1<<(ptype-PTYPEO_MIN);
	  shortMoveMask[1][dir]|=1<<(ptype-16-PTYPEO_MIN);
	  assert(offset32-OFFSET32_MIN>=0 || ((std::cout << dir << "," << offset32 << "," << OFFSET32_MIN << std::endl),0) );
	  assert(-offset-OFFSET32_MIN>=0 || ((std::cout << dir << "," << offset32 << "," << OFFSET32_MIN << std::endl),0) );
	  effectTable[ptype-PTYPEO_MIN][offset32-OFFSET32_MIN]=directEffectVal;
	  effectTable[ptype-16-PTYPEO_MIN][-offset32-OFFSET32_MIN]=directEffectVal;
	}
      }
    }
  }
  //    std::cout << "moveMasks[ROOK]=" << moveMasks[ROOK] << std::endl;
}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
