#include "osl/record/csa.h"
#include "osl/record/csaIOError.h"
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <cassert>

namespace osl
{
  namespace record
  {
  Player csaCharToPlayer(char c){
    if(c=='+') 
      return BLACK;
    if(c=='-') 
      return WHITE;
    throw CsaIOError("not a csa PlayerCharacter "+std::string(1,c));
  }

  Position csaStrToPos(const std::string& s){
    int x=s.at(0)-'0';
    int y=s.at(1)-'0';
    if(x==0 && y==0) return OFFBOARD;
    return newPosition(x,y);
  }

  Ptype csaStrToPtype(const std::string& s){
    for(int i=0;i<16;i++){
      if(s == ptypeTable.getCsaName(static_cast<Ptype>(i))) 
	return static_cast<Ptype>(i);
    }
    throw CsaIOError("unknown string in csaStrToPtype "+s);
  }

  Move csaStrToMove(const std::string& s,const PawnMaskState& state){
    Player pl=csaCharToPlayer(s.at(0));
    Position fromPos=csaStrToPos(s.substr(1,2));
    Position toPos=csaStrToPos(s.substr(3,2));
    Ptype ptype=csaStrToPtype(s.substr(5,2));
    if(fromPos==OFFBOARD){
      if (isPromoted(ptype))
	throw CsaIOError("drop with promote ?! in csaStrToMove "+s);
      return newDrop(toPos,ptype,pl);
    }
    else{
      Piece p0=state.getPieceAt(fromPos);
      Piece p1=state.getPieceAt(toPos);
      Ptype capturePtype=getPtype(p1);
      bool isPromote=(getPtype(p0)!=ptype);
      if (! ((getPtype(p0)==ptype)||(getPtype(p0)==unpromote(ptype))))
	throw CsaIOError("bad promotion in csaStrToMove "+s);
      return newMove(fromPos,toPos,ptype,
		     capturePtype,isPromote,pl);
    }
  }

  /**
   *  csa io error Τ褦ʤΤꤲ٤
   */
  static void csaParseLine(RecordVisitor& rv,const std::string& s)
  {
    Record *rec=rv.getRecord();
    PawnMaskState* state=rv.getState();
    //  std::cout << s << std::endl;
    if (s.length()==0) 
      return;
    switch(s.at(0)){
    case '\'': /* ȹ */
      return;
    case 'V': /* Сֹ */
      rec->setVersion(s.substr(1));
      return;
    case 'N': /* жɼ̾ */
      switch(s.at(1)){
      case '+':
      case '-':
	rec->setPlayer(csaCharToPlayer(s.at(1)),s.substr(2));
	break;
      default:
	std::cerr << "Illegal csa line " << s << std::endl;
	throw CsaIOError("illegal csa line "+s);
      }
      break;
    case 'P': /*  */
      switch(s.at(1)){
      case 'I': /* ʿ */
	state->init(HIRATE);
	break;
      case '+': /* ζ */
      case '-':{ /* ζ */
	Player pl=csaCharToPlayer(s.at(1));
	for(unsigned int i=2;i<=s.length()-4;i+=4){
	  Position pos=csaStrToPos(s.substr(i,2));
	  if(s.substr(i+2,2) == "AL"){
	    state->setPieceAll(pl);
	  }
	  else{
	    Ptype ptype=csaStrToPtype(s.substr(i+2,2));
	    state->setPiece(pl,pos,ptype);
	  }
	}
	break;
      }
      default:
	if(isdigit(s.at(1))){
	  int y=s.at(1)-'0';
	  for(unsigned int x=9,i=2;i<s.length();i+=3,x--){
	    if(s.at(i)==' ') continue;
	    Player pl=csaCharToPlayer(s.at(i));
	    Position pos=newPosition(x,y);
	    Ptype ptype=csaStrToPtype(s.substr(i+1,2));
	    state->setPiece(pl,pos,ptype);
	  }
	}
      }
      break;
    case '+':
    case '-':{
      Player pl=csaCharToPlayer(s.at(0));
      if(s.length()==1){
	state->setTurn(pl);
	rec->setInitialState(*state);
      }
      else{ // actual moves
	const Move m = csaStrToMove(s,*state);
	if (! state->isValidMove(m))
	{
	  std::cerr << "Illegal move " << m << std::endl;
	  throw CsaIOError("illegal move "+s);
	}
	rv.addMoveAndAdvance(m);
	return;
      }
      break;
    }
    case 'T':
      (rv.getLastMove())->setTime(atoi(s.c_str()+1));
      return;
    case '%':
      return;
    default:
      throw CsaIOError("unknown character in csaParseLine "+s);
    }
  }

  ICsaRecordStream::ICsaRecordStream(std::istream& is) : is(is)
  {
    if (! is)
    {
      std::cerr << "ICsaRecordStream::ICsaRecordStream cannot read \n";
      abort();
    }
  }
  
  ICsaRecordStream::~ICsaRecordStream(){}
  
  void ICsaRecordStream::load(Record* rec){
    //  rec->init();
    state.init();
    rv.setState(&state);
    rv.setRecord(rec);
    std::string s;
    while (std::getline(is,s)) 
    {
      //    std::cout << "Read: " << s << std::endl;
      csaParseLine(rv,s);
    }
  }

  CsaFile::CsaFile(std::string fileName){
    std::ifstream ifs(fileName.c_str());
    if (! ifs)
    {
      const std::string msg = "CsaFile::CsaFile file cannot read ";
      std::cerr << msg << fileName << "\n";
      throw CsaIOError(msg + fileName);
    }
    ICsaRecordStream irs(ifs);
    irs.load(&rec);
  }
  Record const& CsaFile::getRecord() const{
    return rec;
  }
  PawnMaskState const& CsaFile::getInitialState() const{
    return rec.getInitialState();
  }

} // namespace record
} // namespace osl
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
