#include "osl/record/kisen.h"
#include "osl/record/csaIOError.h"
#include <iostream>

namespace osl
{
namespace record
{
  Position KisenUtils::convertPosition( int pos ){
    assert(1<=pos && pos<=0x51);
    int y=((pos-1)/9)+1, x=((pos-1)%9)+1;
    return newPosition(x,y);
  }

  Move KisenUtils::convertMove(PawnMaskState const& state,int c0,int c1){
    Move move;

    if(1<=c1 && c1<=0x51){
      Position from=convertPosition(c1),to;
      Piece fromPiece=state.getPieceAt(from);
      assert(getOwner(fromPiece)==state.getTurn() ||
	     (std::cerr << c1 << "," << from << "," << fromPiece << std::endl,0)
	     );
      bool isPromote=false;
      if(1<=c0 && c0<=0x51){
	to=convertPosition(c0);
      }
      else if(0x65<=c0 && c0<=0xb5){
	to=convertPosition(c0-0x64);
	isPromote=true;
      }
      else{
	abort();
      }
      Piece toPiece=state.getPieceAt(to);
      assert(isEmpty(toPiece) || getOwner(toPiece)==alt(state.getTurn()));
      Ptype ptype=getPtype(fromPiece);
      if(isPromote)ptype=promote(ptype);
      move=newMove(from,to,ptype,getPtype(toPiece),isPromote,state.getTurn());
    }
    else{
      assert(0x65<=c1);
      assert(1<=c0&&c0<=0x51);
      Position to=convertPosition(c0);
      Ptype ptype=PTYPE_EMPTY;
      const Ptype ptypes[]={ROOK,BISHOP,GOLD,SILVER,KNIGHT,LANCE,PAWN};
      for(size_t i=0;i<sizeof(ptypes)/sizeof(Ptype);i++){
	int count=state.countOffBoard(state.getTurn(),ptypes[i]);
	if(count>0){
	  if(c1>0x64){
	    c1-=count;
	    if(c1<=0x64) ptype=ptypes[i];
	  }
	}
      }
      assert(ptype!=PTYPE_EMPTY);
      move=newDrop(to,ptype,state.getTurn());
    }
    assert(state.isValidMove(move,true) ||
	   (std::cerr << c0 << "," << c1 << std::endl,0) );
    return move;
  }


  KisenFile::KisenFile(std::string fileName) 
    :ifs(fileName.c_str()),initialState(HIRATE) 
  {
    if (! ifs)
      throw CsaIOError("KisenFile not found");
    ifs.seekg(0,std::ios::end);
    assert((ifs.tellg() % 512)==0);
    numberOfGames=ifs.tellg()/512;
  }

  std::vector<Move> KisenFile::getMoves(size_t index)
  {
    assert(0<=index && index<size());
    std::vector<Move> moves;
    if (index == 172518) 	// ɥǡ
	return moves;
    ifs.seekg(index*512,std::ios::beg);
    unsigned char cbuf[512];
    ifs.read(reinterpret_cast<char *>(&cbuf[0]),512);
    PawnMaskState state(HIRATE);
    //
    Player turn=BLACK;
    for(size_t turnCount=0; 
	cbuf[turnCount*2]!=0 && cbuf[turnCount*2+1]!=0;
	turnCount++, turn=alt(turn)){
      if(turnCount==KisenFile::maxMoves || cbuf[  turnCount *2 ] == 0 || cbuf[ turnCount * 2 + 1 ] == 0 ){ break; }
      int c0=cbuf[turnCount*2], c1=cbuf[turnCount*2+1];
      Move move=KisenUtils::convertMove(state,c0,c1);
      moves.push_back(move);
      state.doMove(move);
      assert(state.isConsistent( true ) );
    }
    return moves;
  }
} // namespace record
} // namespace osl

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
