#include "osl/record/record.h"
#include "osl/record/csaIOError.h"
#include "osl/doUndoMoveLock.h"
#include <stack>
#include <iostream>

namespace osl
{
namespace record
{

  IRecordStream::~IRecordStream(){}
  
  ORecordStream::~ORecordStream(){}
  
  Move MoveRecord::getMove() const { return move; }
  
  int MoveRecord::getNodeIndex() const { return nodeIndex; }
  
  void MoveRecord::setTime(int t){
    time=t;
  }
  
  void NodeRecord::addMoveRecord(int moveIndex){
    moves.push_back(moveIndex);
  }
  
  
  Record::Record(){ init(); }
  
  void Record::init(){
    version="";
    playerNames[0]=playerNames[1]="";
    nrs.clear();
    nrs.push_back(NodeRecord());
    initialState.init();
  }

  void Record::load(IRecordStream& irs){
    irs.load(this);
  }
  void Record::save(ORecordStream& irs){
    // irs.save(this);
  }
  
  void Record::setVersion(const std::string& str){
    version=str;
  }
  std::string Record::getVersion() const{
    return version;
  }
  void Record::setPlayer(Player player,const std::string& str){
    playerNames[playerToIndex(player)]=str;
  }
  std::string Record::getPlayer(Player player) const{
    return playerNames[playerToIndex(player)];
  }
  void Record::setInitialState(const PawnMaskState& state){
    initialState=state;
  }
  PawnMaskState const& Record::getInitialState() const {
    if (! initialState.isConsistent(true))
    {
      const char *msg = "Record: bad initial state";
      std::cerr << msg << "\n";
      throw CsaIOError(msg);
    }
    return initialState;
  }
  int Record::addNodeRecord(){
    nrs.push_back(NodeRecord());
    return nrs.size()-1;
  }
  int Record::addMoveRecord(const MoveRecord& moveRecord){
    mrs.push_back(moveRecord);
    return mrs.size()-1;
  }
  NodeRecord* Record::nodeOf(int index){
    return &nrs[index];
  }
  const NodeRecord* Record::nodeOf(int index) const{
    return &nrs[index];
  }
  MoveRecord* Record::moveOf(int index){
    return &mrs[index];
  }
  const MoveRecord* Record::moveOf(int index) const {
    return &mrs[index];
  }
  NodeRecord& Record::operator[](int index){
    return nrs[index];
  }
  
  void RecordVisitor::addMoveAndAdvance(Move move){
    assert(state->isValidMove(move));

    int newNode=rec->addNodeRecord();
    int newMove=rec->addMoveRecord(MoveRecord(move,newNode));
    (*rec)[nodeIndex].addMoveRecord(newMove);
    nodeIndex=newNode;
    lastMoveIndex=newMove;

    state->doMove(move);
    assert(state->isConsistent() || ((std::cout << move <<"\n"<< *state),0));
  }
  
  
  std::ostream& operator<<(std::ostream& os,const MoveRecord & mr){
    return os << "MoveRecord(" << // mr.getMove() << ',' <<
      mr.getNodeIndex()  << ')';
  }
  
  std::ostream& operator<<(std::ostream& os,Record & r){
    os << "Record(";
    os << "version=" << r.getVersion()
       << ",BLACK=" << r.getPlayer(BLACK)
       << ",WHITE=" << r.getPlayer(WHITE);
    os << ",initial=" << std:: endl << r.getInitialState() << std::endl;
    PawnMaskState state=r.getInitialState();
    RecordVisitor visitor;
    std::stack<DoUndoMoveLock*> moveStack;
    visitor.setState(&state);
    visitor.setRecord(&r);
    NodeRecord* node=visitor.getNode();
    while(node->size()>0){
      int moveIndex=node->at(0);
      MoveRecord* mr=r.moveOf(moveIndex);
      Move move=mr->getMove();
      os << move << "," << mr->getTime() << "," << mr->getComment() << std::endl;
      node=r.nodeOf(mr->getNodeIndex());
      moveStack.push(new DoUndoMoveLock(state, move)); // doMove
      assert(state.isConsistent());
    }
    os << state;
    while(!moveStack.empty())
    {
      DoUndoMoveLock *lock = moveStack.top();
      moveStack.pop();
      delete lock; // state.undoMove(move);
      assert(state.isConsistent());
    }
    os << state;
    return os << ')';
  }
  
  
  std::vector<Move> Record::getMoves() const {
    std::vector<Move> moves;
    const NodeRecord* node=nodeOf(0);
    while(node->size()>0){
      int moveIndex=node->at(0);
      const MoveRecord* mr=moveOf(moveIndex);
      Move move=mr->getMove();
      moves.push_back(move);
      //    std::cout << move << "," << mr->getTime() << "," << mr->getComment() << std::endl;
      node=nodeOf(mr->getNodeIndex());
    }
    return moves;
  }
  
} // namespace record
} // namespace osl
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
