/* mtdfRecorder.cc
 */
#include "osl/search/searchRecorder.h"
#include "osl/moveLogProb.h"
#include "osl/evalTraits.h"
#include "osl/pawnMaskState.h"
#include "osl/search/realizationProbability.h"
#include <fstream>
#include <iostream>
#include <cassert>
#include <unistd.h>

const char *checkMateFileName = "currentCheck.csa";

/** ʲ define ʤ ;η̤ƵϿ */
#define SELECT_CHECKMATE_LOG

/** ʲȵ;äζ̤̥ե¸ */
// #define CHECKMATE_SEARCHER_DEBUG

osl::search::NullRecorder::~NullRecorder()
{
}

/* ------------------------------------------------------------------------- */

osl::search::CountRecorder::CountRecorder()
  : numTerminalNodes(0), numLeafNodes(0), numTotalNodes(0)
{
}

osl::search::CountRecorder::~CountRecorder()
{
}

void osl::search::
CountRecorder::finishSearch(const MoveLogProb& bestMove, int lastF) const
{
#if 0
  // iterative deepening 򤹤ȤäȤ뤵
  std::cerr << "search finished\t" << bestMove << "\t" << lastF << "\n";
  std::cerr << "  terminal: " << numTerminalNodes << "\n";
  std::cerr << "      leaf: " << numLeafNodes << "\n";
  std::cerr << "     total: " << numTotalNodes << "\n";
#endif
  std::cerr << "total number of nodes: " << numTotalNodes << "\n";
}

/* ------------------------------------------------------------------------- */

static bool showAllValues = false;
struct osl::search::SearchRecorder::Recorder
{
  std::ofstream os;
  /** ߤο(ɽù) */
  int currentDepth;
  /** õ򳫻ϤǤlimit */
  int initialLimit;
  int logMargin;
  
  Recorder(const char *filename) 
    : os(filename),
      currentDepth(0), logMargin(RealizationProbability::tableMove)
  {
  }
  std::ostream& stream()
  {
    assert(os);
    os << currentDepth << ':';
    for (int i=0; i<currentDepth; ++i)
      os << "  ";
    return os;
  }
  /** Ѥ̾ɵϿ뿼˼ޤäƤĴ٤ */
  bool notSoDeep(int limit) const
  {
    return 
#ifdef SELECT_CHECKMATE_LOG
      (limit <= initialLimit) // SearchTable::CheckMateSpecialDepth 뤳Ȥ
      && 
#endif
      (initialLimit - limit) <= logMargin;
  }
  void flush()
  {
#if 1
    os << std::flush;
#endif    
  }
};

osl::search::
SearchRecorder::SearchRecorder(const char *filename)
  : recorder(new Recorder(filename))
{
}

osl::search::
SearchRecorder::~SearchRecorder()
{
}

void osl::search::
SearchRecorder::setLogMargin(int margin)
{
  recorder->logMargin = margin;
}

void osl::search::
SearchRecorder::recordInvalidMoveInTable(const MoveLogProb& move, int limit) const
{
#if 0
  if (recorder->notSoDeep(limit))
#endif
    stream() << "?? invalid move in table " << move.getMove() 
	     << " limit " << limit << "\n";
}
void osl::search::
SearchRecorder::newCategory(const char *name, int limit) const
{
  if (recorder->notSoDeep(limit))
  {
    ++recorder->currentDepth;
    stream() << "(" << name << ")\n";
    --recorder->currentDepth;
    recorder->flush();
  }
}


void osl::search::
SearchRecorder::tryMove(const MoveLogProb& m, int lastF, int limit) const
{
  ++recorder->currentDepth;
  if (recorder->notSoDeep(limit-100)) // üǤtryMove ̵
  {
    std::ostream& os = stream();
    os << "==> ";
    csaShow(os, m.getMove());
    os << " " << m.getLogProb() << "\t" 
       << "lastF: " << lastF << " limit: " << limit << "\n";
    recorder->flush();
  }
}

void osl::search::
SearchRecorder::retryMove(const MoveLogProb& m, int lastF, int limit,
			int retryCount) const
{
  ++recorder->currentDepth;
  if (recorder->notSoDeep(limit)) // õɬϿʤȤ䤳
  {
    std::ostream& os = stream();
    os << "ex" << retryCount << "> ";
    csaShow(os, m.getMove());
    os << " " << m.getLogProb() << "\t" 
       << "lastF: " << lastF << " limit: " << limit << "\n";
    recorder->flush();
  }
}

void osl::search::
SearchRecorder::recordValue(const MoveLogProb& m, int val, bool betterMove, int limit) const
{
  if (recorder->notSoDeep(limit)
      && (showAllValues || betterMove))
  {
    std::ostream& os = stream();
    os << "<== " << val << "\t";
    csaShow(os, m.getMove());
    os << "\n";
    recorder->flush();    
  }
  CountRecorder::recordValue(m,val,betterMove,limit);
  --recorder->currentDepth;
}

void osl::search::
SearchRecorder::recordLeaf(int val, int limit) const
{
  CountRecorder::recordLeaf(val, limit);
#if 0
  if (recorder->notSoDeep(limit))
  {
    stream() << "--- " << val << "\n";
    recorder->flush();    
  }
#endif
}

static const char *lowerChar(osl::Player p) 
{
  return (p == osl::BLACK) ? "B (lb)>" : "W (lb)<";
}
static const char *higherChar(osl::Player p) 
{
  return (p == osl::BLACK) ? "B (ub)<" : "W (ub)>";
}

void osl::search::
SearchRecorder::tableHitLowerBound(Player p, int val, int lastF, int limit) const
{
  if (recorder->notSoDeep(limit))
  {
    stream() << "==| table answered " << lowerChar(p) << val
	     << " for " << p << " lastF " << lastF << "\n";
    recorder->flush();
  }
}

void osl::search::
SearchRecorder::tableHitUpperBound(Player p, int val, int lastF, int limit) const
{
  if (recorder->notSoDeep(limit))
  {
    stream() << "==| table answered " << higherChar(p) << val
	     << " for " << p << " lastF " << lastF << "\n";
    recorder->flush();
  }  
}

void osl::search::
SearchRecorder::tableStoreLowerBound(Player p, const MoveLogProb& bestMove, int val, int lastF, int limit) const
{
  assert(isValid(bestMove.getMove()));
  if (recorder->notSoDeep(limit-100)) // ü̵
  {
    std::ostream& os = stream();
    os << "|== table store " << lowerChar(p) << val
       << " lastF " << lastF << " ";
    csaShow(os, bestMove.getMove());
    os << " limit " << limit << "\n";
    recorder->flush();
  }
}

void osl::search::
SearchRecorder::tableStoreUpperBound(Player p, const MoveLogProb& bestMove, int val, int lastF, int limit) const
{
  assert(isValid(bestMove.getMove()));
  if (recorder->notSoDeep(limit-100)) // ü̵
  {
    std::ostream& os = stream();
    os << "|== table store " << higherChar(p) << val
       << " lastF " << lastF << " ";
    csaShow(os, bestMove.getMove());
    os << " limit " << limit << "\n";
    recorder->flush();
  }
}

void osl::search::
SearchRecorder::recordTopLevelLowFail(const MoveLogProb& /* best */, int lastF) const
{
  stream() << "low fail,  lastF=" << lastF << "\n";
  stream() << "       terminal: " << numTerminalNodes << "\n";
  stream() << "           leaf: " << numLeafNodes << "\n";
  stream() << "          total: " << numTotalNodes << "\n";
}
void osl::search::
SearchRecorder::recordTopLevelHighFail(const MoveLogProb& bestMove, int lastF) const
{
  stream() << "high fail, lastF=" << lastF << " " << bestMove << "\n";
  stream() << "       terminal: " << numTerminalNodes << "\n";
  stream() << "           leaf: " << numLeafNodes << "\n";
  stream() << "          total: " << numTotalNodes << "\n";
}

void osl::search::
SearchRecorder::startSearch(int limit) const
{
  stream() << "\nnew search\n";
  recorder->initialLimit = limit;
  CountRecorder::startSearch(limit);
}
void osl::search::
SearchRecorder::finishSearch(const MoveLogProb& bestMove, int lastF) const
{
  stream() << "search finished\t" << bestMove << "\t" << lastF << "\n";
  stream() << "  terminal: " << numTerminalNodes << "\n";
  stream() << "      leaf: " << numLeafNodes << "\n";
  stream() << "     total: " << numTotalNodes << "\n";
  CountRecorder::finishSearch(bestMove, lastF);
}

void osl::search::
SearchRecorder::gotoCheckMateSearch(const PawnMaskState& state, int nodeLimit) const
{
#ifdef CHECKMATE_SEARCHER_DEBUG
  std::ofstream os(checkMateFileName, std::ios::app);
  os << state;
  os << nodeLimit << "\n";
#endif
}

void osl::search::
SearchRecorder::backFromCheckMateSearch() const
{
#ifdef CHECKMATE_SEARCHER_DEBUG
  std::ofstream os(checkMateFileName, std::ios::app);
  os << "done\n";
#endif
}

void osl::search::
SearchRecorder::checkMateStatus(int nodeCount) const
{
  stream() << "checkMateStatus\t" << nodeCount << "\n";
}

std::ostream& osl::search::
SearchRecorder::stream() const
{
  return recorder->stream();
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
