/* simpleHashTable.cc
 */
#include "osl/search/simpleHashTable.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/moveLogProb.h"
#include "osl/hashKeyEqual.h"
#include "osl/evalTraits.h"

#include <algorithm>
#if __GNUG__ >= 3
#  include <ext/hash_map>
   using namespace __gnu_cxx;
#else
#  include <hash_map>
#endif

namespace osl
{
  namespace search
  {
#if __GNUG__ >= 3
    typedef hash_map<HashKey,SimpleHashRecord,hash<HashKey>,HashKeyEqual> hash_map_t;
#else
    typedef hash_map<HashKey,SimpleHashRecord> hash_map_t;
#endif

    class SimpleHashTable::Table : public hash_map_t
    {
      int minimumRecordLimit;
    public:
      /** icc hash_map capacity äƤʤäΤǼʬǻ */
      const unsigned int capacity;
      int numCacheHit, numRecordAfterFull;
      bool verboseDestructor;
    
      Table(unsigned int c) : hash_map_t(c),
			      minimumRecordLimit(1), capacity(c), 
			      numCacheHit(0), numRecordAfterFull(0),
			      verboseDestructor(true)
      {
      }
      ~Table()
      {
	if (verboseDestructor)
	{
	  std::cerr << "SimpleHashTable size: " << size() 
		    << ", cache hit " << numCacheHit
		    << ", table full " << numRecordAfterFull << "\n";
	}
      }
      void setVerbose(bool v) { verboseDestructor = v; }
      void setMinimumRecordLimit(int m)
      {
	minimumRecordLimit = std::max(m, 1);
      }
      int getMinimumRecordLimit() const { return minimumRecordLimit; }
    
      SimpleHashRecord *find(const HashKey& key)
      {
	iterator pos = hash_map_t::find(key);
	if (pos == end())
	  return 0;

	++numCacheHit;
	return &pos->second;
      }
    
      SimpleHashRecord *allocate(const HashKey& key, int limit)
      {
	if (limit < getMinimumRecordLimit())
	  return find(key);
	if (size() < capacity)
	{
	  SimpleHashRecord *record = &operator[](key);
	  if ((record->lowerLimit >= limit)
	      || (record->upperLimit >= limit))
	    ++numCacheHit;
	  return record;
	}
	// 䤵ʤ褦õ
	SimpleHashRecord *result = find(key);
	if (numRecordAfterFull++ == 0)
	  std::cerr << "table full!\n";
	return result;
      }
    };
  
    SimpleHashTable::SimpleHashTable(unsigned int capacity, int minimumRecordLimit, bool verboseDestructor) 
      : table(new Table(capacity))
    {
      table->setMinimumRecordLimit(minimumRecordLimit);
      table->setVerbose(verboseDestructor);
    }
  
    SimpleHashTable::~SimpleHashTable() {
    }
    void SimpleHashTable::clear()
    {
      table->clear();
    }
    void SimpleHashTable::setMinimumRecordLimit(int newLimit)
    {
      table->setMinimumRecordLimit(newLimit);
    }

    SimpleHashRecord * 
    SimpleHashTable::allocate(const HashKey& key, int limit)
    {
      return table->allocate(key, limit);
    }
    SimpleHashRecord * 
    SimpleHashTable::find(const HashKey& key)
    {
      return table->find(key);
    }
    const SimpleHashRecord * 
    SimpleHashTable::find(const HashKey& key) const
    {
      return table->find(key);
    }
    unsigned int SimpleHashTable::size() const
    {
      return table->size();
    }
    unsigned int SimpleHashTable::capacity() const
    {
      return table->capacity;
    }

  } // namespace search
} // namespace osl

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
