/* simpleState.cc
 */

#include "osl/simpleState.h"
#include "osl/effect_action/checkMultiple.h"
#include "osl/effect_action/getPiece.h"
#include <iostream>

namespace osl 
{
  template bool SimpleState:: 
  hasEffectByWithRemove<BLACK>(Position, Position) const;
  template bool SimpleState:: 
  hasEffectByWithRemove<WHITE>(Position, Position) const;
  template bool SimpleState:: 
  hasEffectNotBy<BLACK>(Piece, Position) const;
  template bool SimpleState:: 
  hasEffectNotBy<WHITE>(Piece, Position) const;

  /**
   * 餫ϿPositionȰפemptyȸʤ functor
   */
  class PositionIsEmpty{
    Position position;
  public:
    explicit PositionIsEmpty(Position position) : position(position){}
    bool operator()(Position pos) const{ return position==pos; }  
  };


  SimpleState::SimpleState(){
    init();
  }

  SimpleState::SimpleState(Handicap h){
    init(h);
  }

  Piece SimpleState::hasEffectPiece(Player pl,Position target) const
  {
    if(pl==BLACK) return SimpleState::template hasEffectPiece<BLACK>(target);
    else return SimpleState::template hasEffectPiece<WHITE>(target);
  }
  template<Player P>
  Piece SimpleState::hasEffectPiece(Position target) const
  {
    Piece ret=PIECE_EMPTY;
    typedef effect_action::GetPiece<SimpleState> action_t;
    action_t getPieceAction(ret);
    forEachEffectWithCheck<P,action_t,true,NotEmpty>(target,getPieceAction,NotEmpty());
    return ret;
  }

  template<Player P>
  bool SimpleState::hasMultipleEffectBy(Position target) const 
  {
    Piece attackerPiece=PIECE_EMPTY;
    if (!hasEffectBy<P>(target,attackerPiece)) 
      return false;
    return attackerPiece==PIECE_EMPTY;
  }
  bool SimpleState::hasMultipleEffectBy(Player player,Position target) const 
  {
    if(player==BLACK)
      return SimpleState::template hasMultipleEffectBy<BLACK>(target);
    else
      return SimpleState::template hasMultipleEffectBy<WHITE>(target);
  }

  template<Player P>
  bool SimpleState::hasEffectBy(Position target,Piece& attackerPiece) const 
  {
    attackerPiece=PIECE_EMPTY;
    typedef effect_action::CheckMultiple<SimpleState> action_t;
    action_t checkMultipleAction(attackerPiece);
    forEachEffectWithCheck<P,action_t,true,NotEmpty>(target,checkMultipleAction,NotEmpty());
    switch(checkMultipleAction.count)
    {
    case 0: return false;
    case 1: return true;
    case 2: attackerPiece=PIECE_EMPTY; return true;
    default: 
      ;
    }
    assert(0);
    return true;		// to prevent useless warning
  }

  template<Player P>
  bool SimpleState::
  hasEffectByWithRemove(Position target,Position removed) const
  {
    Piece ret=PIECE_EMPTY;
    typedef effect_action::GetPiece<SimpleState> action_t;
    action_t getPieceAction(ret);
    forEachEffectWithCheck<P,action_t,true,PositionIsEmpty>(target,getPieceAction,PositionIsEmpty(removed));
    return ret!=PIECE_EMPTY;
  }

  template<Player P>
  bool SimpleState::hasEffectNotBy(Piece piece,Position target) const 
  {
    Piece ret=PIECE_EMPTY;
    typedef effect_action::GetPiece<SimpleState> action_t;
    action_t getPieceAction(ret);
    forEachEffectNotBy<P,action_t>(target,piece,getPieceAction);
    return ret!=PIECE_EMPTY;
  }

  void SimpleState::init(){
    turn=BLACK;
    for (int ipos=0;ipos<POSITION_SIZE;ipos++){
      setBoard(static_cast<Position>(ipos),PIECE_EDGE);
    }
    for (int y=1;y<=9;y++)
      for (int x=9;x>0;x--){
	setBoard(newPosition(x,y),PIECE_EMPTY);
      }
    //  promoteMask.clearAll();
    offBoardMask[playerToIndex(BLACK)].clearAll();
    offBoardMask[playerToIndex(WHITE)].clearAll();
    usedMask.clearAll();
  }
  

  void SimpleState::init(Handicap h){
    init();
    // 
    for (int x=9;x>0;x--){
      setPiece(BLACK,newPosition(x,7),PAWN);
      setPiece(WHITE,newPosition(x,3),PAWN);
    }
    // 
    setPiece(BLACK,newPosition(1,9),LANCE);
    setPiece(BLACK,newPosition(9,9),LANCE);
    setPiece(WHITE,newPosition(1,1),LANCE);
    setPiece(WHITE,newPosition(9,1),LANCE);
    //
    setPiece(BLACK,newPosition(2,9),KNIGHT);
    setPiece(BLACK,newPosition(8,9),KNIGHT);
    setPiece(WHITE,newPosition(2,1),KNIGHT);
    setPiece(WHITE,newPosition(8,1),KNIGHT);
    //
    setPiece(BLACK,newPosition(3,9),SILVER);
    setPiece(BLACK,newPosition(7,9),SILVER);
    setPiece(WHITE,newPosition(3,1),SILVER);
    setPiece(WHITE,newPosition(7,1),SILVER);
    //
    setPiece(BLACK,newPosition(4,9),GOLD);
    setPiece(BLACK,newPosition(6,9),GOLD);
    setPiece(WHITE,newPosition(4,1),GOLD);
    setPiece(WHITE,newPosition(6,1),GOLD);
    //
    setPiece(BLACK,newPosition(5,9),KING);
    setPiece(WHITE,newPosition(5,1),KING);
    //
    setPiece(BLACK,newPosition(8,8),BISHOP);
    setPiece(WHITE,newPosition(2,2),BISHOP);
    //
    setPiece(BLACK,newPosition(2,8),ROOK);
    setPiece(WHITE,newPosition(8,2),ROOK);
  }
  

  SimpleState::~SimpleState(){}

  void SimpleState::setPiece(Player player,Position pos,Ptype ptype){
    int num;
    for (num=0;num<40;num++){
      if (!usedMask.isSet(num) && numTable.getPtype(num)==unpromote(ptype)
	  && (ptype!=KING || 
	      num==PtypeTraits<KING>::indexMin+playerToIndex(player))){
	usedMask.setNum(num);
	
	Piece p=newPiece(player,ptype,num,pos);
	setPieceOf(num,p);
	setBoard(pos,p);
	if (pos==OFFBOARD)
	  offBoardMask[playerToIndex(player)].setNum(num);
	else{
	}
	return;
      }
    }
    std::cerr << "SimpleState::setPiece! maybe too many pieces " 
	      << ptype << " " << pos << " " << player << "\n";
    abort();
  }

  void SimpleState::setPieceAll(Player player){
    for (int num=0;num<40;num++){
      if (!usedMask.isSet(num)){
	usedMask.setNum(num);
	offBoardMask[playerToIndex(player)].setNum(num);
	Piece p=newPiece(player,numTable.getPtype(num),num,OFFBOARD);
	setPieceOf(num,p);
      }
    }
  }
  
  // check

  bool SimpleState::isConsistent(bool showError) const
  {
    // boardǤconsistency
    for (int y=1;y<=9;y++)
    {
      for (int x=9;x>=1;x--)
      {
	const Position pos=newPosition(x,y);
	const Piece p0=getPieceAt(pos);
	if (isPiece(p0))
	{
	  if (getPosition(p0)!=pos)
	  {
	    if (showError) 
	      std::cerr << "Position " << getPosition(p0) << "!=" << pos << std::endl;
	    return false;
	  }
	  int num=getNum(p0);
	  if (!validNum(num) || !usedMask.isSet(num)){
	    if (showError) std::cerr << "NotUsed, num=" << num << std::endl;
	    return false;
	  }
	  Piece p1=getPieceOf(num);
	  if (p0!=p1){
	    if (showError) std::cerr << "board[" << pos << "]!=" 
				     << "piece[" << num << "]" << std::endl;
	    return false;
	  }
	}
      }
    }
    // piecesconsistency
    for (int num0=0;num0<PIECE_SIZE;num0++)
    {
      if (isOnBoardNum(num0))
      {
	Piece p0=getPieceOf(num0);
	Ptype ptype=getPtype(p0);
	if (unpromote(ptype)!=numTable.getPtype(num0)){
	  if (showError) std::cerr << "ptype of piece[" << num0 << "]=" 
				   << ptype << std::endl;
	  return false;
	}
	if (!isOnBoard(p0)){
	  if (showError) std::cerr << "mochigoma[" << num0 << "]=true" << std::endl;
	  return false;
	}
	Position pos=getPosition(p0);
	if (!isOnBoard(pos)) {
	  if (showError) std::cerr << "position " << pos << " is not onboard" << std::endl;
	  return false;
	}
	Piece p1=getPieceAt(pos);
	int num1=getNum(p1);
	if (num0 !=num1){
	  if (showError) std::cerr << "pieces[" << num0 << "]=" << p0 << ",board[" << pos << "] is " << p1 << std::endl;
	  return false;
	}
      }
      else
      {
	Piece p0=getPieceOf(num0);
	Ptype ptype=getPtype(p0);
	if (getNum(p0)!=num0){
	  if (showError) 
	    std::cerr << "pieces[" << num0 << "="
		      << p0 << std::endl;
	  return false;
	  
	}
	if (ptype!=numTable.getPtype(num0)){
	  if (showError) std::cerr << "ptype of piece[" << num0 << "]=" 
				   << ptype << std::endl;
	  return false;
	}
	if (getPosition(p0)!=OFFBOARD){
	  if (showError) std::cerr << p0 << " must be offboard" << std::endl;
	  return false;
	}
      }
    }
    return true;
  }

  bool SimpleState::isAlmostValidMove(Move move,bool showError) const
  {
    if (showError)
    {
      const bool valid = SimpleState::template isAlmostValidMove<true>(move);
      if (! valid)
	std::cerr << *this << " " << move << std::endl;
      return valid;
    }
    else
      return SimpleState::template isAlmostValidMove<false>(move);
  }
  template <bool showError>
  bool SimpleState::isAlmostValidMove(Move move) const
  {
    assert(move);		// guard for MOVE_INVALID
    assert(getTurn() == getPlayer(move));

    const Position from=getFrom(move);
    const Position to=getTo(move);
    const Piece toPiece=getPieceAt(to);
    const Ptype ptype=getPtype(move);
    const Player turn = getPlayer(move);
    
    if (isOffBoard(from)) // Ǥļ
    {      
      // åȤ
      if (!isEmpty(toPiece)){
	if (showError) std::cerr << "drop on to piece : " << move << std::endl;
	return false;
      }
      // ⤽⤽ζäƤ뤫?
      if (!hasOffBoard(turn,ptype)){
	if (showError) std::cerr << turn << " don't have : " << ptype << std::endl;
	return false;
      }
      // Υå
      if (ptype==PAWN && isPawnMaskSet(turn,getX(to))){
	if (showError) std::cerr << " Double Pawn : " << move << std::endl;
	return false;
      }
    }
    else
    {
      // fromˤΤζ
      const Piece fromPiece=getPieceAt(from);
      if (isEmpty(fromPiece) || getOwner(fromPiece)!=turn ||
	  getPtype(checkPromote(fromPiece,isPromote(move)))!=getPtype(move)){
	if (showError) std::cerr << " No such piece  : " << move << std::endl;
	return false;
      }
      // toˤΤζ𤫶?
      if (!isEmpty(toPiece) && getOwner(toPiece)==turn){
	if (showError) std::cerr << " No move on  : " << move << std::endl;
	return false;
      }
      //  offsetưptype˴ؤvalid?
      if (ptypeTable.getEffect(getPtypeO(fromPiece),from,to)!=directEffectVal){
	const Offset o=effectOffset(ptypeTable.getEffect(getPtypeO(fromPiece),from,to));
	if (o==0){
	  if (showError){
	    std::cerr << " No such move : " << move << std::endl;
	  }
	  return false;
	}
	// Υ줿ưλ˴֤Ƥ뤫?
	for (Position p=from+o;p!=to;p+=o){
	  if (!isEmpty(getPieceAt(p))){
	    if (showError) 
	      std::cerr << " Not space to move : " << move << std::endl;
	    return false;
	  }
	}
      }
      // capturePtypeפƤ뤫?
      if (getPtype(toPiece)!=getCapturePtype(move)){
	if (showError) std::cerr << " Not such capture : " << move << std::endl;
	return false;
      }
      // promoteƤpromoteǽ
      if (isPromote(move) && isPromoted(fromPiece))
      {
	if (showError) 
	  std::cerr << " can't promote promoted piece : " << move << std::endl;
	return false;
      }
    }

    assert(isValidMoveByRule(move, true)
	   || (std::cerr << *this << "\n", 0));
    return true;
  }

  bool SimpleState::isValidMoveByRule(Move move,bool showError) const
  {
    const Position from=getFrom(move);
    const Position to=getTo(move);
    const Ptype ptype=getPtype(move);
    const Player turn = getPlayer(move);
    
    if (from==OFFBOARD) // Ǥļ
    { 
      // ưʤǤϤʤ?
      if (!canDropTo(turn,ptype,to))
      {
	if (showError) std::cerr << " can't drop to : " << move << std::endl;
	return false;
      }
    }
    else
    {
      const Piece fromPiece=getPieceAt(from);
      //  offsetưptype˴ؤvalid?
      if (ptypeTable.getEffect(getPtypeO(fromPiece),positionDiff(to,from))
	  != directEffectVal)
      {
	const Offset o = 
	  effectOffset(ptypeTable.getEffect(getPtypeO(fromPiece),positionDiff(to,from)));
	if (o==0){
	  if (showError){
	    std::cerr << " No such move : " << move << std::endl;
	  }
	  return false;
	}
      }
      // promoteƤpromoteǽ
      if (isPromote(move))
      {
	if (! (canPromote(unpromote(getPtype(move)))
	       && (canPromotePosition(getPlayer(move), to) 
		   || canPromotePosition(getPlayer(move), from))))
	{
	  if (showError) 
	    std::cerr << " illegal promote type or position : " << move << std::endl;
	  return false;
	}
      }
      // promoteƤʤ˶promoteǤʤ?
      if ((ptype!=KING && ptype!=GOLD && !isPromoted(fromPiece) 
	   && !canDropTo(turn,getPtype(fromPiece),to)) && !isPromote(move)){
	if (showError) 
	  std::cerr << " must promote to this position : " << move << std::endl;
	return false;
      }
    }
    return true;
  }

  bool SimpleState::isValidMove(Move move,bool showError) const
  {
    if (getTurn() != getPlayer(move)){
      if (showError) std::cerr << "invalid player move : " << move << std::endl;
      return false;
    }
    if (! isValidMoveByRule(move, showError))
      return false;
    return isAlmostValidMove(move,showError);
  }
  
  bool SimpleState::dump()
  {
    return std::cerr << *this << "\n";
  }
  
  bool operator==(const SimpleState& st1,const SimpleState& st2){
    assert(st1.isConsistent(false));
    assert(st2.isConsistent(false));
    if (st1.turn!=st2.turn) 
      return false;
    for (int y=1;y<=9;y++)
      for (int x=9;x>0;x--){
	Piece p1=st1.getPieceAt(newPosition(x,y));
	Piece p2=st2.getPieceAt(newPosition(x,y));
	if (getPtypeO(p1)!=getPtypeO(p2)) return false;
      }
    return true;
      
  }

  std::ostream& operator<<(std::ostream& os,const SimpleState& state){
    for (int y=1;y<=9;y++){
      os << 'P' << y;  
      for (int x=9;x>0;x--){
	csaShow(os,state.getPieceAt(newPosition(x,y)));
      }
      os << std::endl;
    }
    // ɽ
    for (int num=0;num<PIECE_SIZE;num++){
      if (state.offBoardMask[0].isSet(num)){
	os << "P+00";
	csaShow(os,numTable.getPtype(num));
	os << std::endl;
      }
      else if (state.offBoardMask[1].isSet(num)){
	os << "P-00";
	csaShow(os,numTable.getPtype(num));
	os << std::endl;
      }
    }
    os << state.getTurn() << std::endl;
    return os;
  }


} // namespace osl


// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
