/* histogram.cc
 */
#include "osl/stat/histogram.h"
#include <iostream>
#include <iomanip>

osl::stat::
Histogram::Histogram(size_t w, size_t len, int s)
  : data(new size_t[len]), length(len), width(w), start(s)
{
  std::fill(&data[0], &data[0]+length, 0);
}
osl::stat::
Histogram::~Histogram()
{
}
void osl::stat::
Histogram::merge(const Histogram& o)
{
  if ((width == o.width)
      && (length == o.length)
      && (start == o.start))
  {
    for (size_t i=0; i<length; ++i)
    {
      data[i] += o.data[i];
    }
  }
}

void osl::stat::
Histogram::show(std::ostream& os) const
{
  int value=start;
  for (size_t i=0; i<length; ++i, value+=width)
  {
    os << std::setw(5) << value << " - " << std::setw(5);
    if (i+1 < length)
      os << value+(int)width;
    else
      os << " ";
    os << std::setw(8) << data[i] << "\n";
  }
}

void osl::stat::
Histogram::showRatio(std::ostream& os, const Histogram& o) const
{
  if ((width == o.width)
      && (length == o.length)
      && (start == o.start))
  {
    int value=start;
    for (size_t i=0; i<length; ++i, value+=width)
    {
      os << std::setw(5) << value << " - " << std::setw(5);
      if (i+1 < length)
	os << value+(int)width;
      else
	os << " ";
      os << std::setw(8) << o.data[i]
	 << std::setw(8) << data[i]
	 << std::setw(8) << std::setprecision(2)
	 << static_cast<double>(o.data[i])/data[i] << "\n";
    }
  }
}
 
/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
