#include "osl/checkmate/dualCheckMateSearcher.h"
#include "osl/checkmate/checkTableAnalyzer.h"
#include "osl/checkmate/checkHashTable.h"
#include "osl/checkmate/checkHashRecord.h"
#include "osl/record/csaString.h"
#include "osl/record/csa.h"
#include "osl/hashEffectState.h"
#include "osl/perfmon.h"

#include <sstream>
#include <string>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <unistd.h>

using namespace osl;
using record::ICsaRecordStream;
void usage(const char *prog)
{
  using namespace std;
  cerr << "Usage: " << prog << " [-vpE] [-A depth(>0)] [-a depth(>0)] csa-filenames "
       << endl
       << "-v show result\n"
       << "-p show proof tree\n"
       << "-A depth  show all tree (expanding solved branch)\n"
       << "-a depth  show tree\n"
       << "-E show filnames that are solved and found to be 'escape'\n"
       << endl;
  exit(1);
}

void search(const char *filename);
bool verbose=false;
unsigned long long totalCycles=0;
size_t defaultLimit = 0;
bool showProofTree = false;
int showAllTreeDepth = 0;
int showTreeDepth = 0;
bool showEscapeFilename = false;

int numCheckMate=0, numEscape=0, numUnkown=0;

int main(int argc, char **argv)
{
  const char *program_name = argv[0];
  bool error_flag = false;
  extern char *optarg;
  extern int optind;
    
  char c;
  while ((c = getopt(argc, argv, "A:a:El:pvh")) != EOF)
  {
    switch(c)
    {
    case 'A':	showAllTreeDepth = atoi(optarg);
      break;
    case 'a':	showTreeDepth = atoi(optarg);
      break;
    case 'E':	showEscapeFilename = true;
      break;
    case 'l':	defaultLimit = atoi(optarg);
      break;
    case 'p':	showProofTree = true;
      break;
    case 'v':	verbose = true;
      break;
    default:	error_flag = true;
    }
  }
  argc -= optind;
  argv += optind;

  if (error_flag || (argc < 1))
    usage(program_name);

  try
  {
    for (int i=0; i<argc; ++i)
    {
      if (verbose)
	std::cerr << "\nsolving " << argv[i] << "\n";
      search(argv[i]);
      totalCycles = 0;
    }
    std::cerr << "check " << numCheckMate << " escape " << numEscape
	      << " unknown " << numUnkown << "\n";
  }
  catch (std::exception& e)
  {
    std::cerr << e.what() << "\n";
    return 1;
  }
}

/**
 * MtdfRecorder  currentCheck.csa ϶̰ʳ CSA Ǥʤ
 * üɤ򤹤롥 interface  IOؿ  lib 
 * դΤɤΤȤ
 *
 * 
 * limit
 * "done"
 * 
 * η֤
 */
bool readStates(std::istream& is, PawnMaskState& state, int& limit)
{
  std::string line;
  std::ostringstream ss;
  while (std::getline(is, line) && (line[0] == 'P'))
    ss << line << "\n";
  ss << line << "\n";		// 
  state = CsaString(ss.str().c_str()).getInitialState(); // str().c_str() ϼ¸⤷ʤ?
  if (! is)
    return false;
  std::stringstream ssl;
  std::getline(is, line);
  ssl << line;
  ssl >> limit;
  const bool result = is;
  
  std::getline(is, line);	// "done" ϺǸϽ񤤤Ƥʤ
  return result;
}

/**
 * CSA Υե뤫 stream 1()ɤ
 */
bool readCsaState(std::istream& is, PawnMaskState& state, int& limit)
{
  assert(defaultLimit);

  record::Record rec;
  record::ICsaRecordStream irs(is);
  irs.load(&rec);
  state = rec.getInitialState();

  limit = defaultLimit;
  is.setstate(std::ios_base::badbit); // ʹߤζ̤ɤޤʤ
  return true;
}

bool readState(std::istream& is, PawnMaskState& state, int& limit)
{
  if (defaultLimit)
    return readCsaState(is, state, limit);
  else
    return readStates(is, state, limit);
}

template <Player P>
void testWinOrLose(const char *curFilename,
		   DualCheckMateSearcher<HashEffectState>& searcher,
		   const PawnMaskState& sstate, int limit)
{
  limit = searcher.limitToCheckCount(limit); // quick hack ΡɿѴ
  std::cerr << "limit " << limit << "\n";
  HashEffectState state((NumEffectState(sstate)));

  if(state.hasEffectBy(alt(P),state.template getKingPosition<P>()))
  {
    // ꤫鲦꤬äƤ
    clock_start();
    const bool lose = searcher.
      template isLosingState<P>(limit, state);
    totalCycles += clock_stop();
    if (verbose)
    {
      const CheckHashTable& table = searcher.getTable(alt(P));
      const CheckHashRecord *record = table.find(state.getHash());
      CheckTableAnalyzer analyzer;
      if (showAllTreeDepth)
	analyzer.showTree(record,std::cout,showAllTreeDepth,true);
      if (showTreeDepth)
	analyzer.showTree(record,std::cout,showTreeDepth,false);
      if (lose)
      {
	++numCheckMate;
	std::cerr << "lose\n";
	std::cerr << "proof tree " << analyzer.proofTreeSize(record,false)
		  << "\n";
	if (showProofTree)
	  analyzer.showProofTree(record,false,std::cerr);
	return;
      }
      else
      {
	assert(record);
	if (showEscapeFilename)
	  std::cerr << curFilename << "\n";
	if (isFinal(record->proofDisProof))
	{
	  ++numEscape;
	  std::cerr << "escape\n";
	  assert(isCheckmateFail(record->proofDisProof));
	  std::cerr << "disproof tree " << analyzer.proofTreeSize(record,true)
		    << "\n";
	  if (showProofTree)
	    analyzer.showProofTree(record,true,std::cerr);
	  return;
	}
      }
    }
  }
  else
  {
    Move checkMateMove;
    clock_start();
    const bool win = searcher.
      template isWinningState<P>(limit, state, checkMateMove);
    totalCycles += clock_stop();
    if (verbose)
    {
      const CheckHashTable& table = searcher.getTable(P);
      const CheckHashRecord *record = table.find(state.getHash());
      CheckTableAnalyzer analyzer;
      if (showAllTreeDepth)
	analyzer.showTree(record,std::cout,showAllTreeDepth,true);
      if (showTreeDepth)
	analyzer.showTree(record,std::cout,showTreeDepth,false);
      if (win)
      {
	++numCheckMate;
	std::cerr << "win by " << checkMateMove << "\n";
	std::cerr << "proof tree " << analyzer.proofTreeSize(record,true)
		  << "\n";
	if (showProofTree)
	  analyzer.showProofTree(record,true,std::cerr);
	return;
      }
      else
      {
	assert(record);
	if (isFinal(record->proofDisProof))
	{
	  ++numEscape;
	  std::cerr << "no checkmate\n";
	  assert(isCheckmateFail(record->proofDisProof));
	  std::cerr << "disproof tree " << analyzer.proofTreeSize(record,false)
		    << "\n";
	  if (showProofTree)
	    analyzer.showProofTree(record,false,std::cerr);
	  return;
	}
      }
    }
  }
  if (verbose)
  {
    ++numUnkown;
    std::cerr << "unknown\n";
  }
}

void search(const char *filename)
{
  std::ifstream is(filename);
  DualCheckMateSearcher<HashEffectState> searcher(CHECKMATE_DEFAULT_TOTAL_NODE_LIMIT,
						  verbose);
#if 0
  int count=0;
#endif  
  while (is)
  {
    PawnMaskState state;
    int limit;
    if (readState(is, state, limit))
    {
#if 0
      if(++count==3388){
	std::cout << state << std::endl;
	checkDebug=true;
      }
#endif
      if (state.getTurn() == BLACK)
	testWinOrLose<BLACK>(filename, searcher, state, limit);
      else
	testWinOrLose<WHITE>(filename, searcher, state, limit);
    }
  }
  clock_message(totalCycles, "total ", 
		searcher.black().getTotalNodeCount()
		+ searcher.white().getTotalNodeCount());
  clock_message(totalCycles, "unique", 
		searcher.getTableBlack().size()
		+ searcher.getTableWhite().size());
}


/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
