#include "osl/category/standardMoveGenerator.h"
#include "osl/category/categoryListUtil.h"
#include "osl/category/probOfEachCategory.h"
#include "osl/category/orderOfMove.h"
#include "osl/category/categoryUtil.h"
#include "osl/stat/histogram.h"
#include "osl/stat/average.h"
#include "osl/moveStack.h"
#include "osl/moveLogProbSet.h"
#include "osl/pieceEval.h"
#include "osl/numEffectState.h"
#include "osl/moveLogProb.h"
#include "osl/record/csa.h"
#include "osl/record/csaIOError.h"
#include "osl/kanjiPrint.h"
#include <boost/format.hpp>
#include <iostream>
#include <string>
#include <map>
using namespace osl;
using namespace osl::category;

/**
 * @file 
 * ƥӤ¬
 */

void usage(const char *prog)
{
  using namespace std;
  cerr << "Usage: " << prog << " [-l limit] [-O order-limit] [-f skip] csafiles"
       << endl;
  exit(1);
}

/** print state if generated move has a greater probability limit than this */
int limit = 0;
/** print state if generated move has a greater order than this */
int orderLimit=0;
/** skip first n moves */
size_t firstSkip = 0;
bool verbose = false;

stat::Average totalProbAverage, totalAllProbAverage,
  totalOrderAverage, totalGeneratedAverage;
const size_t phWidth = 200, phLength=8;
/** λؼγΨ */
stat::Histogram totalProbHistogram(phWidth, phLength);
/** ƤλؼγΨ */
stat::Histogram totalAllProbHistogram(phWidth, phLength);
const size_t ohWidth = 10, ohLength=10;
stat::Histogram totalOrderHistogram(ohWidth, ohLength);
stat::Histogram totalGeneratedHistogram(ohWidth, ohLength);

void testFile(const char *fileName);

struct CategoryRollBook : private OrderOfMove::Output
{
  struct Data
  {
    /** ƥ꤬դΨ */
    stat::Average prob;
    /** λν */
    stat::Average order;
    /** ɤΨμäƤޤä */
    int redherring;
    Data() : redherring(0)
    {
    }
  };
private:
  typedef std::map<std::string, Data> map_t;
  map_t book;
public:
  /** ǡ̤Υ饹ʬ٤ */
  Move analysisTarget;
  /** OrderOfMove::Output ȤƤƯ Ϥ̤Υ饹ʬ٤ */
  void generatorOfEachMove(const char *name, const MoveLogProb& move)
  {
    if (analysisTarget == move.getMove())
    {
      book[name].prob.add(move.getLogProb());
      book[name].order.add(order);
    }
    else
    {
      book[name].redherring++;
    }
  }
public:
  CategoryRollBook();
  template <class MG>
  MoveLogProb analyzeMove(const CategoryEnv& env, Move target,
			  int *order, int *generated)
  {
    this->order = 0;
    this->generated = 0;
    analysisTarget = target;
    OrderOfMove orderAnalyzer(env, target, *this);
    const MoveLogProb result = orderAnalyzer.generate(StandardMoveGenerator());
    *order = this->order;
    *generated = this->generated;
    return result;
  }
  typedef std::multimap<int, std::pair<std::string, const Data*>, 
			std::greater<int> > sorter_t;
  void show(std::ostream& os, const sorter_t& sorter) const
  {
    os << "\nname  Ψ ʾγΨμΨ  Ψ\n";
    for (sorter_t::const_iterator p=sorter.begin(); p!=sorter.end(); ++p)
    {
      const int hit = p->first;
      const int red = p->second.second->redherring;
      const double ratio = 100.0*static_cast<double>(hit)/(hit+red);
      os << boost::format("%22s  % 8d  % 8d  %5.1f  % 6.1f  % 6.1f\n") 
	% p->second.first
	% hit % red % ratio
	% p->second.second->order.getAverage()
	% p->second.second->prob.getAverage();
    }
  }
  void show(std::ostream& os) const
  {
    sorter_t sorter;
    for (map_t::const_iterator p=book.begin(); p!=book.end(); ++p)
    {
      sorter.insert(std::make_pair(p->second.order.numElements(),
				   std::make_pair(p->first, &p->second)));
    }
    show(os, sorter);
  }
};
CategoryRollBook::CategoryRollBook()
{
}

CategoryRollBook book;

int main(int argc, char **argv)
{
  const char *program_name = argv[0];
  bool error_flag = false;
  extern char *optarg;
  extern int optind;
    
  char c;
  while ((c = getopt(argc, argv, "l:f:O:svh")) != EOF)
  {
    switch(c)
    {
    case 'l':	limit = atoi(optarg);
      break;
    case 'f':	firstSkip = atoi(optarg);
      break;
    case 'O':	orderLimit = atoi(optarg);
      break;
    case 'v':	verbose = true;
      break;
    default:	error_flag = true;
    }
  }
  argc -= optind;
  argv += optind;

  if (error_flag || (argc < 1))
    usage(program_name);

  for (int i=0; i<argc; ++i)
  {
    if ((! verbose) && (i % 128 == 0))
      std::cerr << '.';
    testFile(argv[i]);
  }

  std::cout << "\n"
	    << "average order " << totalOrderAverage.getAverage() << "\n";
  totalOrderHistogram.show(std::cout);
  std::cout << "average generated moves " << totalGeneratedAverage.getAverage() << "\n";
  totalGeneratedHistogram.show(std::cout);
  std::cout << "average probability of recorded moves " << totalProbAverage.getAverage() << "\n";
  totalProbHistogram.show(std::cout);
  std::cout << "average probability of all moves " << totalAllProbAverage.getAverage() << "\n";
  totalAllProbHistogram.show(std::cout);
  book.show(std::cout);
}

void testFile(const char *filename)
{
  Record rec;
  try
  {
    rec = CsaFile(filename).getRecord();
  }
  catch (CsaIOError& e)
  {
    std::cerr << "skip " << filename <<"\n";
    std::cerr << e.what() << "\n";
    return;
  }
  catch (...)
  {
    throw;
  }
  
  NumEffectState state(rec.getInitialState());
  std::vector<osl::Move> moves=rec.getMoves();

  KanjiPrint printer(std::cout);
  MoveStack history;
  CategoryEnv env(&state, 2000, &history);

  stat::Histogram probHistogram(phWidth,phLength);
  stat::Histogram orderHistogram(ohWidth, ohLength);
  stat::Histogram generatedHistogram(ohWidth, ohLength);
  stat::Average orderAverage, generatedAverage, probAverage;
  ProbOfEachCategory::Reporter reporter(std::cout);

  for (size_t i=0; i<moves.size(); ++i)
  {
    const Move move = moves[i];
    assert(state.isValidMove(move));
    /** פäƤ */
    if (i >= firstSkip)
    {
      int order=0, generated = 0;
      const MoveLogProb found = 
	book.analyzeMove<StandardMoveGenerator>(env, move, &order, &generated);

      assert(found.validMove() 
	     || (std::cerr << state << move << " " << found << " " << filename << "\n",0));
      orderHistogram.add(order);
      orderAverage.add(order);
      generatedHistogram.add(generated);
      generatedAverage.add(generated);
      probHistogram.add(found.getLogProb());
      probAverage.add(found.getLogProb());
      if ((limit && (found.getLogProb() > limit))
	  || (orderLimit && (order > orderLimit)))
      {
	printer.print(state);
	std::cout << found << "\n";
	const int diff
	  = PieceEval::computeDiffAfterMoveForRP(state, move);
	std::cout << "soma " << diff << "\n";
	ProbOfEachCategory a(env, move, &reporter);
	a.analyze(StandardMoveGenerator());
	std::cout << "\n";
      }

      MoveLogProbSet allMoves;
      CategoryListUtil::gatherAllMoves<StandardMoveGenerator>(env, allMoves);
      for (MoveLogProbSet::const_iterator p=allMoves.begin(); 
	   p!=allMoves.end(); ++p)
      {
	totalAllProbHistogram.add(p->getLogProb());
	totalAllProbAverage.add(p->getLogProb());
      }
    }
    state.doMove(move);
    history.push(move);
  }
  if (verbose)
  {
    std::cout << "arevage order " << orderAverage.getAverage() << "\t";
    std::cout << "arevage prob. " << probAverage.getAverage() << "\n";
  }
  
  totalOrderAverage.merge(orderAverage);
  totalOrderHistogram.merge(orderHistogram);
  totalGeneratedAverage.merge(generatedAverage);
  totalGeneratedHistogram.merge(generatedHistogram);
  totalProbAverage.merge(probAverage);
  totalProbHistogram.merge(probHistogram);
}


// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
