/**
 * ؼ®򸫤
 */
#include "osl/record/csaString.h"
#include "osl/evalHashEffectState.h"
#include "osl/moveLogProbSet.h"
#include "osl/applyMove.h"

#include "osl/category/standardMoveGenerator.h"
#include "osl/category/categoryListUtil.h"
#include "osl/perfmon.h"
#include <iostream>
using namespace osl;
using namespace osl::category;

int moveCount;
int maxLevel;

MoveStack history;
int limit = 800;

template<Player P ,typename State>
void nextMoves(State& state, int level);

template<class State, Player P>
struct DoUndoHelper
{
  State& state;
  int level;
  Move move;
    
  DoUndoHelper(State& s, int level) 
    : state(s), level(level), move(MOVE_INVALID)
  {
  }
  void operator()(Position p)
  {
    history.push(move);
    assert(move!=MOVE_INVALID);
    nextMoves<P,State>(state, level);
    history.pop();
  }
};

template <Player P, typename State>
void nextMoves(State& state,int level)
{
  typedef typename State::effect_state_t effect_t;
  if (level>maxLevel) 
    return;
  const Position opKingPosition
    =state.template getKingPosition<PlayerTraits<P>::opponent>();
  if (state.hasEffectBy(P,opKingPosition)) // ƨˤʤäƤʤ
    return;

  MoveLogProbSet moves;
  CategoryEnv env(&state, limit, &history);
  CategoryListUtil::gatherAllMoves<StandardMoveGenerator>(env, moves);

  typedef DoUndoHelper<State,PlayerTraits<P>::opponent> helper_t;
  helper_t helper(state, level+1);

  for (MoveLogProbSet::const_iterator p=moves.begin(); p!=moves.end(); ++p)
  {
    const Move move = p->getMove();
    moveCount++;
    helper.move = move;
    ApplyMove<P>::doUndoMove(state, move, helper);
  }
}

int main(int argc,char **argv)
{
  // bool effectMode=false;
  bool hashMode=false;
  bool evalMode=false;
  int level=3;
  extern char *optarg;

  char c;
  while ((c = getopt(argc, argv, "l:L:hE")) != EOF)
  {
    switch(c)
    {
    case 'l':	level=atoi(optarg);
      break;
    case 'L':	limit=atoi(optarg);
      break;
    case 'h':	hashMode=true;
      break;
    case 'E':	evalMode=true;
      break;
    default:
      std::cerr << "unknown option\n";
      std::cerr << "Usage: " << argv[0] << "[-l depth] [-L limit] [-hE]\n";
      return 1;
    }
  }
  PawnMaskState state=CsaString(
    "P1-KY *  *  * -KY * -FU-KE * \n"
    "P2 *  *  *  * -OU *  *  *  * \n"
    "P3 *  *  * -FU-FU+RY *  * -KY\n"
    "P4-FU *  * -GI *  *  *  *  * \n"
    "P5 *  *  *  *  *  *  *  *  * \n"
    "P6+FU *  * +RY *  * +FU *  * \n"
    "P7 * +FU * +FU+FU+FU *  *  * \n"
    "P8 *  * +OU * -TO *  *  *  * \n"
    "P9+KY *  *  *  *  *  * +KE * \n"
    "P+00KI00GI00GI00GI00KE00KE00FU00FU00FU00KI\n"
    "P-00KA00KA00KI00FU00FU00FU00FU00KI\n"
    "-\n").getInitialState();
  maxLevel=level;
  moveCount=0;
  clock_start();

  Move lastMove=newMove(newPosition(4,4),newPosition(4,3),PROOK,
			PTYPE_EMPTY,false,BLACK);
  history.push(lastMove);
  if(evalMode)
  {
    std::cerr << "evalMode" << std::endl;
    EvalHashEffectState eState(state);
    
    nextMoves<WHITE,EvalHashEffectState>(eState,0);
  }
  else if(hashMode)
  {
    std::cerr << "hashMode" << std::endl;
    HashEffectState hState(state);
    
    nextMoves<WHITE,HashEffectState>(hState,0);
  }
  else // effectMode
  {
    std::cerr << "effectMode" << std::endl;
    NumEffectState neState(state);
    nextMoves<WHITE,NumEffectState>(neState,0);
  }
  clock_stop("total", moveCount+1);
  std::cerr << "maxLevel=" << maxLevel << ",moveCount=" << moveCount 
	    << std::endl;

}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
