#include "osl/board.h"
#include "osl/bitXmask.h"
#include <iomanip>
#include <iostream>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

using namespace osl;
extern bool isShortTest;

class boardTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( boardTest );
  CPPUNIT_TEST( testConversion );
  CPPUNIT_TEST( testOnBoard );
  CPPUNIT_TEST( testIsValid );
  CPPUNIT_TEST( testOffsetShow );
  CPPUNIT_TEST( testOffsetSub );
  CPPUNIT_TEST( testOffsetAdd );
  CPPUNIT_TEST_SUITE_END();
 public:
  void testConversion();
  void testOnBoard();
  void testIsValid();
  void testOffsetShow();
  void testOffsetSub();
  void testOffsetAdd();
};


CPPUNIT_TEST_SUITE_REGISTRATION(boardTest);

void boardTest::testConversion(){
  for(int x=9;x>=1;x--)
    for(int y=1;y<=9;y++){
      Position pos=newPosition(x,y);
      CPPUNIT_ASSERT( getX(pos) == x  || (std::cout << pos << "," << std::setbase(16) << static_cast<int>(pos) << std::setbase(10) << x << "," << getX(pos) << std::endl,0));
      CPPUNIT_ASSERT( getY(pos) == y || (std::cout << pos << "," << y << "," << getY(pos) << std::endl,0));
    }
}
void boardTest::testOnBoard(){
  CPPUNIT_ASSERT( isOnBoard(OFFBOARD) ==false);
  for(int x=9;x>0;x--)
    for(int y=1;y<=9;y++){
      Position pos=newPosition(x,y);
      CPPUNIT_ASSERT( isOnBoard(pos) == true );
    }
  CPPUNIT_ASSERT( isOnBoard(newPosition(0,4)) ==false);
  CPPUNIT_ASSERT( isOnBoard(newPosition(10,5)) ==false);
  CPPUNIT_ASSERT( isOnBoard(newPosition(4,0)) ==false);
  CPPUNIT_ASSERT( isOnBoard(newPosition(5,10)) ==false);
}
void boardTest::testIsValid(){
  CPPUNIT_ASSERT( isValid(OFFBOARD) ==true);
  for(int x=9;x>0;x--)
    for(int y=1;y<=9;y++){
      Position pos=newPosition(x,y);
      CPPUNIT_ASSERT( isValid(pos) == true );
    }
  CPPUNIT_ASSERT( isValid(newPosition(0,4)) ==false);
  CPPUNIT_ASSERT( isValid(newPosition(10,5)) ==false);
  CPPUNIT_ASSERT( isValid(newPosition(4,0)) ==false);
  CPPUNIT_ASSERT( isValid(newPosition(5,10)) ==false);
}

void boardTest::testOffsetShow(){
  if (! isShortTest)
  {
    std::cout << newOffset(-3,5) << std::endl;
    std::cout << newOffset(9,9) << std::endl;
  }
}
void boardTest::testOffsetSub(){
  CPPUNIT_ASSERT( newPosition(9,9)-newPosition(7,6)==newOffset(2,3));
  CPPUNIT_ASSERT( newPosition(6,7)-newPosition(9,9)==newOffset(-3,-2));
}
void boardTest::testOffsetAdd(){
  CPPUNIT_ASSERT( newPosition(9,9)-newOffset(2,3)==newPosition(7,6));
  CPPUNIT_ASSERT( newPosition(6,7)+newOffset(-3,-2)==newPosition(3,5));
  Position p=newPosition(6,7);
  p+=newOffset(-2,2);
  CPPUNIT_ASSERT( p==newPosition(4,9));
  p-=newOffset(-5,3);
  CPPUNIT_ASSERT( p==newPosition(9,6));

}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
