#include "osl/pawnMaskState.h"
#include "osl/moveLogProb.h"
#include "osl/csaString.h"
#include "osl/numEffectState.h"
#include "osl/eval.tcc"
#include "osl/captureEscapeMoveGen.h"
#include "osl/moveLogProbVector.h"
#include "osl/moveStack.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <algorithm>

class captureEscapeMoveGenTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE(captureEscapeMoveGenTest);
  CPPUNIT_TEST(testGenerate);
  CPPUNIT_TEST(testGenerate2);
  CPPUNIT_TEST(testBug37to);
  CPPUNIT_TEST_SUITE_END();
 private:
  osl::PawnMaskState state;
 public:
  void setUp();
  void testGenerate();
  void testGenerate2();
  void testBug37to();
};


CPPUNIT_TEST_SUITE_REGISTRATION(captureEscapeMoveGenTest);

using namespace osl;
extern bool isShortTest;

void captureEscapeMoveGenTest::setUp(){
  state=CsaString(
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU\n"
"P-00AL\n"
"+\n").getInitialState();
}

void captureEscapeMoveGenTest::testGenerate(){
  NumEffectState neState(state);

  MoveLogProbVector moves;
  CaptureEscapeMoveGenerator generator;
  MoveStack history;
  generator.generate(neState,history,1000,moves);
  CPPUNIT_ASSERT(moves.size() > 0);
  // Ȥ̣٤ɤʤΤǸ
  if (! isShortTest)
    std::cout << moves;
  moves.clear();

  CaptureEscapeMoveGenerator generator1;
  generator1.generate(neState,history,500,moves);
  CPPUNIT_ASSERT(moves.size() > 0);
  if (! isShortTest)
    std::cout << moves;
}

void captureEscapeMoveGenTest::testGenerate2(){
  PawnMaskState bugState=CsaString(
"P1-KY-KE * -KI *  *  * -KE-KY\n"
"P2 * -OU-GI * -KI * -HI *  * \n"
"P3 * -FU *  * -FU-GI-KA-FU-FU\n"
"P4-FU * -FU-FU * -FU-FU *  * \n"
"P5 *  *  *  *  *  *  * +FU * \n"
"P6+FU * +FU * +FU+GI+FU * +FU\n"
"P7 * +FU * +FU * +FU *  *  * \n"
"P8 * +KA+OU * +KI+GI * +HI * \n"
"P9+KY+KE * +KI *  *  * +KE+KY\n"
"+\n").getInitialState();
  bugState.doMove(newMove(newPosition(9,6),newPosition(9,5),PAWN,PTYPE_EMPTY,false,BLACK));
  bugState.doMove(newMove(newPosition(9,4),newPosition(9,5),PAWN,PAWN,false,WHITE));
  bugState.doMove(newMove(newPosition(8,8),newPosition(4,4),BISHOP,PAWN,false,BLACK));
  
  NumEffectState neState(bugState);
  MoveLogProbVector moves;
  
  CaptureEscapeMoveGenerator generator1;
  MoveStack history;
  generator1.generate(neState,history,500,moves);

  const MoveLogProb expected1(newMove(newPosition(4,3),newPosition(4,4),SILVER,BISHOP,false,WHITE),44);
  const MoveLogProb expected2(newMove(newPosition(3,3),newPosition(4,4),BISHOP,BISHOP,false,WHITE),44);

  CPPUNIT_ASSERT(std::find(moves.begin(), moves.end(), expected1));
  CPPUNIT_ASSERT(std::find(moves.begin(), moves.end(), expected2));
}

void captureEscapeMoveGenTest::testBug37to(){
  PawnMaskState bugState=CsaString(
"P1 * -KE * -KI *  *  * -KE-KY\n"
"P2 *  * -GI *  * -HI-GI *  * \n"
"P3-KY-OU-KI * -FU * -KA-FU * \n"
"P4-FU-FU-FU-FU *  *  *  * -FU\n"
"P5 *  *  *  * +FU+FU * +FU * \n"
"P6+FU * +FU+FU+GI+GI-FU * +FU\n"
"P7 * +FU+KA *  * +KI+KE *  * \n"
"P8 *  * +OU *  *  *  * +HI * \n"
"P9+KY+KE * +KI *  *  *  * +KY\n"
"P-00FU\n"
"P-00FU\n"
"-\n").getInitialState();
  NumEffectState neState(bugState);

  MoveLogProbVector moves;
  CaptureEscapeMoveGenerator generator1;
  MoveStack history;
  generator1.generate(neState,history,500,moves);
  const MoveLogProb expected
      (newMove(newPosition(3,6),newPosition(3,7),PPAWN,KNIGHT,true,WHITE),
       270);
  CPPUNIT_ASSERT(moves.isMember(expected));
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
