#include "osl/category/check.h"
#include "osl/category/categoryUtil.h"
#include "osl/record/csaString.h"
#include "osl/numEffectState.h"
#include "osl/moveLogProbVector.h"
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <algorithm>
#include <iostream>

using namespace osl;
using namespace osl::category;
extern bool isShortTest;

class CheckTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(CheckTest);
  CPPUNIT_TEST(testCount);
  CPPUNIT_TEST_SUITE_END();
  PawnMaskState state;
public:
  void setUp();
  void testCount();
};

CPPUNIT_TEST_SUITE_REGISTRATION(CheckTest);

void CheckTest::setUp(){
  state=CsaString(
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU\n"
"P-00AL\n"
"+\n").getInitialState();
}

void CheckTest::testCount()
{
  NumEffectState neState(state);

  MoveLogProbVector moves;
  CategoryEnv env(&neState, 1000, 0);
  CategoryUtil<Check>::generate(env,moves);
  CPPUNIT_ASSERT_EQUAL(static_cast<size_t>(8u), moves.size());
#if 0
  if (! isShortTest)
    std::cout << moves;
#endif
// Drop(+,GOLD,Position(22))
// Move(+,PBISHOP,capture=SILVER,Position(53)->Position(42),promote=0)
// Drop(+,GOLD,Position(41))
// Drop(+,ROOK,Position(41))
// Drop(+,ROOK,Position(51))
// Drop(+,ROOK,Position(61))
// Drop(+,ROOK,Position(71))
// Drop(+,KNIGHT,Position(43))
  moves.clear();

  env.limit = 420;
  CategoryUtil<Check>::generate(env,moves);
  if (! isShortTest)
    std::cout << moves;
  CPPUNIT_ASSERT_EQUAL(static_cast<size_t>(2u), moves.size());
// Drop(+,ROOK,Position(61))
// Drop(+,ROOK,Position(71))
}

#if 0
void CheckTest::testShow2()
{
  PawnMaskState bugState=CsaString(
"P1-KY-KE * -KI *  *  * -KE-KY\n"
"P2 * -OU-GI * -KI * -HI *  * \n"
"P3 * -FU *  * -FU-GI-KA-FU-FU\n"
"P4-FU * -FU-FU * -FU-FU *  * \n"
"P5 *  *  *  *  *  *  * +FU * \n"
"P6+FU * +FU * +FU+GI+FU * +FU\n"
"P7 * +FU * +FU * +FU *  *  * \n"
"P8 * +KA+OU * +KI+GI * +HI * \n"
"P9+KY+KE * +KI *  *  * +KE+KY\n"
"+\n").getInitialState();
  bugState.doMove(newMove(newPosition(9,6),newPosition(9,5),PAWN,PTYPE_EMPTY,false,BLACK));
  bugState.doMove(newMove(newPosition(9,4),newPosition(9,5),PAWN,PAWN,false,WHITE));
  bugState.doMove(newMove(newPosition(8,8),newPosition(4,4),BISHOP,PAWN,false,BLACK));
  
  NumEffectState neState(bugState);

  MoveLogProbVector moves;
  CaptureCheckMoveGenerator action1;
  MoveStack history;
  history.push(newDrop(newPosition(4,4),PAWN,WHITE));
  action1.generate(neState,history,500, moves);
  // std::cout << action1;

  const MoveLogProb expected1(newMove(newPosition(4,3),newPosition(4,4),SILVER,BISHOP,false,WHITE),44);
  const MoveLogProb expected2(newMove(newPosition(3,3),newPosition(4,4),BISHOP,BISHOP,false,WHITE),44);

  CPPUNIT_ASSERT(std::find(moves.begin(), moves.end(), expected1));
  CPPUNIT_ASSERT(std::find(moves.begin(), moves.end(), expected2));
}
#endif
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
