#include "osl/category/standardMoveGenerator.h"
#include "osl/category/categoryListUtil.h"
#include "osl/category/categoryUtil.h"
#include "osl/category/probOfEachCategory.h"
#include "osl/category/orderOfMove.h"
#include "osl/stat/histogram.h"
#include "osl/stat/average.h"
#include "osl/moveStack.h"
#include "osl/moveLogProbSet.h"
#include "osl/pieceEval.h"
#include "osl/record/record.h"
#include "osl/record/csa.h"
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <boost/scoped_array.hpp>
#include <string>
#include <fstream>
#include <iostream>

class StandardMoveGeneratorTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(StandardMoveGeneratorTest);
  CPPUNIT_TEST(testCover);
  CPPUNIT_TEST_SUITE_END();
public:
  void testCover();
};

CPPUNIT_TEST_SUITE_REGISTRATION(StandardMoveGeneratorTest);

using namespace osl;
using namespace osl::category;
extern bool isShortTest;


static void testFile(std::string const& fileName)
{
  const Record rec=CsaFile(fileName).getRecord();
  NumEffectState state(rec.getInitialState());
  const std::vector<osl::Move> moves=rec.getMoves();
  MoveStack history;
  const size_t limit = 1800;	// ۤ800?
  CategoryEnv env(&state, limit, &history);

  stat::Histogram stat(200,10);
  stat::Average orderAverage;
  ProbOfEachCategory::Reporter reporter(std::cerr);
  for (size_t i=0; i<moves.size(); ++i)
  {
    const Move move = moves[i];

    // ˡΥƥ
    {
      MoveLogProbSet allMoves;
      CategoryListUtil::gatherAllMoves<StandardMoveGenerator>(env, allMoves);
      for (MoveLogProbSet::const_iterator p=allMoves.begin(); p!=allMoves.end(); ++p)
      {
	CPPUNIT_ASSERT(isValid(p->getMove()));
	CPPUNIT_ASSERT(state.isValidMove(p->getMove()));
      }
    }
    // ΨΥƥ
    OrderOfMove::Output order;
    OrderOfMove orderAnalyzer(env, move, order);
#if 0
    const 
#endif
      MoveLogProb found = orderAnalyzer.generate(StandardMoveGenerator());
    CPPUNIT_ASSERT(found.validMove()
		   || (std::cerr << state << move << "\n", 0));
    stat.add(found.getLogProb());
    // std::cerr << "order " << order << " " << category << "\n";
    orderAverage.add(order.order);

    if ((! isShortTest) && (found.getLogProb() > 1500))
    {
      std::cerr << state << found << "\n";
      ProbOfEachCategory a(env, move, &reporter);
      a.analyze(StandardMoveGenerator());
      std::cerr << "\n";
    }
    CPPUNIT_ASSERT(found.getLogProb() < 1600);

    state.doMove(move);
    history.push(move);
  }
  if (! isShortTest)
  {
    std::cout << "average order " << orderAverage.getAverage() << "\n";
    stat.show(std::cout);
  }
  CPPUNIT_ASSERT(orderAverage.getAverage() < 20);
}

void StandardMoveGeneratorTest::testCover()
{
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  std::string fileName;
  for(int i=0;i<100 && (ifs >> fileName) ; i++){
    testFile("../data/kifdat/" + fileName);
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
