/* arrayCheckHashTableTest.cc
 */
#ifndef USE_ESTIMATION_BY_DOMINANCE
#  define USE_ESTIMATION_BY_DOMINANCE
#endif

#include "osl/checkmate/arrayCheckHashTable.h"
#include "osl/hashEffectState.h"
#include "osl/record/csaString.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

class arrayCheckHashTableTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(arrayCheckHashTableTest);
  CPPUNIT_TEST(testAllocateBlack);
  CPPUNIT_TEST(testAllocateWhite);
  CPPUNIT_TEST_SUITE_END();
public:
  void testAllocateBlack();
  void testAllocateWhite();
};
CPPUNIT_TEST_SUITE_REGISTRATION(arrayCheckHashTableTest);

using namespace osl;
using namespace osl::checkmate;

/**  */
void arrayCheckHashTableTest::testAllocateBlack()
{
  PawnMaskState state=CsaString(
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU\n"
"P-00AL\n"
"+\n").getInitialState();
  HashEffectState hState((NumEffectState(state)));

  ArrayCheckHashTable table(BLACK);
  CheckHashRecord *record = table.allocate(hState.getHash());
  CPPUNIT_ASSERT(record);
  CPPUNIT_ASSERT_EQUAL(1u, getProof(record->proofDisProof));
  CPPUNIT_ASSERT_EQUAL(1u, getDisProof(record->proofDisProof));
  record->proofDisProof = makeProofDisProof(3,5);

  PawnMaskState state2=CsaString( // ⤬1羯ʤ
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU00FU\n"
"P-00AL\n"
"+\n").getInitialState();
  HashEffectState hState2((NumEffectState(state2)));
  CheckHashRecord *record2 = table.allocate(hState2.getHash());
  CPPUNIT_ASSERT(record2);
  CPPUNIT_ASSERT(record2 != record);
  CPPUNIT_ASSERT_EQUAL(3u, getProof(record2->proofDisProof));
  CPPUNIT_ASSERT_EQUAL(1u, getDisProof(record2->proofDisProof));

  PawnMaskState state3=CsaString( // ⤬1羯ʤ
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU\n"
"P-00AL\n"
"+\n").getInitialState();
  HashEffectState hState3((NumEffectState(state3)));
  CheckHashRecord *record3 = table.allocate(hState3.getHash());
  CPPUNIT_ASSERT(record3);
  CPPUNIT_ASSERT(record3 != record);
  CPPUNIT_ASSERT_EQUAL(1u, getProof(record3->proofDisProof));
  CPPUNIT_ASSERT_EQUAL(5u, getDisProof(record3->proofDisProof));
}

/**  */
void arrayCheckHashTableTest::testAllocateWhite()
{
  PawnMaskState state=CsaString(
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU\n"
"P-00AL\n"
"-\n").getInitialState();
  HashEffectState hState((NumEffectState(state)));

  ArrayCheckHashTable table(WHITE);
  CheckHashRecord *record = table.allocate(hState.getHash());
  CPPUNIT_ASSERT(record);
  CPPUNIT_ASSERT_EQUAL(1u, getProof(record->proofDisProof));
  CPPUNIT_ASSERT_EQUAL(1u, getDisProof(record->proofDisProof));
  record->proofDisProof = makeProofDisProof(3,5);

  PawnMaskState state2=CsaString( // ⤬1羯ʤ
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU00FU\n"
"P-00AL\n"
"-\n").getInitialState();
  HashEffectState hState2((NumEffectState(state2)));
  CheckHashRecord *record2 = table.allocate(hState2.getHash());
  CPPUNIT_ASSERT(record2);
  CPPUNIT_ASSERT(record2 != record);
  CPPUNIT_ASSERT_EQUAL(1u, getProof(record2->proofDisProof));
  CPPUNIT_ASSERT_EQUAL(5u, getDisProof(record2->proofDisProof));

  PawnMaskState state3=CsaString( // ⤬1羯ʤ
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU\n"
"P-00AL\n"
"-\n").getInitialState();
  HashEffectState hState3((NumEffectState(state3)));
  CheckHashRecord *record3 = table.allocate(hState3.getHash());
  CPPUNIT_ASSERT(record3);
  CPPUNIT_ASSERT(record3 != record);
  CPPUNIT_ASSERT_EQUAL(3u, getProof(record3->proofDisProof));
  CPPUNIT_ASSERT_EQUAL(1u, getDisProof(record3->proofDisProof));
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
