/* checkHashRecordTest.cc
 */
#include "osl/checkmate/checkHashRecord.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

class checkHashRecordTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(checkHashRecordTest);
  CPPUNIT_TEST(testFindDagOrigin);
  CPPUNIT_TEST_SUITE_END();
public:
  void testFindDagOrigin();
};
CPPUNIT_TEST_SUITE_REGISTRATION(checkHashRecordTest);

using namespace osl;
using namespace osl::checkmate;

void checkHashRecordTest::testFindDagOrigin()
{
  {
    // ꤬ 0 ʤ 0
    CheckHashRecord record;
    CPPUNIT_ASSERT_EQUAL((CheckHashRecord*)0, record.findDagOrigin(0,true));
  }
  {
    //   a    
    //  / \   к
    // b   c  
    //  \ /   
    //   d    
    CheckHashRecord a, b, c, d;
    a.moves.push_back(CheckMove(MOVE_INVALID, &b));
    a.moves.push_back(CheckMove(MOVE_INVALID, &c));
    b.moves.push_back(CheckMove(MOVE_INVALID, &d));
    b.parent = &a;
    c.moves.push_back(CheckMove(MOVE_INVALID, &d));
    c.parent = &a;
    CPPUNIT_ASSERT_EQUAL((CheckHashRecord*)0, b.findDagOrigin(&c,true));
  }
  {
    //   a    
    //  / \   кɬ
    // b   c  
    // |   |
    // d   e
    //  \ /   
    //   f  
    CheckHashRecord a, b, c, d, e, f;
    a.moves.push_back(CheckMove(MOVE_INVALID, &b));
    a.moves.push_back(CheckMove(MOVE_INVALID, &c));
    b.moves.push_back(CheckMove(MOVE_INVALID, &d));
    b.parent = &a;
    c.moves.push_back(CheckMove(MOVE_INVALID, &e));
    c.parent = &a;
    d.parent = &b;
    e.parent = &c;
    d.proofDisProof = b.proofDisProof;
    e.proofDisProof = c.proofDisProof;
    CPPUNIT_ASSERT_EQUAL(&a, d.findDagOrigin(&e,true));
    CPPUNIT_ASSERT_EQUAL(&a, d.findDagOrigin(&e,false));
    CPPUNIT_ASSERT_EQUAL((CheckHashRecord*)0, a.findDagOrigin(0,true));

    b.proofDisProof = PawnCheckMate;
    CPPUNIT_ASSERT_EQUAL((CheckHashRecord*)0, d.findDagOrigin(&e,true));
    CPPUNIT_ASSERT_EQUAL((CheckHashRecord*)0, d.findDagOrigin(&e,false));
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
