/* checkmateSearcherTest.cc
 */

#define CHECKMATE_DEBUG

#include "osl/checkmate/checkmateSearcher.h"
#include "osl/checkmate/checkmateSearcher.tcc"

#include "osl/pawnMaskState.h"
#include "osl/record/csa.h"
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
// եʬޤ?
#if 1
#  include "osl/hashEffectState.h"
typedef osl::HashEffectState state_t;
#else
#  include "osl/generateAddEffectMoves.tcc"
#  include "osl/generateEscapeMoves.tcc"
#  include "osl/generateCaptureMoves.tcc"
#  include "osl/generateAllMoves.tcc"
#  include "osl/storeMoveAction.h"
#  include "osl/checkMoveGenerator.tcc"
#  if 1
#    include "osl/hashBoardEffectState.h"
typedef osl::HashBoardEffectState state_t;
#  else
#    include "osl/rotBitmapEffect.h"
#    include "osl/hash.h"
typedef osl::HashState<osl::RotBitmapEffect<osl::PawnMaskState> > state_t;
#  endif
#endif
#include "osl/record/csaString.h"

#include <fstream>
#include <iostream>

class checkmateSearcherTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( checkmateSearcherTest );
  CPPUNIT_TEST( testNoCheck );
  CPPUNIT_TEST( testLimit );
  CPPUNIT_TEST( test30c );
  CPPUNIT_TEST( testFilesCheck );
  CPPUNIT_TEST( testFilesNoCheck );
  CPPUNIT_TEST_SUITE_END();
public:
  void testNoCheck();
  void test30c();
  void testFilesCheck();
  void testFilesNoCheck();
  void testLimit();
};
CPPUNIT_TEST_SUITE_REGISTRATION(checkmateSearcherTest);

using namespace osl;
extern bool isShortTest;

template<typename State>
static void testNoCheck(){
  {
    PawnMaskState state=CsaString(
"P1 *  *  *  * -KI *  * -KE-KY\n"
"P2 * -OU *  *  *  *  * -HI * \n"
"P3 * -FU * -KI-FU-FU-GI * -FU\n"
"P4+FU * +FU-FU *  *  *  *  * \n"
"P5-FU *  *  *  *  *  * -FU * \n"
"P6 * -UM *  * +UM *  *  * +FU\n"
"P7 *  *  * +FU+FU+FU+FU+FU * \n"
"P8 *  *  * +KI *  *  * +HI * \n"
"P9+KY *  *  * +OU * +GI+KE+KY\n"
"P+00FU\n"
"P-00FU\n"
"P-00FU\n"
"P+00KE\n"
"P-00KE\n"
"P-00GI\n"
"P-00GI\n"
"P-00KI\n"
"P+00KY\n"
"-\n").getInitialState();
    typename State::effect_state_t neState(state);
    State hState(neState);
    CheckHashTable table(WHITE);
    CheckmateSearcher<WHITE,State> searcher(table);
    Move checkMove=MOVE_INVALID;
    ProofDisProof proofDisProof=searcher.hasCheckmateMove(hState,1000,checkMove);
    CPPUNIT_ASSERT(isUnknown(proofDisProof));
    proofDisProof=searcher.hasCheckmateMove(hState,100000,checkMove);
    CPPUNIT_ASSERT(!isCheckmateSuccess(proofDisProof));
  }
  {
    PawnMaskState state=CsaString(
"P1 *  *  *  *  *  *  * -KE-OU\n"
"P2 *  *  *  *  *  *  *  *  * \n"
"P3 *  *  *  *  *  *  * +TO-FU\n"
"P4 *  *  *  *  *  * +RY *  * \n"
"P5 *  *  *  *  *  *  *  *  * \n"
"P6 *  *  *  *  *  *  *  *  * \n"
"P7 *  *  *  *  *  *  *  *  * \n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 *  * +OU *  *  *  *  *  * \n"
"P+00FU\n"
"P-00AL\n"
"+\n").getInitialState();
    typename State::effect_state_t neState(state);
    State hState(neState);
    CheckHashTable table(BLACK);
    CheckmateSearcher<BLACK,State> searcher(table);
    Move lastMove=newMove(newPosition(2,4),newPosition(2,3),PPAWN,PAWN,true,BLACK);
    ProofDisProof proofDisProof=searcher.hasCheckmateMove(hState,1500,lastMove);
    if (PawnCheckMate != proofDisProof)
      abort();
    CPPUNIT_ASSERT_EQUAL(PawnCheckMate,proofDisProof);
  }
#if 1
  {
    PawnMaskState state=CsaString(
"P1 *  *  *  *  *  *  * -KE * \n"
"P2 *  *  *  *  *  *  * -OU * \n"
"P3 *  *  *  *  *  *  * -FU-FU\n"
"P4 *  *  *  *  *  * +RY+FU * \n"
"P5 *  *  *  *  *  *  *  *  * \n"
"P6 *  *  *  *  *  *  *  *  * \n"
"P7 *  *  *  *  *  *  *  *  * \n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 *  * +OU *  *  *  *  *  * \n"
"P-00AL\n"
"+\n").getInitialState();
    typename State::effect_state_t efState(state);
    State hState(efState);
    CheckHashTable table(BLACK);
    CheckmateSearcher<BLACK,State> searcher(table);
    Move checkMove=MOVE_INVALID;
    ProofDisProof proofDisProof=searcher.hasCheckmateMove(hState,2000,checkMove);
    CPPUNIT_ASSERT(isCheckmateSuccess(proofDisProof) ||
		   (std::cout << hState,abort(),0));
  }
#endif
  {
    PawnMaskState state=CsaString(
"P1 *  *  *  *  *  *  * -KE * \n"
"P2 *  *  *  *  *  *  * -OU * \n"
"P3 *  *  *  *  *  *  * +TO-FU\n"
"P4 *  *  *  *  *  * +RY *  * \n"
"P5 *  *  *  *  *  *  *  *  * \n"
"P6 *  *  *  *  *  *  *  *  * \n"
"P7 *  *  *  *  *  *  *  *  * \n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 *  * +OU *  *  *  *  *  * \n"
"P+00FU\n"
"P-00AL\n"
"-\n").getInitialState();
    typename State::effect_state_t neState(state);
    State hState(neState);
    CheckHashTable table(BLACK);
    CheckmateSearcher<BLACK,State> searcher(table);
    Move lastMove=newMove(newPosition(2,4),newPosition(2,3),PPAWN,PAWN,true,BLACK);
    ProofDisProof proofDisProof=searcher.hasEscapeMove(hState,1500,lastMove);
    CPPUNIT_ASSERT_EQUAL(PawnCheckMate,proofDisProof);
  }
  {
    PawnMaskState state=CsaString(
"P1-KY-KE * -KI *  *  * -KE-KY\n"
"P2 * -OU-GI * -KI * -HI *  * \n"
"P3 * -FU *  * -FU-GI-KA-FU-FU\n"
"P4-FU * -FU-FU * -FU-FU *  * \n"
"P5 *  *  *  *  *  *  * +FU * \n"
"P6+FU * +FU * +FU+GI+FU * +FU\n"
"P7 * +FU * +FU * +FU *  *  * \n"
"P8 * +KA+OU * +KI+GI * +HI * \n"
"P9+KY+KE * +KI *  *  * +KE+KY\n"
"+\n").getInitialState();
    typename State::effect_state_t neState(state);
    State hState(neState);
    CheckHashTable table(BLACK);
    CheckmateSearcher<BLACK,State> searcher(table);
    Move checkMove=MOVE_INVALID;
    CPPUNIT_ASSERT(isCheckmateFail(searcher.hasCheckmateMove(hState,1000,checkMove)));
  }
  {
    PawnMaskState state=CsaString(
"P1 *  *  *  *  *  *  * -KE-OU\n"
"P2 *  *  *  *  *  *  *  *  * \n"
"P3 *  *  *  *  *  *  * +TO-FU\n"
"P4 *  *  *  *  *  * +RY *  * \n"
"P5 *  *  *  *  *  *  *  *  * \n"
"P6 *  *  *  *  *  *  *  *  * \n"
"P7 *  *  *  *  *  *  *  *  * \n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 *  * +OU *  *  *  *  *  * \n"
"P+00FU\n"
"P-00AL\n"
"+\n").getInitialState();
    typename State::effect_state_t neState(state);
    State hState(neState);
    CheckHashTable table(BLACK);
    CheckmateSearcher<BLACK,State> searcher(table);
    Move checkMove=MOVE_INVALID;
    ProofDisProof proofDisProof=searcher.hasCheckmateMove(hState,1000,checkMove);
    CPPUNIT_ASSERT(proofDisProof==PawnCheckMate);
  }
}

/**
 * promoteǤͤpromoteʤȵͤߤȤ
 */

/**
 * Τʤ̤Ǥlimitʾʤ
 * CheckmateFail֤
 * 
 */
void checkmateSearcherTest::testNoCheck(){
  ::testNoCheck<state_t>();
}

template<typename State>
void testLimit(){
  PawnMaskState state0=CsaString(
"P1-KY-KE * -KI *  *  * -KE-KY\n"
"P2 * -OU-GI * -KI * -HI *  * \n"
"P3 * -FU *  * -FU-GI-KA-FU-FU\n"
"P4-FU * -FU-FU * -FU-FU *  * \n"
"P5 *  *  *  *  *  *  * +FU * \n"
"P6+FU * +FU * +FU+GI+FU * +FU\n"
"P7 * +FU * +FU * +FU *  *  * \n"
"P8 * +KA+OU * +KI+GI * +HI * \n"
"P9+KY+KE * +KI *  *  * +KE+KY\n"
"+\n").getInitialState();
  typename State::effect_state_t neState0(state0);
  State hState0(neState0);


  PawnMaskState state1=CsaString(
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU\n"
"P-00AL\n"
"+\n").getInitialState();
  typename State::effect_state_t neState1(state1);
  State hState1(neState1);

  CheckHashTable table(BLACK);
  CheckmateSearcher<BLACK,State> searcher(table,100);
  /* ˵ͤޤʤȤ狼̤ɤ */
  {
    Move checkMove=MOVE_INVALID;
    ProofDisProof proofDisProof=searcher.hasCheckmateMove(hState0,1000,checkMove);
    CPPUNIT_ASSERT(isCheckmateFail(proofDisProof));
  }
  /* ͤߤΥå˻֤̤ɤ */
  {
    Move checkMove=MOVE_INVALID;
    ProofDisProof proofDisProof=searcher.hasCheckmateMove(hState1,1000,checkMove);
    CPPUNIT_ASSERT(isUnknown(proofDisProof));
  }
  /* ˵ͤޤʤȤ狼̤ɤ */
  {
    Move checkMove=MOVE_INVALID;
    ProofDisProof proofDisProof=searcher.hasCheckmateMove(hState0,1000,checkMove);
    CPPUNIT_ASSERT(isCheckmateFail(proofDisProof));
  }
  /* ͤߤΥå˻֤̤ɤ */
  {
    Move checkMove=MOVE_INVALID;
    ProofDisProof proofDisProof=searcher.hasCheckmateMove(hState1,1000,checkMove);
    CPPUNIT_ASSERT(isUnknown(proofDisProof));
  }
}

/**
 * 
 */
void checkmateSearcherTest::testLimit(){
  ::testLimit<state_t>();
}

static int numSolved[2] = { 0 };

template<typename State,bool isCheck,bool isDefense>
static void testFile(const std::string& fileName){
  Record rec=CsaFile(fileName).getRecord();
  std::vector<osl::Move> moves=rec.getMoves();
  PawnMaskState state=rec.getInitialState();
  typename State::effect_state_t bState(state);

  State hState(bState);
  ProofDisProof proofDisProof;
  Move checkMove=MOVE_INVALID;
  if ((hState.getTurn()==BLACK) ^ isDefense) {
    CheckHashTable table(BLACK);
    CheckmateSearcher<BLACK,State> searcher(table,CHECKMATE_DEFAULT_TOTAL_NODE_LIMIT,false);
    if (isDefense)
      proofDisProof=searcher.hasEscapeMove(hState,100000,checkMove);
    else
      proofDisProof=searcher.hasCheckmateMove(hState,100000,checkMove);
  }
  else{
    CheckHashTable table(WHITE);
    CheckmateSearcher<WHITE,State> searcher(table,CHECKMATE_DEFAULT_TOTAL_NODE_LIMIT,false);
    if (isDefense)
      proofDisProof=searcher.hasEscapeMove(hState,100000,checkMove);
    else
      proofDisProof=searcher.hasCheckmateMove(hState,100000,checkMove);
  }

  if (! isShortTest)
    std::cout << "proof=" <<getProof(proofDisProof)
	      << ",disProof=" << getDisProof(proofDisProof) << std::endl;
  if(isFinal(proofDisProof)){
    ++numSolved[isCheck];
    if(isCheck){
      CPPUNIT_ASSERT(isCheckmateSuccess(proofDisProof) ||
		     (std::cout << fileName << "\n" << hState, abort(), 0));
      
      if((! moves.empty()) && (checkMove!=moves[0]))
      {
	std::cout << "different solution in " << fileName << "\n";
#if 0
	std::cout << "checkMove=" << checkMove << ",correct answer=" << moves[0] << std::endl;
	std::cout << state;
#endif
      }
    }
    else{
      CPPUNIT_ASSERT(isCheckmateFail(proofDisProof) ||
		     (std::cout << hState,0));
    }
  }
}

void checkmateSearcherTest::test30c(){
  if (! isShortTest)
    std::cerr << "testFilesCheck: 30c.csa" << std::endl;
  testFile<state_t,true,true>("../data/ProblemProgress2/checkmate/30c.csa");
}

void checkmateSearcherTest::testFilesCheck(){
  std::ifstream ifs("../data/kisenCheckProblems/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  int count=100;
  if (! isShortTest)
      count=10;
  std::string fileName;
  while((ifs >> fileName) && (++i<count)) {
    if(fileName == "") 
	break;
    if (! isShortTest)
      std::cerr << "testFilesCheck: " << fileName << " ";
    testFile<state_t,true,false>("../data/kisenCheckProblems/" + fileName);
  }
  std::cerr << numSolved[true] << " / " << i << " = " 
	    << (double)numSolved[true]/i << "\n";
}

void checkmateSearcherTest::testFilesNoCheck(){
  std::ifstream ifs("../data/kisenNoCheckProblems/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  int count=100;
  if (! isShortTest)
      count=10;
  std::string fileName;
  while((ifs >> fileName) && (++i<count)) {
    if(fileName == "") 
	break;
    if (! isShortTest)
      std::cerr << "testFilesNoCheck: " << fileName << " ";
    testFile<state_t,false,false>("../data/kisenNoCheckProblems/" + fileName);
  }
  std::cerr << numSolved[false] << " / " << i << " = " 
	    << (double)numSolved[false]/i << "\n";
}
// P1 * -KE * 
// P2 * -OU * 
// P3-FU *  * 
// P4 * +FU+RY
// P+00FU
// 8483FU ʤ
// 8483TO Ե


// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
