/* dualCheckMateSearcherTest.cc
 */
#include "osl/checkmate/dualCheckMateSearcher.h"
#include "osl/checkmate/checkTableAnalyzer.h"
#include "osl/hashEffectState.h"
#include "osl/record/csaString.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

#include <iostream>

class dualCheckMateSearcherTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE(dualCheckMateSearcherTest);
  CPPUNIT_TEST(testCheck);
  CPPUNIT_TEST(testInfty);
  CPPUNIT_TEST(testInfty2);
  CPPUNIT_TEST_SUITE_END();
public:
  void testCheck();
  void testInfty();
  void testInfty2();
};
CPPUNIT_TEST_SUITE_REGISTRATION(dualCheckMateSearcherTest);

using namespace osl;
using namespace osl::checkmate;

/**
 */
void dualCheckMateSearcherTest::testCheck()
{
  PawnMaskState state=CsaString(
    "P1+NY+TO *  *  *  * -OU-KE-KY\n"
    "P2 *  *  *  *  * -GI-KI *  *\n"
    "P3 * +RY *  * +UM * -KI-FU-FU\n"
    "P4 *  * +FU-FU *  *  *  *  *\n"
    "P5 *  * -KE * +FU *  * +FU *\n"
    "P6-KE *  * +FU+GI-FU *  * +FU\n"
    "P7 *  * -UM *  *  *  *  *  *\n"
    "P8 *  *  *  *  *  *  *  *  * \n"
    "P9 * +OU * -GI *  *  *  * -NG\n"
    "P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
    "P-00KI00KY00FU00FU\n"
    "P-00AL\n"
    "+\n").getInitialState();

  state.doMove(newDrop(newPosition(9,2),PAWN,BLACK));
  HashEffectState hState(state);

  size_t rootTreeSize = 0;
  {
    DualCheckMateSearcher<HashEffectState> checker;
    Move checkMateMove = MOVE_INVALID;
    // 3 ͤʤΤǤ˵ͤƤۤȤ
    const bool win = checker.isWinningState<WHITE>(50, hState, checkMateMove);
    CPPUNIT_ASSERT(win);
    CPPUNIT_ASSERT(checkMateMove);
    CheckTableAnalyzer analyzer;
    const CheckHashTable& table = checker.getTableWhite();
    const CheckHashRecord *record = table.find(hState.getHash());
    rootTreeSize = analyzer.treeSize(record);
    CPPUNIT_ASSERT_EQUAL(table.size(), rootTreeSize);
    CPPUNIT_ASSERT(analyzer.showProofTree(record,true,std::cerr) < rootTreeSize);
    std::cerr << analyzer.proofTreeSize(record,true)
	      << " < " << rootTreeSize << "\n";
  }
  
  {
    // ʤФͤ
    hState.doMove(newDrop(newPosition(8,8),GOLD,WHITE));
    DualCheckMateSearcher<HashEffectState> checker;

    const bool lose = checker.isLosingState<BLACK>(10, hState);
    CPPUNIT_ASSERT(lose);
    const CheckHashTable& table = checker.getTableWhite();
    CheckTableAnalyzer analyzer;
    const CheckHashRecord *record = table.find(hState.getHash());
    CPPUNIT_ASSERT(analyzer.treeSize(record) < rootTreeSize);
    // analyzer.showTree(record, std::cerr, 10);
    CPPUNIT_ASSERT(analyzer.showProofTree(record,false,std::cerr)
		   <= analyzer.treeSize(record));
  }
}

void dualCheckMateSearcherTest::testInfty(){

  PawnMaskState state=CsaString(
"P1-KY * -KI-KE *  *  *  * +UM\n"
"P2 *  *  *  * -OU * -GI *  * \n"
"P3-FU * -KE-FU-FU-FU * -KA-FU\n"
"P4 * -HI-FU *  *  *  *  *  * \n"
"P5 *  * +KI *  * +FU *  *  * \n"
"P6 *  *  * +FU+KI *  *  *  * \n"
"P7+FU * +OU * +FU * +KI+GI+FU\n"
"P8-TO *  *  *  *  *  *  *  * \n"
"P9 *  *  *  *  *  *  *  * +KY\n"
"P-00FU\n"
"P-00FU\n"
"P-00FU\n"
"P-00FU\n"
"P-00FU\n"
"P-00FU\n"
"P+00KE\n"
"P+00KE\n"
"P-00GI\n"
"P+00GI\n"
"P+00KY\n"
"P+00KY\n"
"P+00HI\n"
"-\n").getInitialState();
  HashEffectState hState(state);

  DualCheckMateSearcher<HashEffectState> checker;
  Move checkMateMove = MOVE_INVALID;
  const bool win = checker.isWinningState<WHITE>(100, hState, checkMateMove);
  CPPUNIT_ASSERT(!win);
  // cppunit ˤtimeout Ϥ? boost::test Ϥɤ
}

void dualCheckMateSearcherTest::testInfty2(){
  PawnMaskState state=CsaString(
"P1-KY-KE *  *  * -OU * -KE+KA\n"
"P2 *  * +FU * -KI *  *  *  * \n"
"P3-FU *  * -FU-GI-KI * -FU-FU\n"
"P4 *  *  *  * -FU *  *  *  * \n"
"P5 * +GI * +FU * -FU *  *  * \n"
"P6 *  *  *  * +FU *  *  *  * \n"
"P7+FU+FU * +KI * +FU * +FU+FU\n"
"P8 * +OU-NY *  *  *  *  * -RY\n"
"P9+KY+KE * +KI *  * +FU *  * \n"
"P+00FU\n"
"P+00FU\n"
"P-00FU\n"
"P-00KE\n"
"P+00GI\n"
"P+00GI\n"
"P-00KY\n"
"P+00KA\n"
"P-00HI\n"
"+\n").getInitialState();
  
  HashEffectState hState((NumEffectState(state)));

  DualCheckMateSearcher<HashEffectState> checker;
  const bool lose = checker.isLosingState<BLACK>(100, hState);
  CPPUNIT_ASSERT(! lose);
}


/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
