/* oracleProverTest.cc
 */
#include "osl/checkmate/oracleProver.h"
#include "osl/checkmate/dualCheckMateSearcher.h"
#include "osl/hashEffectState.h"
#include "osl/record/csaString.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

class oracleProverTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE(oracleProverTest);
  CPPUNIT_TEST(testNoCheck);
  CPPUNIT_TEST(testCheck);
  CPPUNIT_TEST(testCheckAfterAttack);
  CPPUNIT_TEST(testPawnCheckMate);
  CPPUNIT_TEST_SUITE_END();
public:
  void testNoCheck();
  void testCheck();
  void testPawnCheckMate();
  void testCheckAfterAttack();
};
CPPUNIT_TEST_SUITE_REGISTRATION(oracleProverTest);

using namespace osl;
using namespace osl::checkmate;

static PawnMaskState state=CsaString(
    "P1+NY+TO *  *  *  * -OU-KE-KY\n"
    "P2+FU *  *  *  * -GI-KI *  *\n"
    "P3 * +RY *  * +UM * -KI-FU-FU\n"
    "P4 *  * +FU-FU *  *  *  *  *\n"
    "P5 *  * -KE * +FU *  * +FU *\n"
    "P6-KE *  * +FU+GI-FU *  * +FU\n"
    "P7 *  * -UM *  *  *  *  *  *\n"
    "P8 *  *  *  *  *  *  *  *  * \n"
    "P9 * +OU * -GI *  *  *  * -NG\n"
    "P+00HI00KI00KE00KY00FU00FU00FU00FU00FU\n"
    "P-00KI00KY00FU00FU\n"
    "P-00AL\n"
    "-\n").getInitialState();
static HashEffectState hState((NumEffectState(state)));

/** oracle ʤǤ ͤޤʤ */
void oracleProverTest::testNoCheck()
{
  CheckHashTable table(WHITE);
  OracleProver<WHITE, HashEffectState> prover(table);
  ProofOracleAttack oracle;
  Move bestMove = MOVE_INVALID;
  CPPUNIT_ASSERT_EQUAL(false, prover.proofWin(hState, oracle, bestMove));
  CPPUNIT_ASSERT_EQUAL(MOVE_INVALID, bestMove);
}

void oracleProverTest::testCheck()
{
  DualCheckMateSearcher<HashEffectState> checker;
  CheckHashTable& table = checker.getTableWhite();
  Move checkMateMove = MOVE_INVALID;
  const bool win = checker.isWinningState<WHITE>(500, hState, checkMateMove);
  CPPUNIT_ASSERT(win);
  const CheckHashRecord *record = table.find(hState.getHash());
  CPPUNIT_ASSERT(record);
  ProofOracleAttack oracle(record);

  // ̤ͤϵͤ
  Move bestMove = MOVE_INVALID;
  OracleProver<WHITE, HashEffectState> prover(table);
  const bool result = prover.proofWin(hState, oracle, bestMove);
  CPPUNIT_ASSERT_EQUAL(true, result);
  CPPUNIT_ASSERT_EQUAL(bestMove, record->bestMove->move);

  // ̤ͤ
  bestMove = MOVE_INVALID;
  hState.doMove(newDrop(newPosition(3,7),PAWN,WHITE));
  hState.doMove(newDrop(newPosition(8,2),PAWN,BLACK));
  CPPUNIT_ASSERT_EQUAL(true, prover.proofWin(hState, oracle, bestMove));
  CPPUNIT_ASSERT_EQUAL(bestMove, record->bestMove->move);

  // ­ʤȵͤޤʤ
  {
    HashEffectState h2 = hState;
    h2.doMove(newDrop(newPosition(3,8),GOLD,WHITE));
    h2.doMove(newDrop(newPosition(4,3),PAWN,BLACK));
    CPPUNIT_ASSERT_EQUAL(false, prover.proofWin(h2, oracle, bestMove));
  }
  
  // ̤ưƤͤޤʤ
  {
    HashEffectState h2 = hState;
    h2.doMove(newDrop(newPosition(9,3),PAWN,WHITE));
    h2.doMove(newMove(newPosition(8,9),newPosition(7,9),KING,PTYPE_EMPTY,false,BLACK));
    CPPUNIT_ASSERT_EQUAL(false, prover.proofWin(h2, oracle, bestMove));
  }
}

void oracleProverTest::testPawnCheckMate()
{
  PawnMaskState state=CsaString(
    "P1+NY+TO *  *  *  * -OU-KE-KY\n"
    "P2 *  *  *  *  * -GI-KI *  *\n"
    "P3 * +RY *  * +UM * -KI-FU-FU\n"
    "P4 *  * -KE-FU *  *  *  *  *\n"
    "P5 *  *  *  * +FU *  * +FU *\n"
    "P6 * -KE * +FU+GI-FU *  * +FU\n"
    "P7 * +FU-UM *  *  *  *  *  *\n"
    "P8 *  *  *  *  *  *  *  *  * \n"
    "P9 * +OU+FU * -GI *  *  * -NG\n"
    "P+00KI00HI00KI00KE00KY00FU00FU00FU00FU00FU\n"
    "P-00KY00FU00FU\n"
    "P+00AL\n"
    "-\n").getInitialState();
  HashEffectState hState((NumEffectState(state)));

  DualCheckMateSearcher<HashEffectState> checker;
  CheckHashTable& table = checker.getTableWhite();
  Move checkMateMove = MOVE_INVALID;
  const bool win = checker.isWinningState<WHITE>(500, hState, checkMateMove);
  CPPUNIT_ASSERT(win);
  const CheckHashRecord *record = table.find(hState.getHash());
  ProofOracleAttack oracle(record);

  // ̤ͤϵͤ
  Move bestMove = MOVE_INVALID;
  OracleProver<WHITE, HashEffectState> prover(table);
  CPPUNIT_ASSERT_EQUAL(true, prover.proofWin(hState, oracle, bestMove));
  CPPUNIT_ASSERT_EQUAL(bestMove, record->bestMove->move);

  // ̤ͤ
  bestMove = MOVE_INVALID;
  hState.doMove(newDrop(newPosition(3,7),PAWN,WHITE));
  hState.doMove(newDrop(newPosition(4,3),PAWN,BLACK));
  CPPUNIT_ASSERT_EQUAL(true, prover.proofWin(hState, oracle, bestMove));
  CPPUNIT_ASSERT_EQUAL(bestMove, record->bestMove->move);

  // ƱǤͤˤʤ
  bestMove = MOVE_INVALID;
  hState.doMove(newDrop(newPosition(7,3),PAWN,WHITE));
  hState.doMove(newDrop(newPosition(9,9),PAWN,BLACK));
  CPPUNIT_ASSERT_EQUAL(false, prover.proofWin(hState, oracle, bestMove));
}

void oracleProverTest::testCheckAfterAttack()
{
  PawnMaskState state=CsaString(
    "P1+NY+TO *  *  *  * -OU-KE-KY\n"
    "P2+FU *  *  *  * -GI-KI *  *\n"
    "P3 * +RY *  * +UM * -KI-FU-FU\n"
    "P4 *  * +FU-FU *  *  *  *  *\n"
    "P5 *  * -KE * +FU *  * +FU *\n"
    "P6-KE *  * +FU+GI-FU *  * +FU\n"
    "P7 *  * -UM *  *  *  *  *  *\n"
    "P8 *  *  *  *  *  *  *  *  * \n"
    "P9-KI+OU * -GI *  *  *  * -NG\n"
    "P+00HI00KI00KE00KY00FU00FU00FU00FU00FU\n"
    "P-00KY00FU00FU\n"
    "P-00AL\n"
    "+\n").getInitialState();
  HashEffectState hState((NumEffectState(state)));

  DualCheckMateSearcher<HashEffectState> checker;
  CheckHashTable& table = checker.getTableWhite();
  const bool lose = checker.isLosingState<BLACK>(500, hState);
  CPPUNIT_ASSERT(lose);
  const CheckHashRecord *record = table.find(hState.getHash());
  ProofOracleDefense oracle(record);

  // ̤ͤϵͤ
  OracleProver<WHITE, HashEffectState> prover(table);
  CPPUNIT_ASSERT_EQUAL(true, prover.proofLose(hState, oracle));

  // ̤ͤ
  hState.doMove(newDrop(newPosition(4,3),PAWN,BLACK));
  hState.doMove(newDrop(newPosition(3,7),PAWN,WHITE));
  CPPUNIT_ASSERT_EQUAL(true, prover.proofLose(hState, oracle));
  CPPUNIT_ASSERT_EQUAL(true, checker.isLosingState<BLACK>(1, hState));

  // ̤ưƱǤϵͤޤʤ
  {
    HashEffectState h2 = hState;
    h2.doMove(newMove(newPosition(8,9),newPosition(7,9),KING,PTYPE_EMPTY,false,BLACK));
    h2.doMove(newDrop(newPosition(9,3),PAWN,WHITE));
    CPPUNIT_ASSERT_EQUAL(false, prover.proofLose(h2, oracle));
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
