/* compositEvalTest.cc
 */
#include "recursiveDoUndoMove.h"
#include "osl/compositEval.h"
#include "osl/compositEvalHashEffectState.h"
#include "osl/record/csaString.h"
#include "osl/record/csa.h"
#include "osl/doUndoMoveLock.h"
#include "doUndoMoveLockTest.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <bitset>
#include <fstream>

class compositEvalTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(compositEvalTest);
  CPPUNIT_TEST(testLoad);
  // ʹߤ testLoad θ˥ƥȤ뤳
  CPPUNIT_TEST(testConsistent);
  CPPUNIT_TEST(testDoUndoMove);
  CPPUNIT_TEST_SUITE_END();
public:
  void testLoad();
  void testConsistent();
  void testDoUndoMove();
};

CPPUNIT_TEST_SUITE_REGISTRATION(compositEvalTest);

using namespace osl;

void compositEvalTest::testLoad()
{
  const char *filename = "../data/sibling.pair";
  CPPUNIT_ASSERT(piecePairEvalTable.loadFromBinaryFile(filename));
}

void compositEvalTest::testConsistent()
{
  extern bool isShortTest;
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  std::string fileName;
  for (int i=0;i<(isShortTest ? 100 : 10) && (ifs >> fileName) ; i++)
  {
    if (fileName == "") 
      break;
    fileName = "../data/kifdat/" + fileName;

    Record rec=CsaFile(fileName).getRecord();
    CompositEvalHashEffectState eState(rec.getInitialState());
    std::vector<osl::Move> moves=rec.getMoves();

    DoUndoMoveLockTest<CompositEvalHashEffectState> lockTest(eState, moves);
    lockTest.run();

    std::vector<DoUndoMoveLock*> locks(moves.size(), 0);
    for (unsigned int i=0;i<moves.size();i++)
    {
      locks[i] = new DoUndoMoveLock(eState, moves[i]); // doMove
      CPPUNIT_ASSERT(eState.isConsistent(true));
    }
    for (int i=(int)(moves.size()-1);i>=0;i--)
    {
      delete locks[i];		// undoMove
      CPPUNIT_ASSERT(eState.isConsistent());
    }
  }
}

void compositEvalTest::testDoUndoMove(){
  extern bool isShortTest;
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  int count=100;
  if (isShortTest)
    count=10;
  std::string fileName;
  while ((ifs >> fileName) && (++i<count)) {
    if (fileName == "") 
      break;
    if (! isShortTest)
      std::cerr << fileName << " ";

    CsaFile file("../data/kifdat/"+fileName);
    Record rec=file.getRecord();
    CompositEvalHashEffectState eState(rec.getInitialState());

    std::vector<osl::Move> moves=rec.getMoves();
  
    size_t depth=0;
    if(getPlayer(*moves.begin())==BLACK){
      typedef RecursiveDoUndoMove<BLACK,CompositEvalHashEffectState> rec_t;
      rec_t func(eState, &*moves.begin(), &*moves.end(), &depth);
      ApplyMove<BLACK>::doUndoMove(eState, *moves.begin(), func);
    }
    else{
      typedef RecursiveDoUndoMove<WHITE,CompositEvalHashEffectState> rec_t;
      rec_t func(eState, &*moves.begin(), &*moves.end(), &depth);
      ApplyMove<WHITE>::doUndoMove(eState, *moves.begin(), func);
    }
    CPPUNIT_ASSERT_EQUAL(moves.size(), depth);
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
