#include "osl/applyMove.h"
#include "osl/record/record.h"
#include "osl/record/csa.h"
#include "osl/record/csaString.h"
#if 0
#  include "osl/numEffectState.h"
typedef osl::NumEffectState state_t;
#else
#    include "osl/pawnMaskState.h"
// typedef osl::SimpleState state_t;
typedef osl::PawnMaskState state_t;
#endif

#include "osl/pieceVector.h"
#include "osl/effect_action/storePiece.h"
#include "osl/doUndoMoveLock.h"
#include "doUndoMoveLockTest.h"

#include <iomanip>
#include <fstream>
#include <vector>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include "recursiveDoUndoMove.h"

class effectTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( effectTest );
  CPPUNIT_TEST( testMove );
  CPPUNIT_TEST( testPieceHasEffectTo );
  /*
   * foreachEffect 
   */
  CPPUNIT_TEST( testForeachEffect );
  CPPUNIT_TEST( testForeachEffectPtype );
  CPPUNIT_TEST( testForeachEffectNotBy );
  CPPUNIT_TEST( testForeachEffectOfPiece );
  CPPUNIT_TEST( testForSortedEffect );
  /*
   * hasEffect
   */
  CPPUNIT_TEST( testHasEffectBy );
  CPPUNIT_TEST( testHasEffectNotBy );
  CPPUNIT_TEST( testHasMultipleEffectBy );
  CPPUNIT_TEST( testHasEffectPiece );
  CPPUNIT_TEST( testHasEffectDir );
  
  CPPUNIT_TEST( testKingCanBeCaptured );

  CPPUNIT_TEST( testCsaFiles );
  CPPUNIT_TEST_SUITE_END();
 private:
  osl::PawnMaskState state;
  state_t eState;
 public:
  effectTest();
  void testMove();
  void testPieceHasEffectTo();

  void testForeachEffect();
  void testForeachEffectPtype();
  void testForeachEffectNotBy();
  void testForeachEffectOfPiece();
  void testForSortedEffect();

  void testHasEffectBy();
  void testHasEffectNotBy();
  void testHasMultipleEffectBy();
  void testHasEffectPiece();
  void testHasEffectDir();
  void testKingCanBeCaptured();
  void testCsaFiles();
};


CPPUNIT_TEST_SUITE_REGISTRATION(effectTest);

using namespace osl;
using namespace osl::effect_action;

effectTest::effectTest() :
  state(CsaString(
"P1-KY-KE *  *  *  *  *  *  * \n"
"P2 *  *  *  *  *  * -KI-OU * \n"
"P3 *  * -FU *  * -KI *  *  * \n"
"P4-FU-HI *  *  *  *  * +GI-KY\n"
"P5 * -FU+FU * -FU-FU-FU-FU+FU\n"
"P6+FU *  *  *  *  * +KE *  * \n"
"P7 * +FU * +FU * +FU+FU+FU-KY\n"
"P8+KY *  *  *  * +KI+KI *  * \n"
"P9 * +KE-UM * +KA *  * +KE+OU\n"
"P-00FU\n"
"P+00FU\n"
"P-00FU\n"
"P-00GI\n"
"P-00GI\n"
"P-00GI\n"
"P-00HI\n"
"+\n"
).getInitialState()),eState(state)
{
#if 0
  state=CsaString(
"P1-KY-KE *  *  *  *  *  *  * \n"
"P2 *  *  *  *  *  * -KI-OU * \n"
"P3 *  * -FU *  * -KI *  *  * \n"
"P4-FU-HI *  *  *  *  * +GI-KY\n"
"P5 * -FU+FU * -FU-FU-FU-FU+FU\n"
"P6+FU *  *  *  *  * +KE *  * \n"
"P7 * +FU * +FU * +FU+FU+FU-KY\n"
"P8+KY *  *  *  * +KI+KI *  * \n"
"P9 * +KE-UM * +KA *  * +KE+OU\n"
"P-00FU\n"
"P+00FU\n"
"P-00FU\n"
"P-00GI\n"
"P-00GI\n"
"P-00GI\n"
"P-00HI\n"
"+\n"
).getInitialState();
  eState=state_t(state);
#endif
}



void effectTest::testMove(){
}


void effectTest::testPieceHasEffectTo(){
  {
    // (7,9)
    Piece p=eState.getPieceAt(newPosition(7,9));
    CPPUNIT_ASSERT(eState.hasEffectTo(p,newPosition(3,5)));
    CPPUNIT_ASSERT(eState.hasEffectTo(p,newPosition(8,9)));
    CPPUNIT_ASSERT(eState.hasEffectTo(p,newPosition(9,7)));
    CPPUNIT_ASSERT(!eState.hasEffectTo(p,newPosition(7,9)));
    CPPUNIT_ASSERT(!eState.hasEffectTo(p,newPosition(2,4)));
    CPPUNIT_ASSERT(!eState.hasEffectTo(p,newPosition(9,9)));
  }
  {
    Piece p=eState.getPieceAt(newPosition(2,2));
    CPPUNIT_ASSERT(eState.hasEffectTo(p,newPosition(1,3)));
    CPPUNIT_ASSERT(eState.hasEffectTo(p,newPosition(1,1)));
    CPPUNIT_ASSERT(!eState.hasEffectTo(p,newPosition(2,2)));
    CPPUNIT_ASSERT(!eState.hasEffectTo(p,newPosition(4,4)));
  }
}

void effectTest::testForeachEffect(){
  PieceVector myPieces;
  StorePiece myPiecesStore(&myPieces);
  {
    Position to=newPosition(2,8);
    eState.template forEachEffect<BLACK,StorePiece>
      (to, myPiecesStore);
    CPPUNIT_ASSERT(myPieces.isMember(eState.getPieceAt(newPosition(1,9))));
    CPPUNIT_ASSERT(myPieces.isMember(eState.getPieceAt(newPosition(3,8))));
    CPPUNIT_ASSERT(!myPieces.isMember(eState.getPieceAt(newPosition(2,8))));
  }
}

void effectTest::testForeachEffectPtype(){
  PieceVector myPieces;
  StorePiece myPiecesStore(&myPieces);
  {
    Position to=newPosition(2,8);
    eState.template forEachEffectPtype<BLACK,GOLD,StorePiece>
      (to, myPiecesStore);
    CPPUNIT_ASSERT(!myPieces.isMember(eState.getPieceAt(newPosition(1,9))));
    CPPUNIT_ASSERT(myPieces.isMember(eState.getPieceAt(newPosition(3,8))));
    CPPUNIT_ASSERT(!myPieces.isMember(eState.getPieceAt(newPosition(2,8))));
  }
}

void effectTest::testForeachEffectNotBy(){
  PieceVector myPieces;
  StorePiece myPiecesStore(&myPieces);
  {
    Position to=newPosition(2,8);
    Piece p=eState.getPieceAt(newPosition(1,9));
    eState.template forEachEffectNotBy<BLACK,StorePiece>
      (to, p, myPiecesStore);
    CPPUNIT_ASSERT(!myPieces.isMember(eState.getPieceAt(newPosition(1,9))));
    CPPUNIT_ASSERT(myPieces.isMember(eState.getPieceAt(newPosition(3,8))));
    CPPUNIT_ASSERT(!myPieces.isMember(eState.getPieceAt(newPosition(2,8))));
  }
}

void effectTest::testForeachEffectOfPiece(){
  PawnMaskState state1(CsaString(
"P1+NY * +TO *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  * \n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  * -FU *  * \n"
"P5 *  * -KE * +FU * +KY+FU * \n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  * \n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00FU\n"
"P+00FU\n"
"P+00FU\n"
"P+00FU\n"
"P+00FU\n"
"P+00FU\n"
"P+00FU\n"
"P+00KE\n"
"P+00KI\n"
"P-00KI\n"
"P-00KY\n"
"P+00HI\n"
"+\n"
).getInitialState());
  state_t neState(state);
  PieceVector myPieces;
  StorePiece myPiecesStore(&myPieces);
  neState.template forEachEffectOfPiece<StorePiece>(neState.getPieceAt(newPosition(3,5)),myPiecesStore);
  CPPUNIT_ASSERT(!myPieces.isMember(eState.getPieceAt(newPosition(1,3))));
}

void effectTest::testForSortedEffect(){
}

void effectTest::testHasEffectBy(){
  // 7-3 ˤϸ롥
  CPPUNIT_ASSERT( (eState.hasEffectBy(WHITE,newPosition(7,3))));
  // 7-1 ˤϸʤ
  CPPUNIT_ASSERT( !(eState.hasEffectBy(WHITE,newPosition(7,1))));
  // 9-5 ˤ롥
  CPPUNIT_ASSERT( (eState.hasEffectBy(BLACK,newPosition(9,5))));
  // 9-8 ˤϸʤ
  CPPUNIT_ASSERT( !(eState.hasEffectBy(BLACK,newPosition(9,8))));
}

void effectTest::testHasEffectNotBy(){
  // 9-3 ϸι(9-1),(8-1)롥
  CPPUNIT_ASSERT( (eState.hasEffectNotBy(WHITE,eState.getPieceAt(newPosition(9,1)),newPosition(9,3))));
  // 9-2 ϸι(9-1)Τ
  CPPUNIT_ASSERT( !(eState.hasEffectNotBy(WHITE,eState.getPieceAt(newPosition(9,1)),newPosition(9,2))));
}

void effectTest::testHasMultipleEffectBy(){
  // 9-3 ˤϸʣ롥
  CPPUNIT_ASSERT( (eState.hasMultipleEffectBy(WHITE,newPosition(9,3))));
  // 5-7 ˤϸʣʤ
  CPPUNIT_ASSERT( !(eState.hasMultipleEffectBy(WHITE,newPosition(5,7))));
  // 2-8 ˤʣ롥
  CPPUNIT_ASSERT( (eState.hasMultipleEffectBy(BLACK,newPosition(2,8))));
  // 2-3 ˤʣʤ
  CPPUNIT_ASSERT( !(eState.hasMultipleEffectBy(BLACK,newPosition(2,3))));
}

void effectTest::testHasEffectPiece(){
  {
    Piece p=eState.hasEffectPiece(WHITE,newPosition(5,7));
    CPPUNIT_ASSERT(p==eState.getPieceAt(newPosition(7,9)));
  }
  {
    // 2,3ؤϤɤǤɤ
    Piece p=eState.hasEffectPiece(WHITE,newPosition(2,3));
    CPPUNIT_ASSERT(p==eState.getPieceAt(newPosition(2,2)) ||
		   p==eState.getPieceAt(newPosition(3,2)));
  }
  {
    //ʤϸƤ󤸤ᡩ
    //    Piece p=eState.hasEffectPiece(WHITE,newPosition(5,9));
    //    CPPUNIT_ASSERT(p==PIECE_EMPTY);
  }
}

void effectTest::testHasEffectDir(){
  CPPUNIT_ASSERT( (eState.template hasEffectDir<LONG_DR,WHITE>(newPosition(9,7))));
  CPPUNIT_ASSERT( (eState.template hasEffectDir<LONG_UL,BLACK>(newPosition(7,7))));
  CPPUNIT_ASSERT( (!eState.template hasEffectDir<LONG_UL,BLACK>(newPosition(4,8))));
}




static void testFile(const std::string& fileName)
{
  Record rec=CsaFile(fileName).getRecord();
  PawnMaskState state=rec.getInitialState();
  state_t neState(state);
  const std::vector<osl::Move> moves=rec.getMoves();

  DoUndoMoveLockTest<state_t> lockTest(neState, moves);
  lockTest.run();

  size_t depth=0;
  if(getPlayer(*moves.begin())==BLACK){
    RecursiveDoUndoMove<BLACK,state_t>
      func(neState, &*moves.begin(), &*moves.end(), &depth);
    ApplyMove<BLACK>::doUndoMove(neState, *moves.begin(), func);
  }
  else{
    RecursiveDoUndoMove<WHITE,state_t>
      func(neState, &*moves.begin(), &*moves.end(), &depth);
    ApplyMove<WHITE>::doUndoMove(neState, *moves.begin(), func);
  }
  CPPUNIT_ASSERT_EQUAL(moves.size(), depth);
}

void effectTest::testCsaFiles(){
  extern bool isShortTest;
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  int count=100;
  if (isShortTest)
      count=10;
  std::string fileName;
  while((ifs >> fileName) && (++i<count)) {
    if(fileName == "") 
	break;
    if (! isShortTest)
      std::cerr << fileName << " ";
    testFile("../data/kifdat/" + fileName);
  }
}

class MyAction
{
  std::vector<Piece> pieces;
public:
  void doAction(state_t& state,Piece p,Position pos){
    pieces.push_back(p);
  }
  std::vector<Piece> getPieces() const{
    return pieces;
  }
};
std::ostream& operator<<(std::ostream& os,MyAction const& a){
  std::vector<Piece> pieces=a.getPieces();
  os << "MyAction: " << std::endl;
  for(size_t i=0;i<pieces.size();i++)
    os << i << " : " << pieces[i] << std::endl;
  return os << std::endl;
}
  
template <Player P>
bool kingCanBeCaptured(state_t& s)
{
  const int kingIndex=ptypeTable.getKingIndex(P);
  const Position kingPosition=getPosition(s.getPieceOf(kingIndex));
  return s.hasEffectBy(PlayerTraits<P>::opponent, kingPosition);
}
bool kingCanBeCaptured(Player P, state_t& s)
{
  if (P==BLACK)
    return kingCanBeCaptured<BLACK>(s);
  else
    return kingCanBeCaptured<WHITE>(s);
}

void effectTest::testKingCanBeCaptured()
{
  PawnMaskState state=CsaString(
    "P1+NY+TO *  *  *  * -OU-KE-KY\n"
    "P2 *  *  *  *  * -GI-KI *  *\n"
    "P3 * +RY *  * +UM * -KI-FU-FU\n"
    "P4 *  * +FU-FU *  *  *  *  *\n"
    "P5 *  * -KE * +FU *  * +FU *\n"
    "P6-KE *  * +FU+GI-FU *  * +FU\n"
    "P7 *  * -UM *  *  *  *  *  *\n"
    "P8 *  *  *  *  *  *  *  *  * \n"
    "P9 * +OU * -GI *  *  *  * -NG\n"
    "P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
    "P-00KI00KY00FU00FU\n"
    "P-00AL\n"
    "+\n").getInitialState();
  state_t neState(state);
  CPPUNIT_ASSERT(! kingCanBeCaptured(BLACK,neState));
  CPPUNIT_ASSERT(! kingCanBeCaptured(WHITE,neState));
  neState.doMove(newMove(newPosition(8,1),newPosition(7,1),PPAWN,PTYPE_EMPTY,false,BLACK));
  CPPUNIT_ASSERT(! kingCanBeCaptured(BLACK,neState));
  CPPUNIT_ASSERT(! kingCanBeCaptured(WHITE,neState));
  neState.doMove(newMove(OFFBOARD,newPosition(8,8),PAWN,PTYPE_EMPTY,false,WHITE));
  CPPUNIT_ASSERT(kingCanBeCaptured(BLACK,neState));
  CPPUNIT_ASSERT(! kingCanBeCaptured(WHITE,neState));
  neState.doMove(newMove(newPosition(7,1),newPosition(8,1),PPAWN,PTYPE_EMPTY,false,BLACK));
  CPPUNIT_ASSERT(kingCanBeCaptured(BLACK,neState));
  CPPUNIT_ASSERT(! kingCanBeCaptured(WHITE,neState));
}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
