/* effectUtil.t.cc
 */
#include "osl/effectUtil.h"
#include "osl/effectUtil.tcc"
#include "osl/numEffectState.h"
#include "osl/record/csaString.h"
#include "osl/fixedCapacityVector.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <iostream>

class EffectUtilTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(EffectUtilTest);
  CPPUNIT_TEST(testMostValuableEffectOfMoveNotKing);
  CPPUNIT_TEST(testForEachEffectOfMove);
  CPPUNIT_TEST(testCheapest);
  CPPUNIT_TEST(testPromotable);
  CPPUNIT_TEST_SUITE_END();
public:
  void testMostValuableEffectOfMoveNotKing();
  void testCheapest();
  void testPromotable();
  void testForEachEffectOfMove();
};

CPPUNIT_TEST_SUITE_REGISTRATION(EffectUtilTest);

using namespace osl;

struct PositionVector : public FixedCapacityVector<Position,81>
{
  
};

std::ostream& operator<<(std::ostream& os, const PositionVector& vec)
{
  for (size_t i=0; i<vec.size(); ++i)
  {
    os << vec[i] << "\n";
  }
  return os;
}

struct StorePosition
{
  PositionVector& out;
  StorePosition(PositionVector& o) : out(o)
  {
  }
  void operator()(Position pos)
  {
    out.push_back(pos);
  }
};

void EffectUtilTest::testForEachEffectOfMove()
{
  const PawnMaskState state(CsaString(
			       "P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
			       "P2 * -HI *  *  *  *  * -KA * \n"
			       "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
			       "P4 *  *  *  *  *  *  *  *  * \n"
			       "P5 *  *  *  *  *  *  *  *  * \n"
			       "P6 *  *  *  *  *  *  *  *  * \n"
			       "P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
			       "P8 * +KA *  *  *  *  * +HI * \n"
			       "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
			       "+\n").getInitialState());
  PositionVector result;
  StorePosition store(result);
  const Move m2726 = newMove(newPosition(2,7),newPosition(2,6),
			     PAWN,PTYPE_EMPTY,false,BLACK);
  EffectUtil::forEachEffectPositionOfMove(state, m2726, store);
  CPPUNIT_ASSERT_EQUAL(1u, result.size());
  CPPUNIT_ASSERT_EQUAL(newPosition(2,5), result[0]);
  
  // validMove ǤʤƤưΤǼȴ
  const Move m3334 = newMove(newPosition(3,3),newPosition(3,4),
			     PAWN,PTYPE_EMPTY,false,WHITE);
  result.clear();
  EffectUtil::forEachEffectPositionOfMove(state, m3334, store);
  CPPUNIT_ASSERT_EQUAL(1u, result.size());
  CPPUNIT_ASSERT_EQUAL(newPosition(3,5), result[0]);

  const Move m55gin = newDrop(newPosition(5,5), SILVER, BLACK);
  result.clear();
  EffectUtil::forEachEffectPositionOfMove(state, m55gin, store);
  CPPUNIT_ASSERT_EQUAL(5u, result.size());

  const Move m15ryu = newMove(newPosition(1,4), newPosition(1,5), 
			      PROOK, PTYPE_EMPTY, false, WHITE);
  result.clear();
  EffectUtil::forEachEffectPositionOfMove(state, m15ryu, store);
  CPPUNIT_ASSERT_EQUAL(14u, result.size());

  const Move m37ke = newMove(newPosition(2,9), newPosition(3,7), 
			     KNIGHT, PTYPE_EMPTY, false, BLACK);
  result.clear();
  EffectUtil::forEachEffectPositionOfMove(state, m37ke, store);
  CPPUNIT_ASSERT_EQUAL(2u, result.size());
  CPPUNIT_ASSERT_EQUAL(newPosition(4,5), result[0]);
  CPPUNIT_ASSERT_EQUAL(newPosition(2,5), result[1]);
}

void EffectUtilTest::testMostValuableEffectOfMoveNotKing()
{
  const char *endGameState =
  "P1+NY * +TO *  *  * -OU-KE-KY\n"
  "P2 *  *  *  *  * -GI-KI *  *\n"
  "P3 * +RY *  * +UM * -KI-FU-FU\n"
  "P4 *  * +FU-FU *  *  *  *  *\n"
  "P5 *  * -KE * +FU *  * +FU *\n"
  "P6-KE *  * +FU+GI-FU *  * +FU\n"
  "P7 *  * -UM *  *  *  *  *  *\n"
  "P8 *  *  *  *  *  *  *  *  * \n"
  "P9 * +OU * -GI *  *  *  * -NG\n"
  "P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU00FU\n"
  "P-00KI00KY00FU\n"
  "P-00AL\n"
  "+\n";
  const PawnMaskState state(CsaString(endGameState).getInitialState());
  const Move m64um = newMove(newPosition(5,3), newPosition(5,4), 
			     PBISHOP, PTYPE_EMPTY, false, BLACK);
  {
    const Piece highestB 
      = EffectUtil::mostValuableEffectOfMoveNotKing(BLACK, state, m64um);
    CPPUNIT_ASSERT_EQUAL(state.getPieceAt(newPosition(5,5)), highestB);
    const Piece highestW
      = EffectUtil::mostValuableEffectOfMoveNotKing(WHITE, state, m64um);
    CPPUNIT_ASSERT_EQUAL(state.getPieceAt(newPosition(3,2)), highestW);
  }   
  const Move m78um = newMove(newPosition(7,7), newPosition(7,8), 
			     PBISHOP, PTYPE_EMPTY, false, BLACK);
  {
    const Piece highestB 
      = EffectUtil::mostValuableEffectOfMoveNotKing(BLACK, state, m78um);
    CPPUNIT_ASSERT_EQUAL(state.getPieceAt(newPosition(5,6)), highestB);
    const Piece highestW
      = EffectUtil::mostValuableEffectOfMoveNotKing(WHITE, state, m78um);
    CPPUNIT_ASSERT_EQUAL(state.getPieceAt(newPosition(6,9)), highestW);
  }   
}


void EffectUtilTest::testCheapest()
{
  const char *endGameState =
  "P1+NY * +TO *  *  * -OU-KE-KY\n"
  "P2 *  *  *  *  * -GI-KI *  *\n"
  "P3 * +RY *  * +UM * -KI-FU-FU\n"
  "P4 *  * +FU-FU *  *  *  *  *\n"
  "P5 *  * -KE * +FU *  * +FU *\n"
  "P6-KE *  * +FU+GI-FU *  * +FU\n"
  "P7 *  * -UM *  *  *  *  *  *\n"
  "P8 *  *  *  *  *  *  *  *  * \n"
  "P9 * +OU * -GI *  *  *  * -NG\n"
  "P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU00FU\n"
  "P-00KI00KY00FU\n"
  "P-00AL\n"
  "+\n";
  const NumEffectState state(CsaString(endGameState).getInitialState());
  CPPUNIT_ASSERT_EQUAL(state.getPieceOnBoard(newPosition(8,3)),
		       EffectUtil::cheapestPiece
		       (BLACK, state, newPosition(9,3)));
  CPPUNIT_ASSERT_EQUAL(state.getPieceOnBoard(newPosition(9,6)),
		       EffectUtil::cheapestPiece
		       (WHITE, state, newPosition(8,8)));
  CPPUNIT_ASSERT_EQUAL(PIECE_EMPTY,
		       EffectUtil::cheapestPiece
		       (BLACK, state, newPosition(2,8)));
  CPPUNIT_ASSERT_EQUAL(state.getPieceOnBoard(newPosition(3,2)),
		       EffectUtil::cheapestPiece
		       (WHITE, state, newPosition(2,2)));
}

void EffectUtilTest::testPromotable()
{
  const char *endGameState =
  "P1+NY * +TO *  *  * -OU-KE-KY\n"
  "P2 *  *  *  *  * -GI-KI *  *\n"
  "P3 * +RY *  * +UM * -KI-FU-FU\n"
  "P4 *  * +FU-FU *  *  *  *  *\n"
  "P5 *  * -KE * +FU *  * +FU *\n"
  "P6-KE *  * +FU+GI-FU *  * +FU\n"
  "P7 *  * -UM *  *  *  *  *  *\n"
  "P8 *  *  *  *  *  *  *  *  * \n"
  "P9 * +OU * -GI *  *  *  * -NG\n"
  "P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU00FU\n"
  "P-00KI00KY00FU\n"
  "P-00AL\n"
  "+\n";
  const NumEffectState state(CsaString(endGameState).getInitialState());
  CPPUNIT_ASSERT_EQUAL(PIECE_EMPTY,
		       EffectUtil::promotablePiece
		       (BLACK, state, newPosition(9,3)));
  CPPUNIT_ASSERT_EQUAL(state.getPieceOnBoard(newPosition(9,6)),
		       EffectUtil::promotablePiece
		       (WHITE, state, newPosition(8,8)));
  CPPUNIT_ASSERT_EQUAL(state.getPieceOnBoard(newPosition(7,4)),
		       EffectUtil::promotablePiece
		       (BLACK, state, newPosition(7,3)));
  CPPUNIT_ASSERT_EQUAL(state.getPieceOnBoard(newPosition(7,5)),
		       EffectUtil::promotablePiece
		       (WHITE, state, newPosition(6,7)));
}


/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
