#include "recursiveDoUndoMove.h"
#include "osl/pawnMaskState.h"
#include "osl/evalHashEffectState.h"
#include "osl/record/record.h"
#include "osl/record/csaString.h"
#include "osl/record/csa.h"
#include "doUndoMoveLockTest.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <fstream>

class evalTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( evalTest );
  CPPUNIT_TEST( testSoma );
  CPPUNIT_TEST( testBug );
  CPPUNIT_TEST( testShow );
  CPPUNIT_TEST( testConsistent );
  CPPUNIT_TEST( testSimpleEval );
  CPPUNIT_TEST( testIsEvalValueConsistent );
  CPPUNIT_TEST( testDoUndoMove );
  CPPUNIT_TEST_SUITE_END();
private:
  osl::PawnMaskState state;
public:
  void setUp();
  void testBug();
  void testShow();
  void testConsistent();
  void testSimpleEval();
  void testSoma();
  void testIsEvalValueConsistent();
  void testDoUndoMove();
};

CPPUNIT_TEST_SUITE_REGISTRATION(evalTest);

using namespace osl;
extern bool isShortTest;

void evalTest::setUp(){
  state=CsaString(
"P1-KY-KE-GI-KI *  * -HI-KE * \n"
"P2 * -HI *  * -KA-OU *  *  * \n"
"P3-FU * -FU-FU-FU * -GI-FU-KY\n"
"P4 * +KE *  *  * -FU *  * -FU\n"
"P5 * -FU+KE+KA *  *  *  * +FU\n"
"P6 *  * +FU *  *  * +FU *  * \n"
"P7+FU+FU * +FU+FU+FU *  *  * \n"
"P8 * +GI+KI *  *  * +GI *  * \n"
"P9+KY *  *  * +OU+KI *  * +KY\n"
"P+00FU\n"
"P+00FU\n"
"P+00KI\n"
"+\n").getInitialState();
}

void evalTest::testShow(){
  for(int i=PTYPEO_MIN;i<=PTYPEO_MAX;i++){
    PtypeO ptypeO=static_cast<PtypeO>(i);
    if (! isShortTest)
      std::cout << ptypeO << ",val=" << ptypeEvalTable.getVal(ptypeO);
    if(isPiece(ptypeO))
    {
      CPPUNIT_ASSERT(ptypeEvalTable.getVal(ptypeO)
		     *playerToMul(getOwner(ptypeO))>0);
      if (! isShortTest)
	std::cout << ", captureVal=" <<  ptypeEvalTable.getCaptureVal(ptypeO) << std::endl;
      CPPUNIT_ASSERT(ptypeEvalTable.getCaptureVal(ptypeO)
		     *playerToMul(getOwner(ptypeO))<0);
      if(isPromoted(ptypeO))
      {
      
	if (! isShortTest)
	  std::cout << ", promoteVal=" <<  ptypeEvalTable.getPromoteVal(ptypeO);
	CPPUNIT_ASSERT(ptypeEvalTable.getPromoteVal(ptypeO)
		       *playerToMul(getOwner(ptypeO))>0);
      }
    }
    if (! isShortTest)
      std::cout << std::endl;
  }
}

static void testFileConsistent(std::string const& fileName){
  //  std::cout << "testFileConsistent(" << fileName << ")" << std::endl;
  Record rec=CsaFile(fileName).getRecord();
  EvalHashEffectState eState(rec.getInitialState());
  std::vector<osl::Move> moves=rec.getMoves();

  DoUndoMoveLockTest<EvalHashEffectState> lockTest(eState, moves);
  lockTest.run();
}

void evalTest::testConsistent(){
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  std::string fileName;
  for(int i=0;i<100 && (ifs >> fileName) ; i++){
    if(fileName == "") 
      break;
    testFileConsistent("../data/kifdat/" + fileName);
  }
}


void evalTest::testSimpleEval(){
  EvalHashEffectState eState(state);
  int ret=eState.getVal();
  CPPUNIT_ASSERT(-500==ret ||(std::cout << "ret\t" << ret << std::endl,0));
}


void evalTest::testSoma(){
  EvalHashEffectState eState(state);
  {
    int ret=eState.computeValWithMove(newMove(newPosition(1,5),newPosition(1,4),PAWN,PAWN,false,BLACK) );
    /** ꤬ؤʤΤ */
    CPPUNIT_ASSERT_EQUAL(-300, ret);
  }
  {
    int ret=eState.computeValWithMove(newMove(newPosition(6,5),newPosition(9,2),PBISHOP,PTYPE_EMPTY,true,BLACK) );
    /** ּ֤ only */
    CPPUNIT_ASSERT_EQUAL(-2100, ret);
  }
  {
    int ret=eState.computeValWithMove(newMove(newPosition(6,5),newPosition(8,3),PBISHOP,PTYPE_EMPTY,true,BLACK) );
    /**  */
    CPPUNIT_ASSERT_EQUAL(-150, ret);
  }
  {
    int ret=eState.computeValWithMove(newMove(newPosition(6,5),newPosition(2,1),PBISHOP,KNIGHT,true,BLACK) );
    CPPUNIT_ASSERT_EQUAL(-1300, ret);
  }
  {
    int ret=eState.computeValWithMove(newMove(newPosition(8,4),newPosition(7,2),PKNIGHT,PTYPE_EMPTY,true,BLACK) );
    CPPUNIT_ASSERT_EQUAL(-1300, ret);
  }
  {
    int ret1=eState.computeValWithMove(newMove(newPosition(6,5),newPosition(5,4),BISHOP,PTYPE_EMPTY,false,BLACK) );
    CPPUNIT_ASSERT_EQUAL(-2100, ret1);
  }
}
void evalTest::testBug(){
  PawnMaskState bugState=CsaString(
"P1-KY-KE * -KI *  *  * -KE-KY\n"
"P2 * -OU-GI * -KI * -HI *  * \n"
"P3 * -FU *  * -FU-GI-KA * -FU\n"
"P4 *  * -FU-FU * -FU-FU-FU * \n"
"P5-FU *  *  *  *  *  *  *  * \n"
"P6 *  * +FU * +FU+GI+FU * +FU\n"
"P7 * +FU * +FU * +FU *  *  * \n"
"P8 * +KA+OU * +KI+GI * +HI * \n"
"P9+KY+KE * +KI *  *  * +KE+KY\n"
"P-00FU00FU\n"
"+\n").getInitialState();
  EvalHashEffectState eState(bugState);
  int ret=eState.template computeValWithMove<BLACK>( newMove(newPosition(5,6),newPosition(5,5),PAWN,PTYPE_EMPTY,false,BLACK) );
  CPPUNIT_ASSERT(-400==ret ||(std::cout << "ret\t" << ret << std::endl,0));
}


static 
void testIsEvalValueConsistent(std::string const& fileName){
  Record rec=CsaFile(fileName).getRecord();
  EvalHashEffectState eState(rec.getInitialState());
  std::vector<osl::Move> moves=rec.getMoves();
  std::vector<DoUndoMoveLock*> locks(moves.size(), 0);
  for(unsigned int i=0;i<moves.size();i++)
  {
    const Move move=moves[i];
    const int valWithMove = eState.computeValWithMove(move);
    CPPUNIT_ASSERT(EvalHashEffectState::eval_t::isEvalValueConsistent(valWithMove));
    locks[i] = new DoUndoMoveLock(eState, move); // doMove
    CPPUNIT_ASSERT(eState.isConsistent(true));
    CPPUNIT_ASSERT(EvalHashEffectState::eval_t::isEvalValueConsistent
		   (eState.getVal()));
    const int valAfterMove 
      = eState.computeValWithTakeBack(getPlayer(move),getTo(move));
    CPPUNIT_ASSERT(EvalHashEffectState::eval_t::isEvalValueConsistent
		   (valAfterMove));
  }
  for(int i=(int)(moves.size()-1);i>=0;i--)
  {
    delete locks[i];		// undoMove
    CPPUNIT_ASSERT(eState.isConsistent());
    CPPUNIT_ASSERT(EvalHashEffectState::eval_t::isEvalValueConsistent(eState.getVal()));
  }
}

void evalTest::testIsEvalValueConsistent()
{
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  std::string fileName;
  for(int i=0;i<100 && (ifs >> fileName); i++){
    if(fileName == "") 
      break;
    ::testIsEvalValueConsistent("../data/kifdat/" + fileName);
  }
}

void evalTest::testDoUndoMove(){
  extern bool isShortTest;
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  int count=100;
  if (isShortTest)
    count=10;
  std::string fileName;
  while((ifs >> fileName) && (++i<count)) {
    if(fileName == "") 
      break;
    if (! isShortTest)
      std::cerr << fileName << " ";

    CsaFile file("../data/kifdat/"+fileName);
    Record rec=file.getRecord();
    EvalHashEffectState eState(rec.getInitialState());

    std::vector<osl::Move> moves=rec.getMoves();
  
    size_t depth=0;
    typedef EvalHashEffectState state_t;
    if(getPlayer(*moves.begin())==BLACK){
      RecursiveDoUndoMove<BLACK,state_t>
	func(eState, &*moves.begin(), &*moves.end(), &depth);
      ApplyMove<BLACK>::doUndoMove(eState, *moves.begin(), func);
    }
    else{
      RecursiveDoUndoMove<WHITE,state_t>
	func(eState, &*moves.begin(), &*moves.end(), &depth);
      ApplyMove<WHITE>::doUndoMove(eState, *moves.begin(), func);
    }
    CPPUNIT_ASSERT_EQUAL(moves.size(), depth);
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
