#include "hashTest.h"
#include "recursiveDoUndoMove.h"
#include "osl/simpleHashState.h"
#include "osl/record/record.h"
#include "osl/record/csa.h"
#include "osl/record/csaString.h"
#include "osl/doUndoMoveLock.h"
#include "osl/doUndoMoveLock.tcc"
#include "doUndoMoveLockTest.h"
#include <boost/scoped_array.hpp>
#include <fstream>
#include <iostream>
#include <map>
#if __GNUG__ >= 3
#  include <ext/hash_map>
#else
#  include <hash_map>
#endif

CPPUNIT_TEST_SUITE_REGISTRATION(hashTest);
extern bool isShortTest;
using namespace osl;

void hashTest::testNotEqual(){
  SimpleHashState hState(CsaString(
"P1-KY *  * -KI-HI *  * -KE-KY\n"
"P2 * -OU-GI *  *  * -KI *  * \n"
"P3-FU-FU *  *  *  *  * -FU-FU\n"
"P4 *  * -FU * +FU * -FU *  * \n"
"P5 *  *  * -FU+HI+FU * +FU * \n"
"P6 *  * +FU-GI *  *  *  *  * \n"
"P7+FU+FU+KI *  *  * +FU * +FU\n"
"P8 * +OU+KI *  *  *  *  *  * \n"
"P9+KY+KE *  *  *  *  * +KE+KY\n"
"P+00FU\n"
"P+00FU\n"
"P+00FU\n"
"P+00KE\n"
"P+00GI\n"
"P+00GI\n"
"P-00KA\n"
"P+00KA\n"
"+\n").getInitialState());
  // Ѥλˤ
  SimpleHashState hState2(CsaString(
"P1-KY *  * -KI-HI *  * -KE-KY\n"
"P2 * -OU-GI *  *  * -KI *  * \n"
"P3-FU-FU *  *  *  *  * -FU-FU\n"
"P4 *  * -FU * +FU * -FU *  * \n"
"P5 *  *  * -FU+HI+FU * +FU * \n"
"P6 *  * +FU-GI *  *  *  *  * \n"
"P7+FU+FU+KI *  *  * +FU * +FU\n"
"P8 * +OU+KI *  *  *  *  *  * \n"
"P9+KY+KE *  *  *  *  * +KE+KY\n"
"P+00FU\n"
"P+00FU\n"
"P+00FU\n"
"P+00KE\n"
"P+00GI\n"
"P+00GI\n"
"P-00KA\n"
"P-00KA\n"
"+\n").getInitialState());
  CPPUNIT_ASSERT(hState.getHash() != hState2.getHash());
}


static void testDiffFile(const std::string& fileName){
  Record rec=CsaFile(fileName).getRecord();
  SimpleHashState hState(rec.getInitialState());
  std::vector<osl::Move> moves=rec.getMoves();

  DoUndoMoveLockTest<SimpleHashState> lockTest(hState, moves);
  lockTest.run();

  std::vector<DoUndoMoveLock*> locks(moves.size(), 0);
  for(unsigned int i=0;i<moves.size();i++)
  {
    const Move move=moves[i];

    // isConsistent ǽʬʵ뤬 do/undo Ƥhash ѤʤȤ򶯤ǧƤߤ
    const HashKey before = hState.getHash();
    {
      DoUndoMoveLock lock(hState, move);
    }
    CPPUNIT_ASSERT_EQUAL(before, hState.getHash());
    
    locks[i] = new DoUndoMoveLock(hState, move); // doMove
    CPPUNIT_ASSERT( hState.isConsistent() );
    CPPUNIT_ASSERT_EQUAL(hState.getHash(), before.newHashWithMove(move));
  }
  for(int i=(int)(moves.size()-1);i>=0;i--)
  {
    delete locks[i]; // undoMove
    CPPUNIT_ASSERT( hState.isConsistent() );
  }
}


void hashTest::testDiff(){
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  std::string fileName;
  while((ifs >> fileName) && ++i<100){
    if(fileName == "") break;
    testDiffFile("../data/kifdat/" + fileName);
  }
}

/** ʬζ̤򤤤Ƥhash ͤʤΤƥ */
static void testHashUniqueness(const std::string& fileName){
  typedef std::map<HashKey, SimpleHashState*> map_t;
  map_t stateMap;
  
  if (!isShortTest)
    std::cerr << fileName << "\n";
  
  Record rec=CsaFile(fileName).getRecord();
  SimpleHashState hState(rec.getInitialState());
  std::vector<osl::Move> moves=rec.getMoves();
  for(unsigned int i=0;i<moves.size();i++){
    Move move=moves[i];
    hState.doMove(move);
    CPPUNIT_ASSERT(hState.isConsistent());

    const HashKey& key = hState.getHash();
    map_t::const_iterator pos = stateMap.find(key);
    if ((pos != stateMap.end()) && (*pos->second != hState))
    {
      std::cerr << "collsion detected " << key << "\n"
		<< hState
		<< pos->second;
    }
    else
    {
	// memory leak ɡ100MBԤʤȤꤢ
	stateMap[key] = new SimpleHashState(hState);
    }
  }
}

void hashTest::testHashUniqueness(){
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  std::string fileName;
  while((ifs >> fileName) && ++i<100){
    if(fileName == "") 
      break;
    ::testHashUniqueness("../data/kifdat/" + fileName);
  }
}

void hashTest::testDoUndoMove(){
  extern bool isShortTest;
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  int count=100;
  if (isShortTest)
    count=10;
  std::string fileName;
  while((ifs >> fileName) && (++i<count)) {
    if(fileName == "") 
      break;
    if (! isShortTest)
      std::cerr << fileName << " ";

    CsaFile file("../data/kifdat/"+fileName);
    Record rec=file.getRecord();
    PawnMaskState state=rec.getInitialState();
    SimpleHashState hState(state);

    std::vector<osl::Move> moves=rec.getMoves();
  
    size_t depth=0;
    typedef SimpleHashState state_t;
    if(getPlayer(*moves.begin())==BLACK){
      RecursiveDoUndoMove<BLACK,state_t>
	func(hState, &*moves.begin(), &*moves.end(), &depth);
      ApplyMove<BLACK>::doUndoMove(hState, *moves.begin(), func);
    }
    else{
      RecursiveDoUndoMove<WHITE,state_t>
	func(hState, &*moves.begin(), &*moves.end(), &depth);
      ApplyMove<WHITE>::doUndoMove(hState, *moves.begin(), func);
    }
    CPPUNIT_ASSERT_EQUAL(moves.size(), depth);
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
