#include "osl/move_generator/addEffect.h"
#include "osl/move_generator/allMoves.h"
#include "osl/move_action/store.h"
#include "osl/record/record.h"
#include "osl/record/csa.h"
#include "osl/record/csaString.h"
#include "osl/perfmon.h"
#include "osl/moveVector.h"
#include "osl/pawnMaskState.h"
#include "osl/doUndoMoveLock.h"
#if 1
#  include "osl/numEffectState.h"
typedef osl::NumEffectState test_state_t;
#else
// ƥȥե̤ˤޤ
#  include "osl/generateAddEffectMoves.tcc"
#  include "osl/generateEscapeMoves.tcc"
#  include "osl/generateCaptureMoves.tcc"
#  include "osl/generateAllMoves.tcc"
typedef osl::BoardEffectState test_state_t;
#endif


#include <algorithm>
#include <fstream>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

class AddEffectTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( AddEffectTest );
  CPPUNIT_TEST( testMove );
  CPPUNIT_TEST( testDrop );
  CPPUNIT_TEST( testOpen );
  CPPUNIT_TEST( testMoveMember );
  CPPUNIT_TEST_SUITE_END();
private:
  osl::PawnMaskState state;
public:
  AddEffectTest();
  void testMove();
  void testDrop();
  void testOpen();
  void testMoveMember();
};

CPPUNIT_TEST_SUITE_REGISTRATION(AddEffectTest);

using namespace osl;
using namespace osl::move_action;
using namespace osl::move_generator;

AddEffectTest::AddEffectTest(){
  state=CsaString(
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6+KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU\n"
"P-00AL\n"
"+\n"
).getInitialState();
}

template<typename State>
static void testDrop(PawnMaskState& state){
  State eState(state);
  MoveVector moves;
  move_action::Store store(moves);
  AddEffect<State,Store>::
    generateMoves(BLACK,eState,newPosition(3,1),store);
  moves.unique();
  
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(4,3),KNIGHT,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(4,1),GOLD,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(2,2),GOLD,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(4,1),ROOK,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(7,1),ROOK,BLACK)));
  CPPUNIT_ASSERT(!moves.isMember(newDrop(newPosition(4,4),ROOK,BLACK)));
  for(size_t i=0;i<moves.size();i++)
    CPPUNIT_ASSERT(eState.isValidMove(moves[i]) ||
		   (std::cout << eState << moves[i] << std::endl,0));
  moves.clear();
  AddEffect<State,Store >::
    generateMoves(BLACK,eState,newPosition(5,3),store);
  // 2ʤ
  CPPUNIT_ASSERT(!moves.isMember(newDrop(newPosition(5,4),PAWN,BLACK)));
}

void AddEffectTest::testDrop(){
  ::testDrop<test_state_t>(state); 
}

template<typename State>
static void testMove(PawnMaskState& state){
  {
    PawnMaskState state1=CsaString(
"P1-KY-KE *  *  * -OU * -KE-KY\n"
"P2 *  *  *  *  *  * -KI-KA * \n"
"P3-FU * -FU-FU-HI-GI * +RY * \n"
"P4 * -FU *  *  * -FU+FU * -FU\n"
"P5 *  *  *  * -GI *  *  *  * \n"
"P6 *  * +FU+FU *  *  *  *  * \n"
"P7+FU+FU+GI *  * +FU+GI * +FU\n"
"P8 *  * +KI * +KI *  *  *  * \n"
"P9+KY+KE * +OU *  *  * +KE+KY\n"
"P+00FU\n"
"P-00FU\n"
"P-00FU\n"
"P-00FU\n"
"P-00FU\n"
"P+00KI\n"
"P-00KA\n"
"+\n").getInitialState();
    State eState(state1);
    MoveVector moves;
    Store store(moves);
  // ζ
    AddEffect<State,Store>::
      generateMoves(BLACK,eState,newPosition(3,1),store);
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(2,3),newPosition(2,2),PROOK,BISHOP,false,BLACK)) ||
		   (std::cerr << moves << std::endl,0));
  }
  {
    PawnMaskState state1=CsaString(
"P1-KY-KE * -KI *  * -GI+RY-KY\n"
"P2 *  *  * -OU *  * -KI *  * \n"
"P3-FU * -GI-FU-FU-FU *  * -FU\n"
"P4 *  * -FU *  *  *  *  *  * \n"
"P5 *  *  *  * +KA *  *  *  * \n"
"P6 *  *  *  *  *  *  *  *  * \n"
"P7+FU * +FU+FU+FU+FU+FU * +FU\n"
"P8 * +GI+KI *  *  *  *  *  * \n"
"P9+KY+KE *  * +OU+KI+GI * +KY\n"
"P+00FU\n"
"P+00FU\n"
"P-00FU\n"
"P-00FU\n"
"P-00FU\n"
"P+00KE\n"
"P-00KE\n"
"P-00KA\n"
"P-00HI\n"
"+\n").getInitialState();
    State eState(state1);
    MoveVector moves;
    Store store(moves);
  // ζ
    AddEffect<State,Store>::
      generateMoves(BLACK,eState,newPosition(4,1),store);
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(2,1),newPosition(3,2),PROOK,GOLD,false,BLACK)) ||
		   (std::cerr << moves << std::endl,0));
  }
  State eState(state);
  MoveVector moves;
  Store store(moves);
  // 
  AddEffect<State,Store>::
    generateMoves(BLACK,eState,newPosition(9,2),store);
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(9,6),newPosition(8,4),KNIGHT,PTYPE_EMPTY,false,BLACK)));

  // promoteĤ
  moves.clear();
  AddEffect<State,Store>::
    generateMoves(BLACK,eState,newPosition(6,2),store);
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(7,4),newPosition(7,3),PPAWN,PTYPE_EMPTY,true,BLACK)));

  // äƲ
  moves.clear();
  AddEffect<State,Store>::
    generateMoves(BLACK,eState,newPosition(3,1),store);
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(5,3),newPosition(4,2),PBISHOP,SILVER,false,BLACK)) ||
		 (std::cerr << moves << std::endl,0)
		 );
}

void AddEffectTest::testMove(){
  ::testMove<test_state_t>(state);
}

template<typename State>
static void testOpen(PawnMaskState& state){
  {
    State eState(state);
    MoveVector moves;
    Store store(moves);
    AddEffect<State,Store>::
      generateMoves(BLACK,eState,newPosition(3,3),store);
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(5,3),newPosition(5,2),PBISHOP,PTYPE_EMPTY,false,BLACK))||
		   (std::cerr << moves << std::endl,0));
  }
  {
    PawnMaskState state1=CsaString(
"P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
"P2 * -HI *  *  *  *  * -KA * \n"
"P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
"P4 *  *  *  *  *  *  *  *  * \n"
"P5 *  *  *  *  *  *  *  *  * \n"
"P6 *  * +FU *  *  *  *  *  * \n"
"P7+FU+FU * +FU+FU+FU+FU+FU+FU\n"
"P8 * +KA *  *  *  *  * +HI * \n"
"P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
"-\n"
).getInitialState();
    State eState(state1);
    MoveVector moves;
    Store store(moves);
    AddEffect<State,Store>::
      generateMoves(WHITE,eState,newPosition(5,5),store);
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(3,3),newPosition(3,4),PAWN,PTYPE_EMPTY,false,WHITE))||
		   (std::cerr << moves << std::endl,0));
  }
}

// 
void AddEffectTest::testOpen(){
  ::testOpen<test_state_t>(state);
}

template<typename State> 
static bool isAddEffectMove(State& state,Position pos,Move move){
  DoUndoMoveLock lock(state, move);
  const bool flag=state.hasEffectBy(getPlayer(move),pos);
  return flag;
}

template<typename State>
static void testMoveFile(const std::string& fileName){
  Record rec=CsaFile(fileName).getRecord();
  PawnMaskState state=rec.getInitialState();
  State eState(state);
  std::vector<osl::Move> moves=rec.getMoves();
  for(unsigned int i=0;i<moves.size();i++){
    MoveVector allMoves;
    Store store(allMoves);
    AllMoves<SimpleState,Store>::
      generateMoves(eState.getTurn(),eState,store);
    for(int y=1;y<=9;y++)
      for(int x=9;x>0;x--){
	Position pos=newPosition(x,y);
	MoveVector effectMoves;
	Store storeEffect(effectMoves);
	AddEffect<State,Store>::
	  generateMoves(eState.getTurn(),eState,pos,storeEffect);
	{
	  MoveVector effectMoves1=effectMoves;
	  effectMoves.unique();
	  CPPUNIT_ASSERT( (effectMoves.size()==effectMoves1.size()) ||
			  (std::cout << std::endl << eState << pos << std::endl << effectMoves1 << std::endl,0));
	}
	
	for(size_t j=0;j<effectMoves.size();j++){
	  CPPUNIT_ASSERT( (eState.isValidMove(effectMoves[j],true)  &&
			   isAddEffectMove<State>(eState,pos,effectMoves[j]))||
			  (std::cout << eState << pos << std::endl << effectMoves[j] << std::endl,0)
			  );
	}
	if(eState.hasEffectBy(eState.getTurn(),pos)) continue;
	// 
	size_t count=0;
	MoveVector tmpMoves;
	for(size_t j=0;j<allMoves.size();j++){
	  if(isAddEffectMove<State>(eState,pos,allMoves[j])){
	    tmpMoves.push_back(allMoves[j]);
	    count++;
	    CPPUNIT_ASSERT(effectMoves.isMember(allMoves[j])||
			   (std::cout << eState << pos << std::endl << allMoves[j] << std::endl,0));
	  }
	}
	CPPUNIT_ASSERT(count == effectMoves.size() ||
		       (std::cout << eState << pos << std::endl << count << std::endl << tmpMoves << effectMoves << effectMoves.size() << std::endl,0)
		       );
      }
    Move move=moves[i];
    eState.doMove(move);
  }
}

void AddEffectTest::testMoveMember(){
  extern bool isShortTest;
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  int count=20;
  if (isShortTest) 
    count=2;
  std::string fileName;
  while((ifs >> fileName) && ++i<count){
    if(fileName == "") 
      break;
    if (! isShortTest)
      std::cerr << fileName << " ";
    testMoveFile<test_state_t>("../data/kifdat/" + fileName);
  }
}


// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
