#include "osl/move_generator/addEffectWithoutEffect.h"
#include "osl/move_generator/allMoves.h"
#include "osl/move_action/store.h"
#include "osl/moveVector.h"
#include "osl/pawnMaskState.h"
typedef osl::PawnMaskState test_state_t;
#include "osl/record/csa.h"
#include "osl/record/csaString.h"
#include "osl/moveVector.h"
#include "osl/doUndoMoveLock.h"
#include <iostream>
#include <fstream>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>


class AddEffectWithoutEffectTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( AddEffectWithoutEffectTest );
  CPPUNIT_TEST( testLongOpen );
  CPPUNIT_TEST( testShortMove );
  CPPUNIT_TEST( testShortDrop );
  CPPUNIT_TEST( testAddEffectTable );
  CPPUNIT_TEST( testLongMove );
  CPPUNIT_TEST( testLongDrop );
  CPPUNIT_TEST( testMoveMember );
  CPPUNIT_TEST_SUITE_END();
private:
  osl::PawnMaskState state0,state1,state2,state3,state4;
public:
  AddEffectWithoutEffectTest();
  void testShortMove();
  void testShortDrop();
  void testAddEffectTable();
  void testLongMove();
  void testLongOpen();
  void testLongDrop();
  void testMoveMember();
};

CPPUNIT_TEST_SUITE_REGISTRATION(AddEffectWithoutEffectTest);

using namespace osl;
using namespace osl::move_action;
using namespace osl::move_generator;

AddEffectWithoutEffectTest::AddEffectWithoutEffectTest(){
  state0=CsaString(
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  * +KI-GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6+KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KY00FU00FU\n"
"P-00AL\n"
"+\n"
).getInitialState();
  state1=CsaString(
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  * +KI-GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6+KE *  * +FU * -FU *  * +FU\n"
"P7 *  * -UM *  * -GI *  *  *\n"
"P8-FU *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KY00FU\n"
"P-00AL\n"
"-\n"
).getInitialState();
  state2=CsaString(
"P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
"P2 * -HI *  *  *  *  * -KA * \n"
"P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
"P4 *  *  *  *  *  *  *  *  * \n"
"P5 *  *  *  *  *  *  *  *  * \n"
"P6 *  *  *  *  *  *  *  *  * \n"
"P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
"P8 * +KA *  *  *  *  * +HI * \n"
"P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
"+\n"
).getInitialState();
  state3=CsaString(
"P1-KY-KE-OU+KA *  * +RY * -KY\n"
"P2 *  *  *  * -KI * +NK *  * \n"
"P3-FU * -GI-FU-FU-FU *  * -FU\n"
"P4 *  * -FU *  *  *  *  *  * \n"
"P5 *  *  *  * +KA *  *  *  * \n"
"P6 *  *  *  *  *  *  *  *  * \n"
"P7+FU * +FU+FU+FU+FU+FU * +FU\n"
"P8 *  * -NK * +OU *  *  *  * \n"
"P9+KY+KE * -HI * +KI+GI * +KY\n"
"P+00FU\n"
"P+00FU\n"
"P+00FU\n"
"P-00FU\n"
"P-00FU\n"
"P-00GI\n"
"P-00GI\n"
"P-00KI\n"
"P-00KI\n"
"+\n"
).getInitialState();
  state4=CsaString(
"P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
"P2 * -HI *  *  *  *  * -KA * \n"
"P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
"P4 *  *  *  *  *  *  *  *  * \n"
"P5 *  *  *  *  *  *  *  *  * \n"
"P6 *  * +FU *  *  *  *  *  * \n"
"P7+FU+FU * +FU+FU+FU+FU+FU+FU\n"
"P8 * +KA *  *  *  *  * +HI * \n"
"P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
"-\n"
).getInitialState();

}

void AddEffectWithoutEffectTest::testShortMove(){
  typedef osl::PawnMaskState State;
  MoveVector moves;
  Store store(moves);
  {
    // gold
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(BLACK,state0,newPosition(3,1),store);
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(5,2),newPosition(4,1),GOLD,PTYPE_EMPTY,false,BLACK)));
  }
  {
    // 
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(BLACK,state0,newPosition(9,2),store);
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(9,6),newPosition(8,4),KNIGHT,PTYPE_EMPTY,false,BLACK)));
  }
  moves.clear();
  {
    // promoteĤ
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(BLACK,state0,newPosition(6,2),store);
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(7,4),newPosition(7,3),PPAWN,PTYPE_EMPTY,true,BLACK)));
  }
#if 0
  moves.clear();
  {
    // äƲ
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(BLACK,state0,newPosition(3,1),store);
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(5,3),newPosition(4,2),PBISHOP,SILVER,false,BLACK)) ||
		   (std::cerr << moves << std::endl,0)
		   );
  }
#endif
  /* ֤Υåɬ*/
  moves.clear();
  {
    // 
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(WHITE,state1,newPosition(8,9),store);
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(6,9),newPosition(7,8),SILVER,PTYPE_EMPTY,false,WHITE)));
    CPPUNIT_ASSERT(!moves.isMember(newMove(newPosition(6,9),newPosition(7,9),PSILVER,PTYPE_EMPTY,true,WHITE)));
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(6,9),newPosition(7,8),PSILVER,PTYPE_EMPTY,true,WHITE)));
  }
  moves.clear();
  {
    // promoteǤʤpositionǤ
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(WHITE,state1,newPosition(6,3),store);
    CPPUNIT_ASSERT(!moves.isMember(newMove(newPosition(4,2),newPosition(5,3),PSILVER,PBISHOP,true,WHITE)));
  }
  moves.clear();
  {
    // promoteǤʤpositionǤ
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(WHITE,state1,newPosition(6,6),store);
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(4,7),newPosition(5,6),PSILVER,PTYPE_EMPTY,true,WHITE)));
  }
}

void AddEffectWithoutEffectTest::testShortDrop(){
  MoveVector moves;
  Store store(moves);
  {
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(BLACK,state0,newPosition(3,1),store);
    // 
    CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(4,1),GOLD,BLACK)));
    CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(2,2),GOLD,BLACK)));
    // ʤϤʤ
    CPPUNIT_ASSERT(!moves.isMember(newDrop(newPosition(2,2),SILVER,BLACK)));
    // 
    CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(4,3),KNIGHT,BLACK)));
    // ξˤϤʤ
    CPPUNIT_ASSERT(!moves.isMember(newDrop(newPosition(2,3),KNIGHT,BLACK)));
  }
  moves.clear();
  {
    // 
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(WHITE,state1,newPosition(8,9),store);
    CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,8),PAWN,WHITE)));
  }
}
void AddEffectWithoutEffectTest::testAddEffectTable(){
  {
    // offset 3,2ROOKDLΩäƤϤ
    NearMask nearMask=addEffectTable.getNearMaskLong<BLACK>(ROOK,newPosition(1,1),newPosition(4,3));
    //
    CPPUNIT_ASSERT( nearMask == (((1<<L)|(1<<D))<<16) ||
		    (std::cout << nearMask << std::endl,0) );
  }
  {
    // offset -3,-1NearMaskPROOKRΩäƤϤ
    NearMask nearMaskPROOK=addEffectTable.getNearMaskPROOK<BLACK>(newPosition(4,2),newPosition(1,1));
    CPPUNIT_ASSERT( nearMaskPROOK == (((1<<R))<<16) ||
		    (std::cout << nearMaskPROOK << std::endl,0) );
    NearMask nearMask=addEffectTable.getNearMaskLong<BLACK>(PROOK,newPosition(4,2),newPosition(1,1));
    // offset -3,-1nearMaskUΩäƤϤ
    // ǤOKʤΤ ... <<16
    CPPUNIT_ASSERT( nearMask == (1<<U) ||
		    (std::cout << nearMask << std::endl,0) );
  }
}

void AddEffectWithoutEffectTest::testLongMove(){
  typedef osl::PawnMaskState State;
  MoveVector moves;
  Store store(moves);
  {
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(BLACK,state0,newPosition(5,2),store);
    // ε
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(8,3),newPosition(8,2),PROOK,PTYPE_EMPTY,false,BLACK)));
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(8,3),newPosition(7,2),PROOK,PTYPE_EMPTY,false,BLACK)));
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(8,3),newPosition(9,2),PROOK,PTYPE_EMPTY,false,BLACK)));
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(8,3),newPosition(6,3),PROOK,PTYPE_EMPTY,false,BLACK)));
  }
  moves.clear();
  {
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(WHITE,state1,newPosition(8,9),store);
    /**  */
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(7,7),newPosition(6,7),PBISHOP,PTYPE_EMPTY,false,WHITE)));
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(7,7),newPosition(7,8),PBISHOP,PTYPE_EMPTY,false,WHITE)));
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(7,7),newPosition(8,8),PBISHOP,PTYPE_EMPTY,false,WHITE)));
  }
  moves.clear();
  {
    PawnMaskState state=CsaString(
"P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
"P2 * -HI *  *  *  *  * -KA * \n"
"P3-FU-FU-FU-FU-FU-FU * -FU-FU\n"
"P4 *  *  *  *  *  * -FU *  * \n"
"P5 *  *  *  *  *  *  *  *  * \n"
"P6 *  * +FU *  *  *  *  *  * \n"
"P7+FU+FU * +FU+FU+FU+FU+FU+FU\n"
"P8 * +KA *  *  *  *  * +HI * \n"
"P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
"+\n"
).getInitialState();
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(BLACK,state,newPosition(3,1),store);
    /** Ѥ */
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(8,8),newPosition(2,2),PBISHOP,BISHOP,true,BLACK)) ||
		   (std::cout << moves << std::endl,0));
  }
}
void AddEffectWithoutEffectTest::testLongOpen(){
  typedef osl::PawnMaskState State;
  MoveVector moves;
  Store store(moves);
  {
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(BLACK,state3,newPosition(7,1),store);
    // εopen
    CPPUNIT_ASSERT(moves.size() == 8);
  }
  moves.clear();
  {
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(BLACK,state2,newPosition(7,1),store);
    // ֤
    CPPUNIT_ASSERT(!moves.isMember(newMove(newPosition(2,8),newPosition(7,8),ROOK,PTYPE_EMPTY,false,BLACK)));
  }
  moves.clear();
  {
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(BLACK,state0,newPosition(3,3),store);
    // εblockƤ
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(5,3),newPosition(5,4),PBISHOP,PTYPE_EMPTY,false,BLACK)));
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(5,3),newPosition(4,4),PBISHOP,PTYPE_EMPTY,false,BLACK)));
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(5,3),newPosition(6,2),PBISHOP,PTYPE_EMPTY,false,BLACK)));
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(5,3),newPosition(7,1),PBISHOP,PTYPE_EMPTY,false,BLACK)));
    // negative
    CPPUNIT_ASSERT(!moves.isMember(newMove(newPosition(5,3),newPosition(5,2),PBISHOP,PTYPE_EMPTY,false,BLACK)));
    CPPUNIT_ASSERT(!moves.isMember(newMove(newPosition(5,3),newPosition(6,3),PBISHOP,PTYPE_EMPTY,false,BLACK)));
  }
  moves.clear();
  {
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(WHITE,state4,newPosition(5,5),store);
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(3,3),newPosition(3,4),PAWN,PTYPE_EMPTY,false,WHITE))||
		   (std::cerr << moves << std::endl,0));

  }
}
void AddEffectWithoutEffectTest::testLongDrop(){
  typedef osl::PawnMaskState State;
  MoveVector moves;
  Store store(moves);
  {
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(BLACK,state0,newPosition(3,1),store);
    // 
    CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(4,1),ROOK,BLACK)));
    CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(5,1),ROOK,BLACK)));
    CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(6,1),ROOK,BLACK)));
    CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(7,1),ROOK,BLACK)));
    // ξˤϤʤ
    CPPUNIT_ASSERT(!moves.isMember(newDrop(newPosition(2,1),ROOK,BLACK)));
    // Ƥʤ
    CPPUNIT_ASSERT(!moves.isMember(newDrop(newPosition(2,2),ROOK,BLACK)));
    // ֤Ƥʤʤ
    CPPUNIT_ASSERT(!moves.isMember(newDrop(newPosition(3,4),ROOK,BLACK)));
    // ֤Ƥʤʤ
    CPPUNIT_ASSERT(!moves.isMember(newDrop(newPosition(3,4),LANCE,BLACK)));
    // 𤬤ʤʤ
    CPPUNIT_ASSERT(!moves.isMember(newDrop(newPosition(2,2),BISHOP,BLACK)));
  }
  moves.clear();
  {
    AddEffectWithoutEffect<SimpleState,Store>::
      generateMoves(WHITE,state1,newPosition(8,9),store);
    /**  */
    CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,8),LANCE,WHITE)));
    CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,7),LANCE,WHITE)));
    CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,6),LANCE,WHITE)));
    CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,5),LANCE,WHITE)));
    CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,4),LANCE,WHITE)));
    CPPUNIT_ASSERT(!moves.isMember(newDrop(newPosition(8,3),LANCE,WHITE)));
    CPPUNIT_ASSERT(!moves.isMember(newDrop(newPosition(8,2),LANCE,WHITE)));
  }
}

#if 0
template<typename State>
static bool hasEffectBy(State & state,Player pl,Position pos){
  NullBoardEffect<State> eState(state);
  return eState.hasEffectBy(pl,pos);
}
#endif

template<typename State> 
static bool isAddEffectMove(Player pl,State& state,Position pos,Move move)
{
  DoUndoMoveLock lock(state, move);
  bool flag=state.hasEffectBy(pl,pos);
  return flag;
}

template<typename State>
static void testMoveFile(const std::string& fileName){
  Record rec=CsaFile(fileName).getRecord();
  PawnMaskState state=rec.getInitialState();
  State eState(state);
  std::vector<osl::Move> moves=rec.getMoves();
  for(unsigned int i=0;i<moves.size();i++){
    MoveVector allMoves;
    Store store(allMoves);
    AllMoves<SimpleState,Store>::
      generateMoves(eState.getTurn(),eState,store);
    for(int y=1;y<=9;y++)
      for(int x=9;x>0;x--){
	Position pos=newPosition(x,y);
#if 0
	std::cout << pos << std::endl;
#endif
	/** ˤϸƤФʤ */
	if(eState.hasEffectBy(eState.getTurn(),pos)) continue;
	MoveVector effectMoves;
	Store storeEffect(effectMoves);
	/**
	 * KINGĤΤfalse
	 */
	AddEffectWithoutEffect<SimpleState,Store,false>::
	  generateMoves(eState.getTurn(),eState,pos,storeEffect);
	{
	  MoveVector effectMovesUnique=effectMoves;
	  effectMovesUnique.unique();
	  CPPUNIT_ASSERT( (effectMoves.size()==effectMovesUnique.size()) ||
			  (std::cout << std::endl << eState << pos << std::endl << effectMovesUnique << std::endl,0));
	}
	for(size_t j=0;j<effectMoves.size();j++){
	  CPPUNIT_ASSERT( (eState.isValidMove(effectMoves[j],true)  &&
			   isAddEffectMove<State>(eState.getTurn(),eState,pos,effectMoves[j]))||
			  (std::cout << std::endl << eState << pos << std::endl << effectMoves[j] << std::endl,0)
			  );
	}
	// 
	size_t count=0;
	MoveVector tmpMoves;
	for(size_t j=0;j<allMoves.size();j++){
	  /** ʬưƸΤȤĤʤ */
	  if(isAddEffectMove<State>(eState.getTurn(),eState,pos,allMoves[j])
	     && getFrom(allMoves[j])!=pos){
	    tmpMoves.push_back(allMoves[j]);
	    count++;
	    CPPUNIT_ASSERT(effectMoves.isMember(allMoves[j])||
			   (std::cout << eState << pos << std::endl << allMoves[j] << std::endl,0));
	  }
	}
	CPPUNIT_ASSERT(count == effectMoves.size() ||
		       (std::cout << eState << pos << std::endl << count << std::endl << tmpMoves << effectMoves << effectMoves.size() << std::endl,0)
		       );
      }
    Move move=moves[i];
    eState.doMove(move);
  }
}

void AddEffectWithoutEffectTest::testMoveMember(){
  extern bool isShortTest;
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  int count=20;
  if (isShortTest) 
    count=2;
  std::string fileName;
  while((ifs >> fileName) && ++i<count){
    if(fileName == "") 
      break;
    if (! isShortTest)
      std::cerr << fileName << " ";
    testMoveFile<test_state_t>("../data/kifdat/" + fileName);
  }
}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
