#include "osl/move_generator/allMoves.h"
#include "osl/move_action/store.h"
#include "osl/move_action/voidAction.h"
#include "osl/pawnMaskState.h"
#include "osl/record/record.h"
#include "osl/record/csaString.h"
#include "osl/record/csa.h"
#include "osl/perfmon.h"
#include "osl/move.h"
#include "osl/moveVector.h"
#include <fstream>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

class generateAllMovesTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( generateAllMovesTest );
  CPPUNIT_TEST( testMoveGen );
  CPPUNIT_TEST( testMoveMember );
  CPPUNIT_TEST_SUITE_END();
 private:
  osl::PawnMaskState state;
 public:
  generateAllMovesTest::generateAllMovesTest();
  void testMoveGen();
  void testMoveMember();
};

using namespace osl;
using namespace osl::move_generator;
using namespace osl::move_action;

CPPUNIT_TEST_SUITE_REGISTRATION(generateAllMovesTest);

generateAllMovesTest::generateAllMovesTest(){
  state=CsaString(
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU\n"
"P-00AL\n"
"+\n").getInitialState();
}

template<class State>
static void myStateTest(State& state0){
  typedef AllMoves<typename State::simple_board_t,Store> DoAll;

  MoveVector moves;
  Store action(moves);
  DoAll::generateMoves(BLACK,state0,action);
  //  std::cout << state0 << std::endl;
  for(size_t i=0;i<moves.size();i++){
    CPPUNIT_ASSERT(state0.isValidMove(moves[i]));
  }
  Move tmpMove=moves[0];
  state0.doMove(tmpMove);
  moves.clear();
  DoAll::generateMoves(WHITE,state0,action);
  for(size_t i=0;i<moves.size();i++){
    CPPUNIT_ASSERT(state0.isValidMove(moves[i]));
  }
}

void generateAllMovesTest::testMoveGen(){
  // std::cerr << std::endl << " generateAllMovesTest::testMoveGen" << std::endl;
  PawnMaskState state1(state);
  myStateTest<PawnMaskState>(state1);
}

static void testMoveFile(const std::string& fileName){
  Record rec=CsaFile(fileName).getRecord();
  PawnMaskState state=rec.getInitialState();
  const std::vector<osl::Move> moves=rec.getMoves();
  for(unsigned int i=0;i<moves.size();i++){
    MoveVector curMoves;
    Store action(curMoves);
    AllMoves<SimpleState,Store>::generateMoves(state.getTurn(),state,action);
    curMoves.unique();
#if 1
    // ư̲Ƥޤ
    for(size_t j=0;j<curMoves.size();j++)
      CPPUNIT_ASSERT(state.isValidMove(curMoves[j]) ||
		     (std::cout << state << curMoves[j] << std::endl,0));
#endif
    const Move move=moves[i];
    CPPUNIT_ASSERT(curMoves.isMember(move));
    state.doMove(move);
  }
}

void generateAllMovesTest::testMoveMember(){
  // std::cerr << std::endl << " generateAllMovesTest::testMoveMember" << std::endl;
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  std::string fileName;
  while((ifs >> fileName) && ++i<100){
    if(fileName == "") break;
    testMoveFile("../data/kifdat/" + fileName);
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
