#include "osl/move_generator/allMoves.h"
#include "osl/move_generator/capture.h"
#include "osl/move_action/store.h"
#include "osl/record/record.h"
#include "osl/record/csaString.h"
#include "osl/record/csa.h"
#include "osl/moveVector.h"
#include "osl/numEffectState.h"
typedef osl::NumEffectState test_state_t;

#include <iostream>
#include <fstream>

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

using namespace osl;

class generateCaptureMovesTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( generateCaptureMovesTest );
  CPPUNIT_TEST( testMove );
  CPPUNIT_TEST( testFile );
  CPPUNIT_TEST_SUITE_END();
private:
  PawnMaskState state;
public:
  generateCaptureMovesTest();
  void testMove();
  void testFile();
};

CPPUNIT_TEST_SUITE_REGISTRATION(generateCaptureMovesTest);

using namespace osl;
using namespace osl::move_generator;
using namespace osl::move_action;

generateCaptureMovesTest::generateCaptureMovesTest(){
  state=CsaString(
"P1+NY+TO * +HI *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 *  * +GI * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE+FU+FU *  * +FU *\n"
"P6+KE *  *  *  * -FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU\n"
"P-00AL\n"
"+\n"
).getInitialState();
}

template<typename State>
static void testMove(PawnMaskState& state){
  State eState(state);

  MoveVector moves;
  Store store(moves);
  Capture<State,Store >::
    template generateMoves<BLACK>(eState,newPosition(6,4),store);
  moves.unique();
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(6,1),newPosition(6,4),PROOK,PAWN,true,BLACK)) ||
		 (std::cerr << eState << moves << std::endl,0));
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(6,1),newPosition(6,4),ROOK,PAWN,false,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(7,3),newPosition(6,4),PSILVER,PAWN,true,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(7,3),newPosition(6,4),SILVER,PAWN,false,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(6,5),newPosition(6,4),PAWN,PAWN,false,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(5,3),newPosition(6,4),PBISHOP,PAWN,false,BLACK)));
  CPPUNIT_ASSERT(moves.size()==6);

  for(size_t i=0;i<moves.size();i++)
    CPPUNIT_ASSERT(eState.isValidMove(moves[i]) ||
		   (std::cout << eState << moves[i] << std::endl,0));
  moves.clear();
}

void generateCaptureMovesTest::testMove(){
#if 0
  MoveVector moves;
  Store store(moves);
  test_state_t eState(state);
  Capture<test_state_t,Store >::
    generateMoves<BLACK>(eState,newPosition(6,4),store);
#endif
#if 1
  ::testMove<test_state_t>(state);
#endif
}

template<typename State>
static void testFile(const std::string& fileName){
  Record rec=CsaFile(fileName).getRecord();
  PawnMaskState state=rec.getInitialState();
  State eState(state);
  std::vector<osl::Move> moves=rec.getMoves();
  for(unsigned int i=0;i<moves.size();i++){
    MoveVector allMoves;
    Store store(allMoves);
    AllMoves<SimpleState,Store >::
      generateMoves(eState.getTurn(),eState,store);
    for(int y=1;y<=9;y++)
      for(int x=9;x>0;x--){
	Position pos=newPosition(x,y);
	Piece p=eState.getPieceAt(pos);
	if(!isEmpty(p) && getOwner(p)==alt(eState.getTurn())){
	  MoveVector captureMoves;
	  Store storeCapture(captureMoves);
	  Capture<State,Store >::
	    generateMoves(eState.getTurn(),eState,pos,storeCapture);
	  captureMoves.unique();
	  for(size_t j=0;j<captureMoves.size();j++){
	    CPPUNIT_ASSERT( (eState.isValidMove(captureMoves[j],true)  &&
			     getTo(captureMoves[j])==pos ) ||
			    (std::cerr << eState << captureMoves[j] << std::endl,0)
			    );
	  }
	  for(size_t j=0;j<allMoves.size();j++){
	    if(getTo(allMoves[j])==pos){
	      CPPUNIT_ASSERT(captureMoves.isMember(allMoves[j])||
			     (std::cerr << eState << pos << std::endl << allMoves[j] << std::endl << captureMoves << std::endl,0));
	    }
	  }
	}
      }
    Move move=moves[i];
    eState.doMove(move);
  }
}

void generateCaptureMovesTest::testFile(){
  extern bool isShortTest;
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  int count=100;
  if (isShortTest) 
      count=10;
  std::string fileName;
  while((ifs >> fileName) && ++i<count){
    if(fileName == "") 
      break;
    if (! isShortTest)
      std::cerr << fileName << " ";
    ::testFile<test_state_t>("../data/kifdat/" + fileName);
  }
}
