#include "osl/move_generator/escape.h"
#include "osl/move_generator/allMoves.h"
#include "osl/move_action/store.h"
#include "osl/record/csa.h"
#include "osl/record/csaString.h"
#include "osl/moveVector.h"
#include "osl/pawnMaskState.h"
#include "osl/doUndoMoveLock.h"

typedef osl::PawnMaskState test_state_t;

#include <algorithm>
#include <iostream>
#include <fstream>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

class generateEscapeMovesTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( generateEscapeMovesTest );
  CPPUNIT_TEST( testValid );
  CPPUNIT_TEST( testCapture );
  CPPUNIT_TEST( testEscape );
  CPPUNIT_TEST( testBlocking );
  CPPUNIT_TEST( testMoveMember );
  CPPUNIT_TEST_SUITE_END();
private:
  osl::PawnMaskState state;
public:
  void setUp();
  void testValid();
  void testCapture();
  void testEscape();
  void testBlocking();
  void testMoveMember();
};

CPPUNIT_TEST_SUITE_REGISTRATION(generateEscapeMovesTest);

using namespace osl;
using namespace osl::move_action;
using namespace osl::move_generator;

void generateEscapeMovesTest::setUp(){
  state=CsaString(
"P1+NY *  *  *  *  * -OU * -KY\n"
"P2 * +TO *  *  * -GI-KI *  *\n"
"P3 * -RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5+KE * -KE * +FU *  * +FU *\n"
"P6 *  *  * +FU+GI-FU *  * +FU\n"
"P7+KE * -UM *  *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU\n"
"P-00AL\n"
"+\n"
).getInitialState();
}


template<typename State>
static void testValid(PawnMaskState& state){
  {
    PawnMaskState state1=CsaString(
"P1-KY-KE-GI-KI-OU * -GI-KE-KY\n"
"P2 *  *  *  *  *  * -KI *  * \n"
"P3-FU * -FU-FU-FU-FU *  * -FU\n"
"P4 *  * +HI *  *  *  *  *  * \n"
"P5 *  *  *  *  *  *  *  *  * \n"
"P6 *  *  *  *  *  *  *  *  * \n"
"P7+FU * +FU+FU+FU+FU+FU * +FU\n"
"P8 * +GI+KI *  *  *  *  *  * \n"
"P9+KY+KE *  * +OU+KI+GI+KE+KY\n"
"P+00FU\n"
"P+00FU\n"
"P-00FU\n"
"P-00FU\n"
"P-00FU\n"
"P+00KA\n"
"P-00KA\n"
"P+00HI\n"
"-\n"
).getInitialState();
    test_state_t eState(state1);
    MoveVector moves;
    Store store(moves);
    Escape<WHITE,test_state_t::effect_state_t,Store >::
      generateMoves(eState,eState.getPieceAt(newPosition(7,3)),eState.getPieceAt(newPosition(7,4)),store);
    moves.unique();
    
    CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(7,3),newPosition(7,4),PAWN,ROOK,false,WHITE)));
  }
  test_state_t eState(state);
  MoveVector moves;
  Store store(moves);

  Escape<BLACK,test_state_t::effect_state_t,Store >::
    generateMoves(eState,eState.getPieceAt(newPosition(8,9)),eState.getPieceAt(newPosition(8,3)),store);
  moves.unique();
  //  std::cerr << moves << std::endl;
  for (size_t i=0;i<moves.size();i++)
    CPPUNIT_ASSERT(eState.isValidMove(moves[i]) ||
		   (std::cerr << eState << moves[i] << std::endl,0));
  //  escape ˤʤäƤ뤫ϥåʤ
}

void generateEscapeMovesTest::testValid(){
  ::testValid<test_state_t>(state);
}

template<typename State>
static void testCapture(PawnMaskState& state){

  test_state_t eState(state);
  MoveVector moves;
  Store store(moves);
  Escape<BLACK,test_state_t::effect_state_t,Store >::
    generateMoves(eState,eState.getPieceAt(newPosition(8,9)),
		  eState.getPieceAt(newPosition(8,3)),store);
  moves.unique();
  
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(9,5),newPosition(8,3),PKNIGHT,PROOK,true,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(9,5),newPosition(8,3),KNIGHT,PROOK,false,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(8,2),newPosition(8,3),PPAWN,PROOK,false,BLACK)));
}

void generateEscapeMovesTest::testCapture(){
  ::testCapture<test_state_t>(state);
}

/**
 * ĤƤƨ
 */
template<typename State>
static void testEscape(PawnMaskState& state){
  test_state_t eState(state);
  MoveVector moves;
  Store store(moves);
  Escape<BLACK,test_state_t::effect_state_t,Store >::
    generateMoves(eState,eState.getPieceAt(newPosition(8,9)),
		  eState.getPieceAt(newPosition(8,3)),store);
  moves.unique();
  // std::cerr << moves << std::endl;
  // valid escape
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(8,9),newPosition(9,8),KING,PTYPE_EMPTY,false,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(8,9),newPosition(7,9),KING,PTYPE_EMPTY,false,BLACK)));
  // 
  CPPUNIT_ASSERT(!moves.isMember(newMove(newPosition(8,9),newPosition(9,9),KING,PTYPE_EMPTY,false,BLACK)));
  //
  CPPUNIT_ASSERT(eState.hasEffectBy(WHITE,newPosition(8,8)));
  CPPUNIT_ASSERT(!moves.isMember(newMove(newPosition(8,9),newPosition(8,8),KING,PTYPE_EMPTY,false,BLACK)));
  CPPUNIT_ASSERT(!moves.isMember(newMove(newPosition(8,9),newPosition(7,8),KING,PTYPE_EMPTY,false,BLACK)));
}

void generateEscapeMovesTest::testEscape(){
  ::testEscape<test_state_t>(state);
}


template<typename State>
static void testBlocking(PawnMaskState& state){
  test_state_t eState(state);
  MoveVector moves;
  Store store(moves);
  Escape<BLACK,test_state_t::effect_state_t,Store >::
    generateMoves(eState,eState.getPieceAt(newPosition(8,9)),
		  eState.getPieceAt(newPosition(8,3)),store);
  moves.unique();
  
  CPPUNIT_ASSERT(moves.isMember(newMove(newPosition(9,7),newPosition(8,5),KNIGHT,PTYPE_EMPTY,false,BLACK)));
  //
  CPPUNIT_ASSERT(!moves.isMember(newMove(newPosition(9,8),newPosition(8,8),KING,PTYPE_EMPTY,false,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,4),PAWN,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,5),PAWN,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,6),PAWN,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,7),PAWN,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,8),PAWN,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,4),LANCE,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,5),LANCE,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,6),LANCE,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,7),LANCE,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,8),LANCE,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,4),KNIGHT,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,5),KNIGHT,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,6),KNIGHT,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,7),KNIGHT,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,8),KNIGHT,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,4),GOLD,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,5),GOLD,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,6),GOLD,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,7),GOLD,BLACK)));
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(8,8),GOLD,BLACK)));
  {
  PawnMaskState state1=CsaString(
"P1-KY-KE-GI-KI-OU * -GI-KE-KY\n"
"P2 *  *  *  *  *  * -KI-KA * \n"
"P3-FU * -FU-FU-FU-FU *  * -FU\n"
"P4 *  *  *  *  *  * +HI *  * \n"
"P5 *  *  *  *  *  *  *  *  * \n"
"P6 * -HI+FU *  *  *  *  *  * \n"
"P7+FU *  * +FU+FU+FU+FU * +FU\n"
"P8 * +KA+KI *  *  *  *  *  * \n"
"P9+KY+KE+GI * +OU+KI+GI+KE+KY\n"
"P+00FU\n"
"P+00FU\n"
"P+00FU\n"
"P-00FU\n"
"P-00FU\n"
"-\n"
).getInitialState();
  moves.clear();
  test_state_t eState1(state1);
  Escape<WHITE,test_state_t::effect_state_t,Store >::
    generateMoves(eState1,eState1.getPieceAt(newPosition(3,2)),
		  eState1.getPieceAt(newPosition(3,4)),store);
  moves.unique();
  CPPUNIT_ASSERT(moves.isMember(newDrop(newPosition(3,3),PAWN,WHITE)) ||
		 (std::cerr << eState1 << moves << std::endl,0));
  }
}
// 
void generateEscapeMovesTest::testBlocking(){
  ::testBlocking<test_state_t>(state);
}

template<typename State>
static void testMoveFile(const std::string& fileName){
  Record rec=CsaFile(fileName).getRecord();
  PawnMaskState state=rec.getInitialState();
  test_state_t neState(state);
  std::vector<osl::Move> moves=rec.getMoves();
  for (unsigned int i=0;i<moves.size();i++){
    MoveVector allMoves;
    Store store(allMoves);
    //    std::cerr << "Before AllGene: " << neState.pawnMask[0] << "," << neState.pawnMask[1] << std::endl;
    AllMoves<SimpleState,Store >::
      generateMoves(neState.getTurn(),neState,store);
    //    std::cerr << "After AllGene: " << neState.pawnMask[0] << "," << neState.pawnMask[1] << std::endl;
    allMoves.unique();
    for (int num=0;num<40;num++){
      Piece p=neState.getPieceOf(num);
      if (isOnBoard(p) && getOwner(p)==neState.getTurn()){
	Position pos=getPosition(p);
	if (neState.hasEffectBy(alt(neState.getTurn()),pos)){
	  //
	  Piece p1=neState.hasEffectPiece(alt(neState.getTurn()),pos);
	  bool canSelfCapture=neState.hasEffectBy(alt(neState.getTurn()),getPosition(p1));
	  //
	  MoveVector escapeMoves;
	  Store storeEscape(escapeMoves);
	  //	  std::cerr << "Before GenEscape: " << pos << neState.pawnMask[0] << "," << neState.pawnMask[1] << std::endl;
	  if(neState.getTurn()==BLACK)
	    Escape<BLACK,test_state_t::effect_state_t,Store >::
	      generateMoves(neState,p,p1,storeEscape);
	  else
	    Escape<WHITE,test_state_t::effect_state_t,Store >::
	      generateMoves(neState,p,p1,storeEscape);
	  //	  std::cerr << "After GenEscape: " << pos << neState.pawnMask[0] << "," << neState.pawnMask[1] << std::endl;
	  for (size_t i=0;i<allMoves.size();i++)
	  {
	    DoUndoMoveLock lock(neState, allMoves[i]);
	    //	    std::cerr << "InMoveLock: " << allMoves[i] << neState.pawnMask[0] << "," << neState.pawnMask[1] << std::endl;
	    if (!neState.hasEffectBy(neState.getTurn(),
				    getPosition(neState.getPieceOf(num))))
	    {
	      CPPUNIT_ASSERT(escapeMoves.isMember(allMoves[i]) ||
			     (std::cerr << std::endl << neState << allMoves[i] << std::endl << escapeMoves << p << std::endl << p1 << std::endl << canSelfCapture << std::endl,0));
	    }
	  }
	}
      }
    }
    Move move=moves[i];
    //    std::cerr << "Before : " << neState.pawnMask[0] << "," << neState.pawnMask[1] << std::endl;
    neState.doMove(move);
    //    std::cerr << "After : " << neState.pawnMask[0] << "," << neState.pawnMask[1] << std::endl;
    assert(neState.isConsistent(true) ||
	   (std::cerr << move << std::endl << neState << std::endl,0)
	   );
  }
}

void generateEscapeMovesTest::testMoveMember(){
  extern bool isShortTest;
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  int count=100;
  if (isShortTest) 
      count=10;
  std::string fileName;
  while((ifs >> fileName) && ++i<count){
    if (fileName == "") 
	break;
    testMoveFile<test_state_t>("../data/kifdat/" + fileName);
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
