#include "osl/record/csaString.h"
#include "osl/record/csa.h"
#include "recursiveDoUndoMove.h"
#include "osl/perfmon.h"
#include "osl/applyMove.h"
#include "osl/pawnMaskState.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <fstream>
class pawnMaskStateTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE( pawnMaskStateTest );
  CPPUNIT_TEST( testShow );
  CPPUNIT_TEST( testIsValid );
  CPPUNIT_TEST( testHasEffectFromTo );
  CPPUNIT_TEST( testForEachEffectOfPiece );
  CPPUNIT_TEST( testPieceHasEffectTo );
  CPPUNIT_TEST( testDoUndoMove );
  CPPUNIT_TEST( testDoUndoMoveFiles );
  CPPUNIT_TEST( testPieceConsistent );
  CPPUNIT_TEST( testCaptureBug );
  CPPUNIT_TEST( testPieceConsistent );
  CPPUNIT_TEST_SUITE_END();
 private:
  osl::PawnMaskState state;
 public:
  void setUp();
  void testShow();
  void testIsValid();
  void testHasEffectFromTo();
  void testForEachEffectOfPiece();
  void testPieceHasEffectTo();
  void testDoUndoMove();
  void testDoUndoMoveFiles();
  void testCaptureBug();
  void testPieceConsistent();
};
CPPUNIT_TEST_SUITE_REGISTRATION(pawnMaskStateTest);

using namespace osl;
extern bool isShortTest;

void pawnMaskStateTest::setUp(){
  state=CsaString(
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU\n"
"P-00AL\n"
"+\n").getInitialState();
}

void pawnMaskStateTest::testShow(){
  CPPUNIT_ASSERT(state.isConsistent() || state.isConsistent(true) );
  if (! isShortTest)
  {
    std::cout << state;
    PawnMaskState state1(HIRATE);
    std::cout << state1;
  }
}

void pawnMaskStateTest::testIsValid(){
  // μ
  CPPUNIT_ASSERT(state.isValidMove(newMove(newPosition(6,9),newPosition(7,8),SILVER,PTYPE_EMPTY,false,WHITE),false)==false);
  // ʳؤput
  CPPUNIT_ASSERT(state.isValidMove(newDrop(newPosition(2,1),PAWN,BLACK),false)==false);
  // äƤʤ
  CPPUNIT_ASSERT(state.isValidMove(newDrop(newPosition(9,9),SILVER,BLACK),false)==false);
  // 
  CPPUNIT_ASSERT(state.isValidMove(newDrop(newPosition(7,1),PAWN,BLACK),false)==false);
  // ǤϤʤ
  CPPUNIT_ASSERT(state.isValidMove(newDrop(newPosition(8,2),PAWN,BLACK),false)==true);
  // ưʤ
  CPPUNIT_ASSERT(state.isValidMove(newDrop(newPosition(4,1),PAWN,BLACK),false)==false);
  // fromˤΤζ
  CPPUNIT_ASSERT(state.isValidMove(newMove(newPosition(8,2),newPosition(7,3),PROOK,PTYPE_EMPTY,false,BLACK),false)==false);
  // toˤΤζ𤫶?
  CPPUNIT_ASSERT(state.isValidMove(newMove(newPosition(8,1),newPosition(9,1),PPAWN,PLANCE,false,BLACK),false)==false);
  CPPUNIT_ASSERT(state.isValidMove(newMove(newPosition(8,1),newPosition(7,1),PPAWN,PTYPE_EMPTY,false,BLACK),false)==true);
  CPPUNIT_ASSERT(state.isValidMove(newMove(newPosition(5,3),newPosition(4,2),PBISHOP,SILVER,false,BLACK),false)==true);

      //  offsetưptype˴ؤvalid?
  CPPUNIT_ASSERT(state.isValidMove(newMove(newPosition(8,1),newPosition(9,2),PPAWN,PTYPE_EMPTY,false,BLACK),false)==false);
      // Υ줿ưλ˴֤Ƥ뤫?
  CPPUNIT_ASSERT(state.isValidMove(newMove(newPosition(8,3),newPosition(6,3),PROOK,PTYPE_EMPTY,false,BLACK),false)==true);
  CPPUNIT_ASSERT(state.isValidMove(newMove(newPosition(8,3),newPosition(4,3),PROOK,PTYPE_EMPTY,false,BLACK),false)==false);
      // capturePtypeפƤ뤫?
  CPPUNIT_ASSERT(state.isValidMove(newMove(newPosition(5,3),newPosition(4,2),PBISHOP,PTYPE_EMPTY,false,BLACK),false)==false);
      // promoteƤpromoteǽ
  CPPUNIT_ASSERT(state.isValidMove(newMove(newPosition(8,1),newPosition(7,1),PPAWN,PTYPE_EMPTY,true,BLACK),false)==false);
      // promoteƤʤ˶promoteǤʤ?
#if 0
  // 򤫤
  // ߤpawnMaskStateȽǤǤʤ
  CPPUNIT_ASSERT(state.isValidMove(newMove(newPosition(8,9),newPosition(8,8),KING,PTYPE_EMPTY,false,BLACK),false)==false);
#endif

  const PawnMaskState s2 = CsaString(
"P1-KY *  *  *  *  *  * -KE-KY\n"
"P2 *  *  *  *  *  * -KI * -OU\n"
"P3-FU *  *  *  *  * -GI-GI * \n"
"P4 *  * -FU * -FU * -FU-FU-FU\n"
"P5 * -FU *  *  * +FU * +FU * \n"
"P6 *  * +FU * +FU+KA+FU * +FU\n"
"P7+FU * -TO *  * +KI+KE *  * \n"
"P8+HI *  *  *  *  * +GI+OU * \n"
"P9+KY *  * -RY * +KI *  * +KY\n"
"P+00FU00FU00FU00KE00KI\n"
"P-00KE00GI00KA\n"
"+\n").getInitialState();
  const Move illegal = newMove(newPosition(2,5),newPosition(2,4),PPAWN,PAWN,true,BLACK);
  CPPUNIT_ASSERT(! s2.isValidMove(illegal,false));
}

void pawnMaskStateTest::testHasEffectFromTo(){
  // (8,3)ε
  CPPUNIT_ASSERT(state.hasEffectFromTo(newPtypeO(BLACK,PROOK),newPosition(8,3),newPosition(8,9)));
  CPPUNIT_ASSERT(!state.hasEffectFromTo(newPtypeO(BLACK,PROOK),newPosition(8,3),newPosition(3,3)));
  // (4,2)ζ
  CPPUNIT_ASSERT(state.hasEffectFromTo(newPtypeO(WHITE,SILVER),newPosition(4,2),newPosition(5,3)));
  CPPUNIT_ASSERT(!state.hasEffectFromTo(newPtypeO(WHITE,SILVER),newPosition(4,2),newPosition(3,2)));
  // (8,2)ε, Ǥǽ
  CPPUNIT_ASSERT(state.hasEffectFromTo(newPtypeO(BLACK,PROOK),newPosition(8,2),newPosition(7,1)));
}

void pawnMaskStateTest::testPieceHasEffectTo(){
  {
    // (7,7)
    Piece p=state.getPieceAt(newPosition(7,7));
    CPPUNIT_ASSERT(state.hasEffectTo(p,newPosition(9,5)));
    CPPUNIT_ASSERT(state.hasEffectTo(p,newPosition(7,6)));
    CPPUNIT_ASSERT(state.hasEffectTo(p,newPosition(6,6)));
    CPPUNIT_ASSERT(!state.hasEffectTo(p,newPosition(5,5)));
    CPPUNIT_ASSERT(!state.hasEffectTo(p,newPosition(7,5)));
    CPPUNIT_ASSERT(!state.hasEffectTo(p,newPosition(7,7)));
  }
  {
    Piece p=state.getPieceAt(newPosition(1,1));
    CPPUNIT_ASSERT(state.hasEffectTo(p,newPosition(1,3)));
    CPPUNIT_ASSERT(!state.hasEffectTo(p,newPosition(1,1)));
    CPPUNIT_ASSERT(!state.hasEffectTo(p,newPosition(2,3)));
    CPPUNIT_ASSERT(!state.hasEffectTo(p,newPosition(1,4)));
  }
}
struct StorePositionVector{
  std::vector<Position> store;
  template<Player P,Ptype Type>
  void doActionPtype(Piece p,Position pos){
    store.push_back(pos);
  }
  template<Player P>
  void doAction(Piece p,Position pos){
    store.push_back(pos);
  }
  bool isMember(Position pos) const{
    return std::find(store.begin(),store.end(),pos)!=store.end();
  }
};
void pawnMaskStateTest::testForEachEffectOfPiece(){
  {
    // 53
    StorePositionVector store;
    state.forEachEffectOfPiece<BLACK,PBISHOP,StorePositionVector>(newPosition(5,3),store);
    CPPUNIT_ASSERT(store.isMember(newPosition(4,2)));
    CPPUNIT_ASSERT(store.isMember(newPosition(2,6)));
    // block
    CPPUNIT_ASSERT(!store.isMember(newPosition(3,1)));
    CPPUNIT_ASSERT(!store.isMember(newPosition(5,1)));
  }
  {
    // 53
    StorePositionVector store;
    Piece p=state.getPieceAt(newPosition(5,3));
    state.forEachEffectOfPiece<StorePositionVector>(p,store);
    CPPUNIT_ASSERT(store.isMember(newPosition(4,2)));
    CPPUNIT_ASSERT(store.isMember(newPosition(2,6)));
    // block
    CPPUNIT_ASSERT(!store.isMember(newPosition(3,1)));
    CPPUNIT_ASSERT(!store.isMember(newPosition(5,1)));
  }
}

template<typename State>
struct CheckOnBoard{
  State& state;
  int ret;
  CheckOnBoard(State& s) : state(s)
  {
  }
  void operator()(Position to){
    CPPUNIT_ASSERT(state.isConsistent());
    int count=0;
    for(int num=0;num<40;num++)
      if(state.isOnBoardNum(num)) count++;
    ret=count;
  }
};

const Move b5364uma = newMove(newPosition(5,3),newPosition(6,4),PBISHOP,PAWN,false,BLACK);
void pawnMaskStateTest::testDoUndoMove(){
  CheckOnBoard<PawnMaskState> checkOnBoard(state);
  ApplyMoveOfTurn::doUndoMove(state, b5364uma, checkOnBoard);
  CPPUNIT_ASSERT(state.isConsistent());
  CPPUNIT_ASSERT_EQUAL(25, checkOnBoard.ret);
}



void pawnMaskStateTest::testDoUndoMoveFiles(){
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  int count=100;
  if (isShortTest)
      count=10;
  std::string fileName;
  while((ifs >> fileName) && (++i<count)) {
    if(fileName == "") 
	break;
    if (! isShortTest)
      std::cerr << fileName << " ";

    fileName = "../data/kifdat/" + fileName;
    Record rec=CsaFile(fileName).getRecord();
    PawnMaskState state=rec.getInitialState();
    std::vector<osl::Move> moves=rec.getMoves();
  
    size_t depth=0;
    if(getPlayer(*moves.begin())==BLACK){
      RecursiveDoUndoMove<BLACK,PawnMaskState>
	func(state, &*moves.begin(), &*moves.end(), &depth);
      ApplyMove<BLACK>::doUndoMove(state, *moves.begin(), func);
    }
    else{
      RecursiveDoUndoMove<WHITE,PawnMaskState>
	func(state, &*moves.begin(), &*moves.end(), &depth);
      ApplyMove<WHITE>::doUndoMove(state, *moves.begin(), func);
    }
    CPPUNIT_ASSERT_EQUAL(moves.size(), depth);
  }
}


void pieceNumConsistentBeforeCapture(PawnMaskState& state)
{
  CPPUNIT_ASSERT(state.isConsistent());
  for (int i=0; i<PIECE_SIZE; ++i)
  {
    CPPUNIT_ASSERT(state.isOnBoardNum(i));
    CPPUNIT_ASSERT(! state.isOffBoardNum(i));

    const Piece target = state.getPieceOf(i);
    const Position pos = getPosition(target);
    const Piece pieceOfSameLocation = state.getPieceAt(pos);
    CPPUNIT_ASSERT_EQUAL(target, pieceOfSameLocation);
  }
}

void pawnMaskStateTest::testPieceConsistent()
{
#if 0
  PawnMaskState s(HIRATE); // Ǥ褤?
#endif
  PawnMaskState s=CsaString(
"P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
"P2 * -HI *  *  *  *  * -KA * \n"
"P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
"P4 *  *  *  *  *  *  *  *  * \n"
"P5 *  *  *  *  *  *  *  *  * \n"
"P6 *  *  *  *  *  *  *  *  * \n"
"P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
"P8 * +KA *  *  *  *  * +HI * \n"
"P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
"+\n").getInitialState();
  pieceNumConsistentBeforeCapture(s);
  s.doMove(newMove(newPosition(7,7),newPosition(7,6),PAWN,PTYPE_EMPTY,false,BLACK));
  pieceNumConsistentBeforeCapture(s);
  s.doMove(newMove(newPosition(3,3),newPosition(3,4),PAWN,PTYPE_EMPTY,false,WHITE));
  pieceNumConsistentBeforeCapture(s);
  const Move move = newMove(newPosition(7,6),newPosition(7,5),PAWN,PTYPE_EMPTY,false,BLACK);
  s.doMove(move);
  const Piece expected = s.getPieceAt(getTo(move));
  pieceNumConsistentBeforeCapture(s);
  for (int i=0; i<PIECE_SIZE; ++i)
  {
    CPPUNIT_ASSERT(s.isOnBoardNum(i));
    CPPUNIT_ASSERT(! s.isOffBoardNum(i));
    const Piece target = s.getPieceOf(i);
    CPPUNIT_ASSERT((getPosition(target) != getTo(move))
		   || (expected == target));
    // ư𼫿ȤǤʤС𤬽Ťʤ뤳ȤϤʤ
  }
}

void pawnMaskStateTest::testCaptureBug()
{
  PawnMaskState s=CsaString(
    "P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
    "P2 * -HI *  *  *  *  * -KA * \n"
    "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
    "P4 *  *  *  *  *  *  *  *  * \n"
    "P5 *  *  *  *  *  *  *  *  * \n"
    "P6 *  *  *  *  *  *  *  *  * \n"
    "P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
    "P8 * +KA *  *  *  *  * +HI * \n"
    "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
    "+\n").getInitialState();
  s.doMove(newMove(newPosition(7,7),newPosition(7,6),PAWN,PTYPE_EMPTY,false,BLACK));
  s.doMove(newMove(newPosition(3,3),newPosition(3,4),PAWN,PTYPE_EMPTY,false,WHITE));
  CPPUNIT_ASSERT(s.isConsistent());

  const Position from = newPosition(8,8);
  const Position to = newPosition(2,2);
  const Move move = newMove(from,to,PBISHOP,BISHOP,true,BLACK);
  CPPUNIT_ASSERT(s.isValidMove(move,true));
  s.doMove(move);
  CPPUNIT_ASSERT(s.isConsistent());
  const Piece expected = s.getPieceAt(getTo(move));
  for (int i=0; i<PIECE_SIZE; ++i)
  {
    if(s.isOnBoardNum(i)){
      const Piece target = s.getPieceOf(i);
      CPPUNIT_ASSERT((getPosition(target) != getTo(move))
		     || (expected == target)); // ư𼫿ȤǤʤС𤬽Ťʤ뤳ȤϤʤ
    }
  }
}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
