/* piecePairEvalTest.cc
 */
#include "recursiveDoUndoMove.h"
#define PIECE_PAIR_EXTRA_DEBUG
#include "osl/piecePairEval.h"
#include "osl/pairEvalHashEffectState.h"
#include "osl/record/csaString.h"
#include "osl/record/csa.h"
#include "osl/doUndoMoveLock.h"
#include "osl/doUndoMoveLock.tcc"
#include "osl/eval.tcc"
#include "osl/doUndoMoveStack.h"
#include "doUndoMoveLockTest.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <bitset>
#include <fstream>

class PiecePairEvalTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(PiecePairEvalTest);
  CPPUNIT_TEST(testForEachRelation);
#if 0
  CPPUNIT_TEST(testUniqIndex);
  CPPUNIT_TEST(testIndex);
  CPPUNIT_TEST(testLoad);
  // testEval ʹߤ testLoad θ˥ƥȤ뤳
  CPPUNIT_TEST(testConsistent);
  CPPUNIT_TEST(testEval);
  CPPUNIT_TEST(testDoUndoMove);
#endif
  CPPUNIT_TEST_SUITE_END();
public:
  void testUniqIndex();
  void testForEachRelation();
  void testIndex();
  void testLoad();
  void testEval();
  void testConsistent();
  void testDoUndoMove();
};

CPPUNIT_TEST_SUITE_REGISTRATION(PiecePairEvalTest);

using namespace osl;
extern bool isShortTest;

typedef std::bitset<PiecePairEvalTable::maxPairIndex> set_t;
set_t visited;
void visitRelation(size_t index)
{
  size_t i1, i2;
  PiecePairEvalTable::meltIndex(index, i1, i2);
  const size_t i = PiecePairEvalTable::canonicalIndexOf(i1, i2);

  if (visited[i])
  {
    std::cerr << "oops " << i << "\n";
    size_t i1, i2;
    PiecePairEvalTable::meltIndex(i, i1, i2);
    std::cerr << i1 << " " << i2 << "\n";
    Position pos;
    PtypeO ptypeo;
    PiecePairEvalTable::meltIndex(i1, pos, ptypeo);
    std::cerr << pos << " " << ptypeo << "\n";
    PiecePairEvalTable::meltIndex(i2, pos, ptypeo);
    std::cerr << pos << " " << ptypeo << "\n";
  }
  CPPUNIT_ASSERT(! visited[i]);
  visited.flip(i);
}

void PiecePairEvalTest::testForEachRelation()
{
  PiecePairEvalTable::forEachRelation(visitRelation);
  size_t numCanons = 0;
  for (size_t i=0; i<PiecePairEvalTable::maxPairIndex; ++i)
  {
    size_t i1, i2;
    PiecePairEvalTable::meltIndex(i, i1, i2);
    if (PiecePairEvalTable::canonicalIndexOf(i1, i2) != i)
      goto confirm_not_visited;

    Position pos1, pos2;
    PtypeO ptypeo1, ptypeo2;
    PiecePairEvalTable::meltIndex(i1, pos1, ptypeo1);
    if (! isValid(pos1))
      goto confirm_not_visited;
    if (! isPiece(ptypeo1))
      goto confirm_not_visited;
    PiecePairEvalTable::meltIndex(i2, pos2, ptypeo2);
    if (! isValid(pos2))
      goto confirm_not_visited;
    if (! isPiece(ptypeo2))
      goto confirm_not_visited;
    if ((pos1 == pos2)
	&& isOnBoard(pos1)
	&& (ptypeo1 != ptypeo2))
      goto confirm_not_visited;
    if (isOffBoard(pos1) && isPromoted(ptypeo1))
      goto confirm_not_visited;
    if (isOffBoard(pos2) && isPromoted(ptypeo2))
      goto confirm_not_visited;

    // must be visited
    if (! visited[i])
    {
      std::cerr << "must visit " << i1 << " " << i2 << "\n"
		<< pos1 << " " << ptypeo1 << " " << pos2 << " " << ptypeo2
		<< "\n";
    }
    ++numCanons;
    continue;
  confirm_not_visited:
    if (visited[i])
    {
      std::cerr << "must not visit " << i1 << " " << i2 << "\n"
		<< pos1 << " " << ptypeo1 << " " << pos2 << " " << ptypeo2
		<< "\n";
    }
  }
  CPPUNIT_ASSERT_EQUAL(2578852u, numCanons);
  CPPUNIT_ASSERT_EQUAL(numCanons, visited.count());
}

void PiecePairEvalTest::testEval()
{
  PawnMaskState state(HIRATE);
  const int start_value = PiecePairEval(state).getVal();
  int value1, value12, value2;
  const Move m1 = newMove(newPosition(7,7),newPosition(7,6),PAWN,PTYPE_EMPTY,false,BLACK);
  {
    DoUndoMoveLock lock1(state, m1);
    CPPUNIT_ASSERT(state.isConsistent());

    value1 = PiecePairEval(state).getVal();
    const Move m12 = newMove(newPosition(3,3),newPosition(3,4),PAWN,PTYPE_EMPTY,false,WHITE);
    {
      DoUndoMoveLock lock12(state, m12);
      CPPUNIT_ASSERT(state.isConsistent());
      value12 = PiecePairEval(state).getVal();
      if (! isShortTest)
	std::cerr << "76FU34FU " << start_value 
		  << " ==> " << value1 
		  << " ==> " << value12 
		  << "\n";
    } //   state.undoMove(m12);
    CPPUNIT_ASSERT(state.isConsistent());
  } // state.undoMove(m1);
  CPPUNIT_ASSERT(state.isConsistent());

  const Move m2 = newMove(newPosition(2,7),newPosition(2,6),PAWN,PTYPE_EMPTY,false,BLACK);
  {
    DoUndoMoveLock lock2(state, m2);
    CPPUNIT_ASSERT(state.isConsistent());
    value2 = PiecePairEval(state).getVal();
    if (! isShortTest)
      std::cerr << "26FU " << start_value << " ==> " << value2 << "\n";
  } //  state.undoMove(m2);
  CPPUNIT_ASSERT(state.isConsistent());

  const Move m3 = newMove(newPosition(8,7),newPosition(8,6),PAWN,PTYPE_EMPTY,false,BLACK);
  {
    DoUndoMoveLock lock2(state, m3);
    CPPUNIT_ASSERT(state.isConsistent());
    const int value3 = PiecePairEval(state).getVal();
    if (! isShortTest)
      std::cerr << "86FU " << start_value << " ==> " << value3 << "\n";

    CPPUNIT_ASSERT(start_value < value1);
    CPPUNIT_ASSERT(value1 > value12);
    CPPUNIT_ASSERT(start_value < value2);
    CPPUNIT_ASSERT(value3 < value1);
    CPPUNIT_ASSERT(value3 < value2);
  }
#if 0
  PawnMaskState state=CsaString(
    "P1+NY+TO *  *  *  * -OU-KE-KY\n"
    "P2 *  *  *  *  * -GI-KI *  *\n"
    "P3 * +RY *  * +UM * -KI-FU-FU\n"
    "P4 *  * +FU-FU *  *  *  *  *\n"
    "P5 *  * -KE * +FU *  * +FU *\n"
    "P6-KE *  * +FU+GI-FU *  * +FU\n"
    "P7 *  * -UM *  *  *  *  *  *\n"
    "P8 *  *  *  *  *  *  *  *  * \n"
    "P9 * +OU * -GI *  *  *  * -NG\n"
    "P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
    "P-00KI00KY00FU00FU\n"
    "P-00AL\n"
    "+\n").getInitialState();
  PiecePairEval eval(state);
  std::cerr << "value is " << eval.getVal() << "\n";

  state=CsaString(
    "P1-KY-KE-GI-KI *  * -HI-KE-KY\n"
    "P2 * -HI *  * -KA-OU *  *  * \n"
    "P3-FU * -FU-FU-FU * -GI-FU-FU\n"
    "P4 *  *  *  *  * -FU *  *  * \n"
    "P5 * -FU * +KA *  *  *  *  * \n"
    "P6 *  * +FU *  *  * +FU * +FU\n"
    "P7+FU+FU * +FU+FU+FU+KE *  * \n"
    "P8 * +GI+KI *  *  * +GI *  * \n"
    "P9+KY+KE *  * +OU+KI *  * +KY\n"
    "P+00FU\n"
    "P+00FU\n"
    "P+00KI\n"
    "+\n").getInitialState();
  PiecePairEval eval2(state);
  std::cerr << "value is " << eval2.getVal() << "\n";
#endif
}

void PiecePairEvalTest::testLoad()
{
  const char *filename = "../data/sibling.pair";
  CPPUNIT_ASSERT(piecePairEvalTable.loadFromBinaryFile(filename));
}

void PiecePairEvalTest::testUniqIndex()
{
  set_t s;
  s.flip(PiecePairEvalTable::positionIndexOf(OFFBOARD));
  for (int x=1; x<=9; ++x)
  {
    for (int y=1; y<=9; ++y)
    {
      const Position pos1 = newPosition(x,y);
      const unsigned int index = PiecePairEvalTable::positionIndexOf(pos1);
      CPPUNIT_ASSERT(index < PiecePairEvalTable::maxPositionIndex);
      CPPUNIT_ASSERT(! s.test(index));
      s.flip(index);
    }
  }
  s.reset();
  for (int p=PPAWN; p<=ROOK; ++p)
  {
    const Ptype ptype = static_cast<Ptype>(p);
    const unsigned int indexBlack = PiecePairEvalTable::ptypeOIndexOf(newPtypeO(BLACK,ptype));
    const unsigned int indexWhite = PiecePairEvalTable::ptypeOIndexOf(newPtypeO(WHITE,ptype));
    
    CPPUNIT_ASSERT(indexBlack < PiecePairEvalTable::maxPtypeOIndex);
    CPPUNIT_ASSERT(! s.test(indexBlack));
    s.flip(indexBlack);
    CPPUNIT_ASSERT(indexWhite < PiecePairEvalTable::maxPtypeOIndex);
    CPPUNIT_ASSERT(! s.test(indexWhite));
    s.flip(indexWhite);
  }
}

/** member ѿˤ٤ȴ*/
set_t dejavu;
void testPieces(Piece p1, Piece p2)
{
  const unsigned int index12 = PiecePairEvalTable::indexOf(p1, p2);
  // std::cerr << index12 << " " << p1 << " " << p2 << "\n";
  CPPUNIT_ASSERT(index12 < PiecePairEvalTable::maxPairIndex);
  CPPUNIT_ASSERT(! dejavu.test(index12)
		 || (std::cerr << p1 << p2, 0));
  dejavu.flip(index12);
  if ((getPosition(p1) == getPosition(p2))
      && (getPtype(p1) == getPtype(p2))
      && (getOwner(p1) == getOwner(p2)))
    return;
  
  const unsigned int index21 = PiecePairEvalTable::indexOf(p2, p1);
  // std::cerr << index21 << " " << p2 << "\n";
  CPPUNIT_ASSERT(index21 < PiecePairEvalTable::maxPairIndex);
  CPPUNIT_ASSERT(! dejavu.test(index21)
		 || (std::cerr << p1 << p2, 0));
  dejavu.flip(index21);
}

void PiecePairEvalTest::testIndex()
{
  for (int x=1; x<=9; ++x)
  {
    std::cerr << x << " ";
    for (int y=1; y<=9; ++y)
    {
      const Position pos1 = newPosition(x,y);
      for (int ptype=PPAWN; ptype<=PTYPE_MAX; ++ptype)
      {
	const Piece p1 = newPiece(BLACK, static_cast<Ptype>(ptype), 0/*?*/, pos1);
	const unsigned int index = PiecePairEvalTable::indexOf(p1);
	CPPUNIT_ASSERT(index < PiecePairEvalTable::maxPieceIndex
		       || (std::cerr << p1 << "\n" << index << "\n" << PiecePairEvalTable::maxPieceIndex << "\n", 0));
	{
	  PtypeO ptypeo;
	  Position pos;
	  PiecePairEvalTable::meltIndex(index, pos, ptypeo);
	  CPPUNIT_ASSERT_EQUAL(newPtypeO(BLACK, static_cast<Ptype>(ptype)),
			       ptypeo);
	  CPPUNIT_ASSERT_EQUAL(pos1, pos);
	}
	const Piece p1w = newPiece(WHITE, static_cast<Ptype>(ptype), 0/*?*/, pos1);
	CPPUNIT_ASSERT(PiecePairEvalTable::indexOf(p1w) < PiecePairEvalTable::maxPieceIndex);
	{
	  PtypeO ptypeo;
	  Position pos;
	  PiecePairEvalTable::meltIndex(PiecePairEvalTable::indexOf(p1w),
					pos, ptypeo);
	  CPPUNIT_ASSERT_EQUAL(newPtypeO(WHITE, static_cast<Ptype>(ptype)),
			       ptypeo);
	  CPPUNIT_ASSERT_EQUAL(pos1, pos);
	}

	testPieces(p1, p1);
	testPieces(p1, p1w);
	testPieces(p1w, p1w);

	for (int x2=x; x2<=9; ++x2)
	{
	  for (int y2=y; y2<=9; ++y2)
	  {
	    Position pos2 = newPosition(x2,y2);
	    if (pos1 == pos2)
	      continue;
	    
	    for (int ptype2=PPAWN; ptype2<=PTYPE_MAX; ++ptype2)
	    {
	      const Piece p2 = newPiece(BLACK, static_cast<Ptype>(ptype2), 0/*?*/, pos2);
	      const unsigned int index2 = PiecePairEvalTable::indexOf(p2);
	      CPPUNIT_ASSERT(index2 < PiecePairEvalTable::maxPieceIndex
			     || (std::cerr << p1 << "\n" << index << "\n" << PiecePairEvalTable::maxPieceIndex << "\n", 0));
	      const Piece p2w = newPiece(WHITE, static_cast<Ptype>(ptype2), 0/*?*/, pos2);
	      CPPUNIT_ASSERT(PiecePairEvalTable::indexOf(p2w) < PiecePairEvalTable::maxPieceIndex);

	      testPieces(p1, p2);
	      testPieces(p1, p2w);
	      testPieces(p1w, p2);
	      testPieces(p1w, p2w);
	    }
	  }
	}

	for (int ptype2=KING; ptype2<=PTYPE_MAX; ++ptype2)
	{
	  const Piece p2 = newPiece(BLACK, static_cast<Ptype>(ptype2), 0/*?*/, OFFBOARD);
	  const unsigned int index2 = PiecePairEvalTable::indexOf(p2);
	  CPPUNIT_ASSERT(index2 < PiecePairEvalTable::maxPieceIndex
			 || (std::cerr << p1 << "\n" << index << "\n" << PiecePairEvalTable::maxPieceIndex << "\n", 0));
	  const Piece p2w = newPiece(WHITE, static_cast<Ptype>(ptype2), 0/*?*/, OFFBOARD);
	  CPPUNIT_ASSERT(PiecePairEvalTable::indexOf(p2w) < PiecePairEvalTable::maxPieceIndex);

	  testPieces(p1, p2);
	  testPieces(p1, p2w);
	  testPieces(p1w, p2);
	  testPieces(p1w, p2w);
	}
      }	// ptype
    } // x
  } // y

  // ƱΤϺǸ
  for (int ptype2=KING; ptype2<=PTYPE_MAX; ++ptype2)
  {
    const Piece p2 = newPiece(BLACK, static_cast<Ptype>(ptype2), 0/*?*/, OFFBOARD);
    const unsigned int index2 = PiecePairEvalTable::indexOf(p2);
    CPPUNIT_ASSERT(index2 < PiecePairEvalTable::maxPieceIndex);
    const Piece p2w = newPiece(WHITE, static_cast<Ptype>(ptype2), 0/*?*/, OFFBOARD);
    CPPUNIT_ASSERT(PiecePairEvalTable::indexOf(p2w) < PiecePairEvalTable::maxPieceIndex);
    testPieces(p2, p2);
    testPieces(p2, p2w);
    testPieces(p2w, p2w);
  }
#if 0
  std::cerr << PiecePairEvalTable::maxPieceIndex << "\n";
  std::cerr << PiecePairEvalTable::maxPairIndex << "\n";
#endif
}



void PiecePairEvalTest::testConsistent()
{
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  std::string fileName;
  for(int i=0;i<(isShortTest ? 100 : 10) && (ifs >> fileName) ; i++){
    if(fileName == "") 
      break;
    fileName = "../data/kifdat/" + fileName;

    Record rec=CsaFile(fileName).getRecord();
    PairEvalHashEffectState eState(rec.getInitialState());
    const std::vector<osl::Move> moves=rec.getMoves();
#if 0
    DoUndoMoveLockTest<PairEvalHashEffectState> lockTest(eState, moves);
    lockTest.run();
#endif
    DoUndoMoveStack<PairEvalHashEffectState> lockStack;
    for (size_t i=0; i<moves.size(); ++i)
    {
      const int prev = eState.getEval().getRawVal();
      const int next = prev + PiecePairEval::diffWithMove(eState, moves[i]);
      lockStack.push(eState, moves[i]);

      if (next != eState.getEval().getRawVal())
      {
	std::cerr << prev << " " << next << " " << eState.getVal() << "\n";
	std::cerr << i << " " << moves[i] << "\n" << eState;
      }
      CPPUNIT_ASSERT_EQUAL(next, eState.getEval().getRawVal());
    }
  }
}

void PiecePairEvalTest::testDoUndoMove(){
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  int count=100;
  if (isShortTest)
    count=10;
  std::string fileName;
  while((ifs >> fileName) && (++i<count)) {
    if(fileName == "") 
      break;
    if (! isShortTest)
      std::cerr << fileName << " ";

    CsaFile file("../data/kifdat/"+fileName);
    Record rec=file.getRecord();
    PairEvalHashEffectState eState(rec.getInitialState());

    std::vector<osl::Move> moves=rec.getMoves();
  
    size_t depth=0;
    typedef PairEvalHashEffectState state_t;
    if(getPlayer(*moves.begin())==BLACK){
      RecursiveDoUndoMove<BLACK,state_t>
	func(eState, &*moves.begin(), &*moves.end(), &depth);
      ApplyMove<BLACK>::doUndoMove(eState, *moves.begin(), func);
    }
    else{
      RecursiveDoUndoMove<WHITE,state_t>
	func(eState, &*moves.begin(), &*moves.end(), &depth);
      ApplyMove<WHITE>::doUndoMove(eState, *moves.begin(), func);
    }
    CPPUNIT_ASSERT_EQUAL(moves.size(), depth);
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
