/* pieceStandTest.cc
 */
#include "osl/pieceStand.h"
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <iostream>

using namespace osl;

class pieceStandTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(pieceStandTest);
  CPPUNIT_TEST(testCarry);
  CPPUNIT_TEST(testSet);
  CPPUNIT_TEST(testSuperior);
  CPPUNIT_TEST_SUITE_END();
public:
  void testCarry();
  void testSet();
  void testSuperior();
};

CPPUNIT_TEST_SUITE_REGISTRATION(pieceStandTest);


void pieceStandTest::testCarry()
{
  unsigned int flag=0;
  flag |= (1<<30);
  flag |= (1<<27);
  flag |= (1<<23);
  flag |= (1<<17);
  flag |= (1<<13);
  flag |= (1<<9);
  flag |= (1<<5);
  flag |= (1<<2);
  CPPUNIT_ASSERT_EQUAL(flag, PieceStand::carryMask);

  PieceStand pieces;
  pieces.carriesOn();
  for (int i=KING; i<=PTYPE_MAX; ++i)
    CPPUNIT_ASSERT_EQUAL(0u, pieces.get((Ptype)i));
  pieces.carriesOff();
  for (int i=KING; i<=PTYPE_MAX; ++i)
    CPPUNIT_ASSERT_EQUAL(0u, pieces.get((Ptype)i));
}


void pieceStandTest::testSet() 
{
  PieceStand pieces;
  for (int i=KING; i<=PTYPE_MAX; ++i)
  {
    const Ptype pi = (Ptype)i;
    const unsigned int num =	// ˤʤθĿ
      ptypeTable.getIndexLimit(pi) - ptypeTable.getIndexMin(pi);
    CPPUNIT_ASSERT_EQUAL(0u, pieces.get(pi));
    pieces.add(pi, num);
    for (int j=KING; j<=PTYPE_MAX; ++j)
    {
      const Ptype pj = (Ptype)j;
      if (i==j)
	CPPUNIT_ASSERT_EQUAL(num, pieces.get(pj));
      else
	CPPUNIT_ASSERT_EQUAL(0u, pieces.get(pj));
    }
    pieces.sub(pi, num);
    CPPUNIT_ASSERT_EQUAL(0u, pieces.get(pi));
  }
}

void pieceStandTest::testSuperior() 
{
  PieceStand p1, p2;
  CPPUNIT_ASSERT(p1.isSuperiorOrEqualTo(p2));
  CPPUNIT_ASSERT(p2.isSuperiorOrEqualTo(p1));
  
  p1.add(PAWN, 3);

  CPPUNIT_ASSERT(p1.isSuperiorOrEqualTo(p2));
  CPPUNIT_ASSERT(! p2.isSuperiorOrEqualTo(p1));

  CPPUNIT_ASSERT_EQUAL(3u, p1.get(PAWN));

  p2.add(GOLD, 1);
  
  CPPUNIT_ASSERT(! p1.isSuperiorOrEqualTo(p2));
  CPPUNIT_ASSERT(! p2.isSuperiorOrEqualTo(p1));

  p1.add(GOLD, 1);

  CPPUNIT_ASSERT(p1.isSuperiorOrEqualTo(p2));
  CPPUNIT_ASSERT(! p2.isSuperiorOrEqualTo(p1));
}

  
/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
