#include "osl/piece.h"
#include "osl/pieceMask.h"
#include <iomanip>
#include <iostream>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

class pieceTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( pieceTest );
  CPPUNIT_TEST( testShow );
  CPPUNIT_TEST( testSize );
  CPPUNIT_TEST( testBsf );
  CPPUNIT_TEST( testAddPosition );
  CPPUNIT_TEST( testIsEmpty);
  CPPUNIT_TEST( testIsEdge);
  CPPUNIT_TEST( testPromote);
  CPPUNIT_TEST( testPieceMask);
  CPPUNIT_TEST_SUITE_END();
 public:
  void testSize();
  void testShow();
  void testBsf();
  void testAddPosition();
  void testIsEmpty();
  void testIsEdge();
  void testPromote();
  void testPieceMask();
};

using namespace osl;
extern bool isShortTest;

CPPUNIT_TEST_SUITE_REGISTRATION(pieceTest);

void pieceTest::testSize(){
  CPPUNIT_ASSERT_EQUAL(sizeof(int), sizeof(Piece));
}

void pieceTest::testShow(){
  if (! isShortTest)
  {
    for(int num=0;num<PIECE_SIZE;num++)
      std::cerr << num << ":" << numTable.getPtype(num) 
	//	      << "," << std::setbase(16) << numTable.getMaskLow(num) 
		<< std::setbase(10) << std::endl;
  }
  //  std::cerr << newPiece(BLACK,PPAWN,0,newPosition(7,5)) << std::endl;
  //  std::cerr << newPiece(WHITE,KING,30,newPosition(1,9)) << std::endl;
}
void pieceTest::testBsf(){
  // std::cerr << std::endl << "pieceTest::testBsf()" << std::endl;
  for(unsigned int i=1;i<i+11111;i+=11111){
    CPPUNIT_ASSERT( bsf(i)==slowbsf(i) || (std::cerr << std::setbase(16)  << "," << i << "," << bsf(i) << "," << slowbsf(i) << std::setbase(10) << std::endl,0));
  }
}
void pieceTest::testAddPosition(){
  Piece p1=newPiece(BLACK,PPAWN,0,newPosition(7,5));
  CPPUNIT_ASSERT( getPosition(p1)==newPosition(7,5));
  p1+=newOffset(1,2);
  CPPUNIT_ASSERT( p1==newPiece(BLACK,PPAWN,0,newPosition(8,7)));
  p1=setPosition(p1,newPosition(6,4));
  CPPUNIT_ASSERT( p1==newPiece(BLACK,PPAWN,0,newPosition(6,4)));
}
void pieceTest::testIsEmpty(){
  CPPUNIT_ASSERT( !isEmpty(newPiece(BLACK,PPAWN,0,newPosition(8,7))));
  CPPUNIT_ASSERT( !isEmpty(newPiece(WHITE,KING,30,newPosition(6,4))));
  CPPUNIT_ASSERT( isEmpty(PIECE_EMPTY));
  CPPUNIT_ASSERT( !isEmpty(PIECE_EDGE));
}
void pieceTest::testIsEdge(){
  CPPUNIT_ASSERT( !isEdge(newPiece(BLACK,PPAWN,0,newPosition(8,7))));
  CPPUNIT_ASSERT( !isEdge(newPiece(WHITE,KING,30,newPosition(6,4))));
  CPPUNIT_ASSERT( !isEdge(PIECE_EMPTY));
  CPPUNIT_ASSERT( isEdge(PIECE_EDGE));
}
void pieceTest::testPromote(){
  Piece p1=newPiece(BLACK,PAWN,0,newPosition(8,7));
  Piece p2=promote(p1);
  CPPUNIT_ASSERT( p2==newPiece(BLACK,PPAWN,0,newPosition(8,7)));
  Piece p3=unpromote(p2);
  CPPUNIT_ASSERT( p1==p3 );
  Piece p4=checkPromote(p1,0); 
  CPPUNIT_ASSERT( p1==p4 );
  Piece p5=checkPromote(p1,1); 
  CPPUNIT_ASSERT( p2==p5 );
  Piece p6=checkUnpromote(p2,0);
  CPPUNIT_ASSERT( p2==p6 );
  Piece p7=checkUnpromote(p2,1);
  CPPUNIT_ASSERT( p1==p7 );
}
void pieceTest::testPieceMask(){
  PieceMask pm;
  // std::cerr << pm << std::endl;
  for(int i=0;i<40;i++)
    pm.setNum(i);
  for(int i=5;i<30;i++)
    pm.clearNum(i);
  CPPUNIT_ASSERT(pm.getMask(0)==0xc000001f);
  CPPUNIT_ASSERT(pm.getMask(1)==0xff);
  for(int i=0;i<5;i++)
    CPPUNIT_ASSERT(pm.isSet(i));
  for(int i=5;i<30;i++)
    CPPUNIT_ASSERT(!pm.isSet(i));
  for(int i=30;i<40;i++)
    CPPUNIT_ASSERT(pm.isSet(i));
}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
