#include "osl/ptype.h"
#include <iostream>
#include <iomanip>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

class ptypeTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( ptypeTest );
  CPPUNIT_TEST( testCanDropTo );
  CPPUNIT_TEST( testIsPromoted );
  CPPUNIT_TEST( testCanPromote );
  CPPUNIT_TEST( testPromote );
  CPPUNIT_TEST( testUnpromote );
  CPPUNIT_TEST( testPromoteUnpromote );
  CPPUNIT_TEST( testShow );
  CPPUNIT_TEST_SUITE_END();
 public:
  void testCanDropTo();
  void testIsPromoted();
  void testCanPromote();
  void testPromote();
  void testUnpromote();
  void testPromoteUnpromote();
  void testShow();
};

using namespace osl;
extern bool isShortTest;

CPPUNIT_TEST_SUITE_REGISTRATION(ptypeTest);

void ptypeTest::testCanDropTo(){
  CPPUNIT_ASSERT_EQUAL( false, (PtypePlayerTraits<PAWN,BLACK>::canDropTo(newPosition(4,1))));
  CPPUNIT_ASSERT_EQUAL( false, canDropTo(BLACK,PAWN,newPosition(4,1)));
}
void ptypeTest::testIsPromoted(){
  CPPUNIT_ASSERT( isPromoted(PPAWN) == true);
  CPPUNIT_ASSERT( isPromoted(PROOK) == true);
  CPPUNIT_ASSERT( isPromoted(KING) == false);
  CPPUNIT_ASSERT( isPromoted(ROOK) == false);
}
static void checkCanPromote(Ptype ptype){
  if(isPromoted(ptype) || ptype==GOLD || ptype==KING){
    CPPUNIT_ASSERT( canPromote(ptype) == false);
  }
  else {
    CPPUNIT_ASSERT( canPromote(ptype) == true);
  }
}
void ptypeTest::testCanPromote(){
  for_each_piece_ptype(checkCanPromote);
}
void ptypeTest::testPromote(){
  CPPUNIT_ASSERT( promote(PAWN) == PPAWN);
  CPPUNIT_ASSERT( promote(ROOK) == PROOK);
}
void ptypeTest::testUnpromote(){
  CPPUNIT_ASSERT( unpromote(PPAWN) == PAWN);
  CPPUNIT_ASSERT( unpromote(PROOK) == ROOK);
  CPPUNIT_ASSERT( unpromote(KING) == KING);
  CPPUNIT_ASSERT( unpromote(ROOK) == ROOK);
}
// promoteǽʤ
static void checkPromoteUnpromote(Ptype ptype){
  if(canPromote(ptype)){
    CPPUNIT_ASSERT( unpromote(promote(ptype)) == ptype);
  }
}
void ptypeTest::testPromoteUnpromote(){
  for_each_piece_ptype(checkPromoteUnpromote);
}
static void checkTable(Ptype ptype){
  std::cout << ptype << ",Index=" << ptypeTable.getIndex(ptype) 
	    << ",maskLow=" << ptypeTable.getMaskLow(ptype) 
	    << ",moveMask=" << std::setbase(16) 
	    << ptypeTable.getMoveMask(ptype) 
	    << std::setbase(10) << std::endl;
}

void ptypeTest::testShow(){
  if (! isShortTest)
  {
    for_each_piece_ptype(checkTable);
    for(int i=0;i<2;i++){
      Player p=static_cast<Player>(-i);
      for(int j=0;j<SHORT_DIRECTION_SIZE;j++){
	Direction dir=static_cast<Direction>(j);
	for(int k=PPAWN;k<=PTYPE_MAX;k++){
	  Ptype ptype=static_cast<Ptype>(k);
	  std::cout << p << "," << dir << "," << ptype << "," << std::setbase(16) 
		    << ptypeTable.getShortMoveMask(p,static_cast<PtypeO>(k),dir)
		    << ","
		    << ptypeTable.getShortMoveMask(p,static_cast<PtypeO>(k-16),dir)
		    << std::setbase(10) << std::endl;
	}
      }
    }
  }
}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
