#include "csaTest.h"
#include "osl/record/csaString.h"
#include "osl/perfmon.h"
#include "osl/pawnMaskState.h"
#include "osl/doUndoMoveLock.h"
#include <fstream>
using namespace osl;
using namespace osl::record;

CPPUNIT_TEST_SUITE_REGISTRATION(csaTest);

void csaTest::setUp(){
  rec=CsaFile("../test/kcc2.csa").getRecord();
  state=rec.getInitialState();
  moves=rec.getMoves();
}

void csaTest::testShow(){
}

void csaTest::testLoad(){
  //  std::cout << rec << std::endl;
  // st/d::cout << irs.getState() << std::endl;
}

struct posToInt {
  int operator()(){
    return 1;
  }
};


inline void myStateTest(PawnMaskState& state){
  // std::cout << state << std::endl;
  {
    const Position pos0=csaStrToPos("55"),pos1=csaStrToPos("54");
    const Move move=newMove(pos0,pos1,
			    PAWN,PTYPE_EMPTY,false,BLACK);
    const PawnMaskState stateCopy(state);
    {
      DoUndoMoveLock lock(state, move);
    }
    CPPUNIT_ASSERT_EQUAL(stateCopy, state);

    for (int i=0;i<10000;i++)
    {
      DoUndoMoveLock lock(state, move);
    }
  }
  {
    const Position pos1=csaStrToPos("99");
    const Move move=newDrop(pos1,PAWN,BLACK);
    const PawnMaskState stateCopy(state);
    {
      DoUndoMoveLock lock(state, move);
    }
    CPPUNIT_ASSERT_EQUAL(stateCopy, state);

    for (int i=0;i<10000;i++)
    {
      DoUndoMoveLock lock(state, move);
    }
  }
  {
    const Position pos0=csaStrToPos("53"),pos1=csaStrToPos("42");
    const Move move=newMove(pos0,pos1,
			    PBISHOP,SILVER,false,BLACK);
    const PawnMaskState stateCopy(state);
    {
      DoUndoMoveLock lock(state, move);
    }
    CPPUNIT_ASSERT_EQUAL(stateCopy, state);

    for (int i=0;i<10000;i++)
    {
      DoUndoMoveLock lock(state, move);
    }
  }
}

void csaTest::testSpeed(){
  Record rec=CsaString(
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU\n"
"P-00AL\n"
"+\n").getRecord();
  PawnMaskState state0=rec.getInitialState();
  //  std::cout << state0;
  // std::cout << "BitXmask" << std::endl;
  {
    const Position pos0=csaStrToPos("55"),pos1=csaStrToPos("54");
    const Move move=newMove(pos0,pos1,
			    PAWN,PTYPE_EMPTY,false,BLACK);
    const PawnMaskState stateCopy(state0);
    {
      DoUndoMoveLock lock(state0, move);
    }
    CPPUNIT_ASSERT_EQUAL(stateCopy, state0);

    for (int i=0;i<10000;i++)
    {
      DoUndoMoveLock lock(state0, move);
    }
  }
  {
    const Position pos1=csaStrToPos("99");
    const Move move=newDrop(pos1,PAWN,BLACK);
    const PawnMaskState stateCopy(state0);
    {
      DoUndoMoveLock lock(state0, move);
    }
    CPPUNIT_ASSERT_EQUAL(stateCopy, state0);

    for (int i=0;i<10000;i++)
    {
      DoUndoMoveLock lock(state0, move);
    }
  }
  {
    const Position pos0=csaStrToPos("53"),pos1=csaStrToPos("42");
    const Move move=newMove(pos0,pos1,
			     PBISHOP,SILVER,false,BLACK);
    const PawnMaskState stateCopy(state0);
    {
      DoUndoMoveLock lock(state0, move);
    }
    CPPUNIT_ASSERT_EQUAL(stateCopy, state0);

    for (int i=0;i<10000;i++)
    {
      DoUndoMoveLock lock(state0, move);
    }
  }
}

void csaTest::testSpeedGame(){
}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
