#include "endGameState.h"
#include "osl/search/mtdf.h"
#include "osl/search/mtdf.tcc"
#include "osl/search/searchRecorder.h"
#include "osl/search/simpleHashTable.h"
#include "osl/search/tableAnalyzer.h"
#include "osl/category/standardMoveGenerator.h"
#include "osl/record/csaString.h"
#include "osl/compositEvalHashEffectState.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <iostream>

class compositMtdfCaptureCheckIDTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE(compositMtdfCaptureCheckIDTest);
  CPPUNIT_TEST(testDriverOneMove);
  CPPUNIT_TEST(testDriverTwoMove);
  CPPUNIT_TEST(testDriverThreeMove);
  CPPUNIT_TEST(testBugSearch);
#if 1
  CPPUNIT_TEST(testInfty);
#endif
  CPPUNIT_TEST_SUITE_END();
private:
  osl::PawnMaskState state,state1;
public:
  compositMtdfCaptureCheckIDTest();
  void testDriverOneMove();
  void testDriverTwoMove();
  void testDriverThreeMove();
  void testBugSearch();
  void testInfty();
};


CPPUNIT_TEST_SUITE_REGISTRATION(compositMtdfCaptureCheckIDTest);

using namespace osl;

namespace
{
struct Loader
{
  Loader()
  {
    const char *filename = "../data/sibling.pair";
    CPPUNIT_ASSERT(piecePairEvalTable.loadFromBinaryFile(filename));
  }
} loader;
}

const Move move61rook = newMove(OFFBOARD,newPosition(6,1),ROOK,PTYPE_EMPTY,false,BLACK);

compositMtdfCaptureCheckIDTest::compositMtdfCaptureCheckIDTest()
{
  state=CsaString(endGameState).getInitialState();
  state1=CsaString(
"P1-KY-KE-GI-KI *  * -HI-KE-KY\n"
"P2 * -HI *  * -KA-OU *  *  * \n"
"P3-FU * -FU-FU-FU * -GI-FU-FU\n"
"P4 *  *  *  *  * -FU *  *  * \n"
"P5 * -FU * +KA *  *  *  *  * \n"
"P6 *  * +FU *  *  * +FU * +FU\n"
"P7+FU+FU * +FU+FU+FU+KE *  * \n"
"P8 * +GI+KI *  *  * +GI *  * \n"
"P9+KY+KE *  * +OU+KI *  * +KY\n"
"P+00FU\n"
"P+00FU\n"
"P+00KI\n"
"+\n").getInitialState();
}

typedef MTDF<CompositEvalHashEffectState,StandardMoveGenerator,
	     void,SearchRecorder,RealizationProbability> Mtdf;
typedef MTDF<CompositEvalHashEffectState,StandardMoveGenerator,
	     SimpleHashTable,SearchRecorder,RealizationProbability> MtdfWithTable;

void compositMtdfCaptureCheckIDTest::testDriverOneMove(){
  std::cerr << "\n testDriverOneMove\n";
  osl::SearchRecorder recorder("compositMtdfCaptureCheckIDTest.testDriverOneMove");

  CompositEvalHashEffectState eState(state);

  Move best500, best500ID;
  {
    recorder.stream() << "\ntestDriverOneMove + SOMA + TABLE\n";
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    best500 = mtdfSearcher.computeBestMove(500);

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 10);
  }
  {
    recorder.stream() << "\ntestDriverOneMove + SOMA + ID\n";
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    best500ID = mtdfSearcher.computeBestMoveIteratively(500);

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 10);
  }
  CPPUNIT_ASSERT_EQUAL(MOVE_INVALID, best500);
  CPPUNIT_ASSERT_EQUAL(MOVE_INVALID, best500ID);
}

void compositMtdfCaptureCheckIDTest::testDriverTwoMove(){
  std::cerr << "\n testDriverTwoMove\n";
  osl::SearchRecorder recorder("compositMtdfCaptureCheckIDTest.testDriverTwoMove");

  CompositEvalHashEffectState eState(state);
  Move bestMoveWithoutID;
  // recorder.setLogMargin(1000);
  {
    recorder.stream() << "\ntestDriverTwoMove 1000 + SOMA + TABLE\n";
    recorder.stream() << eState;
    
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    mtdfSearcher.setInitialGuess(2500); // id ̵Ǥ initial guess ʤǽ
    bestMoveWithoutID=mtdfSearcher.computeBestMove(700);

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 10);
  }
  Move bestMoveWithID;
  {
    recorder.stream() << "\ntestDriverTwoMove 1000 + SOMA + TABLE + ID\n";
    recorder.stream() << eState;

    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMoveWithID=mtdfSearcher.computeBestMoveIteratively(700);

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 10);
  }
  CPPUNIT_ASSERT_EQUAL(move61rook, bestMoveWithoutID);
  CPPUNIT_ASSERT_EQUAL(move61rook, bestMoveWithID);
}

void compositMtdfCaptureCheckIDTest::testDriverThreeMove(){
  std::cerr << "\n testDriverThreeMove\n";
  osl::SearchRecorder recorder("compositMtdfCaptureCheckIDTest.testDriverThreeMove");

  CompositEvalHashEffectState eState(state);
  Move bestMove, bestMoveID;
  {
    recorder.stream() << "\ntestDriverThreeMove 900 + SOMA + TABLE\n";
    // recorder.setLogMargin(1000);
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    mtdfSearcher.setInitialGuess(2500); // id ̵Ǥ initial guess ʤǽ

    bestMove=mtdfSearcher.computeBestMove(900);

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 10);
  }
  {
    recorder.stream() << "\ntestDriverThreeMove 900 + SOMA + TABLE + ID\n";
    SimpleHashTable table(1000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMoveID=mtdfSearcher.computeBestMoveIteratively(900);

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 10);
  }
  CPPUNIT_ASSERT_EQUAL(MOVE_INVALID, bestMove);
  CPPUNIT_ASSERT_EQUAL(MOVE_INVALID, bestMoveID);
}

void compositMtdfCaptureCheckIDTest::testBugSearch(){
  std::cerr << "\n testBugSearch\n";
  osl::SearchRecorder recorder("compositMtdfCaptureCheckIDTest.testBugSearch");

  PawnMaskState bugState=CsaString(
    "P1-KY-KE * -KI *  *  * -KE-KY\n"
    "P2 * -OU-GI * -KI * -HI *  * \n"
    "P3 * -FU *  * -FU-GI-KA-FU-FU\n"
    "P4-FU * -FU-FU * -FU-FU *  * \n"
    "P5 *  *  *  *  *  *  * +FU * \n"
    "P6+FU * +FU * +FU+GI+FU * +FU\n"
    "P7 * +FU * +FU * +FU *  *  * \n"
    "P8 * +KA+OU * +KI+GI * +HI * \n"
    "P9+KY+KE * +KI *  *  * +KE+KY\n"
    "+\n").getInitialState();
  CompositEvalHashEffectState eState(bugState);
  Move bestMove;
  {
    recorder.stream() << "\ntestBugSearch 800\n";
    SimpleHashTable table(3000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMove=mtdfSearcher.computeBestMove(800); // 2500
    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 10);
  }
  Move bestMoveID;
  {
    recorder.stream() << "\ntestBugSearch 1200 + ID\n";
    // recorder.setLogMargin(1000);
    SimpleHashTable table(3000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMoveID=mtdfSearcher.computeBestMoveIteratively(1000); // 2500
    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 10);
  }
#if 0
  CPPUNIT_ASSERT_EQUAL(bestMove, bestMoveID);
#endif
  // iterative deepening 򤹤ȡɤߤη̤αƶ ? 75ͤóݤ뤳Ȥ򤱤
  // const Move move77ka = newMove(newPosition(8,8),newPosition(7,7),BISHOP,PTYPE_EMPTY,false,BLACK);

  // פȤƤ 46ζ  57 ˵뤦ˤۤ-45⤬ˤʤΤͭޤʻųݤ
  const Move move35fu = newMove(newPosition(3,6),newPosition(3,5),PAWN,PTYPE_EMPTY,false,BLACK);
  CPPUNIT_ASSERT_EQUAL(move35fu, bestMoveID);
}

void compositMtdfCaptureCheckIDTest::testInfty(){
  std::cerr << "\n testInfty\n";
  osl::SearchRecorder recorder("compositMtdfCaptureCheckIDTest.testInfty");

  PawnMaskState state=CsaString(
    "P1-KY * -KI-KE *  *  *  * -KY\n"
    "P2 *  *  *  * -OU * -GI+UM * \n"
    "P3-FU * -KE-FU-FU-FU * -KA-FU\n"
    "P4 * -HI-FU *  *  *  *  *  * \n"
    "P5 *  *  *  *  *  *  *  *  * \n"
    "P6 * +KI * +FU+KI *  *  *  * \n"
    "P7+FU * +GI * +FU * +KI+GI+FU\n"
    "P8-TO * +OU *  *  *  *  *  * \n"
    "P9 *  *  *  *  *  *  *  * +KY\n"
    "P+00FU\n"
    "P-00FU\n"
    "P-00FU\n"
    "P-00FU\n"
    "P-00FU\n"
    "P-00FU\n"
    "P-00FU\n"
    "P+00KE\n"
    "P-00KE\n"
    "P+00GI\n"
    "P-00KY\n"
    "P+00HI\n"
    "+\n").getInitialState();
  CompositEvalHashEffectState eState(state);

  recorder.stream() << "\ntestInfty\n";
  recorder.setLogMargin(150);
  SimpleHashTable table(1000000,1);
  MtdfWithTable mtdfSearcher(eState,recorder,&table);
  Move best500ID = mtdfSearcher.computeBestMoveIteratively(1000,MOVE_INVALID,
							   200,600,10000);
  TableAnalyzer analyzer(table,recorder.stream());
  analyzer.examineTopLevel(eState, 10);

  // checkmateSearcher.tcc 1.18 ǵäƤ褦ˤʤ롣
  // ʻ֤롣
  // ~CheckmateSearcher(715406/10000000)
  // ~CheckmateSearcher(485616/10000000)
  // õ 0:  terminal: 100413
  // ȷ빽¿

  // iteration 100ŤĹԤ᤯뤬ɤߤϼˤʤ
  // 101 ŤĤˤˤƤѤʤɵ
  CPPUNIT_ASSERT_EQUAL(newDrop(newPosition(3,3), PAWN, BLACK), best500ID);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
