#include "endGameState.h"
#include "osl/search/mtdf.h"
#include "osl/search/mtdf.tcc"
#include "osl/search/searchRecorder.h"
#include "osl/search/simpleHashTable.h"
#include "osl/search/tableAnalyzer.h"
#include "osl/category/standardMoveGenerator.h"
#include "osl/evalHashEffectState.h"
#include "osl/category/fWMoveGen.h"
#include "osl/record/csaString.h"
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <iostream>

/**
 * iterative deepening Τ̵
 * ѤʤȤĴ٤
 *
 * Ʊλؤ꤬
 * - id  ¸Ψ㤤꤬ͥ褵
 * - id ʤ move generator ǽ˺ä꤬ͥ褵
 * Ȥ㤤Ǥ뤿ᡤͣɤΤ̤Ѥ뤫
 * ؼɾͤĴ٤ɬפ
 */
class mtdfCaptureCheckIDTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( mtdfCaptureCheckIDTest );
  CPPUNIT_TEST( testDriverOneMove );
  CPPUNIT_TEST( testDriverTwoMove );
#if 1
  CPPUNIT_TEST( testDriverThreeMove );
#endif
  CPPUNIT_TEST( testBugSearch );
  CPPUNIT_TEST_SUITE_END();
private:
  osl::PawnMaskState state,state1;
public:
  mtdfCaptureCheckIDTest();
  void testDriverOneMove();
  void testDriverTwoMove();
  void testDriverThreeMove();
  void testBugSearch();
};

CPPUNIT_TEST_SUITE_REGISTRATION(mtdfCaptureCheckIDTest);
using namespace osl;

typedef MTDF<EvalHashEffectState,StandardMoveGenerator,
	     SimpleHashTable,SearchRecorder,RealizationProbability> RpMtdfWithTable;

const Move move61rook = newMove(OFFBOARD,newPosition(6,1),ROOK,PTYPE_EMPTY,false,BLACK);
const Move move51rook = newMove(OFFBOARD,newPosition(5,1),ROOK,PTYPE_EMPTY,false,BLACK);
const Move move5354uma = newMove(newPosition(5,3),newPosition(5,4),PBISHOP,PTYPE_EMPTY,false,BLACK);

mtdfCaptureCheckIDTest::mtdfCaptureCheckIDTest()
{
  state=CsaString(endGameState).getInitialState();
  state1=CsaString(
"P1-KY-KE-GI-KI *  * -HI-KE-KY\n"
"P2 * -HI *  * -KA-OU *  *  * \n"
"P3-FU * -FU-FU-FU * -GI-FU-FU\n"
"P4 *  *  *  *  * -FU *  *  * \n"
"P5 * -FU * +KA *  *  *  *  * \n"
"P6 *  * +FU *  *  * +FU * +FU\n"
"P7+FU+FU * +FU+FU+FU+KE *  * \n"
"P8 * +GI+KI *  *  * +GI *  * \n"
"P9+KY+KE *  * +OU+KI *  * +KY\n"
"P+00FU\n"
"P+00FU\n"
"P+00KI\n"
"+\n").getInitialState();
}

void mtdfCaptureCheckIDTest::testDriverOneMove(){
  std::cerr << "\n testDriverOneMove\n";
  osl::SearchRecorder recorder("mtdfCaptureCheckIDTest.testDriverOneMove");

  EvalHashEffectState eState(state);
  // const Move move99gold = newMove(OFFBOARD,newPosition(9,9),GOLD,PTYPE_EMPTY,false,BLACK);
  
  {
    recorder.stream() << "\ntestDriverOneMove + SOMA + TABLE\n";
    recorder.setLogMargin(100);
    SimpleHashTable table(1000000,1);
    RpMtdfWithTable mtdfSearcher(eState,recorder,&table);
    const Move best500=mtdfSearcher.computeBestMove(500);

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 5);

    CPPUNIT_ASSERT_EQUAL(move5354uma,best500);
  }
  {
    recorder.stream() << "\ntestDriverOneMove + SOMA + ID\n";
    SimpleHashTable table(1000000,1);
    RpMtdfWithTable mtdfSearcher(eState,recorder,&table);
    Move best500=mtdfSearcher.computeBestMoveIteratively(500);
    //  newMove(newPosition(7,4),newPosition(7,3),PPAWN,PTYPE_EMPTY,true,BLACK)

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 5);

    CPPUNIT_ASSERT_EQUAL(move5354uma, best500);
  }
}

void mtdfCaptureCheckIDTest::testDriverTwoMove(){
  std::cerr << "\n testDriverTwoMove\n";
  osl::SearchRecorder recorder("mtdfCaptureCheckIDTest.testDriverTwoMove");

  EvalHashEffectState eState(state);
  Move bestMoveWithoutID;
  {
    recorder.stream() << "\ntestDriverTwoMove 700 + SOMA + TABLE\n";
    recorder.stream() << eState;
    recorder.setLogMargin(100);
    
    SimpleHashTable table(700000,1);
    RpMtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMoveWithoutID=mtdfSearcher.computeBestMove(700);

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 5);
  }
  Move bestMoveWithID;
  {
    recorder.stream() << "\ntestDriverTwoMove 700 + SOMA + TABLE + ID\n";
    recorder.stream() << eState;

    SimpleHashTable table(700000,1);
    RpMtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMoveWithID=mtdfSearcher.computeBestMoveIteratively(700);

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 5);
  }
  CPPUNIT_ASSERT_EQUAL(move51rook, bestMoveWithID);
  CPPUNIT_ASSERT_EQUAL(move61rook, bestMoveWithoutID);
  
}

void mtdfCaptureCheckIDTest::testDriverThreeMove(){
  std::cerr << "\n testDriverThreeMove\n";
  osl::SearchRecorder recorder("mtdfCaptureCheckIDTest.testDriverThreeMove");

  EvalHashEffectState eState(state);
  Move bestMove, bestMoveID;
  {
    recorder.stream() << "\ntestDriverThreeMove 800 + SOMA + TABLE\n";
    SimpleHashTable table(1000000,501);
    RpMtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMove=mtdfSearcher.computeBestMove(800);

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 5);
  }
  {
    recorder.stream() << "\ntestDriverThreeMove 800 + SOMA + TABLE + ID\n";
    //recorder.setLogMargin(300);
    SimpleHashTable table(1000000,501);
    RpMtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMoveID=mtdfSearcher.computeBestMoveIteratively(800);

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 5);
  }
  CPPUNIT_ASSERT_EQUAL(bestMove, bestMoveID);
}

void mtdfCaptureCheckIDTest::testBugSearch(){
  std::cerr << "\n testBugSearch\n";
  osl::SearchRecorder recorder("mtdfCaptureCheckIDTest.testBugSearch");

  PawnMaskState bugState=CsaString(
    "P1-KY-KE * -KI *  *  * -KE-KY\n"
    "P2 * -OU-GI * -KI * -HI *  * \n"
    "P3 * -FU *  * -FU-GI-KA-FU-FU\n"
    "P4-FU * -FU-FU * -FU-FU *  * \n"
    "P5 *  *  *  *  *  *  * +FU * \n"
    "P6+FU * +FU * +FU+GI+FU * +FU\n"
    "P7 * +FU * +FU * +FU *  *  * \n"
    "P8 * +KA+OU * +KI+GI * +HI * \n"
    "P9+KY+KE * +KI *  *  * +KE+KY\n"
    "+\n").getInitialState();
  EvalHashEffectState eState(bugState);
  Move bestMove;
  {
    recorder.stream() << "\ntestBugSearch 1200\n";
    SimpleHashTable table(3000000,501);
    RpMtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMove=mtdfSearcher.computeBestMove(1200); // 2500

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 5);
  }
  Move bestMoveID;
  {
    recorder.stream() << "\ntestBugSearch 1200 + ID\n";
    // recorder.setLogMargin(1000);
    SimpleHashTable table(3000000,501);
    RpMtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMoveID=mtdfSearcher.computeBestMoveIteratively(1200); // 2500

    TableAnalyzer analyzer(table,recorder.stream());
    analyzer.examineTopLevel(eState, 5);
  }
#if 1
  CPPUNIT_ASSERT_EQUAL(bestMove, bestMoveID);
#endif
  // iterative deepening 򤹤ȡɤߤη̤αƶ ? 75ͤóݤ뤳Ȥ򤱤
  CPPUNIT_ASSERT_EQUAL(newMove(newPosition(4,8),newPosition(3,7),SILVER,PTYPE_EMPTY,false,BLACK),
		       bestMoveID);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
