#include "osl/mtdf.h"
#include "osl/category/standardMoveGenerator.h"
#include "osl/csaString.h"
#include "osl/evalHashEffectState.h"
#include "osl/mtdfRecorder.h"
#include "osl/simpleHashTable.h"
#include "osl/perfmon.h"
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <iostream>

class mtdfCaptureCheckTableTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( mtdfCaptureCheckTableTest );
  CPPUNIT_TEST( testDriverOneMove );
  CPPUNIT_TEST( testDriverTwoMove );
  CPPUNIT_TEST( testDriverThreeMove );
  CPPUNIT_TEST( testBugSearch );
  CPPUNIT_TEST_SUITE_END();
private:
  osl::PawnMaskState state,state1;
public:
  mtdfCaptureCheckTableTest();
  void testDriverOneMove();
  void testDriverTwoMove();
  void testDriverThreeMove();
  void testBugSearch();
};


CPPUNIT_TEST_SUITE_REGISTRATION(mtdfCaptureCheckTableTest);

using namespace osl;

mtdfCaptureCheckTableTest::mtdfCaptureCheckTableTest()
{
  state=CsaString(endGameState).getInitialState();
  state1=CsaString(
"P1-KY-KE-GI-KI *  * -HI-KE-KY\n"
"P2 * -HI *  * -KA-OU *  *  * \n"
"P3-FU * -FU-FU-FU * -GI-FU-FU\n"
"P4 *  *  *  *  * -FU *  *  * \n"
"P5 * -FU * +KA *  *  *  *  * \n"
"P6 *  * +FU *  *  * +FU * +FU\n"
"P7+FU+FU * +FU+FU+FU+KE *  * \n"
"P8 * +GI+KI *  *  * +GI *  * \n"
"P9+KY+KE *  * +OU+KI *  * +KY\n"
"P+00FU\n"
"P+00FU\n"
"P+00KI\n"
"+\n").getInitialState();
}

typedef MTDF<EvalHashEffectState,StandardMoveGenerator,
	     void,SearchRecorder,RealizationProbability> Mtdf;
typedef MTDF<EvalHashEffectState,StandardMoveGenerator,
	     SimpleHashTable,SearchRecorder,RealizationProbability>
  MtdfWithTable;

void mtdfCaptureCheckTableTest::testDriverOneMove(){
  std::cerr << "\n testDriverOneMove\n";
  osl::SearchRecorder recorder("mtdfCaptureCheckTableTest.testDriverOneMove");

  EvalHashEffectState eState(state);
  Move best500, best500Table;
  recorder.stream() << "\ntestDriverOneMove + SOMA\n";
  {
    Mtdf mtdfSearcher(eState,recorder);
    best500=mtdfSearcher.computeBestMove(500);
  }
  {
    recorder.stream() << "\ntestDriverOneMove + SOMA + TABLE\n";
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    best500Table=mtdfSearcher.computeBestMove(500);
  }
  const Move move61rook = newMove(OFFBOARD,newPosition(6,1),ROOK,PTYPE_EMPTY,false,BLACK);
  // const Move move99gold = newMove(OFFBOARD,newPosition(9,9),GOLD,PTYPE_EMPTY,false,BLACK);
  CPPUNIT_ASSERT_EQUAL(move61rook, best500);
  CPPUNIT_ASSERT_EQUAL(/*move99gold*/ move61rook, best500Table);
}

void mtdfCaptureCheckTableTest::testDriverTwoMove(){
  std::cerr << "\n testDriverTwoMove\n";
  osl::SearchRecorder recorder("mtdfCaptureCheckTableTest.testDriverTwoMove");

  EvalHashEffectState eState(state);
  Move bestMoveWithoutTable;
  {
    recorder.stream() << "\ntestDriverTwoMove 800 + SOMA\n";
    recorder.stream() << eState;
    
    Mtdf mtdfSearcher(eState,recorder);
    bestMoveWithoutTable=mtdfSearcher.computeBestMove(800);
  }
  Move bestMove;
  {
    recorder.stream() << "\ntestDriverTwoMove 800 + SOMA + TABLE\n";
    recorder.stream() << eState;

    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMove=mtdfSearcher.computeBestMove(800);
  }
  const Move move61rook = newMove(OFFBOARD,newPosition(6,1),ROOK,PTYPE_EMPTY,false,BLACK);
  const Move move51rook = newMove(OFFBOARD,newPosition(5,1),ROOK,PTYPE_EMPTY,false,BLACK);
  CPPUNIT_ASSERT_EQUAL(move61rook, bestMoveWithoutTable);
  CPPUNIT_ASSERT_EQUAL(move51rook, bestMove);
}

void mtdfCaptureCheckTableTest::testDriverThreeMove(){
  std::cerr << "\n testDriverThreeMove\n";
  osl::SearchRecorder recorder("mtdfCaptureCheckTableTest.testDriverThreeMove");

  EvalHashEffectState eState(state);
  recorder.stream() << eState;
  Move bestMove, bestMoveTable;
  {
    recorder.stream() << "\ntestDriverThreeMove 1000 + SOMA\n";
    recorder.resetNodeCount();
    Mtdf mtdfSearcher(eState,recorder);
    clock_start();
    bestMove=mtdfSearcher.computeBestMove(1000);
    clock_stop("",recorder.currentNodeCount());
  }
  {
    recorder.stream() << "\ntestDriverThreeMove 1000 + SOMA + TABLE\n";
    SimpleHashTable table(1100000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    clock_start();
    bestMoveTable=mtdfSearcher.computeBestMove(1000);
    clock_stop("",1);
  }
  const Move move61rook = newMove(OFFBOARD,newPosition(6,1),ROOK,PTYPE_EMPTY,false,BLACK);
  const Move move43knight = newMove(OFFBOARD,newPosition(4,3),KNIGHT,PTYPE_EMPTY,false,BLACK);
  CPPUNIT_ASSERT_EQUAL(move61rook, bestMove);
  CPPUNIT_ASSERT_EQUAL(move43knight, bestMoveTable);
}

void mtdfCaptureCheckTableTest::testBugSearch(){
  std::cerr << "\n testBugSearch\n";
  osl::SearchRecorder recorder("mtdfCaptureCheckTableTest.testBugSearch");

  PawnMaskState bugState=CsaString(
    "P1-KY-KE * -KI *  *  * -KE-KY\n"
    "P2 * -OU-GI * -KI * -HI *  * \n"
    "P3 * -FU *  * -FU-GI-KA-FU-FU\n"
    "P4-FU * -FU-FU * -FU-FU *  * \n"
    "P5 *  *  *  *  *  *  * +FU * \n"
    "P6+FU * +FU * +FU+GI+FU * +FU\n"
    "P7 * +FU * +FU * +FU *  *  * \n"
    "P8 * +KA+OU * +KI+GI * +HI * \n"
    "P9+KY+KE * +KI *  *  * +KE+KY\n"
    "+\n").getInitialState();
  EvalHashEffectState eState(bugState);

  Move bestMove, bestMoveTable;
  {
    recorder.stream() << "\ntestBugSearch 1200 + SOMA\n";
    Mtdf mtdfSearcher(eState,recorder);
    bestMove=mtdfSearcher.computeBestMove(1200);
  }
  
  {
    recorder.stream() << "\ntestBugSearch 1200\n";
    SimpleHashTable table(3000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMoveTable=mtdfSearcher.computeBestMove(1200);
  }
  CPPUNIT_ASSERT_EQUAL(bestMove, bestMoveTable);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
