#include "osl/mtdf.h"
#include "osl/category/standardMoveGenerator.h"
#include "osl/csaString.h"
#include "osl/perfmon.h"
#include "osl/evalHashEffectState.h"
#include "osl/mtdfRecorder.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <boost/scoped_ptr.hpp>
#include <iostream>

class mtdfCaptureCheckTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( mtdfCaptureCheckTest );
  CPPUNIT_TEST( testDriverOneMove );
  CPPUNIT_TEST( testDriverTwoMove );
  CPPUNIT_TEST( testDriverThreeMove );
  CPPUNIT_TEST( testBugSearch );
  CPPUNIT_TEST_SUITE_END();
private:
  osl::PawnMaskState state,state1;
public:
  mtdfCaptureCheckTest();
  void testDriverOneMove();
  void testDriverTwoMove();
  void testDriverThreeMove();
  void testBugSearch();
};

CPPUNIT_TEST_SUITE_REGISTRATION(mtdfCaptureCheckTest);

using namespace osl;

const Move move61rook = newMove(OFFBOARD,newPosition(6,1),ROOK,PTYPE_EMPTY,false,BLACK);

mtdfCaptureCheckTest::mtdfCaptureCheckTest()
{
  state=CsaString(endGameState).getInitialState();
  state1=CsaString(
"P1-KY-KE-GI-KI *  * -HI-KE-KY\n"
"P2 * -HI *  * -KA-OU *  *  * \n"
"P3-FU * -FU-FU-FU * -GI-FU-FU\n"
"P4 *  *  *  *  * -FU *  *  * \n"
"P5 * -FU * +KA *  *  *  *  * \n"
"P6 *  * +FU *  *  * +FU * +FU\n"
"P7+FU+FU * +FU+FU+FU+KE *  * \n"
"P8 * +GI+KI *  *  * +GI *  * \n"
"P9+KY+KE *  * +OU+KI *  * +KY\n"
"P+00FU\n"
"P+00FU\n"
"P+00KI\n"
"+\n").getInitialState();
}

typedef MTDF<EvalHashEffectState, StandardMoveGenerator,
	     void, SearchRecorder, RealizationProbability> Mtdf;

void mtdfCaptureCheckTest::testDriverOneMove(){
  std::cerr << "\n testDriverOneMove\n";
  osl::SearchRecorder recorder("mtdfCaptureCheckTest.testDriverOneMove");
  
  EvalHashEffectState eState(state);
  Mtdf mtdfSearcher(eState,recorder);
  recorder.stream() << "\ntestDriverOneMove\n";
  recorder.stream() << eState;	// ̤ե˽񤤤Ƥ
  recorder.setLogMargin(100);

  mtdfSearcher.setInitialGuess();
  const Move best500=mtdfSearcher.computeBestMoveWithoutSoma(500);
  CPPUNIT_ASSERT_EQUAL(move61rook, best500);

  recorder.stream() << "\ntestDriverOneMove + SOMA\n";
  recorder.stream() << eState;
  mtdfSearcher.setInitialGuess();
  Move best500Soma=mtdfSearcher.computeBestMove(500);
  CPPUNIT_ASSERT_EQUAL(best500, best500Soma);
  // newMove(newPosition(5,3),newPosition(6,4),PBISHOP,PAWN,false,BLACK),
}

void mtdfCaptureCheckTest::testDriverTwoMove(){
  std::cerr << "\n testDriverTwoMove\n";
  osl::SearchRecorder recorder("mtdfCaptureCheckTest.testDriverTwoMove");

  EvalHashEffectState eState(state);
  Mtdf mtdfSearcher(eState,recorder);

  recorder.stream() << "\ntestDriverTwoMove 500\n";
  recorder.stream() << eState;
  
  mtdfSearcher.setInitialGuess();
  const Move best1000=mtdfSearcher.computeBestMoveWithoutSoma(600);
  CPPUNIT_ASSERT_EQUAL(move61rook, best1000);

//newMove(OFFBOARD,newPosition(6,1),ROOK,PTYPE_EMPTY,false,BLACK),
  recorder.stream() << "\ntestDriverTwoMove 700 + SOMA\n";
  recorder.stream() << eState;

  mtdfSearcher.setInitialGuess();
  Move best1000Soma=mtdfSearcher.computeBestMove(600);
  CPPUNIT_ASSERT_EQUAL(move61rook, best1000Soma);
 //newMove(newPosition(5,3),newPosition(6,4),PBISHOP,PAWN,false,BLACK)
}

void mtdfCaptureCheckTest::testDriverThreeMove(){
  std::cerr << "\n testDriverThreeMove\n";
  osl::SearchRecorder recorder("mtdfCaptureCheckTest.testDriverThreeMove");

  EvalHashEffectState eState(state);
  Mtdf mtdfSearcher(eState,recorder);

#if 0  
  recorder.stream() << "\ntestDriverThreeMove 1200\n";
  recorder.stream() << eState;
  recorder.resetNodeCount();
  clock_start();
  mtdfSearcher.setInitialGuess();
  Move best=mtdfSearcher.computeBestMoveWithoutSoma(700);
  clock_stop("",recorder.currentNodeCount());
  CPPUNIT_ASSERT_EQUAL(newMove(OFFBOARD,newPosition(7,1),ROOK,PTYPE_EMPTY,false,BLACK),
		       best);
#endif
  // std::cout<<eState<<best<<std::endl;
  recorder.stream() << "\ntestDriverThreeMove 1000 + SOMA\n"; // 1200
  recorder.stream() << eState;
  recorder.setLogMargin(100);
  recorder.resetNodeCount();
  clock_start();
  mtdfSearcher.setInitialGuess();
  Move bestSoma=mtdfSearcher.computeBestMove(700);
  clock_stop("",recorder.currentNodeCount());
  // const Move expectedSoma = newMove(newPosition(5,3),newPosition(6,4),PBISHOP,PAWN,false,BLACK);
  CPPUNIT_ASSERT_EQUAL(move61rook, bestSoma);
  //std::cout<<"Soma:" << std::endl <<bestSoma<<std::endl;

}

void mtdfCaptureCheckTest::testBugSearch(){
  std::cerr << "\n testBugSearch\n";
  osl::SearchRecorder recorder("mtdfCaptureCheckTest.testBugSearch");

  PawnMaskState bugState=CsaString(
"P1-KY-KE * -KI *  *  * -KE-KY\n"
"P2 * -OU-GI * -KI * -HI *  * \n"
"P3 * -FU *  * -FU-GI-KA-FU-FU\n"
"P4-FU * -FU-FU * -FU-FU *  * \n"
"P5 *  *  *  *  *  *  * +FU * \n"
"P6+FU * +FU * +FU+GI+FU * +FU\n"
"P7 * +FU * +FU * +FU *  *  * \n"
"P8 * +KA+OU * +KI+GI * +HI * \n"
"P9+KY+KE * +KI *  *  * +KE+KY\n"
"+\n").getInitialState();
  EvalHashEffectState eState(bugState);

  Mtdf mtdfSearcher(eState,recorder);
  recorder.stream() << "\ntestBugSearch 1600 + SOMA\n";
  recorder.stream() << eState;
  
  mtdfSearcher.setInitialGuess();
  Move bestSoma=mtdfSearcher.computeBestMove(1300);
  // 55 ϸǤ϶»ƤФʤɤ꤫?
  CPPUNIT_ASSERT_EQUAL(newMove(newPosition(5,6),newPosition(5,5),PAWN,PTYPE_EMPTY,false,BLACK),
		       bestSoma);

  recorder.stream() << "\ntestBugSearch 1600\n";
  recorder.stream() << eState;
  mtdfSearcher.setInitialGuess();
  Move best=mtdfSearcher.computeBestMoveWithoutSoma(1300);
  // 5655  1615 ɤǤΤ GenerateAllMoves ΥХ? ?
  CPPUNIT_ASSERT_EQUAL(newMove(newPosition(5,6),newPosition(5,5),PAWN,PTYPE_EMPTY,false,BLACK),
		       best);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
