#include "mtdfdebug.h"
#include "endGameState.h"
#include "osl/search/searchRecorder.h"
#include "osl/search/simpleHashTable.h"
#include "osl/record/csaString.h"
#include "osl/evalHashEffectState.h"
#include "osl/category/fWMoveGen.h"
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <iostream>

/**
 * Iterative deepening Τ̵ ؤ꤬ѤʤȤǧ
 *
 * Ʊλؤ꤬ʣ硤
 * - id  iteration κͥ
 * - id ʤ move generator ǽ˺ä꤬ͥ
 * Ȥ㤤Ǥ
 */
class mtdfIterativeDeepeningTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( mtdfIterativeDeepeningTest );
  CPPUNIT_TEST( testDriverOneMove );
  CPPUNIT_TEST( testDriverTwoMove );
  CPPUNIT_TEST( testDriverThreeMove );
  CPPUNIT_TEST( testBugSearch );
  CPPUNIT_TEST_SUITE_END();
private:
  osl::PawnMaskState state,state1;
public:
  mtdfIterativeDeepeningTest();
  void testDriverOneMove();
  void testDriverTwoMove();
  void testDriverThreeMove();
  void testBugSearch();
};


CPPUNIT_TEST_SUITE_REGISTRATION(mtdfIterativeDeepeningTest);

using namespace osl;

mtdfIterativeDeepeningTest::mtdfIterativeDeepeningTest()
{
  state=CsaString(endGameState).getInitialState();
  state1=CsaString(
"P1-KY-KE-GI-KI *  * -HI-KE-KY\n"
"P2 * -HI *  * -KA-OU *  *  * \n"
"P3-FU * -FU-FU-FU * -GI-FU-FU\n"
"P4 *  *  *  *  * -FU *  *  * \n"
"P5 * -FU * +KA *  *  *  *  * \n"
"P6 *  * +FU *  *  * +FU * +FU\n"
"P7+FU+FU * +FU+FU+FU+KE *  * \n"
"P8 * +GI+KI *  *  * +GI *  * \n"
"P9+KY+KE *  * +OU+KI *  * +KY\n"
"P+00FU\n"
"P+00FU\n"
"P+00KI\n"
"+\n").getInitialState();
}

void mtdfIterativeDeepeningTest::testDriverOneMove(){
  std::cerr << "\n testDriverOneMove\n";
  osl::SearchRecorder recorder("mtdfIterativeDeepeningTest.testDriverOneMove");

  EvalHashEffectState eState(state);
  const Move m5364um
    = newMove(newPosition(5,3),newPosition(6,4),PBISHOP,PAWN,false,BLACK);
  {
    recorder.stream() << "\ntestDriverOneMove + SOMA + TABLE\n";
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    Move best500=mtdfSearcher.computeBestMove(500);
    CPPUNIT_ASSERT_EQUAL(m5364um, best500);
  }
  {
    recorder.stream() << "\ntestDriverOneMove + SOMA + ID\n";
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    Move best500=mtdfSearcher.computeBestMoveIteratively(500);
    CPPUNIT_ASSERT_EQUAL(m5364um, best500);
  }
}

void mtdfIterativeDeepeningTest::testDriverTwoMove(){
  std::cerr << "\n testDriverTwoMove\n";
  osl::SearchRecorder recorder("mtdfIterativeDeepeningTest.testDriverTwoMove");

  EvalHashEffectState eState(state);
  Move best1000, best1000Id;
  {
    recorder.stream() << "\ntestDriverTwoMove 1000 + SOMA + TABLE\n";
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    best1000=mtdfSearcher.computeBestMove(1000);
  }
  {
    recorder.stream() << "\ntestDriverTwoMove 1000 + SOMA + TABLE + ID\n";
    recorder.setLogMargin(1000);
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    best1000Id=mtdfSearcher.computeBestMoveIteratively(1000,MOVE_INVALID,500);
  }
  CPPUNIT_ASSERT_EQUAL(best1000, best1000Id);
  // const Move expectedSoma = newMove(OFFBOARD,newPosition(7,1),ROOK,PTYPE_EMPTY,false,BLACK);
}

void mtdfIterativeDeepeningTest::testDriverThreeMove(){
  std::cerr << "\n testDriverThreeMove\n";
  osl::SearchRecorder recorder("mtdfIterativeDeepeningTest.testDriverThreeMove");

  EvalHashEffectState eState(state);
  Move bestMove, bestMoveID;
  {
    recorder.stream() << "\ntestDriverThreeMove 1500 + SOMA + TABLE\n";
    SimpleHashTable table(1000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMove=mtdfSearcher.computeBestMove(1500);
  }
  {
    recorder.stream() << "\ntestDriverThreeMove 1500 + SOMA + TABLE + ID\n";
    SimpleHashTable table(1000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    // recorder.setLogMargin(1500);
    recorder.stream() << "win values are "
		      << MtdfWithTable::winValue(BLACK) << " "
		      << MtdfWithTable::winValue(WHITE) << "\n";
    bestMoveID=mtdfSearcher.computeBestMoveIteratively(1500,MOVE_INVALID,500);
  }
  CPPUNIT_ASSERT_EQUAL(bestMove, bestMoveID);
}

void mtdfIterativeDeepeningTest::testBugSearch(){
  std::cerr << "\n testBugSearch\n";
  osl::SearchRecorder recorder("mtdfIterativeDeepeningTest.testBugSearch");

  PawnMaskState bugState=CsaString(
    "P1-KY-KE * -KI *  *  * -KE-KY\n"
    "P2 * -OU-GI * -KI * -HI *  * \n"
    "P3 * -FU *  * -FU-GI-KA-FU-FU\n"
    "P4-FU * -FU-FU * -FU-FU *  * \n"
    "P5 *  *  *  *  *  *  * +FU * \n"
    "P6+FU * +FU * +FU+GI+FU * +FU\n"
    "P7 * +FU * +FU * +FU *  *  * \n"
    "P8 * +KA+OU * +KI+GI * +HI * \n"
    "P9+KY+KE * +KI *  *  * +KE+KY\n"
    "+\n").getInitialState();
  EvalHashEffectState eState(bugState);
  Move bestMove;
  {
    recorder.stream() << "\ntestBugSearch 2500\n";
    SimpleHashTable table(3000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMove=mtdfSearcher.computeBestMove(1600); // 2500
  }
  Move bestMoveID;
  {
    recorder.stream() << "\ntestBugSearch 2500 + ID\n";
    recorder.setLogMargin(100);
    SimpleHashTable table(3000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMoveID=mtdfSearcher.computeBestMoveIteratively(1600,MOVE_INVALID,500); // 2500
  }
  CPPUNIT_ASSERT_EQUAL(bestMove, bestMoveID);
  CPPUNIT_ASSERT_EQUAL(newMove(newPosition(5,6),newPosition(5,5),PAWN,PTYPE_EMPTY,false,BLACK),
		       bestMoveID);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
