#include "mtdfdebug.h"
#include "endGameState.h"
#include "osl/search/searchRecorder.h"
#include "osl/search/simpleHashTable.h"
#include "osl/record/csaString.h"
#include "osl/evalHashEffectState.h"
#include "osl/category/fWMoveGen.h"
#include "osl/moveLogProb.h"
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <iostream>

/**
 * table ̵ͭǺ꤬ѤʤȤǧ
 */
class mtdfSimpleHashTableTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( mtdfSimpleHashTableTest );
  CPPUNIT_TEST( testDriverOneMove );
  CPPUNIT_TEST( testDriverTwoMove );
  CPPUNIT_TEST( testDriverThreeMove );
  CPPUNIT_TEST( testBugSearch );
  CPPUNIT_TEST_SUITE_END();
private:
  osl::PawnMaskState state,state1;
public:
  mtdfSimpleHashTableTest();
  void testDriverOneMove();
  void testDriverTwoMove();
  void testDriverThreeMove();
  void testBugSearch();
};


CPPUNIT_TEST_SUITE_REGISTRATION(mtdfSimpleHashTableTest);

using namespace osl::search;
using namespace osl;

mtdfSimpleHashTableTest::mtdfSimpleHashTableTest()
{
  state=CsaString(endGameState).getInitialState();
  state1=CsaString(
"P1-KY-KE-GI-KI *  * -HI-KE-KY\n"
"P2 * -HI *  * -KA-OU *  *  * \n"
"P3-FU * -FU-FU-FU * -GI-FU-FU\n"
"P4 *  *  *  *  * -FU *  *  * \n"
"P5 * -FU * +KA *  *  *  *  * \n"
"P6 *  * +FU *  *  * +FU * +FU\n"
"P7+FU+FU * +FU+FU+FU+KE *  * \n"
"P8 * +GI+KI *  *  * +GI *  * \n"
"P9+KY+KE *  * +OU+KI *  * +KY\n"
"P+00FU\n"
"P+00FU\n"
"P+00KI\n"
"+\n").getInitialState();
}

void mtdfSimpleHashTableTest::testDriverOneMove(){
  std::cerr << "\n testDriverOneMove\n";
  osl::SearchRecorder recorder("mtdfSimpleHashTableTest.testDriverOneMove");

  EvalHashEffectState eState(state);

  Move bestMove, bestMoveTable;
  recorder.stream() << "\ntestDriverOneMove + SOMA\n";
  {
    Mtdf mtdfSearcher(eState,recorder);
    bestMove = mtdfSearcher.computeBestMove(500);
  }
  {
    recorder.stream() << "\ntestDriverOneMove + SOMA + TABLE\n";
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMoveTable = mtdfSearcher.computeBestMove(500);
  }
  CPPUNIT_ASSERT_EQUAL(bestMove, bestMoveTable);
}

void mtdfSimpleHashTableTest::testDriverTwoMove(){
  std::cerr << "\n testDriverTwoMove\n";
  osl::SearchRecorder recorder("mtdfSimpleHashTableTest.testDriverTwoMove");
  recorder.stream() << state;
  
  EvalHashEffectState eState(state);
  Move best1000, best1000table;
  {
    recorder.stream() << "\ntestDriverTwoMove 1000 + SOMA\n";
    Mtdf mtdfSearcher(eState,recorder);
    best1000 = mtdfSearcher.computeBestMove(1000);
  }
  {
    recorder.stream() << "\ntestDriverTwoMove 1000 + SOMA + TABLE\n";
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    best1000table = mtdfSearcher.computeBestMove(1000);
  }
#if 0
  const Move m5364uma = newMove(newPosition(5,3),newPosition(6,4),
				PBISHOP,PAWN,false,BLACK);
#endif
  const Move move61rook = newMove(OFFBOARD,newPosition(6,1),ROOK,PTYPE_EMPTY,false,BLACK);
  CPPUNIT_ASSERT_EQUAL(move61rook, best1000);
  CPPUNIT_ASSERT_EQUAL(move61rook, best1000table);
}

void mtdfSimpleHashTableTest::testDriverThreeMove(){
  std::cerr << "\n testDriverThreeMove\n";
  osl::SearchRecorder recorder("mtdfSimpleHashTableTest.testDriverThreeMove");

  EvalHashEffectState eState(state);

  Move bestMove, bestMoveTable;
  {
    recorder.stream() << "\ntestDriverThreeMove 1500 + SOMA\n";
    Mtdf mtdfSearcher(eState,recorder);
    bestMove=mtdfSearcher.computeBestMove(1500);
  }
  {
    recorder.stream() << "\ntestDriverThreeMove 1500 + SOMA + TABLE\n";
    SimpleHashTable table(1000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMoveTable=mtdfSearcher.computeBestMove(1500);
  }
  CPPUNIT_ASSERT_EQUAL(bestMove, bestMoveTable);
}

void mtdfSimpleHashTableTest::testBugSearch(){
  std::cerr << "\n testBugSearch\n";
  osl::SearchRecorder recorder("mtdfSimpleHashTableTest.testBugSearch");

  PawnMaskState bugState=CsaString(
    "P1-KY-KE * -KI *  *  * -KE-KY\n"
    "P2 * -OU-GI * -KI * -HI *  * \n"
    "P3 * -FU *  * -FU-GI-KA-FU-FU\n"
    "P4-FU * -FU-FU * -FU-FU *  * \n"
    "P5 *  *  *  *  *  *  * +FU * \n"
    "P6+FU * +FU * +FU+GI+FU * +FU\n"
    "P7 * +FU * +FU * +FU *  *  * \n"
    "P8 * +KA+OU * +KI+GI * +HI * \n"
    "P9+KY+KE * +KI *  *  * +KE+KY\n"
    "+\n").getInitialState();
  EvalHashEffectState eState(bugState);

  Move bestMove, bestMoveTable;
  {
    recorder.stream() << "\ntestBugSearch 2500 + SOMA\n";
    Mtdf mtdfSearcher(eState,recorder);
    bestMove=mtdfSearcher.computeBestMove(2500);
  }
  
  {
    recorder.stream() << "\ntestBugSearch 2500\n";
    recorder.setLogMargin(100);
    SimpleHashTable table(3000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMoveTable=mtdfSearcher.computeBestMove(2500);
  }
  CPPUNIT_ASSERT_EQUAL(bestMove, bestMoveTable);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
