#include "mtdfdebug.h"
#include "endGameState.h"
#include "osl/search/searchRecorder.h"
#include "osl/record/csaString.h"
#include "osl/category/fWMoveGen.h"
#include "osl/perfmon.h"
#include "osl/evalHashEffectState.h"
#include "osl/moveLogProb.h"
#include "osl/pawnMaskState.h"

#include <iostream>

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

class mtdfTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE(mtdfTest);
  CPPUNIT_TEST(testWinValue);
  CPPUNIT_TEST(testShow);
  CPPUNIT_TEST(testMoveLogProb);
  CPPUNIT_TEST(testDriverOneMove);
  CPPUNIT_TEST(testDriverTwoMove);
  CPPUNIT_TEST(testDriverThreeMove);
  CPPUNIT_TEST(testBugSearch);
  CPPUNIT_TEST_SUITE_END();
private:
  osl::PawnMaskState state,state1;
public:
  mtdfTest();
  void testShow();
  void testMoveLogProb();
  void testDriverOneMove();
  void testDriverTwoMove();
  void testDriverThreeMove();
  void testBugSearch();
  void testWinValue();
};

CPPUNIT_TEST_SUITE_REGISTRATION(mtdfTest);

using namespace osl;

const Move move61rook = newMove(OFFBOARD,newPosition(6,1),ROOK,PTYPE_EMPTY,false,BLACK);
const Move move5364uma = newMove(newPosition(5,3),newPosition(6,4),PBISHOP,PAWN,false,BLACK);

mtdfTest::mtdfTest()
{
  state=CsaString(endGameState).getInitialState();
  state1=CsaString(
"P1-KY-KE-GI-KI *  * -HI-KE-KY\n"
"P2 * -HI *  * -KA-OU *  *  * \n"
"P3-FU * -FU-FU-FU * -GI-FU-FU\n"
"P4 *  *  *  *  * -FU *  *  * \n"
"P5 * -FU * +KA *  *  *  *  * \n"
"P6 *  * +FU *  *  * +FU * +FU\n"
"P7+FU+FU * +FU+FU+FU+KE *  * \n"
"P8 * +GI+KI *  *  * +GI *  * \n"
"P9+KY+KE *  * +OU+KI *  * +KY\n"
"P+00FU\n"
"P+00FU\n"
"P+00KI\n"
"+\n").getInitialState();
}

void mtdfTest::testShow(){
  std::cerr << "\n testShow\n";
  osl::SearchRecorder recorder("mtdfTest.testShow");

  EvalHashEffectState eState(state);
  Mtdf mtdfSearcher(eState,recorder);
  //  Move move=mtdfSearcher.computeBestMove(500);
  //  std::cout << move << std::endl;
}

void mtdfTest::testMoveLogProb(){
  std::cerr << "\n testMoveLogProb\n";

  MoveLogProb moveLogProb1(newMove(newPosition(8,2),newPosition(7,3),PROOK,PTYPE_EMPTY,false,BLACK),45);
  MoveLogProb moveLogProb2(newDrop(newPosition(8,2),PAWN,BLACK),60);
  //  std::cout << moveLogProb1 << std::endl << moveLogProb2 << std::endl;
  CPPUNIT_ASSERT_EQUAL(moveLogProb1.getMove(),
		       newMove(newPosition(8,2),newPosition(7,3),PROOK,PTYPE_EMPTY,false,BLACK));
  CPPUNIT_ASSERT_EQUAL(moveLogProb1.getLogProb(),
		       45);
  CPPUNIT_ASSERT_EQUAL(moveLogProb2.getMove(),
		       newDrop(newPosition(8,2),PAWN,BLACK));
  CPPUNIT_ASSERT_EQUAL(moveLogProb2.getLogProb(),
		       60);
}

void mtdfTest::testDriverOneMove(){
  std::cerr << "\n testDriverOneMove\n";
  osl::SearchRecorder recorder("mtdfTest.testDriverOneMove");
  recorder.setLogMargin(100);
  
  EvalHashEffectState eState(state);
  Mtdf mtdfSearcher(eState,recorder);

  recorder.stream() << "\ntestDriverOneMove + SOMA\n";
  recorder.stream() << eState;
  mtdfSearcher.setInitialGuess();
  Move best500Soma=mtdfSearcher.computeBestMove(500);
  CPPUNIT_ASSERT_EQUAL(move5364uma, best500Soma);
}

void mtdfTest::testDriverTwoMove(){
  std::cerr << "\n testDriverTwoMove\n";
  osl::SearchRecorder recorder("mtdfTest.testDriverTwoMove");

  EvalHashEffectState eState(state);
  Mtdf mtdfSearcher(eState,recorder);

  recorder.stream() << "\ntestDriverTwoMove 1000 + SOMA\n";
  recorder.stream() << eState;

  mtdfSearcher.setInitialGuess();
  const Move best1000Soma=mtdfSearcher.computeBestMove(1000);
  CPPUNIT_ASSERT_EQUAL(move61rook, best1000Soma);
}

void mtdfTest::testDriverThreeMove(){
  std::cerr << "\n testDriverThreeMove\n";
  osl::SearchRecorder recorder("mtdfTest.testDriverThreeMove");

  EvalHashEffectState eState(state);
  Mtdf mtdfSearcher(eState,recorder);

  recorder.stream() << "\ntestDriverThreeMove 1500 + SOMA\n";
  recorder.resetNodeCount();

  clock_start();
  mtdfSearcher.setInitialGuess();
  Move bestSoma=mtdfSearcher.computeBestMove(1500);
  clock_stop("",recorder.currentNodeCount());

  // const Move expectedSoma = newMove(newPosition(5,3),newPosition(6,4),PBISHOP,PAWN,false,BLACK);
  const Move expectedSoma = newMove(OFFBOARD,newPosition(6,1),ROOK,PTYPE_EMPTY,false,BLACK);
  CPPUNIT_ASSERT_EQUAL(expectedSoma, bestSoma);

}

void mtdfTest::testBugSearch(){
  std::cerr << "\n testBugSearch\n";
  osl::SearchRecorder recorder("mtdfTest.testBugSearch");

  PawnMaskState bugState=CsaString(
"P1-KY-KE * -KI *  *  * -KE-KY\n"
"P2 * -OU-GI * -KI * -HI *  * \n"
"P3 * -FU *  * -FU-GI-KA-FU-FU\n"
"P4-FU * -FU-FU * -FU-FU *  * \n"
"P5 *  *  *  *  *  *  * +FU * \n"
"P6+FU * +FU * +FU+GI+FU * +FU\n"
"P7 * +FU * +FU * +FU *  *  * \n"
"P8 * +KA+OU * +KI+GI * +HI * \n"
"P9+KY+KE * +KI *  *  * +KE+KY\n"
"+\n").getInitialState();
  EvalHashEffectState eState(bugState);

  recorder.setLogMargin(100);
  Mtdf mtdfSearcher(eState,recorder);
  recorder.stream() << "\ntestBugSearch 2500 + SOMA\n";
  recorder.stream() << eState;
  
  mtdfSearcher.setInitialGuess();
  Move bestSoma=mtdfSearcher.computeBestMove(2500);
  CPPUNIT_ASSERT_EQUAL(newMove(newPosition(5,6),newPosition(5,5),PAWN,PTYPE_EMPTY,false,BLACK),
		       bestSoma);

}

void mtdfTest::testWinValue()
{
  CPPUNIT_ASSERT(Mtdf::isWinValue(BLACK, Mtdf::winValue(BLACK)));
  CPPUNIT_ASSERT(Mtdf::isWinValue(WHITE, Mtdf::winValue(WHITE)));
  CPPUNIT_ASSERT(! Mtdf::isWinValue(BLACK, Mtdf::winValue(WHITE)));
  CPPUNIT_ASSERT(! Mtdf::isWinValue(WHITE, Mtdf::winValue(BLACK)));
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
