#include "osl/mtdf.h"
#include "osl/csaString.h"
#include "osl/pairEvalHashEffectState.h"
#include "osl/mtdfRecorder.h"
#include "osl/fullWidth.h"
#include "osl/simpleHashTable.h"
#include "osl/moveLogProb.h"
#include "osl/perfmon.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <iostream>

class piecePairMtdfIDTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE( piecePairMtdfIDTest );
  CPPUNIT_TEST( testBugSearch );
  CPPUNIT_TEST( testDriverOneMove );
  CPPUNIT_TEST( testDriverTwoMove );
  CPPUNIT_TEST( testDriverThreeMove );
  CPPUNIT_TEST_SUITE_END();
private:
  osl::PawnMaskState state,state1;
public:
  piecePairMtdfIDTest();
  void testDriverOneMove();
  void testDriverTwoMove();
  void testDriverThreeMove();
  void testBugSearch();
};


CPPUNIT_TEST_SUITE_REGISTRATION(piecePairMtdfIDTest);

using namespace osl;

namespace
{
struct Loader
{
  Loader()
  {
    const char *filename = "../data/sibling.pair";
    CPPUNIT_ASSERT(piecePairEvalTable.loadFromBinaryFile(filename));
  }
} loader;
}

piecePairMtdfIDTest::piecePairMtdfIDTest()
{
  state=CsaString(
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU\n"
"P-00AL\n"
"+\n").getInitialState();
  state1=CsaString(
"P1-KY-KE-GI-KI *  * -HI-KE-KY\n"
"P2 * -HI *  * -KA-OU *  *  * \n"
"P3-FU * -FU-FU-FU * -GI-FU-FU\n"
"P4 *  *  *  *  * -FU *  *  * \n"
"P5 * -FU * +KA *  *  *  *  * \n"
"P6 *  * +FU *  *  * +FU * +FU\n"
"P7+FU+FU * +FU+FU+FU+KE *  * \n"
"P8 * +GI+KI *  *  * +GI *  * \n"
"P9+KY+KE *  * +OU+KI *  * +KY\n"
"P+00FU\n"
"P+00FU\n"
"P+00KI\n"
"+\n").getInitialState();
}

typedef MTDF<PairEvalHashEffectState,FWMoveGen,
	     SimpleHashTable,SearchRecorder,FullWidthMoveProbability> MtdfWithTable;

void piecePairMtdfIDTest::testDriverOneMove(){
  std::cerr << "\n testDriverOneMove\n";
  osl::SearchRecorder recorder("piecePairMtdfIDTest.testDriverOneMove");
  PairEvalHashEffectState eState(state);
  {
    recorder.stream() << "\ntestDriverOneMove + SOMA + TABLE\n";
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    Move best500=mtdfSearcher.computeBestMove(500);
    CPPUNIT_ASSERT_EQUAL(newMove(newPosition(7,4),newPosition(7,3),PPAWN,PTYPE_EMPTY,true,BLACK),
			 best500);
  }
  {
    recorder.stream() << "\ntestDriverOneMove + SOMA + ID\n";
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    Move best500=mtdfSearcher.computeBestMoveIteratively(500);
    CPPUNIT_ASSERT( newMove(newPosition(7,4),newPosition(7,3),PPAWN,PTYPE_EMPTY,true,BLACK)==best500 ||
		    (std::cout << eState<< best500 << std::endl,0) 
      );
  }
}

void piecePairMtdfIDTest::testDriverTwoMove(){
  std::cerr << "\n testDriverTwoMove\n";
  osl::SearchRecorder recorder("piecePairMtdfIDTest.testDriverTwoMove");

  PairEvalHashEffectState eState(state);
  {
    recorder.stream() << "\ntestDriverTwoMove 1000 + SOMA + TABLE\n";
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    Move best1000=mtdfSearcher.computeBestMove(1000);
    CPPUNIT_ASSERT_EQUAL(newMove(newPosition(5,3),newPosition(6,4),PBISHOP,PAWN,false,BLACK),
			 best1000);
  }
  {
    recorder.stream() << "\ntestDriverTwoMove 1000 + SOMA + TABLE + ID\n";
    // recorder.setLogMargin(1000);
    SimpleHashTable table(1000000,1);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    Move best1000=mtdfSearcher.computeBestMoveIteratively(1000);
    // const Move expectedSoma = newMove(newPosition(5,3),newPosition(6,4),PBISHOP,PAWN,false,BLACK);
    // 4/28 ֤Ǥľ꤬71-61 ɤΤϵˤʤ뤬
    // Υ multimap ФƤ֤ˤΤΤ褦
    const Move expectedSoma = newMove(OFFBOARD,newPosition(7,1),ROOK,PTYPE_EMPTY,false,BLACK);
    
    CPPUNIT_ASSERT(expectedSoma==best1000||(std::cout<<eState<<best1000,0));
  }
  
}

void piecePairMtdfIDTest::testDriverThreeMove(){
  std::cerr << "\n testDriverThreeMove\n";
  osl::SearchRecorder recorder("piecePairMtdfIDTest.testDriverThreeMove");

  PairEvalHashEffectState eState(state);

  Move bestMove, bestMoveID;
  {
    recorder.stream() << "\ntestDriverThreeMove 1500 + SOMA + TABLE\n";
    SimpleHashTable table(1000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    clock_start();
    bestMove=mtdfSearcher.computeBestMove(1500);
    clock_stop("",1);
  }
  {
    recorder.stream() << "\ntestDriverThreeMove 1500 + SOMA + TABLE + ID\n";
    SimpleHashTable table(1000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    clock_start();
    bestMoveID=mtdfSearcher.computeBestMoveIteratively(1500);
    clock_stop("",1);
  }
  CPPUNIT_ASSERT_EQUAL(bestMove, bestMoveID);
}

void piecePairMtdfIDTest::testBugSearch(){
  std::cerr << "\n testBugSearch\n";
  osl::SearchRecorder recorder("piecePairMtdfIDTest.testBugSearch");

  PawnMaskState bugState=CsaString(
    "P1-KY-KE * -KI *  *  * -KE-KY\n"
    "P2 * -OU-GI * -KI * -HI *  * \n"
    "P3 * -FU *  * -FU-GI-KA-FU-FU\n"
    "P4-FU * -FU-FU * -FU-FU *  * \n"
    "P5 *  *  *  *  *  *  * +FU * \n"
    "P6+FU * +FU * +FU+GI+FU * +FU\n"
    "P7 * +FU * +FU * +FU *  *  * \n"
    "P8 * +KA+OU * +KI+GI * +HI * \n"
    "P9+KY+KE * +KI *  *  * +KE+KY\n"
    "+\n").getInitialState();
  PairEvalHashEffectState eState(bugState);
  Move bestMove;
  {
    recorder.stream() << "\ntestBugSearch 2500\n";
    SimpleHashTable table(3000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMove=mtdfSearcher.computeBestMove(1600); // 2500
  }
  Move bestMoveID;
  {
    recorder.stream() << "\ntestBugSearch 2500 + ID\n";
    recorder.setLogMargin(100);
    SimpleHashTable table(3000000,501);
    MtdfWithTable mtdfSearcher(eState,recorder,&table);
    bestMoveID=mtdfSearcher.computeBestMoveIteratively(1600); // 2500
  }
#if 0
  CPPUNIT_ASSERT_EQUAL(bestMove, bestMoveID);
#endif
  // iterative deepening 򤹤ȡɤߤη̤αƶ ? 75ͤóݤ뤳Ȥ򤱤
  CPPUNIT_ASSERT_EQUAL(newMove(newPosition(5,6),newPosition(5,5),PAWN,PTYPE_EMPTY,false,BLACK),
		       bestMoveID);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
