/* quiescenceSearch.t.cc
 */
#include "osl/search/quiescenceSearch.h"
#include "osl/search/quiescenceSearch.tcc"
#include "osl/numEffectState.h"
#include "osl/record/csaString.h"
#include "osl/record/record.h"
#include "osl/record/csa.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

#include <fstream>
#include <iostream>

class QuiescenceSearchTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(QuiescenceSearchTest);
  CPPUNIT_TEST(testNoCapture);
  CPPUNIT_TEST(testCaptureBlack);
  CPPUNIT_TEST(testCaptureWhite);
  CPPUNIT_TEST(testCaptureCheck);
  CPPUNIT_TEST(testCaptureCheckMate);
  CPPUNIT_TEST(testEquality);
  CPPUNIT_TEST_SUITE_END();
public:
  void testNoCapture();
  void testCaptureBlack();
  void testCaptureWhite();
  void testCaptureCheck();
  void testCaptureCheckMate();
  /** SimpleState  NumEffectState Ʊ̤Ф뤳Ȥǧ 
   * (ν֤㤦)
   */
  void testEquality();
};

CPPUNIT_TEST_SUITE_REGISTRATION(QuiescenceSearchTest);

using namespace osl;
using namespace osl::search;
extern bool isShortTest;

typedef NumEffectState state_t;
typedef QuiescenceSearch<state_t,PieceEval> qsearch_t;

void QuiescenceSearchTest::testNoCapture()
{
  state_t state(CsaString(
			 "P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
			 "P2 * -HI *  *  *  *  * -KA * \n"
			 "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
			 "P4 *  *  *  *  *  *  *  *  * \n"
			 "P5 *  *  *  *  *  *  *  *  * \n"
			 "P6 *  *  *  *  *  *  *  *  * \n"
			 "P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
			 "P8 * +KA *  *  *  *  * +HI * \n"
			 "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
			 "+\n").getInitialState());
  qsearch_t searcher(state);
  PieceEval ev(state);
  const int val1 = searcher.search(BLACK, ev, newPosition(8,2));
  CPPUNIT_ASSERT_EQUAL(0, val1);
}

void QuiescenceSearchTest::testCaptureBlack()
{
  {
    state_t state1(CsaString(
			    "P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
			    "P2 *  *  *  *  *  *  * -KA * \n"
			    "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
			    "P4 *  *  *  *  *  *  *  *  * \n"
			    "P5 *  *  *  *  *  *  *  *  * \n"
			    "P6 * -HI *  *  *  *  *  *  * \n"
			    "P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
			    "P8 * +KA *  *  *  *  * +HI * \n"
			    "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
			    "+\n").getInitialState());
    qsearch_t searcher1(state1);
    PieceEval ev(state1);
    const int val1 = searcher1.search(BLACK, ev, newPosition(8,6));
    CPPUNIT_ASSERT_EQUAL(PtypeEvalTraits<ROOK>::val*2, val1);
  }
  {    
    // ֤ѤΤ
    state_t state2(CsaString(
			    "P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
			    "P2 *  *  *  *  *  *  *  *  * \n"
			    "P3-FU * -FU-FU-FU-FU-FU-FU-FU\n"
			    "P4 * +FU *  *  *  *  *  *  * \n"
			    "P5 * -FU *  *  *  *  *  *  * \n"
			    "P6 * -HI *  *  *  *  * -KA * \n"
			    "P7+FU+HI+FU+FU+FU+FU+GI+FU+FU\n"
			    "P8 * +KA+GI *  *  * +FU *  * \n"
			    "P9+KY+KE * +KI+OU+KI * +KE+KY\n"
			    "+\n").getInitialState());
    qsearch_t searcher2(state2);
    PieceEval ev2(state2);
    const int val2 = searcher2.search(BLACK, ev2, newPosition(8,6));
    CPPUNIT_ASSERT_EQUAL(PtypeEvalTraits<BISHOP>::val*2, val2);

    NullWindow w(101);
    const int val2null = searcher2.search<BLACK>(w, ev2, newPosition(8,6));
    CPPUNIT_ASSERT_EQUAL(PtypeEvalTraits<BISHOP>::val*2, val2null);
  }
}

void QuiescenceSearchTest::testCaptureWhite()
{
  {
    state_t state1(CsaString(
			    "P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
			    "P2 *  *  *  *  *  *  * -KA * \n"
			    "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
			    "P4 *  *  *  *  *  *  *  *  * \n"
			    "P5 *  *  *  *  *  *  *  *  * \n"
			    "P6 * -HI *  *  *  *  *  *  * \n"
			    "P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
			    "P8 * +KA *  *  *  *  * +HI * \n"
			    "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
			    "-\n").getInitialState());
    qsearch_t searcher1(state1);
    PieceEval ev(state1);
    const int val1 = searcher1.search(WHITE, ev, newPosition(8,7));
    CPPUNIT_ASSERT_EQUAL(-PtypeEvalTraits<PROOK>::val
			 -PtypeEvalTraits<PAWN>::val*2
			 +PtypeEvalTraits<ROOK>::val,
			 val1);
  }
  {    
    // ֤ˤȤж»ʤ
    state_t state2(CsaString(
			    "P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
			    "P2 *  *  *  *  *  *  *  *  * \n"
			    "P3-FU * -FU-FU-FU-FU-FU-FU-FU\n"
			    "P4 * +FU *  *  *  *  *  *  * \n"
			    "P5 * -FU *  *  *  *  *  *  * \n"
			    "P6 * -HI *  *  *  *  * -KA * \n"
			    "P7+FU+HI+FU+FU+FU+FU+GI+FU+FU\n"
			    "P8 * +KA+GI *  *  * +FU *  * \n"
			    "P9+KY+KE * +KI+OU+KI * +KE+KY\n"
			    "-\n").getInitialState());
    qsearch_t searcher2(state2);
    PieceEval ev2(state2);
    const int val2 = searcher2.search(WHITE, ev2, newPosition(7,8));
    CPPUNIT_ASSERT_EQUAL(0, val2);
  }
}

void QuiescenceSearchTest::testCaptureCheck()
{
  {
    // ȶ⤬ȴʤΤ
    state_t state0(CsaString(
		     "P1 * -OU *  * -KI-GI * +RY * \n"
		     "P2 *  *  *  * -KA * +GI *  * \n"
		     "P3-FU-FU-FU-FU * -FU-FU-FU-FU\n"
		     "P4 *  *  *  * -FU * -GI-RY * \n"
		     "P5 *  *  *  * +FU *  *  *  * \n"
		     "P6 *  *  *  *  *  *  *  *  * \n"
		     "P7+FU+FU+FU+FU * +FU+FU+FU+FU\n"
		     "P8 * +KA *  *  *  *  *  *  * \n"
		     "P9+KY+KE+GI+KI+OU+KI * +KE+KY\n"
		     "P-00AL\n"
		     "-\n").getInitialState());
    qsearch_t searcher0(state0);
    PieceEval ev0(state0);
    const int val0 = searcher0.search(WHITE, ev0, newPosition(3,2));
    CPPUNIT_ASSERT_EQUAL(-PtypeEvalTraits<PAWN>::val*2, val0);
  }
  {
    // 32Ȳȴ
    state_t state1(CsaString(
		     "P1 * -OU *  *  * -KI * +RY * \n"
		     "P2 *  *  *  * -GI * +KI *  * \n"
		     "P3-FU-FU-FU-KA * -FU-FU-FU-FU\n"
		     "P4 *  *  * -FU-FU *  * -RY * \n"
		     "P5 *  *  *  * +FU *  *  *  * \n"
		     "P6 *  *  *  *  *  *  *  *  * \n"
		     "P7+FU+FU+FU+FU * +FU+FU+FU+FU\n"
		     "P8 * +KA *  *  *  *  *  *  * \n"
		     "P9+KY+KE+GI+KI+OU * +GI+KE+KY\n"
		     "P-00AL\n"
		     "-\n").getInitialState());
    CPPUNIT_ASSERT_EQUAL(0, PieceEval(state1).getVal());
    qsearch_t searcher1(state1);
    PieceEval ev(state1);
    const int val1 = searcher1.search(WHITE, ev, newPosition(3,2));
    CPPUNIT_ASSERT_EQUAL(-PtypeEvalTraits<PAWN>::val*2, val1);
  }
  {
    state_t state2(CsaString(
		     "P1-KY-KE * -KI *  *  * +RY-KY\n"
		     "P2 * -GI-OU-FU *  *  *  *  * \n"
		     "P3-FU-FU *  * -KI *  *  *  * \n"
		     "P4 *  *  *  * +FU *  *  * -FU\n"
		     "P5 *  *  * +FU *  *  *  *  * \n"
		     "P6 *  * +FU *  *  * +GI * +FU\n"
		     "P7+FU *  *  *  *  * +KI+FU * \n"
		     "P8 *  *  *  *  * -RY+FU+OU * \n"
		     "P9+KY+KE *  *  * -GI-GI+KE+KY\n"
		     "P+00FU00FU00KI00KA\n"
		     "P-00FU00FU00FU00FU00FU00KE00KA\n"
		     "+\n"
		     ).getInitialState());
    qsearch_t searcher2(state2);
    PieceEval ev(state2);
    const int val2 = searcher2.search(BLACK, ev, newPosition(3,9));
    const int val2n = searcher2.search<BLACK>(-1102, -1098, 0, ev, newPosition(3,9));
    CPPUNIT_ASSERT_EQUAL(val2, val2n);
  }
  {
    state_t state3(CsaString(
		     "P1-KY-KE * -KI *  *  * +RY-KY\n"
		     "P2 * -GI-OU-FU *  *  *  *  * \n"
		     "P3-FU-FU *  * -KI *  *  *  * \n"
		     "P4 *  *  *  * +FU *  *  * -FU\n"
		     "P5 *  *  * +FU *  *  *  *  * \n"
		     "P6 *  * +FU *  *  * +GI * +FU\n"
		     "P7+FU *  *  *  *  * +KI+FU * \n"
		     "P8 *  *  *  *  * -RY+FU+OU * \n"
		     "P9+KY+KE *  *  * -GI-KA+KE+KY\n"
		     "P+00FU00FU00KI00KA\n"
		     "P-00FU00FU00FU00FU00FU00KE00GI\n"
		     "+\n"
		     ).getInitialState());
    qsearch_t searcher3(state3);
    PieceEval ev(state3);
    const int val3  = searcher3.search(BLACK, ev, newPosition(3,9));
    const int val3n = searcher3.search<BLACK>(-1102, -1098, 0, ev, newPosition(3,9));
    CPPUNIT_ASSERT_EQUAL(val3, val3n);
  }
}

void QuiescenceSearchTest::testCaptureCheckMate()
{
  {
    // 
    state_t state(CsaString(
		    "P1-KY-KE-GI-KI-KA-KI-GI-KE-KY\n"
		    "P2 * -HI *  *  *  *  * -OU * \n"
		    "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
		    "P4 *  *  *  *  *  *  * +FU * \n"
		    "P5 *  *  *  *  *  *  *  *  * \n"
		    "P6 *  *  *  *  *  *  *  *  * \n"
		    "P7+FU+FU+FU+FU+FU+FU+FU * +FU\n"
		    "P8 * +KA *  *  *  *  * +HI * \n"
		    "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
		    "+\n").getInitialState());
    qsearch_t searcher(state);
    PieceEval ev(state);
    const int val1 = searcher.search(BLACK, ev, newPosition(2,2));
    CPPUNIT_ASSERT_EQUAL(SearchEvalBase<PieceEval>::winValue(BLACK), val1);
  }
  {
    // ꤸʤ֤Ȥ뤬ƨƤ֤Ȥ롥
    state_t state(CsaString(
		    "P1-KY-KE-GI-KI * -KI-GI-KE-KY\n"
		    "P2 * -HI *  *  *  *  * -KA * \n"
		    "P3-FU-FU * -FU-FU-FU-OU-FU-FU\n"
		    "P4 *  *  *  *  *  * -FU *  * \n"
		    "P5 *  * -FU *  *  * +HI *  * \n"
		    "P6 *  * +FU *  *  *  *  *  * \n"
		    "P7+FU+FU * +FU+FU+FU+FU+FU+FU\n"
		    "P8 * +KA *  *  *  *  *  *  * \n"
		    "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
		    "-\n").getInitialState());
    qsearch_t searcher(state);
    PieceEval ev(state);
    const int val1 = searcher.search(WHITE, ev, newPosition(7,6));
    CPPUNIT_ASSERT_EQUAL(200, val1);
  }
  {
    // ֤äƲ
    state_t state(CsaString(
		    "P1-KY-KE-GI-KI * -KI-GI-KE-KY\n"
		    "P2 * -HI *  *  *  *  * -KA * \n"
		    "P3-FU-FU * -FU-FU-FU-OU-FU-FU\n"
		    "P4 *  * -FU *  *  * -FU+HI * \n"
		    "P5 *  *  *  *  *  *  *  *  * \n"
		    "P6 *  * +FU *  *  *  *  *  * \n"
		    "P7+FU+FU * +FU+FU+FU+FU+FU+FU\n"
		    "P8 * +KA *  *  *  *  *  *  * \n"
		    "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
		    "-\n").getInitialState());
    qsearch_t searcher(state);
    PieceEval ev(state);
    const int val1 = searcher.search(WHITE, ev, newPosition(7,6));
    CPPUNIT_ASSERT_EQUAL(-PtypeEvalTraits<ROOK>::val*2, val1);
  }
  {
    // 
    state_t state(CsaString(
		    "P1 *  *  *  *  *  *  *  *  * \n"
		    "P2 * +KI-OU-KI+RY *  *  *  * \n"
		    "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
		    "P4 *  *  * -RY *  *  *  *  * \n"
		    "P5 *  *  *  *  *  *  *  *  * \n"
		    "P6 *  *  *  *  *  *  *  *  * \n"
		    "P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
		    "P8 * +KA * +GI *  *  *  *  * \n"
		    "P9+KY+KE *  * +OU+KI+GI+KE+KY\n"
		    "P-00AL\n"
		    "-\n").getInitialState());
    qsearch_t searcher(state);
    PieceEval ev(state);
    const int val1 = searcher.search(WHITE, ev, newPosition(8,2));
    CPPUNIT_ASSERT_EQUAL(0, val1);
  }
  {
    // Хβǽä QuiescenceSearch ΤǤϤʤ褦
    state_t state(CsaString(
		    "P1-KY *  *  *  * +NK-OU * -KY\n"
		    "P2 *  *  *  * +TO-GI-KI *  * \n"
		    "P3-FU * -HI *  * -FU * -FU * \n"
		    "P4 *  *  *  * -GI * -FU-KY-FU\n"
		    "P5 * -FU *  *  * -KE * +FU * \n"
		    "P6+FU *  * -FU+FU *  *  *  * \n"
		    "P7+KA+FU+KE+GI * +FU+FU * +FU\n"
		    "P8-GI+KI+KI *  *  *  * +HI * \n"
		    "P9-UM * +OU *  *  *  * +KE+KY\n"
		    "P+00FU\n"
		    "P+00FU\n"
		    "P-00FU\n"
		    "P-00KI\n"
		    "-\n").getInitialState());
    qsearch_t searcher(state);
    PieceEval ev(state);
    const int val1 = searcher.search(WHITE, ev, newPosition(4,1));
    CPPUNIT_ASSERT(PieceEval::isEvalValueConsistent(val1));
    CPPUNIT_ASSERT(! SearchEvalBase<PieceEval>::isWinValue(BLACK, val1));
  }
}

static void testIsEvalValueConsistent(std::string const& fileName)
{
  Record rec=CsaFile(fileName).getRecord();
  PawnMaskState state(rec.getInitialState());
  NumEffectState neState(state);
  const std::vector<osl::Move> moves=rec.getMoves();

  typedef QuiescenceSearch<PawnMaskState,PieceEval> qsearch_t;
  typedef QuiescenceSearch<NumEffectState,PieceEval> ne_qsearch_t;

  qsearch_t qs(state);
  ne_qsearch_t neqs(neState);
  PieceEval ev(state);
  for(unsigned int i=0;i<moves.size();i++)
  {
    if (i > 0)
    {
      const Position lastTo = getTo(moves[i-1]);
      const int val = qs.search(state.getTurn(), ev, lastTo);
      const int neval = neqs.search(state.getTurn(), ev, lastTo);
      if (val != neval)
      {
	std::cerr << fileName << " " << i << "\n" << state;
      }
    
      CPPUNIT_ASSERT_EQUAL(val, neval);
      CPPUNIT_ASSERT(PieceEval::isEvalValueConsistent(val));

      CPPUNIT_ASSERT(state.isConsistent(true));
      CPPUNIT_ASSERT(neState.isConsistent(true));
    }

    const Move move = moves[i];
    ApplyMoveOfTurn::doMove(ev, state, move);
    ApplyMoveOfTurn::doMove(state, move);
    ApplyMoveOfTurn::doMove(neState, move);
  }
}

void QuiescenceSearchTest::testEquality()
{
  std::ifstream ifs("../data/kifdat/FILES");
  CPPUNIT_ASSERT(ifs);
  std::string fileName;
  for(int i=0;i<100 && (ifs >> fileName); i++){
    if(fileName == "") 
      break;
    ::testIsEvalValueConsistent("../data/kifdat/" + fileName);
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
