#include "osl/search/simpleHashTable.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/moveLogProb.h"
#include "osl/hashKey.h"
#include "osl/simpleHashState.h"
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

extern bool isShortTest;

//  simpleHashRecordTest ʬΥɬפ

class simpleHashTableTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE(simpleHashTableTest);
  CPPUNIT_TEST(testCreation);
  CPPUNIT_TEST(testSize);
  CPPUNIT_TEST(testAllocate);
  CPPUNIT_TEST(testIsRecordedUpper);
  CPPUNIT_TEST(testIsRecordedLower);
  CPPUNIT_TEST(testOverWrite);
  CPPUNIT_TEST(testCapacity);
  CPPUNIT_TEST(testClear);
  CPPUNIT_TEST(testSetMinimumRecordLimit);
  CPPUNIT_TEST_SUITE_END();
public:
  void testCreation();
  void testSize();
  void testAllocate();
  void testIsRecordedUpper();
  void testIsRecordedLower();
  void testOverWrite();
  void testCapacity();
  void testClear();
  void testSetMinimumRecordLimit();
};

CPPUNIT_TEST_SUITE_REGISTRATION(simpleHashTableTest);

using namespace osl;
using namespace osl::search;

void simpleHashTableTest::testCreation() {
  SimpleHashTable table(1,1,false);
}

void simpleHashTableTest::testSize() {
  SimpleHashTable table(1,1,false);
  CPPUNIT_ASSERT_EQUAL(0u, table.size());
}

void simpleHashTableTest::testAllocate() {
  SimpleHashTable table(100,1,false);
  HashKey k;
  CPPUNIT_ASSERT_EQUAL(0u, table.size());

  SimpleHashRecord *record = table.allocate(k, 100);
  CPPUNIT_ASSERT_EQUAL(1u, table.size());
  CPPUNIT_ASSERT(! record->hasLowerBound(0));
  CPPUNIT_ASSERT(! record->hasUpperBound(0));
}

void simpleHashTableTest::testIsRecordedUpper() {
  SimpleHashTable table(100,1,false);
  MoveLogProb m1;
  const HashKey k;
  const int limit = 100;
  const int value = 8;
  {
    SimpleHashRecord *record = table.allocate(k, limit);
    record->setUpperBound(BLACK, limit, m1, value);
  }
  const SimpleHashRecord *record = table.find(k);
  CPPUNIT_ASSERT(record);
  CPPUNIT_ASSERT_EQUAL(value, record->upperBound);
}

void simpleHashTableTest::testIsRecordedLower() {
  SimpleHashTable table(100,1,false);
  MoveLogProb m1;
  const HashKey k;
  const int limit = 100;
  const int value = 8;
  {
    SimpleHashRecord *record = table.allocate(k, limit);
    record->setLowerBound(BLACK, limit, m1, value);
  }
  const SimpleHashRecord *record = table.find(k);
  CPPUNIT_ASSERT(record);
  CPPUNIT_ASSERT_EQUAL(value, record->lowerBound);
}

/** 
 * 2Ͽ鿼ۤη̤Ĥ 
 */
void simpleHashTableTest::testOverWrite() {
  const HashKey k;
  MoveLogProb m1;
  const int shallow_limit = 100;
  const int shallow_value = 8;
  const int deep_limit = 200;
  const int deep_value = 888;
  {
    // shallow -> deep
    SimpleHashTable table(100,1,false);
    SimpleHashRecord *record = table.allocate(k, shallow_limit);
    record->setLowerBound(BLACK, shallow_limit, m1, shallow_value);
    record->setLowerBound(BLACK, deep_limit, m1, deep_value);
    
    CPPUNIT_ASSERT(record->hasLowerBound(deep_limit));
    CPPUNIT_ASSERT_EQUAL(deep_value, record->lowerBound);
  }
  {
    // deep -> shallow
    SimpleHashTable table(100,1,false);
    SimpleHashRecord *record = table.allocate(k, shallow_limit);
    record->setLowerBound(BLACK, deep_limit, m1, deep_value);
    record->setLowerBound(BLACK, shallow_limit, m1, shallow_value);

    CPPUNIT_ASSERT(record->hasLowerBound(deep_limit));
    CPPUNIT_ASSERT_EQUAL(deep_value, record->lowerBound);
  }
}

/** capacity ۤ insert ʤ */
void simpleHashTableTest::testCapacity() {
  const unsigned int capacity = 100;
  SimpleHashTable table(capacity,1,false);
  CPPUNIT_ASSERT_EQUAL(capacity, table.capacity());
  // MoveLogProb m1;
  
  for (unsigned int i=1; i<=capacity*5; ++i)
  {
    const HashKey k(i,2,3,4);
    const int limit = i;
    // const int value = i*2;
    SimpleHashRecord *r = table.allocate(k, limit);
    if (i <= capacity)
    {
      CPPUNIT_ASSERT_EQUAL(i, table.size());
      const SimpleHashRecord *record = table.find(k);
      CPPUNIT_ASSERT_EQUAL(const_cast<const SimpleHashRecord*>(r), record);
    }
    else 
    {
      CPPUNIT_ASSERT_EQUAL(capacity, table.size());
      CPPUNIT_ASSERT(! table.find(k));
    }
  }
}

void simpleHashTableTest::testClear() {
  SimpleHashTable table(100,1,false);
  const HashKey k;
  CPPUNIT_ASSERT_EQUAL(0u, table.size());
  table.allocate(k, 100);
  CPPUNIT_ASSERT_EQUAL(1u, table.size());
  table.clear();
  CPPUNIT_ASSERT_EQUAL(0u, table.size());
}

void simpleHashTableTest::testSetMinimumRecordLimit() {
  SimpleHashTable table(100,10000,false);
  const HashKey k;
  table.setMinimumRecordLimit(1000);
  table.allocate(k, 100);
  CPPUNIT_ASSERT_EQUAL(0u, table.size());
  table.allocate(k, 10000);
  CPPUNIT_ASSERT_EQUAL(1u, table.size());
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
